#!/bin/sh

# Installscript for Vuurmuur.
#
# Copyright (c) 2004 by Victor Julien
# Licenced under the GPL.
#
VERSION="0.5.54"

# defaults
INSTALLDIR="/usr"
ETCDIR="/etc/vuurmuur"
PLUGINDIR="$ETCDIR/plugins/"
LOGDIR="/var/log/vuurmuur/"
SYSTEMLOG="/var/log/messages"

CURPATH=`pwd`
LOG="$CURPATH/install.log"

INSTALL="0"
UPGRADE="0"
UNINSTALL="0"
DRYRUN="0"

VERBOSE="1"
DEBUG="0"

if [ "$EUID" != "0" ]; then
    echo "Error. This script requires to be run as user root."
    exit 1
fi

# initialize the log
echo "Vuurmuur installer starting..." > $LOG


function Exit
{
    RETVAL="$1"
    
    # gathering some system info
    echo >> $LOG
    echo "---" >> $LOG
    echo >> $LOG
    echo >> $LOG
    echo "Getting some info about the system." >> $LOG
    echo >> $LOG
    gcc --version >> $LOG
    libtoolize --version >> $LOG
    make --version >> $LOG
    aclocal --version >> $LOG
    autoheader --version >> $LOG
    automake --version >> $LOG
    autoconf --version >> $LOG
    echo >> $LOG
    uname -a >> $LOG
    
    echo
    echo
    echo "Installation Failed"
    echo "==================="
    echo
    echo "Please take a look at install.log. If you can't solve the problem"
    echo "mail me at victor@nk.nl. Please include the install.log."
    echo

    exit $RETVAL
}

function PrintL
{
    if [ "$VERBOSE" = "1" ]; then
	echo "$1"
    fi

    echo "$1" >> $LOG
}

function Cp
{
    cp $1 $2 $3 &> tmp.log
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "Cp succeeded."
	fi
    else
	PrintL "cp $1 $2 $3 failed with returncode $RESULT."
	cat tmp.log >> $LOG
	rm -f tmp.log
	Exit 1
    fi

    rm -f tmp.log
}

function Cd
{
    cd $1 &> tmp.log
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "Cd succeeded."
	fi
    else
	PrintL "cd $1 failed with returncode $RESULT."
	cat tmp.log >> $LOG
	rm -f tmp.log
	Exit 1
    fi

    rm -f tmp.log
}

function Libtoolize
{
    touch tmp.log
    libtoolize $1 &> tmp.log
    RESULT="$?"
    cat tmp.log >> $LOG
    rm -f tmp.log
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "Libtoolize succeeded."
	fi
    else
	PrintL "libtoolize $1 failed with returncode $RESULT."
	Exit 1
    fi
}

function Make
{
    touch tmp.log
    make $1 $2 &> tmp.log
    RESULT="$?"
    cat tmp.log >> $LOG
    rm -f tmp.log
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "Make succeeded."
	fi
    else
	PrintL "make $1 $2 failed with returncode $RESULT."
	Exit 1
    fi
}

function Configure
{
    touch tmp.log
    ./configure $1 &> tmp.log
    RESULT="$?"
    cat tmp.log >> $LOG
    rm -f tmp.log
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "configure $1 succeeded."
	fi
    else
	PrintL "./configure $1 failed with returncode $RESULT."
	Exit 1
    fi
}

function CheckFile
{
    if [ ! -f $1 ]; then
	PrintL "Error! Missing file: $1!"
	exit 1
    else
	if [ "$DEBUG" = "1" ]; then
    	    PrintL "Good! $1 exists."
	fi
    fi
}

function UnPack
{
    touch tmp.log
    gzip -cd $1 | tar -xvf - >> $LOG
    RESULT="$?"
    cat tmp.log >> $LOG
    rm -f tmp.log
    if [ "$RESULT" = "0" ]; then
	if [ "$DEBUG" = "1" ]; then
	    PrintL "UnPack $1 succeeded."
	fi
    else
	PrintL "UnPack $1 failed with returncode $RESULT."
	Exit 1
    fi
}

function RevCheckDir
{
    cd $1 &> /dev/null
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	PrintL "Error: the directory '$1' already exists."
	cd $CURPATH
	Exit 1
    fi
}

function CheckDir
{
    cd $1 &> /dev/null
    RESULT="$?"
    if [ "$RESULT" != "0" ]; then
	PrintL "Error: the directory '$1' doesn't exist."
	Exit 1
    fi

    cd $CURPATH
}

function MkDir
{
    mkdir $1 &> /dev/null
    RESULT="$?"
    if [ "$RESULT" != "0" ]; then
	PrintL "Error: the directory '$1' could not be created."
	Exit 1
    fi
}


# commandline options
if [ "$1" = "--dryrun" ]; then
    DRYRUN="1"
    PrintL "Commandline option '--dryrun' enabled."
elif [ "$1" = "--install" ]; then
    INSTALL="1"
    if [ "$DEBUG" = "1" ]; then
	PrintL "Commandline option '--install' enabled."
    fi

    echo
    echo "Vuurmuur installation"
    echo "====================="
    echo
    echo "Welcome to the installation of Vuurmuur. First you will be"
    echo "asked a couple of questions about the location to install the"
    echo "various parts of Vuurmuur. It is recommended that you choose"
    echo "the defaults, by pressing just enter."
    echo

elif [ "$1" = "--upgrade" ]; then
    UPGRADE="1"
    if [ "$DEBUG" = "1" ]; then
	PrintL "Commandline option '--install' enabled."
    fi

    echo
    echo "Vuurmuur installation (upgrade)"
    echo "==============================="
    echo
    echo "Welcome to the installation of Vuurmuur. First you will be"
    echo "asked a couple of questions about the location to install the"
    echo "various parts of Vuurmuur. Make sure that you point to the"
    echo "existing paths."
    echo


elif [ "$1" = "--uninstall" ]; then
    UNINSTALL="1"
    PrintL "Commandline option '--uninstall' enabled."
else
    echo
    echo "Help for Vuurmuur-installer."
    echo
    echo "Commandline options:"
    echo
    echo "	--install	install and setup up the config"
    echo "	--upgrade	install without touching the config"
    echo "	--uninstall	uninstall but leave the config alone"
    echo
    echo "Please read INSTALL for more information."
    echo
    exit 0
fi


if [ "$UNINSTALL" = "0" ]; then
    # get the installdir
    echo "Please enter the installation dir ($INSTALLDIR)."
    read TALK
    if [ "$TALK" != "" ]; then
	INSTALLDIR="$TALK"
    fi
    #echo   "Installdir: $INSTALLDIR ..."
    PrintL "Installdir: $INSTALLDIR ..."
    
    cd $INSTALLDIR &> /dev/null
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	Cd $CURPATH
    else
	MkDir $INSTALLDIR
    fi


    # get the shareddir
    SHAREDDIR="$INSTALLDIR/share/vuurmuur"
    echo
    echo "Please enter the shared files dir ($SHAREDDIR)."
    read TALK
    if [ "$TALK" != "" ]; then
	SHAREDDIR="$TALK"
    fi    
    #echo   "Using Shareddir: $SHAREDDIR ..."
    PrintL "Using Shareddir: $SHAREDDIR ..."

    cd $SHAREDDIR &> /dev/null
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	Cd $CURPATH
    else
	MkDir $SHAREDDIR
    fi

    cd $SHAREDDIR/scripts &> /dev/null
    RESULT="$?"
    if [ "$RESULT" = "0" ]; then
	Cd $CURPATH
    else
        MkDir $SHAREDDIR/scripts
	chmod 0700 $SHAREDDIR/scripts
    fi
fi

if [ "$UPGRADE" = "1" ] || [ "$INSTALL" = "1" ]; then
    if [ "$UPGRADE" = "1" ]; then
	echo
	echo "Please enter the directory containing the current Vuurmuur config ($ETCDIR)."
    elif [ "$INSTALL" = "1" ]; then
	echo
        echo "Please enter the directory where the config is going to be stored ($ETCDIR)."
    fi

    read TALK
    if [ "$TALK" != "" ]; then
        ETCDIR="$TALK"
    fi
    #echo   "Using Etcdir: '$ETCDIR'."
    PrintL "Using Etcdir: '$ETCDIR'."
	
    if [ "$UPGRADE" = "1" ]; then
        CheckDir $ETCDIR
    elif [ "$INSTALL" = "1" ]; then
        RevCheckDir $ETCDIR
    
        MkDir $ETCDIR
	chmod 0700 $ETCDIR

        # create the plugins dir as we need it for the installation of the plugin
        MkDir $ETCDIR/plugins
	chmod 0700 $ETCDIR/plugins
    fi
fi

if [ "$INSTALL" = "1" ]; then

    echo
    echo "Please enter the directory where Vuurmuur will store it's logs ($LOGDIR)."

    read TALK
    if [ "$TALK" != "" ]; then
        LOGDIR="$TALK"
    fi

    #echo   "Using Logdir: '$LOGDIR'."
    PrintL "Using Logdir: '$LOGDIR'."
	
    if [ "$INSTALL" = "1" ]; then
	cd $LOGDIR &> /dev/null
	RESULT="$?"
	if [ "$RESULT" != "0" ]; then
	    MkDir $LOGDIR
	    chmod 0700 $LOGDIR
	fi
	    
	cd $CURPATH
    fi
fi

if [ "$UNINSTALL" = "0" ]; then

    echo
    echo "Ok, thank you. Going to build Vuurmuur now. Depending on your hardware"
    echo "this process will take about 2 to 10 minutes."
    echo

    PrintL "Testing for the installation files..."
    CheckFile vuurmuur-$VERSION.tar.gz
    CheckFile libvuurmuur-$VERSION.tar.gz
    CheckFile vuurmuur_log-$VERSION.tar.gz
    CheckFile vuurmuur_conf-$VERSION.tar.gz
    CheckFile vuurmuur_textdir-$VERSION.tar.gz

    PrintL "Going to extract the files..."
    UnPack vuurmuur-$VERSION.tar.gz
    UnPack vuurmuur_log-$VERSION.tar.gz
    UnPack vuurmuur_conf-$VERSION.tar.gz
    UnPack vuurmuur_textdir-$VERSION.tar.gz
    UnPack libvuurmuur-$VERSION.tar.gz
    PrintL "Extracting the files done..."
fi

Cd libvuurmuur
if [ "$INSTALL" = "1" ] || [ "$UPGRADE" = "1" ]; then

    PrintL "Going to build libvuurmuur... (common code for all parts of Vuurmuur)."
    Libtoolize -f
    Make -f Makefile.cvs
    Configure --prefix=$INSTALLDIR
    Make
    if [ "$DRYRUN" != "1" ]; then
	Make install
    fi
    Make clean
    PrintL "Building and installing libvuurmuur finished."

elif [ "$UNINSTALL" = "1" ]; then

    PrintL "Going to uninstall libvuurmuur..."
    Make uninstall
    PrintL "Un-installing libvuurmuur finished."
fi
Cd ..


Cd vuurmuur
if [ "$INSTALL" = "1" ] || [ "$UPGRADE" = "1" ]; then
    PrintL "Going to build vuurmuur... (the daemon that creates the rules)."
    Libtoolize -f
    Make -f Makefile.cvs
    Configure --prefix=$INSTALLDIR
    Make
    if [ "$DRYRUN" != "1" ]; then
	Make install
	Cp -f scripts/vuurmuur $SHAREDDIR/scripts/vuurmuur-initd.sh
    fi
    Make clean
    PrintL "Building and installing vuurmuur finished."

elif [ "$UNINSTALL" = "1" ]; then

    PrintL "Going to uninstall vuurmuur..."
    Make uninstall
    PrintL "Un-installing vuurmuur finished."
fi
Cd ..


Cd vuurmuur_log
if [ "$INSTALL" = "1" ] || [ "$UPGRADE" = "1" ]; then
    PrintL "Going to build vuurmuur_log... (the log parsing daemon)."
    Libtoolize -f
    Make -f Makefile.cvs
    Configure --prefix=$INSTALLDIR
    Make
    if [ "$DRYRUN" != "1" ]; then
	Make install
	Cp -f scripts/vuurmuur $SHAREDDIR/scripts/vuurmuur-logrotate
    fi
    Make clean
    PrintL "Building and installing vuurmuur_log finished."

elif [ "$UNINSTALL" = "1" ]; then

    PrintL "Going to uninstall vuurmuur_log..."
    Make uninstall
    PrintL "Un-installing vuurmuur_log finished."
fi
Cd ..


Cd textdir
if [ "$INSTALL" = "1" ] || [ "$UPGRADE" = "1" ]; then
    PrintL "Going to build textdir plugin... (handles the way Vuurmuur stores it's config)."
    Make
    if [ "$DRYRUN" != "1" ]; then
	Cp -f textdir.so $ETCDIR/plugins/
	
	if [ "$INSTALL" = "1" ]; then
	    touch $ETCDIR/plugins/textdir.conf
	    chmod 0600 $ETCDIR/plugins/textdir.conf
	    echo "LOCATION=\"$ETCDIR/textdir\"" > $ETCDIR/plugins/textdir.conf
	fi
    fi
    Make clean
    PrintL "Building and installing textdir plugin finished."

elif [ "$UNINSTALL" = "1" ]; then

    PrintL "Going to uninstall textdir plugin..."
    rm -f $ETCDIR/plugins/textdir.so
    PrintL "Un-installing textdir plugin finished."
fi
Cd ..


Cd vuurmuur_conf
if [ "$INSTALL" = "1" ] || [ "$UPGRADE" = "1" ]; then
    PrintL "Going to build vuurmuur_conf... (the Ncurses based user interface)."
    Libtoolize -f
    Make -f Makefile.cvs
    Configure --prefix=$INSTALLDIR
    Make
    if [ "$DRYRUN" != "1" ]; then
	Make install
	#TODO: vuurmuur.hlp and vuurmuur_searchlog.sh
	Cp -f help/vuurmuur.hlp $SHAREDDIR/
	Cp -f scripts/vuurmuur-searchlog.sh $SHAREDDIR/scripts/
    fi
    Make clean
    PrintL "Building and installing vuurmuur_conf finished."

elif [ "$UNINSTALL" = "1" ]; then

    PrintL "Going to uninstall vuurmuur_conf..."
    Make uninstall
    #TODO: vuurmuur.hlp and vuurmuur_searchlog.sh
    PrintL "Un-installing vuurmuur_conf finished."
fi
Cd ..


if [ "$INSTALL" = "1" ]; then

    echo
    echo "Setting up the Vuurmuur config for first-time use."

    MkDir $ETCDIR/textdir
    chown root:root $ETCDIR/textdir
    chmod 0700 $ETCDIR/textdir
    
    MkDir $ETCDIR/textdir/interfaces
    chown root:root $ETCDIR/textdir/interfaces
    chmod 0700 $ETCDIR/textdir/interfaces
    
    MkDir $ETCDIR/textdir/services
    chown root:root $ETCDIR/textdir/services
    chmod 0700 $ETCDIR/textdir/services
    
    MkDir $ETCDIR/textdir/zones
    chown root:root $ETCDIR/textdir/zones
    chmod 0700 $ETCDIR/textdir/zones

    cp -r --preserve=mode zones/* $ETCDIR/textdir/zones
    chown root:root $ETCDIR/textdir/zones -R
    chmod 0700 $ETCDIR/textdir/zones

    cp services/* $ETCDIR/textdir/services
    chown root:root $ETCDIR/textdir/services -R
    chmod 0700 $ETCDIR/textdir/services
    
    touch $ETCDIR/rules.conf
    chown root:root $ETCDIR/rules.conf
    chmod 0600 $ETCDIR/rules.conf

    touch $ETCDIR/blocked.list
    chown root:root $ETCDIR/blocked.list
    chmod 0600 $ETCDIR/blocked.list

    # create the config file
    CONFIGFILE="$ETCDIR/config.conf"
    touch $CONFIGFILE
    chown root:root $CONFIGFILE
    chmod 0600 $ETCDIR/config.conf

    # seek iptables
    which iptables &> /dev/null
    RESULT=$?
    if [ "$RESULT" = "0" ]; then
	IPTABLESLOC=`which iptables`
    else
	echo
	echo "Warning: could not find the location of the 'iptables' command. Please make sure you have"
	echo "iptables installed and set it's location in the Vuurmuur config."
	IPTABLESLOC="/sbin/iptables"
    fi

    # seek ifconfig
    which ifconfig &> /dev/null
    RESULT=$?
    if [ "$RESULT" = "0" ]; then
	IFCONFIGLOC=`which ifconfig`
    else
	echo
	echo "Warning: could not find the location of the 'ifconfig' command. Please make sure you have"
	echo "ifconfig installed and set it's location in the Vuurmuur config."
	IFCONFIGLOC="/sbin/ifconfig"
    fi


    # write the configfile
    echo "# begin of file" > $CONFIGFILE

    echo "SERVICES_BACKEND=\"textdir\"" >> $CONFIGFILE
    echo "ZONES_BACKEND=\"textdir\"" >> $CONFIGFILE
    echo "INTERFACES_BACKEND=\"textdir\"" >> $CONFIGFILE
    echo "PLUGINDIR=\"$PLUGINDIR\"" >> $CONFIGFILE

    echo "RULESFILE=\"$ETCDIR/rules.conf\"" >> $CONFIGFILE
    echo "BLOCKLISTFILE=\"$ETCDIR/blocked.list\"" >> $CONFIGFILE

    echo "IPTABLES=\"$IPTABLESLOC\"" >> $CONFIGFILE
    echo "IFCONFIG=\"$IFCONFIGLOC\"" >> $CONFIGFILE

    echo "SCRIPTS_DIR=\"$SHAREDDIR/scripts\"" >> $CONFIGFILE
    echo "LOGDIR=\"$LOGDIR\"" >> $CONFIGFILE
    echo "SYSTEMLOG=\"$SYSTEMLOG\"" >> $CONFIGFILE
    echo "LOGLEVEL=\"info\"" >> $CONFIGFILE

    echo "# LOG_POLICY controls the logging of the default policy." >> $CONFIGFILE
    echo "LOG_POLICY=\"Yes\"" >> $CONFIGFILE

    echo "# LOG_POLICY_LIMIT sets the maximum number of logs per second." >> $CONFIGFILE
    echo "LOG_POLICY_LIMIT=\"30\"" >> $CONFIGFILE

    echo "# SYN_LIMIT sets the maximum number of SYN-packets per second." >> $CONFIGFILE
    echo "SYN_LIMIT=\"10\"" >> $CONFIGFILE
    echo "SYN_LIMIT_BURST=\"20\"" >> $CONFIGFILE

    echo "# end of file" >> $CONFIGFILE

    # create the config file
    VUURMUURCONFIGFILE="$ETCDIR/vuurmuur_conf.conf"
    touch $VUURMUURCONFIGFILE
    chown root:root $VUURMUURCONFIGFILE
    chmod 0600 $VUURMUURCONFIGFILE

    # write the vuurmuur_conf settingsfile
    echo "# begin of file" > $VUURMUURCONFIGFILE

    echo "ADVANCED_MODE=\"No\"" >> $VUURMUURCONFIGFILE
    echo
    echo "HELPFILE=\"$SHAREDDIR/vuurmuur.hlp\"" >>  $VUURMUURCONFIGFILE
    echo
    echo "NEWRULE_LOG=\"Yes\"" >> $VUURMUURCONFIGFILE
    echo
    echo "NEWRULE_LOGLIMIT=\"30\"" >> $VUURMUURCONFIGFILE
    echo
    echo "LOGVIEW_BUFSIZE=\"1500\"" >> $VUURMUURCONFIGFILE

    echo "# end of file" >> $VUURMUURCONFIGFILE

    # some final information
    echo
    echo "Installation Complete"
    echo "====================="
    echo
    echo "Installing Vuurmuur completed successfully. Please run vuurmuur_conf to"
    echo "complete your configuration. The first step is to define one or more"
    echo "interfaces and attach those to a network. After that rules can be created."
    echo
    echo "Example init.d and logrotate scripts are installed in:"
    echo "'$SHAREDDIR/scripts'."
    echo
    echo

elif [ "$UPGRADE" = "1" ]; then

    echo
    echo "Upgrading Complete"
    echo "=================="
    echo
    echo "Upgrading to version $VERSION completed successfully. Please run"
    echo "vuurmuur_conf and check the configuration for new or changed options."
    echo
    echo "ATTENTION!"
    echo "Please note that the protection rules (e.g. anti-spoofing) have been"
    echo "redesigned in version 0.5.54. Please check your interfaces and"
    echo "networks."
    echo
fi

#EOF
