/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"


int
libvuurmuur_logprint_error(int errorlevel, char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size,
			retval = 0;

	/* start with the head */
	snprintf(long_str, MAX_LOGRULE_SIZE, "%s (%d): ", head, errorlevel);

	/* init size */
	size = strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	/* print in the error log */
	libvuurmuur_logprint(conf.errorlog_location, libvuurmuur_verbose, bash_out, long_str);
	/* and in the info log */
	libvuurmuur_logprint(conf.vuurmuurlog_location, 0, 0, long_str);

	return(0);
}


int
libvuurmuur_logprint(char *logfile, int logconsole, int bash, char *logstring)
{
	int		retval=0;
	pid_t		pid;
	time_t		td;
	struct tm	*dcp;
	FILE		*fp;

	pid = getpid();
	time(&td);
	dcp = localtime(&td);

	if(bash == 1 || logconsole == 1)
	{
		if(bash == 1)
		{
			fprintf(stdout, "# ");
		}
		fprintf(stdout, "%s", logstring);
	}

	fp = fopen(logfile, "a");
	if(!fp)
	{
		fprintf(stdout, "Error opening logfile '%s', %s.\n", logfile, strerror(errno));
		retval=-1;
	}
	else
	{
		fprintf(fp, "%02d/%02d/%04d %02d:%02d:%02d : PID %-5d : %-13s : %s",	dcp->tm_mon +1,		// Month
											dcp->tm_mday,		// Day
											dcp->tm_year + 1900,	// Year
											dcp->tm_hour,		// Hour
											dcp->tm_min,		// Minute
											dcp->tm_sec,		// Second
											pid,
											vrprint.logger,		/* the name of the logger */
											logstring);

		fflush(fp);
		fclose(fp);
	}

	return(retval);
}

int
libvuurmuur_logprint_warning(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size,
			retval=0;

	snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);

	// init size
	size = strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// now print in the warning log
	libvuurmuur_logprint(conf.vuurmuurlog_location, libvuurmuur_verbose, bash_out, long_str);

	return(0);
}

int
libvuurmuur_logprint_info(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size, retval=0;

	snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	libvuurmuur_logprint(conf.vuurmuurlog_location, libvuurmuur_verbose, bash_out, long_str);

	return(0);
}

int
libvuurmuur_logprint_debug(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size=0,
			retval=0;

	if(head)
		snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);
	else
		strcpy(long_str, "");

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// print in the debug log
	libvuurmuur_logprint(conf.debuglog_location, 0, 0, long_str);

	return(0);
}

int libvuurmuur_stdoutprint_error(int errorlevel, char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size,
			retval=0;

	snprintf(long_str, MAX_LOGRULE_SIZE, "%s (%d): ", head, errorlevel);

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// print
	fprintf(stdout, "%s", long_str);
	fflush(stdout);

	return(0);
}

int libvuurmuur_stdoutprint_warning(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size,
			retval=0;

	snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// print
	fprintf(stdout, "%s", long_str);

	return(0);
}

int libvuurmuur_stdoutprint_info(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size,
			retval=0;

	snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// print
	fprintf(stdout, "%s", long_str);

	return(0);
}

int libvuurmuur_stdoutprint_debug(char *head, char *fmt, ...)
{
	va_list		ap;
	char		long_str[MAX_LOGRULE_SIZE];
	char		temp_str[MAX_LOGRULE_SIZE];
	int		size=0,
			retval=0;

	if(head)
		snprintf(long_str, MAX_LOGRULE_SIZE, "%s: ", head);
	else
		strcpy(long_str, "");

	// init size
	size=strlen(long_str);

	va_start(ap, fmt);
	vsnprintf(temp_str, sizeof(temp_str), fmt, ap);

	size=size+strlen(temp_str);
	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, temp_str, sizeof(long_str));
	}
	else
	{
		fprintf(stdout, "Error, logstring to long (size = %d, while max is %d)\n", size, MAX_LOGRULE_SIZE);
		retval=-1;
	}

	va_end(ap);

	if(size < MAX_LOGRULE_SIZE)
	{
		strlcat(long_str, "\n", sizeof(long_str));
		size=size+strlen("\n");
	}

	if(size > MAX_LOGRULE_SIZE)
	{
		fprintf(stdout, "Error, string too long for logging.\n");
		retval=-1;
	}

	// print
	fprintf(stdout, "%s", long_str);

	return(0);
}
