/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "vuurmuur.h"

/* remove_list */
int
query_remove_list(const int debuglvl, int place, d_list *rules_list, int updatenumbers)
{
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!rules_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "query_remove_list: start: place: %d, updatenumbers: %d, listsize: %d", place, updatenumbers, rules_list->len);

	for(d_node = rules_list->top; d_node ; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "query_remove_list: rule_ptr->number: %d, place: %d", rule_ptr->number, place);

		if(rule_ptr->number != place)
		{
			//fprintf(stdout, "do nothing\n");
		}
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "query_remove_list: now we have to remove (query_ptr->number: %d, place: %d)", rule_ptr->number, place);

			if(d_list_node_is_bot(debuglvl, d_node))
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "query_remove_list: removing last entry");

				if(d_list_remove_bot(debuglvl, rules_list) < 0)
				{
					(void)vrprint.error(-1, "Error", "removing bottom node failed (in: query_remove_list).");
					return(-1);
				}

			}
			else
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "query_remove_list: removing normal entry");

				if(d_list_remove_node(debuglvl, rules_list, d_node) < 0)
				{
					(void)vrprint.error(-1, "Error", "removing the node failed (in: query_remove_list).");
					return(-1);
				}

				if(updatenumbers == 1)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "query_remove_list: updatenumbers: %d, %d", place, 0);

					query_update_numbers(place, rules_list, 0);
				}
			}

			// we only remove one rule at a time
			break;
		}
	}

	//fprintf(stdout, "remove_list done\n");
	return(0);
}

/*	query_update_numbers

	action:
		0: decrease
		1: increase
*/
void
query_update_numbers(int place, d_list *rules_list, int action)
{
	struct RuleData_	*query_ptr = NULL;
	d_list_node		*d_node = NULL;
	int			i = 0;

	if(action < 0 || action > 1)
	{
		(void)vrprint.error(-1, "Internal Error", "FIXME action (%d) not 0 or 1.", action);
		return;
	}

	(void)vrprint.debug("Debug", "query_update_numbers: Update higher (or equal) than: %d, action = %d. (list len is %d)", place, action, rules_list->len);

	for(d_node = rules_list->top, i = 1; d_node ; d_node = d_node->next, i++)
	{
		query_ptr = d_node->data;

		if(i > place)
		{
			if(action == 1)
				query_ptr->number++;
		}

		if(i >= place)
		{
			if(action == 0 && query_ptr->number != 0)
				query_ptr->number--;
		}
	}

	return;
}

/*- query_print_list - */
void
query_print_list(const d_list *rules_list)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*query_ptr = NULL;

	for(d_node = rules_list->top; d_node ; d_node = d_node->next)
	{
		query_ptr = d_node->data;

		(void)vrprint.debug("Debug", "%3d, %-8s, %s, %s, %s, %s, %s, %s, status: %d, delta: %d", query_ptr->number, query_ptr->action, query_ptr->service, query_ptr->from, query_ptr->to, query_ptr->who, query_ptr->source, query_ptr->danger, query_ptr->status, query_ptr->number);
	}

	return;
}


void
free_options(const int debuglvl, struct options *opt)
{
	d_list_node		*d_node = NULL;
	struct portdata		*port_ptr = NULL;

	if(!opt)
		return;

	if(opt->RemoteportList.len > 0)
	{
		/*
			free all portranges
		*/
		for(d_node = opt->RemoteportList.top; d_node; d_node = d_node->next)
		{
			port_ptr = d_node->data;
			free(port_ptr);
		}

		d_list_cleanup(debuglvl, &opt->RemoteportList);
	}

	if(opt->ListenportList.len > 0)
	{
		/*
			free all portranges
		*/
		for(d_node = opt->ListenportList.top; d_node; d_node = d_node->next)
		{
			port_ptr = d_node->data;
			free(port_ptr);
		}

		d_list_cleanup(debuglvl, &opt->ListenportList);
	}

	free(opt);
}
