/***************************************************************************
 *   Copyright (C) 2002-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*************************************************************************** 
 * Here we try to analize the actual rule.                                  *
 ***************************************************************************/
#include "vuurmuur.h"


/*	analyse_rule

	Function for gathering the info for creation of the rule
	and for sanity checking the rule.

	Returncodes:
		 0: ok
		-1: error
 */
int
rules_analyze_rule(const int debuglvl, struct RuleData_ *rule_ptr, struct RuleCache_ *create, d_list *ServicesList, d_list *ZonedataList, d_list *InterfacesList)
{
	int	result = 0;
	char	network[MAX_NET_ZONE] = "";

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start **", __FUNCTION__);


	/* safety */
	if(!rule_ptr || !ServicesList || !ZonedataList || !InterfacesList)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* if were on bash mode, alloc mem for the description */
	if(bash_out == 1)
	{
		if(!(create->description = malloc(sizeof(bash_description))))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
			return(-1);
		}
	}
	else
	{
		create->description = NULL;
	}

	/* first the protect rule */
	if(strcasecmp(rule_ptr->action, "protect") == 0)
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug("Debug", "%s: action: %s, who: %s, danger: %s, source: %s", __FUNCTION__, rule_ptr->action, rule_ptr->who, rule_ptr->danger, rule_ptr->source);

		/* description */
		if(bash_out == 1 && create->description != NULL)
		{
			snprintf(create->description, sizeof(bash_description), "rule: action: %s, who: %s, danger: %s, source: %s", rule_ptr->action, rule_ptr->who, rule_ptr->danger, rule_ptr->source);
		}

		/* get who */
		if(strcmp(rule_ptr->who, "") != 0)
		{
			if(rule_ptr->type == PROT_IPTABLES)
			{
				create->who_int = NULL;

				if(!(create->who = search_zonedata(debuglvl, rule_ptr->who, ZonedataList)))
				{
					(void)vrprint.error(-1, "Error", "zone '%s' not found (in: %s).", rule_ptr->who, __FUNCTION__);
					return(-1);
				}
			}
			else if(rule_ptr->type == PROT_PROC_INT)
			{
				create->who = NULL;
				create->who_int = NULL;

				if(!(create->who_int = search_interface(debuglvl, rule_ptr->who, InterfacesList)))
				{
					(void)vrprint.error(-1, "Error", "interface '%s' not found (in: %s).", rule_ptr->who, __FUNCTION__);
					return(-1);
				}
			}
			else
			{
				create->who = NULL;
				(void)vrprint.error(-1, "Error", "don't know what to do with '%s' for rule type '%d' (in: %s).", rule_ptr->who, rule_ptr->type, __FUNCTION__);
				return(-1);
			}
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: calling get_danger_info() for danger...", __FUNCTION__);

		result = get_danger_info(debuglvl, rule_ptr->danger, rule_ptr->source, &create->danger);
		if(result == 0)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: get_danger_info succesfull.", __FUNCTION__);
		}
		else
		{
			(void)vrprint.error(-1, "Error", "getting danger '%s' failed (in: %s).", rule_ptr->danger, __FUNCTION__);
			return(-1);
		}

		/* set the action */
		if(strlcpy(create->action, "protect", sizeof(create->action)) > sizeof(create->action))
		{
			(void)vrprint.error(-1, "Error", "buffer overflow! (in: %s).", __FUNCTION__);
			return(-1);
		}
	}
	/* normal rule */
	else
	{
		/* this is the rule */
		if(debuglvl >= LOW)
			(void)vrprint.debug("Debug", "%s: action: %s, service: %s, from: %s, to: %s", __FUNCTION__, rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);

		/* description */
		if(bash_out == 1 && create->description != NULL)
		{
			snprintf(create->description, sizeof(bash_description), "rule: action: %s, service: %s, from: %s, to: %s", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
		}

		/*
			GATHERING INFO FOR CREATING THE RULE
		*/
		create->active = rule_ptr->active;

		create->from = NULL;
		create->to = NULL;

		/*
			get from if not firewall
		*/
		if(strcmp(rule_ptr->from, "firewall") != 0)
		{
			/* get the pointer to the zonedata in the ZonedataList */
			if(!(create->from = search_zonedata(debuglvl, rule_ptr->from, ZonedataList)))
			{
				(void)vrprint.error(-1, "Error", "'from' zone '%s' not found (in: %s).", rule_ptr->from, __FUNCTION__);
				return(-1);
			}
		}
		else
		{
			/* we get the data later */
			create->from_firewall = 1;
		}

		/* normal network */
		if(strcmp(rule_ptr->to, "firewall") != 0)
		{
			/* get the pointer to the zonedata in the ZonedataList */
			if(!(create->to = search_zonedata(debuglvl, rule_ptr->to, ZonedataList)))
			{
				(void)vrprint.error(-1, "Error", "'to' zone '%s' not found (in: %s).", rule_ptr->to, __FUNCTION__);
				return(-1);
			}
		}
		/* firewall */
		else
		{
			/* first check if we don't have two firewalls */
			if(create->from_firewall)
			{
				(void)vrprint.error(-1, "Error", "'from' and 'to' are both set to firewall (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}

			snprintf(network, sizeof(network), "%s.%s", create->from->network_name, create->from->zone_name);

			if(!(create->to = search_zonedata(debuglvl, network, ZonedataList)))
			{
				(void)vrprint.error(-1, "Error", "'to' zone '%s' not found (in: %s).", network, __FUNCTION__);
				return(-1);
			}

			create->to_firewall = 1;
		}

		/* now get the data for the from-firewall */
		if(create->from_firewall)
		{
			/* first check if we don't have two firewalls */
			if(create->to_firewall)
			{
				(void)vrprint.error(-1, "Error", "'from' and 'to' are both set to firewall (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}
		
			/* get the pointer to the zonedata in the ZonedataList */
			snprintf(network, sizeof(network), "%s.%s", create->to->network_name, create->to->zone_name);
			if(!(create->from = search_zonedata(debuglvl, network, ZonedataList)))
			{
				(void)vrprint.error(-1, "Error", "'from' zone '%s' not found (in: %s).", network, __FUNCTION__);
				return(-1);
			}
		}

		/* get the pointer to the services in the ServicesList */
		if(!(create->service = search_service(debuglvl, rule_ptr->service, ServicesList)))
		{
			(void)vrprint.error(-1, "Error", "service '%s' not found (in: %s).", rule_ptr->service, __FUNCTION__);
			return(-1);
		}

		/* get the rule options */
		if(rule_ptr->opt != NULL)
			create->option = *rule_ptr->opt;

		/* determine which action to take (ACCEPT, DROP, REJECT etc.). */
		if(determine_action(debuglvl, rule_ptr->action, create->action, sizeof(create->action), &create->option) == 0)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: determine_action succesfull, create->action = %s", __FUNCTION__, create->action);
		}
		else
		{
			(void)vrprint.error(-1, "Error", "could not determine action (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* determine which chain to use. */
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: calling determine_chain()...", __FUNCTION__);

		if(determine_chain(debuglvl, rule_ptr, create->chain, sizeof(create->chain), &create->ruletype) == 0)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: determine_chain succesfull, create->chain = %s", __FUNCTION__, create->chain);
		}
		else
		{
			(void)vrprint.error(-1, "Error", "could not determine chain (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* QUEUE-ing can only be in input, output and forward rules */
		if(strcasecmp(create->action, "QUEUE") == 0 && (create->ruletype != RT_INPUT && create->ruletype != RT_OUTPUT && create->ruletype != RT_FORWARD))
		{
			(void)vrprint.error(-1, "Error", "the QUEUE target can only be used in the input, output and forward chains (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* make sure we only porfw to a host */
		if(create->ruletype == RT_PORTFW)
		{
			if(create->from_firewall)
			{
				(void)vrprint.error(-1, "Error", "portforwarding is not allowed from the firewall (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}
			
			if(create->to->type != TYPE_HOST)
			{
				(void)vrprint.error(-1, "Error", "portforwarding is only allowed to a host (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}
		}
	
		/* make sure we only redirect to the firewall */
		if(create->ruletype == RT_REDIRECT)
		{
			if(create->from_firewall)
			{
				(void)vrprint.error(-1, "Error", "redirecting is not allow from the firewall (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}
			
			if(!create->to_firewall)
			{
				(void)vrprint.error(-1, "Error", "redirecting is only possible to the firewall (%s service %s from %s to %s).", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
				return(-1);
			}
		}

	} // end else-protect

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** end **, success", __FUNCTION__);

	return(0);
}


/*	rules_init_list

	Returncodes:
		 0: ok
		-1: error
*/
int
rules_init_list(const int debuglvl, d_list *rules_list, struct rgx_ *reg)
{
	FILE			*fp = NULL;
	int			retval = 0,
				count = 1;
	char			line[MAX_RULE_LENGTH];
	struct RuleData_	*rule_ptr = NULL;
	char			protect_warning_shown = 0;

	
	/* safety */
	if(!rules_list || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/*	setup the list: the cleanup function is set to NULL
		so it's the users responsibility to free memory.
	*/
	if(d_list_setup(debuglvl, rules_list, NULL) < 0)
		return(-1);

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "rules_location: '%s'", conf.rules_location);

	/*
		open the rulesfile
	*/
	if(!(fp = fopen(conf.rules_location, "r")))
	{
		(void)vrprint.error(-1, "Error", "opening rulesfile '%s': %s (in: %s).", conf.rules_location, strerror(errno), __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: opening rulesfile succeded.", __FUNCTION__);

	/*
		run trough the file
	*/
	while(fgets(line, sizeof(line), fp) != NULL)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: strlen(line) = %d", __FUNCTION__, strlen(line));

		/*
			check if the line is a comment
		*/
//TODO what? what? what?
		if((strlen(line) <= 1) || (line[0] == '#'))
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: skipping line because its a comment or its empty.", __FUNCTION__);
		}
		/* no comment */
		else
		{
			/* alloc memory for the rule */
			if(!(rule_ptr = rule_malloc()))
			{
				(void)vrprint.error(-1, "Error", "claiming memory failed: %s (in: %s).", strerror(errno), __FUNCTION__);
				return(-1);
			}

			/* parse the line. We don't really care if it fails, we just ignore it. */
			if(rules_parse_line(debuglvl, line, rule_ptr, reg) < 0)
			{
				(void)vrprint.debug("Debug", "%s: parsing rule failed", __FUNCTION__);
			}
			else
			{
				/* protect rules are no longer supported in the main rules list */
				if(strcasecmp(rule_ptr->action, "protect") == 0)
				{
					if(!protect_warning_shown)
					{
						(void)vrprint.warning("Warning", "please note that the protect rules (e.g. anti-spoof) have been changed. Please recheck your networks and interfaces.");
						protect_warning_shown = 1;
					}
					
					free(rule_ptr);
					rule_ptr = NULL;
				}
				else
				{
					/* append to the rules list */
					if(!(d_list_append(debuglvl, rules_list, rule_ptr)))
					{
						(void)vrprint.error(-1, "Internal Error", "failed to add rule to list (in: %s).", __FUNCTION__);

						free(rule_ptr);
						return(-1);
					}
					else
					{
						/*
							set the rule number
						*/
						rule_ptr->number = count;
						count++;
					}
				}
			}
		}
	}

	(void)vrprint.info("Info", "%d rules loaded.", count-1);

	if(fclose(fp) < 0)
	{
		(void)vrprint.error(-1, "Error", "closing rules file failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		retval = -1;
	}

	return(retval);
}


/*	rules_parse_line

	Returncodes:
		0: ok
		-1: error
*/
int
rules_parse_line(const int debuglvl, const char *line, struct RuleData_ *rule_ptr, struct rgx_ *reg)
{
	int	line_pos = 0,	// position in line
		var_pos=0,	// position in varible
		result=0;
	char	options[MAX_OPTIONS_LENGTH];


	/* safety first */
	if(!line || !rule_ptr || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* this should not happen, but it can't hurt to check, right? */
	if(strlen(line) > MAX_RULE_LENGTH)
	{
		(void)vrprint.error(-1, "Internal Error", "rule is too long (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* see if the rule is active */
	if(line[0] == ';')
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "rule is in-active.");

		rule_ptr->active = 0;

		line_pos = 1; 
	}
	else
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "rule is active.");

		rule_ptr->active = 1;
	}

	/* get the action */
	for(var_pos = 0; line_pos < sizeof(rule_ptr->action)-1 && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
	{
		rule_ptr->action[var_pos] = line[line_pos];
	}
	rule_ptr->action[var_pos] = '\0';

	/*
		now we analyze the action
	*/
	if(strcasecmp(rule_ptr->action, "protect") == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: protect: rule_ptr->action: %s", __FUNCTION__, rule_ptr->action);

		/*
			get the who, or 'against'
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.who) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->who[var_pos] = line[line_pos];
		}
		rule_ptr->who[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: protect: who: '%s'", __FUNCTION__, rule_ptr->who);

		/*
			now check what kind of rule we have
		*/
		if(strcasecmp(rule_ptr->who, "against") == 0)
		{
			/*
				clear who, because we don't use it
			*/
			strcpy(rule_ptr->who, "");

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: rule is a PROT_PROC_SYS", __FUNCTION__);

			/*
				okay, now lets see what kind of danger we are talking about
			*/
			for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.danger) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
			{
				rule_ptr->danger[var_pos] = line[line_pos];
			}
			rule_ptr->danger[var_pos] = '\0';

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: protect: danger: '%s'", __FUNCTION__, rule_ptr->danger);
		}
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: now we know who (%s), let get the danger (but first check who).", __FUNCTION__, rule_ptr->who);

			/*
				validate the who-zone
			*/
			if(validate_zonename(debuglvl, rule_ptr->who, 1, NULL, NULL, NULL, reg->zonename) != 0)
			{
				(void)vrprint.error(-1, "Error", "invalid zonename: '%s' (in: %s).", rule_ptr->who, __FUNCTION__);
				return(-1);
			}

			/*
				get the keyword 'against'
			*/
			for(line_pos++, var_pos = 0; var_pos < strlen("against") && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
			{
				rule_ptr->danger[var_pos] = line[line_pos];
			}
			rule_ptr->danger[var_pos] = '\0';

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: protect: keyword against: '%s'", __FUNCTION__, rule_ptr->danger);

			/*
				if 'against' is missing, the rule is malformed, so we bail out screaming & kicking
			*/
			if(strcasecmp(rule_ptr->danger, "against") != 0)
			{
				(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'against' is missing: %s (in: %s).", line, __FUNCTION__);
				return(-1);
			}

			/*
				okay, now lets see what kind of danger we are talking about
			*/
			for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.danger) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
			{
				rule_ptr->danger[var_pos] = line[line_pos];
			}
			rule_ptr->danger[var_pos] = '\0';

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: protect: danger: '%s'", __FUNCTION__, rule_ptr->danger);

			/*
				now determine if the danger is 'spoofing'
			*/
			if(strcasecmp(rule_ptr->danger, "spoofing") == 0)
			{
				/*
					get the 'from'
				*/
				for(line_pos++, var_pos = 0; var_pos < strlen("from") && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
				{
					rule_ptr->source[var_pos] = line[line_pos];
				}
				rule_ptr->source[var_pos] = '\0';

				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: protect: keyword from: '%s'", __FUNCTION__, rule_ptr->source);

				/*
					if 'from' is missing, the rule is malformed, so we bail out screaming & kicking
				*/
				if(strcasecmp(rule_ptr->source, "from") != 0)
				{
					(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'from' is missing: %s (in: %s).", line, __FUNCTION__);
					return(-1);
				}

				/*
					get the source
				*/
				for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.source) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
				{
					rule_ptr->source[var_pos] = line[line_pos];
				}
				rule_ptr->source[var_pos] = '\0';

				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: protect: source: '%s'", __FUNCTION__, rule_ptr->source);
			}

			/*
				if don't use rule_ptr->source, clear it (just to be sure).
			*/
			else
			{
				strcpy(rule_ptr->source, "");
			}
		}
	}
	else
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: normal rule: action: %s", __FUNCTION__, rule_ptr->action);

		/*
			analize the action
		*/
		result = actiontoi(rule_ptr->action);
		if(result <= AT_ERROR || result >= AT_TOO_BIG)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, unknown action type: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			first check for the keyword 'service'
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.service) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->service[var_pos] = line[line_pos];
		}
		rule_ptr->service[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: keyword service: '%s'.", __FUNCTION__, rule_ptr->service);

		if(strcasecmp(rule_ptr->service, "service") != 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'service' is missing: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			get the service itself
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.service) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->service[var_pos] = line[line_pos];
		}
		rule_ptr->service[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: service: '%s'.", __FUNCTION__, rule_ptr->service);

		if(strcmp(rule_ptr->service, "from") == 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'service' found, but has no value: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			validate the service name
		*/
		if(validate_servicename(debuglvl, rule_ptr->service, reg->servicename) != 0)
		{
			(void)vrprint.error(-1, "Error", "invalid servicename: '%s' (in: %s).", rule_ptr->service, __FUNCTION__);
			return(-1);
		}

		/*
			first check for the keyword 'from'
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.from) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->from[var_pos] = line[line_pos];
		}
		rule_ptr->from[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: keyword from: '%s'.", __FUNCTION__, rule_ptr->from);

		if(strcasecmp(rule_ptr->from, "from") != 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'from' is missing: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			get the from itself
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.from) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->from[var_pos] = line[line_pos];
		}
		rule_ptr->from[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: from: '%s'.", __FUNCTION__, rule_ptr->from);

		/*
			see if the from is actually the to keyword
		*/
		if(strcmp(rule_ptr->from, "to") == 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'from' found, but has no value: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			now validate the from-zone
		*/
		if(validate_zonename(debuglvl, rule_ptr->from, 1, NULL, NULL, NULL, reg->zonename) != 0)
		{
			(void)vrprint.error(-1, "Error", "invalid from-zonename: '%s' (in: %s).", rule_ptr->from, __FUNCTION__);
			return(-1);
		}

		/*
			first check for the keyword 'to'
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.to) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->to[var_pos] = line[line_pos];
		}
		rule_ptr->to[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: keyword to: '%s'.", __FUNCTION__, rule_ptr->to);

		if(strcasecmp(rule_ptr->to, "to") != 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'to' is missing: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			get to
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(RuleData.to) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			rule_ptr->to[var_pos] = line[line_pos];
		}
		rule_ptr->to[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: to: '%s'.", __FUNCTION__, rule_ptr->to);

		/*
			see that to is not the keyword options
		*/
		if(strcmp(rule_ptr->to, "options") == 0)
		{
			(void)vrprint.error(-1, "Error", "bad rule syntax, keyword 'to' found, but has no value: %s (in: %s).", line, __FUNCTION__);
			return(-1);
		}

		/*
			now validate the to-zone
		*/
		if(validate_zonename(debuglvl, rule_ptr->to, 1, NULL, NULL, NULL, reg->zonename) != 0)
		{
			(void)vrprint.error(-1, "Error", "invalid zonename: '%s' (in: %s).", rule_ptr->to, __FUNCTION__);
			return(-1);
		}

		/*
			first check for the keyword 'options'
		*/
		for(line_pos++, var_pos = 0; var_pos < sizeof(options) && line[line_pos] != ' ' && line[line_pos] != '\0' && line[line_pos] != '\n'; line_pos++, var_pos++)
		{
			options[var_pos] = line[line_pos];
		}
		options[var_pos] = '\0';

		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: keyword options: '%s'.", __FUNCTION__, options);

		/*
			if this keyword exists we have options
		*/
		if(strcasecmp(options, "options") == 0)
		{
			/*
				get options: NOTE: whitespaces are allowed!
			*/
			for(line_pos++, var_pos = 0; var_pos < sizeof(options) && line[line_pos] != '\0' && line[line_pos] != '\n' && line_pos < MAX_RULE_LENGTH; line_pos++, var_pos++)
			{
				options[var_pos] = line[line_pos];
			}
			options[var_pos] = '\0';

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: options: '%s'.", __FUNCTION__, options);

			/* alloc options struct */
			if(!(rule_ptr->opt = ruleoption_malloc(debuglvl)))
			{
				(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
				return(-1);
			}

			/*
				now split them up
			*/
			if(rules_read_options(debuglvl, options, rule_ptr->opt) < 0)
			{
				(void)vrprint.error(-1, "Error", "parsing rule options failed for: %s.", line);

				free(rule_ptr->opt);
				rule_ptr->opt = NULL;

				return(-1);
			}
		}
		/*
			no options
		*/
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: rule has no options.", __FUNCTION__);
			
			strcpy(options, "");

			rule_ptr->opt = NULL;
		}
	}

	return(0);
}


/*
	TODO: mask! it should we only read/write for owner root, and nothing to the others
*/
int
rules_write_file(const int debuglvl, const char *rulesfile_location, d_list *rules_list)
{
	FILE			*fp = NULL;
	int			retval = 0;
	d_list_node		*d_node = NULL;
	char			line[512],
				*option_ptr = NULL;
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!rulesfile_location || !rules_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* open the rulesfile */
	if(!(fp = rules_file_open(rulesfile_location, "w+", 0)))
	{
		(void)vrprint.error(-1, "Error", "opening rulesfile '%s' failed: %s (in: %s).", rulesfile_location, strerror(errno), __FUNCTION__);
		return(-1);
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "%s: number of rules %d.", __FUNCTION__, rules_list->len);

	/* starting banner */
	fprintf(fp, "# Vuurmuur configfile, do not place comments in it, for they will be overwritten\n");

	/* loop trough the list */
	for(d_node = rules_list->top; d_node ; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* protect rule */
		if(strcasecmp(rule_ptr->action, "protect") == 0)
		{
			if(strcmp(rule_ptr->who, "") == 0)
			{
				/* proc rule global */
				snprintf(line, sizeof(line), "protect against %s\n", rule_ptr->danger);
			}
			else if(strcmp(rule_ptr->source, "") == 0)
			{
				/* proc rule int */
				snprintf(line, sizeof(line), "protect %s against %s\n", rule_ptr->who, rule_ptr->danger);
			}
			else
			{
				/* danger + source */
				snprintf(line, sizeof(line), "protect %s against %s from %s\n", rule_ptr->who, rule_ptr->danger, rule_ptr->source);
			}
		}
		else
		{
			/* normal rule */
			if(rule_ptr->active)
				snprintf(line, sizeof(line), "%s service %s from %s to %s", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);
			else
				snprintf(line, sizeof(line), ";%s service %s from %s to %s", rule_ptr->action, rule_ptr->service, rule_ptr->from, rule_ptr->to);

			if(!(option_ptr = rules_assemble_options_string(debuglvl, rule_ptr->opt, rule_ptr->action)))
				; /* no options, or error */
			else
			{
				strlcat(line, " ", sizeof(line));
				strlcat(line, option_ptr, sizeof(line));
				free(option_ptr);
			}

			strlcat(line, "\n", sizeof(line));
		}
		
		/* now print the rule to the file */
		fprintf(fp, "%s", line);
	}

	/* print the end-of-file so we know all went fine */
	fprintf(fp, "# end of file\n");
	fflush(fp);

	/* close the rulesfile */
	retval = rules_file_close(fp, rulesfile_location);
	return(retval);
}


/*	cleanup_ruleslist

	O(n) function: with n is the number of rules

	Returncodes:
		 0: ok
		-1: error
*/
int
rules_cleanup_list(const int debuglvl, d_list *rules_list)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!rules_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/*
		loop trough the list to remove the options
	*/
	for(d_node = rules_list->top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/*
			free the options. If there are none the
			'free_options' functions will take care
			of that.
		*/
		free_options(debuglvl, rule_ptr->opt);
		rule_ptr->opt = NULL;

		free(rule_ptr);
		rule_ptr = NULL;
	}

	/*
		cleanup the list
	*/
	if(d_list_cleanup(debuglvl, rules_list) < 0)
		return(-1);

	return(0);
}


/*	rules_insert_list

	inserts a rule into the ruleslist at position 'place'.
	
	Returncodes:
		 0: ok
		-1: error
*/
int
rules_insert_list(const int debuglvl, int place, d_list *rules_list, struct RuleData_ *rule_ptr)
{
	struct RuleData_	*listrule_ptr = NULL;
	int			retval = 0;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!rules_list || !rule_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "insert at: %d. (list len is %d), number: %d, action: %s, service: %s, from: %s, to: %s, danger: %s, who: %s, source: %s.",
										place, 
										rules_list->len,
										rule_ptr->number,
										rule_ptr->action,
										rule_ptr->service,
										rule_ptr->from,
										rule_ptr->to,
										rule_ptr->danger,
										rule_ptr->who,
										rule_ptr->source);


	/* if we insert into empty list, we always insert at the top */
	if(rules_list->len == 0)
	{
		(void)vrprint.debug(__FUNCTION__, "inserting into an empty list. Setting place to 1 (place was: %d).", place);
		place = 1;
	}

	/* handle inserting at the bottom of the list */
	if(place > rules_list->len)
	{
		(void)vrprint.debug(__FUNCTION__, "place > rules_list->len (%d, %d). Setting place to %d.", place, rules_list->len, rules_list->len + 1);
		place = rules_list->len + 1;
	}


	/* handle insertion at the top of the list */
	if(place == 1)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "place to insert: top");

		if(!(d_list_prepend(debuglvl, rules_list, rule_ptr)))
		{
			(void)vrprint.error(-1, "Internal Error", "inserting the data to the top of list failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "d_list_prepend succes, now update numbers (place: %d)", place);

		query_update_numbers(place, rules_list, 1);

		/* set number to 1 */
		rule_ptr->number = 1;

		/* we're done */
		return(0);
	}


	/*	now loop trough the list

		count: counts the number of rules we already processed
	*/
	for(d_node = rules_list->top; d_node ; d_node = d_node->next)
	{
		if(!(listrule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		(void)vrprint.debug(__FUNCTION__, "entry: %s %s %s %s %s", listrule_ptr->action, listrule_ptr->service, listrule_ptr->danger, listrule_ptr->who, listrule_ptr->source);

		if(listrule_ptr->number == place - 1)
		{
			(void)vrprint.debug(__FUNCTION__, "place to insert: place: %d, %s %s %s %s %s", place, listrule_ptr->action, listrule_ptr->service, listrule_ptr->danger, listrule_ptr->who, listrule_ptr->source);

			if(!(d_list_insert_after(debuglvl, rules_list, d_node, rule_ptr)))
			{
				(void)vrprint.error(-1, "Internal Error", "inserting the data into the list failed.");
				return(-1);
			}

			/* update numbers after count */
			(void)vrprint.debug(__FUNCTION__, "d_list_insert_after succes, now update numbers (place: %d)", place);
			query_update_numbers(place - 1, rules_list, 1);

			/* set the number */
			(void)vrprint.debug(__FUNCTION__, "d_list_insert_after succes, now set QueryData_ptr->number to place: %d.", place);
			rule_ptr->number = place;

			/* we're done now */
			return(0);
		}
		/* we're not just there yet */
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "not the right place: %d, %s %s %s %s %s", place, listrule_ptr->action, listrule_ptr->service, listrule_ptr->danger, listrule_ptr->who, listrule_ptr->source);
		}
	}

	return(retval);
}


char *
rules_assemble_options_string(const int debuglvl, struct options *opt, const char *action)
{
	char	*option_ptr = NULL,
		options[MAX_OPTIONS_LENGTH] = "",
		*ports_ptr = NULL;
	char	redirect_port[6] = "",
		loglimit_string[6] = "";

	/* safety - this is not an error! */
	if(!opt || !action)
		return(NULL);

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "action: '%s'.", action);

	/* init */
	strlcpy(options, "options ", sizeof(options));

	/* the log options are also valid for the action LOG */
	if(opt->rule_log == 1 || strncasecmp(action, "LOG", 3) == 0)
	{
		if(opt->rule_log == 1)
			strlcat(options, "log,", sizeof(options));

		/* loglimit */
		if(opt->loglimit > 0)
		{
			snprintf(loglimit_string, sizeof(loglimit_string), "%d", opt->loglimit);

			strlcat(options, "loglimit=\"", sizeof(options));
			strlcat(options, loglimit_string, sizeof(options));
			strlcat(options, "\",", sizeof(options));
		}

		/* log prefix */
		if(opt->rule_logprefix == 1 && strcmp(opt->logprefix, "") != 0)
		{
			strlcat(options, "logprefix=\"", sizeof(options));
			strlcat(options, opt->logprefix, sizeof(options));
			strlcat(options, "\",", sizeof(options));
		}
	}

	if(strcasecmp(action, "portfw") == 0)
	{
		if(opt->listenport == 1)
		{
			if(!(ports_ptr = list_to_portopts(debuglvl, &opt->ListenportList, "listenport")))
				;// no nothing?
			else
			{
				strlcat(options, ports_ptr, sizeof(options));
				free(ports_ptr);
				strlcat(options, ",", sizeof(options));
			}
		}
		if(opt->remoteport == 1)
		{
			if(!(ports_ptr = list_to_portopts(debuglvl, &opt->RemoteportList, "remoteport")))
				;// no nothing?
			else
			{
				strlcat(options, ports_ptr, sizeof(options));
				free(ports_ptr);
				strlcat(options, ",", sizeof(options));
			}
		}
	}

	if(strcasecmp(action, "reject") == 0)
	{
		if(opt->reject_option == 1)
		{
			strlcat(options, "rejecttype=\"", sizeof(options));
			strlcat(options, opt->reject_type, sizeof(options));
			strlcat(options, "\",", sizeof(options));
		}
	}

	if(strcasecmp(action, "redirect") == 0)
	{
		if(opt->redirectport > 0 && opt->redirectport <= 65535)
		{
			snprintf(redirect_port, sizeof(redirect_port), "%d", opt->redirectport);

			strlcat(options, "redirectport=\"", sizeof(options));
			strlcat(options, redirect_port, sizeof(options));
			strlcat(options, "\",", sizeof(options));
		}
	}

	if(strcasecmp(action, "queue") == 0)
	{
		if(opt->markiptstate == 1)
		{
			strlcat(options, "markiptstate,", sizeof(options));
		}
	}
	
	/* comment */
	if(opt->rule_comment == 1 && strcmp(opt->comment, "") != 0)
	{
		strlcat(options, "comment=\"", sizeof(options));
		strlcat(options, opt->comment, sizeof(options));
		strlcat(options, "\",", sizeof(options));
	}

	/* terminate the string */
	options[strlen(options)-1] = '\0';

	/* check if we did anything */
	if(strcmp(options, "options") != 0)
	{
		if(!(option_ptr = malloc(strlen(options)+1)))
		{
			(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
			return(NULL);
		}
		else
		{
			strlcpy(option_ptr, options, strlen(options)+1);
			
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "%s: '%s'.", __FUNCTION__, option_ptr);
		}
	}

	return(option_ptr);
}


/*
	Returncodes:
		-1: error
		 0: no change
		 1: change
*/
int
rules_compare_options(const int debuglvl, struct options *old_opt, struct options *new_opt, char *action)
{
	char	*old_str = NULL,
		*new_str = NULL;
	int	retval = 0;

	//both NULL: no change
	if(!old_opt && !new_opt)
	{
		(void)vrprint.debug("Debug", "compare_options: options not changed (both NULL)");
		return(0);
	}

	// if they are not the same: change
	if((!old_opt && new_opt) || (old_opt && !new_opt))
	{
		(void)vrprint.debug("Debug", "compare_options: options changed! (one NULL, other not)");
		return(1);
	}

	// from here on, we are sure we have two options
	if(!(old_str = rules_assemble_options_string(debuglvl, old_opt, action)))
		return(-1);

	if(!(new_str = rules_assemble_options_string(debuglvl, new_opt, action)))
	{
		free(old_str);
		return(-1);
	}

	if(strcmp(old_str, new_str) == 0)
		retval = 0;
	else
	{
		(void)vrprint.debug("Debug", "compare_options: options changed! (str compare)");
		retval = 1;
	}

	// free the mem
	free(old_str);
	free(new_str);

	return(retval);
}


/*
	TODO: compare active
*/
void *
search_rule(const int debuglvl, d_list *rules_list, struct RuleData_ *searchrule_ptr)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*listrule_ptr = NULL;

	/* safety */
	if(!rules_list || !searchrule_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(NULL);
	}

	for(d_node = rules_list->top; d_node ; d_node = d_node->next)
	{
		if(!(listrule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(NULL);
		}

		if(strcmp(listrule_ptr->action, searchrule_ptr->action) == 0)
		{
			/* protect rule */
			if(strcasecmp(searchrule_ptr->action, "protect") == 0)
			{
				/* compare who */
				if(strcmp(listrule_ptr->who, searchrule_ptr->who) == 0)
				{
					/* compare source */
					if(strcmp(listrule_ptr->source, searchrule_ptr->source) == 0)
					{
						/* compare the danger */
						if(strcmp(listrule_ptr->danger, searchrule_ptr->danger) == 0)
						{
							return(listrule_ptr);
						}
					}
				}
			}

			/* normal rule */
			else
			{
				/* first compare the service-names */
				if(strcmp(listrule_ptr->service, searchrule_ptr->service) == 0)
				{
					/* comparing the 'from'-name */
					if(strcmp(listrule_ptr->from, searchrule_ptr->from) == 0)
					{
						/* comparing the 'to'-name */
						if(strcmp(listrule_ptr->to, searchrule_ptr->to) == 0)
						{
							/* comparing the rule options */
							if(rules_compare_options(debuglvl, listrule_ptr->opt, searchrule_ptr->opt, listrule_ptr->action) == 0)
							{
								return(listrule_ptr);
							}
						}
					}
				}
			}
		}
	}

	return(NULL);
}


/* rules_read_options

    Call with the string with options and ouputs the option structure.

    TODO: this needs to be totally redesigned

*/
int
rules_read_options(const int debuglvl, char *optstr, struct options *op)
{
	int	retval = 0,
		x = 0,
		cur_pos = 0,
		o = 0,
		p = 0,
		trema = 0;
	char	curopt[512] = "",
		portstring[512] = "";

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "** start **");

	/* safety */
	if(!optstr || !op)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* TODO: is this really needed? */
	op->rule_log = 0,
	op->remoteport = 0,
	op->listenport = 0;
	op->rule_logprefix = 0,
	op->redirectport = 0,
	op->reject_option = 0;
	op->ListenportList.len = 0;
	op->RemoteportList.len = 0;

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "options: '%s', strlen(optstr): %d", optstr, strlen(optstr));

	/* check if we even got a string to disassemble */
	if(strlen(optstr) == 0)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug(__FUNCTION__, "no options.");

		return(0);
	}


	while((strlen(optstr) >= x))
	{
		curopt[cur_pos] = optstr[x];
		cur_pos++;

		/* between the trema's (") don't use the comma as a separator. */
		if((optstr[x] == '"') && (trema == 1))
		{
			trema = 2;
		}
		if((optstr[x] == '"') && (trema == 0))
		{
			trema = 1;
		}

		if(((optstr[x] == ',') && ((trema == 0) || (trema == 2))) || (optstr[x] == '\0'))
		{
			curopt[cur_pos - 1] = '\0';
			cur_pos = 0;
		}
		x++;

		/* reset trema, so we can have more trema pairs. */
		if(trema == 2)
			trema = 0;

		/* we are done */
		if(cur_pos == 0)
		{
			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNCTION__, "curopt: '%s'.", curopt);

			/* error message for a missing trema */
			if(trema == 1)
			{
				(void)vrprint.error(-1, "Error", "unbalanced \" in rule (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}


			/*
				start parsing the options
			*/

			/* log - log the rule? */
			if(strcmp(curopt, "log") == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "logging enabled.");

				op->rule_log = 1;
			}
			/* loglimit */
			else if(strncmp(curopt, "loglimit", strlen("loglimit")) == 0)
			{
				for(p = 0, o = strlen("loglimit") + 1; o < strlen(curopt); o++)
				{
					if(curopt[o] != '\"')
					{
						portstring[p] = curopt[o];
						p++;
					}
				}
				portstring[p] = '\0';

				op->loglimit = atoi(portstring);
				op->logburst = op->loglimit * 2;

				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "loglimit: %d, logburst %d.", op->loglimit, op->logburst);
			}
			/* mark the iptablesstate? */
			else if(strcmp(curopt, "markiptstate") == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "marking iptstate enabled.");

				op->markiptstate = 1;
			}
			/* remoteport - for portforwarding */
			else if(strncmp(curopt, "remoteport", strlen("remoteport")) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "remoteport specified.");

				/* copy the string containing the ports */
				for(p = 0, o = strlen("remoteport") + 1; o <= strlen(curopt); o++, p++)
				{
					portstring[p] = curopt[o];
				}

				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "remoteport string: '%s'.", portstring);

				if(portopts_to_list(debuglvl, portstring, &op->RemoteportList) < 0)
				{
					(void)vrprint.error(-1, "Error", "parsing remoteport option failed. Please check the syntax of the rule.");
					return(-1);
				}

				op->remoteport = 1;
			}
			/* listenport - for portforwarding */
			else if(strncmp(curopt, "listenport", strlen("listenport")) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "listenport specified.");

				/* copy the string containing the ports */
				for(p = 0, o = strlen("listenport") + 1; o <= strlen(curopt); o++, p++)
				{
					portstring[p] = curopt[o];
				}

				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "listenport string: '%s'.", portstring);

				if(portopts_to_list(debuglvl, portstring, &op->ListenportList) < 0)
				{
					(void)vrprint.error(-1, "Error", "parsing listenport option failed. Please check the syntax of the rule.");
					return(-1);
				}

				op->listenport = 1;
			}
			/* rule comment */
			else if(strncmp(curopt, "comment", strlen("comment")) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "comment.");

				for(p = 0, o = strlen("comment") + 2; o < strlen(curopt) - 1; o++, p++)
				{
					op->comment[p] = curopt[o];
				}
				op->comment[p] = '\0';
				op->rule_comment = 1;
			}
			/* logprefix, max 29 characters long. */
			else if(strncmp(curopt, "logprefix", strlen("logprefix")) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "logprefix.");

				for(p = 0, o = strlen("logprefix") + 2; p < 12 && o < strlen(curopt) - 1; o++, p++)
				{
					op->logprefix[p] = curopt[o];
				}
				op->logprefix[p] = '\0';

				if(strlen(op->logprefix) > 14)
				{
//TODO: not disable, but truncate */
					(void)vrprint.warning("Warning", "logprefix is too long. Maximum length is 14 characters.");
					op->rule_logprefix = 0;
					op->logprefix[0] = '\0';
				}
				else
				{
					op->rule_logprefix=1;
				}
			}
			/* redirectport */
			else if(strncmp(curopt, "redirectport", strlen("redirectport")) == 0)
			{
				for(p = 0, o = strlen("redirectport") + 1; o < strlen(curopt); o++)
				{
					if(curopt[o] != '\"')
					{
						portstring[p] = curopt[o];
						p++;
					}
				}
				portstring[p] = '\0';

				op->redirectport = atoi(portstring);
				if(op->redirectport <= 0 || op->redirectport > 65535)
				{
					(void)vrprint.error(-1, "Error", "redirectport must be 1-65535.");
					op->redirectport = 0;
					return(-1);
				}

				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "read_options: redirectport: %d, %s", op->redirectport, portstring);
			}
			/* reject type */
			else if(strncmp(curopt, "rejecttype", strlen("rejecttype")) == 0)
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug(__FUNCTION__, "rejecttype.");

				for(p = 0, o = strlen("rejecttype") + 1; o < strlen(curopt) && o < 23 + strlen("rejecttype") + 1; o++)
				{ /* 23 is from the length of the string */

					if(curopt[o] != '\"')
					{
						op->reject_type[p] = curopt[o];
						p++;
					}
				}
				op->reject_type[p] = '\0';
				op->reject_option = 1;

				/* check if the option is valid. */
				if(	strcmp(op->reject_type, "icmp-net-unreachable") == 0);
				else if(strcmp(op->reject_type, "icmp-host-unreachable") == 0);
				else if(strcmp(op->reject_type, "icmp-proto-unreachable") == 0);
				else if(strcmp(op->reject_type, "icmp-port-unreachable") == 0);
				else if(strcmp(op->reject_type, "icmp-net-prohibited") == 0);
				else if(strcmp(op->reject_type, "icmp-host-prohibited") == 0);
				else if(strcmp(op->reject_type, "tcp-reset") == 0);
				else
				{
					(void)vrprint.error(-1, "Error", "%s is not a valid reject-type.", op->reject_type);

					op->reject_option = 0;
					return(-1);
				}
			}
			/* unknown option */
			else
			{
				(void)vrprint.error(-1, "Error", "unknown option '%s'.", curopt);
				return(-1);
			}
		}
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "** end **, return = %d", retval);

	return(retval);
}


static int
blocklist_add_ip_to_list(const int debuglvl, d_list *blocklist, char *ip)
{
	size_t	len = 0;
	char	*ipaddress = NULL;

	/* safety */
	if(!blocklist || !ip)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the length */
	len = strlen(ip);
	if(len <= 0 || len > 15)
	{
		(void)vrprint.error(-1, "Internal Error", "weird ipaddress size %d (in: %s:%d).", len, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* alloc the mem */
	if(!(ipaddress = malloc(len + 1)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/*  copy addr */
	strlcpy(ipaddress, ip, len + 1);

	/* append to list */
	if(d_list_append(debuglvl, blocklist, ipaddress) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "appending into the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


static int
blocklist_add_string_to_list(const int debuglvl, d_list *blocklist, char *str)
{
	size_t	len = 0;
	char	*string = NULL;

	/* safety */
	if(!blocklist || !str)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* get the length */
	len = strlen(str);
	if(len <= 0)
	{
		(void)vrprint.error(-1, "Internal Error", "weird string size %d (in: %s:%d).", len, __FUNCTION__, __LINE__);
		return(-1);
	}

	/* alloc the mem */
	if(!(string = malloc(len + 1)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/*  copy addr */
	strlcpy(string, str, len + 1);

	/* append to list */
	if(d_list_append(debuglvl, blocklist, string) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "appending into the list failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


/*	blocklist_read_file

	if 'load_ips' is set to 1 the ipaddresses of the hosts and the groups will
	be loaded, otherwise just the name in the file
*/
int
blocklist_read_file(const int debuglvl, d_list *zones_list, d_list *block_list, char load_ips)
{
	FILE		*fp = NULL;
	char		line[256];
	size_t		len = 0;
	ZoneData	*zone_ptr = NULL,
			*member_ptr = NULL;
	d_list_node	*d_node = NULL;


	/* safety */
	if(!zones_list || !block_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* setup the blocklist */
	d_list_setup(debuglvl, block_list, free);


	/* check the file: it reports error's by itself */
	if(stat_ok(debuglvl, conf.blocklist_location))
	{
		/* open the blocklist-file */
		if(!(fp = fopen(conf.blocklist_location, "r")))
		{
			(void)vrprint.error(-1, "Error", "opening blockfile '%s' failed: %s (in: %s:%d).", conf.blocklist_location, strerror(errno), __FUNCTION__, __LINE__);
			return(-1);
		}
	}
	else
	{
		/* no error msg, stat_ok takes care of that */
		return(-1);
	}


	/* read the file */
	while(fgets(line, sizeof(line), fp) != NULL)
	{
		len = strlen(line);

		if(len > 0 && line[0] != '#')
		{
			/* cut of the newline */
			if(line[len - 1] == '\n')
				line[len - 1] = '\0';

			/* call check_ipv4address with the quiet flag */
			if(check_ipv4address(debuglvl, NULL, NULL, line, 1) != 1)
			{
				/* search for the name in the zones list */
				if((zone_ptr = search_zonedata(debuglvl, line, zones_list)))
				{
					if(zone_ptr->type != TYPE_HOST && zone_ptr->type != TYPE_GROUP)
					{
						if(zone_ptr->type == TYPE_NETWORK)
						{
							(void)vrprint.warning("Warning", "you can only add an ipaddress, host or group to the blocklist. '%s' is a network.", zone_ptr->name);
						}
						else
						{
							(void)vrprint.warning("Warning", "you can only add an ipaddress, host or group to the blocklist. '%s' is a zone.", zone_ptr->name);
						}
					}
					else
					{
						if(!zone_ptr->active)
						{
							if(!load_ips)
							{
								/* add the string */
								if(blocklist_add_string_to_list(debuglvl, block_list, line) < 0)
								{
									(void)vrprint.error(-1, "Internal Error", "adding string to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
									return(-1);
								}
							}
							else
							{
								(void)vrprint.warning("Warning", "host/group '%s' is not active, so not adding to blocklist.", zone_ptr->name);
							}
						}
						else
						{
							if(zone_ptr->type == TYPE_HOST)
							{
								if(!load_ips)
								{
									/* add the string */
									if(blocklist_add_string_to_list(debuglvl, block_list, line) < 0)
									{
										(void)vrprint.error(-1, "Internal Error", "adding string to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
								}
								else
								{
									/* add the hosts ipaddress */
									if(blocklist_add_ip_to_list(debuglvl, block_list, zone_ptr->ipv4.ipaddress) < 0)
									{
										(void)vrprint.error(-1, "Internal Error", "adding ipaddress to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
								}
							}

							if(zone_ptr->type == TYPE_GROUP)
							{
								if(!load_ips)
								{
									/* add the string */
									if(blocklist_add_string_to_list(debuglvl, block_list, line) < 0)
									{
										(void)vrprint.error(-1, "Internal Error", "adding string to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
								}
								else
								{
									for(d_node = zone_ptr->GroupList.top; d_node; d_node = d_node->next)
									{
										if(!(member_ptr = d_node->data))
										{
											(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
											return(-1);
										}

										if(!member_ptr->active)
										{
											(void)vrprint.warning("Warning", "groupmember '%s' from group '%s' is not active, so not adding to blocklist.", member_ptr->name, zone_ptr->name);
										}
										else
										{
											/* add the groupmembers ipaddress */
											if(blocklist_add_ip_to_list(debuglvl, block_list, member_ptr->ipv4.ipaddress) < 0)
											{
												(void)vrprint.error(-1, "Internal Error", "adding ipaddress to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
												return(-1);
											}
										}
									}
								}
							}
						}
					}
				}
				else
				{
					(void)vrprint.warning("Warning", "'%s' is neither a (valid) ipaddress, host or group. Not adding to blocklist.", line);
				}
			}
			else
			{
				if(!load_ips)
				{
					/* add the string */
					if(blocklist_add_string_to_list(debuglvl, block_list, line) < 0)
					{
						(void)vrprint.error(-1, "Internal Error", "adding string to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
						return(-1);
					}
				}
				else
				{
					/* valid ip, so add to the block list */
					if(blocklist_add_ip_to_list(debuglvl, block_list, line) < 0)
					{
						(void)vrprint.error(-1, "Internal Error", "adding ipaddress to blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
						return(-1);
					}
				}
			}
		}
	}

	if(fclose(fp) < 0)
	{
		(void)vrprint.error(-1, "Error", "closing blockfile failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	(void)vrprint.info("Info", "added %d items to the blocklist.", block_list->len);
	return(0);
}


int
blocklist_write_file(const int debuglvl, d_list *block_list)
{
	d_list_node	*d_node   = NULL;
	char		*itemname = NULL;
	int		retval = 0;
	FILE		*fp = NULL;


	/* safety */
	if(!block_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(!conf.use_blocklist)
		return(0);

	/* open the blockfile */
	if(!(fp = fopen(conf.blocklist_location, "w+")))
	{
		(void)vrprint.error(-1, "Error", "opening blocklistfile '%s' failed: %s (in: %s).", conf.blocklist_location, strerror(errno), __FUNCTION__);
		return(-1);
	}

	fprintf(fp, "# BlockList for Vuurmuur\n");
	fprintf(fp, "#\n");
	fprintf(fp, "# put a list ipaddresses, hosts and groups to be blocked,\n");
	fprintf(fp, "# one per line.\n");
	
	for(d_node = block_list->top; d_node; d_node = d_node->next)
	{
		if(!(itemname = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		fprintf(fp, "%s\n", itemname);
	}

	/* print the end-of-file so we know all went fine */
	fprintf(fp, "# end of file\n");
	fflush(fp);

	/* close the rulesfile */
	retval = fclose(fp);
	return(retval);
}


/*	actiontoi

	Converts the action into an int as defined by enum action_types.
*/
int
actiontoi(const char *action)
{
	if(!action)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(AT_ERROR);
	}

	if(strcasecmp(action, "accept") == 0)
		return(AT_ACCEPT);
	else if(strcasecmp(action, "drop") == 0)
		return(AT_DROP);
	else if(strcasecmp(action, "reject") == 0)
		return(AT_REJECT);
	else if(strcasecmp(action, "log") == 0)
		return(AT_LOG);
	else if(strcasecmp(action, "portfw") == 0)
		return(AT_PORTFW);
	else if(strcasecmp(action, "redirect") == 0)
		return(AT_REDIRECT);
	else if(strcasecmp(action, "snat") == 0)
		return(AT_SNAT);
	else if(strcasecmp(action, "masq") == 0)
		return(AT_MASQ);
	else if(strcasecmp(action, "queue") == 0)
		return(AT_QUEUE);
	else
	{
		(void)vrprint.error(-1, "Error", "unknown action '%s' (in: %s:%d).", action, __FUNCTION__, __LINE__);
		return(AT_ERROR);
	}
}
