/***************************************************************************
 *   Copyright (C) 2002-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "main.h"
 

/*	apply changes

	This function checks all data in memory for changes and applies the changes to the
	rules in memory.

	Returncodes:
		 0: succes, changes applied
		 1: succes, no changes seen //defuct
		-1: error
*/
int
apply_changes(const int debuglvl,	d_list *services_list,
					d_list *zones_list,
					d_list *interfaces_list,
					d_list *rules_list,
					d_list *block_list,
					struct rgx_ *reg)
{
	int			retval=0,		// start at no changes
				result=0;

	d_list_node		*d_node = NULL;

	struct RuleData_	*rule_ptr = NULL;

	/*	these are counters that count how many rules there are in each
		chain above the current rule. This way we always know where
		to insert a new rule.
	*/
	int			ipt_i = 1,		/* input - also has a log rule */
				ipt_o = 1,		/* output - also has a log rule */
				ipt_f = 1,		/* forward - also has a log rule */
				ipt_e = 0,		/* prerouting */
				ipt_p = 0;		/* postrouting */

	int			start_ipt_i = 0,	/* input */
				start_ipt_o = 0,	/* output */
				start_ipt_f = 0,	/* forward */
				start_ipt_e = 0,	/* prerouting */
				start_ipt_p = 0;	/* postrouting */
				
	char			cmd[MAX_PIPE_COMMAND];


	/* safety */
	if(!services_list || !zones_list || !interfaces_list || !rules_list || !block_list || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	(void)vrprint.info("Info", "Reloading config...");


	/* close the current backends */
	result = unload_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "unloading backends failed.");
		return(-1);
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 5);

	/* reload the config
	
	   if it fails it's no big deal, we just keep using the old config.
	*/
	if(reload_config(debuglvl, &conf) < VR_CNF_OK)
	{
		(void)vrprint.warning("Warning", "reloading config failed, using old config.");
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 10);

	/* reopen the backends */
	result = load_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		(void)vrprint.error(-1, "Error", "re-opening backends failed.");
		return(-1);
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 15);

	/*
		reload the services, interfaces, zones and rules.
	*/
	(void)vrprint.info("Info", "Reloading services...");
	result = reload_services(debuglvl, services_list, reg->servicename);
	if(result == 0)
	{
	}
	else if(result == 1)
	{
		(void)vrprint.info("Info", "Services changed.");
		retval=0;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Reloading services failed.");
		return(-1);
	}
	//services_print_list(ServicesList);

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 20);

	(void)vrprint.info("Info", "Reloading interfaces...");
	result = reload_interfaces(debuglvl, interfaces_list);
	if(result == 0)
	{
		(void)vrprint.debug("Debug", "Interfaces didn't change.");
	}
	else if(result == 1)
	{
		(void)vrprint.info("Info", "Interfaces changed.");
		retval=0;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Reloading interfaces failed.");
		return(-1);
	}
	//interface_print_list(ZonedataList);

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 25);
	
	(void)vrprint.info("Info", "Reloading zones...");
	result = reload_zonedata(debuglvl, zones_list, reg);
	if(result == 0)
	{
		(void)vrprint.debug("Debug", "Zones didn't change.");
	}
	else if(result == 1)
	{
		(void)vrprint.info("Info", "Zones changed.");
		retval=0;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Reloading zones failed.");
		return(-1);
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 30);

	/*
		count how many rules are to be removed in each chain
	*/
	for(d_node = rules_list->top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		ipt_i = ipt_i + rule_ptr->rulecache.iptcount.input;
		ipt_o = ipt_o + rule_ptr->rulecache.iptcount.output;
		ipt_f = ipt_f + rule_ptr->rulecache.iptcount.forward;
		ipt_e = ipt_e + rule_ptr->rulecache.iptcount.preroute;
		ipt_p = ipt_p + rule_ptr->rulecache.iptcount.postroute;
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "apply_changes: rule to remove: i: %d, o: %d, f: %d, e: %d, p: %d.", ipt_i, ipt_o, ipt_f, ipt_e, ipt_p);

	/*
		determine where to start
	*/
	start_ipt_i = ipt_input_count + 1;
	start_ipt_o = ipt_output_count + 1;
	start_ipt_f = ipt_forward_count + 1;
	start_ipt_e = ipt_preroute_count + 1;
	start_ipt_p = ipt_postroute_count + 1;

	if(debuglvl >= LOW)
		(void)vrprint.debug("Info", "apply_changes: here we start: i: %d, o: %d, f: %d, e: %d, p: %d.", start_ipt_i, start_ipt_o, start_ipt_f, start_ipt_e, start_ipt_p);

	/* clear the rules_list */
	if(rules_cleanup_list(debuglvl, rules_list) < 0)
		retval = -1;

	/* re-initialize the rules_list */
	result = rules_init_list(debuglvl, rules_list, reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading rulesfile succesfull.");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "init_rulesfile failed.");
		retval = -1;
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 35);

	/*
		analyzing the rules
	*/
	if(analyze_all_rules(debuglvl, rules_list, zones_list, services_list, interfaces_list) != 0)
	{
		(void)vrprint.error(-1, "Error", "analizing rules failed.");
		retval=-1;
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 40);

	/*
		here remove the old INPUT rules
	*/
	if(remove_rule(debuglvl, RT_INPUT, start_ipt_i, ipt_i) < 0)
	{
		(void)vrprint.error(-1, "Error", "unable to remove rules.");
		retval=-1;
	}
	/*
		here remove the old OUTPUT rules
	*/
	if(remove_rule(debuglvl, RT_OUTPUT, start_ipt_o, ipt_o) < 0)
	{
		(void)vrprint.error(-1, "Error", "unable to remove rules.");
		retval=-1;
	}
	/*
		here remove the old FORWARD rules
	*/
	if(remove_rule(debuglvl, RT_FORWARD, start_ipt_f, ipt_f) < 0)
	{
		(void)vrprint.error(-1, "Error", "unable to remove rules.");
		retval=-1;
	}
//TODO: sanatize this name -> portfw means preroute
	/*
		here remove the old PREROUTE rules
	*/
	if(remove_rule(debuglvl, RT_PORTFW, start_ipt_e, ipt_e) < 0)
	{
		(void)vrprint.error(-1, "Error", "unable to remove rules.");
		retval = -1;
	}
//TODO: sanatize this name -> snat means postroute
	/*
		here remove the old POSTROUTE rules
	*/
	if(remove_rule(debuglvl, RT_SNAT, start_ipt_p, ipt_p) < 0)
	{
		(void)vrprint.error(-1, "Error", "unable to remove rules.");
		retval = -1;
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 50);

	/* flush the mangle table */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;

	/* flush anti spoof old rules */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush ANTISPOOF", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 60);

	/* update the blocklist */
	if(conf.use_blocklist)
	{
		/* clear the block_list */
		d_list_cleanup(debuglvl, block_list);

		/* and reload it (with load_ips == 1) */
		if(blocklist_read_file(debuglvl, zones_list, block_list, 1) < 0)
		{
			(void)vrprint.error(-1, "Error","reading the blocklist failed (in: %s:%d).", __FUNCTION__, __LINE__);
			retval = -1;
		}

		/* flush the BLOCKLIST */
		snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush BLOCKLIST", conf.iptables_location);
		result = pipe_command(debuglvl, cmd, 0);
		if(result < 0)
			retval = -1;
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 70);

	/* now create the rules */
	if(create_all_rules(debuglvl, rules_list, block_list, 0) != 0)
	{
		(void)vrprint.error(-1, "Error", "creating rules failed.");
		retval=-1;
	}

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 80);

	/* update the SYN-LIMIT rules */
	if(update_synlimit_rules(debuglvl) < 0)
		retval = -1;

	shm_update_progress(debuglvl, sem_id, &shm_table->reload_progress, 90);

	if(retval == 0)
		(void)vrprint.info("Info", "Reloading config completed successfully.");

	return(retval);
}


/*	reload_services
*/
int
reload_services(const int debuglvl, d_list *services_list, regex_t *servicename_regex)
{
	int			retval=0,
				result;
	struct ServicesData_	*ser_ptr = NULL;
	char			name[MAX_SERVICE];
	int			zonetype;
	d_list_node		*d_node = NULL;

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug",  "reload_services: ** start **");


	/* safety */
	if(!services_list || !servicename_regex)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}


	/* check if we have a backend */
	if(!sf)
	{
		(void)vrprint.error(-1, "Internal Error", "backend not open (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* first reset all statusses */
	for(d_node = services_list->top; d_node; d_node = d_node->next)
	{
		if(!(ser_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		ser_ptr->status = ST_UNTOUCHED;
	}


	/* set special service 'all' to 'keep'. */
	if(!(ser_ptr = search_service(debuglvl, "all", services_list)))
	{
		/* if this happens, something is very, very wrong */
		(void)vrprint.error(-1, "Internal Error", "special service 'all' not found in memory.");
		return(-1);
	}
	else
	{
		ser_ptr->status = ST_KEEP;
	}


	/* loop trough the services in the backend */
	while(sf->list(debuglvl, serv_backend, name, &zonetype, CAT_SERVICES) != NULL)
	{
		if(validate_servicename(debuglvl, name, servicename_regex) == 0)
		{
			ser_ptr = search_service(debuglvl, name, services_list);
			if(ser_ptr == NULL) /* not found */
			{
				(void)vrprint.info("Info", "Service '%s' is added.", name);
				
				retval = 1;
	
				/* new service */
				result = insert_service(debuglvl, services_list, name);
				if(result != 0)
				{
					(void)vrprint.error(-1, "Internal Error", "inserting data for '%s' into the list failed (in: reload_services).", name);
					return(-1);
				}
			}
			else
			{
				/* check the content of the service for changes */
				result = reload_services_check(debuglvl, ser_ptr);
				if(result == 1)
					retval = 1;
				else if(result < 0)
					return(-1);
			}
		}
	}


	/* the untouched services are to be removed */
	for(d_node = services_list->top; d_node; d_node = d_node->next)
	{
		/* get the service */
		if(!(ser_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* if status is UNTOUCHED, mark REMOVED */
		if(ser_ptr->status == ST_UNTOUCHED)
		{
			ser_ptr->status = ST_REMOVED;

			(void)vrprint.info("Info", "Service '%s' is removed.", ser_ptr->name);
		}
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "reload_services: ** end **, result = %d", retval);

	return(retval);
}


/*	reload_services_check

	TODO:
		- replace 'new_ser_ptr->status = changed; ser_ptr->status = changed;' with 'status = changed',
		like in reload_zones_check.

	Returncodes:
		1: changes
		0: no changes
		-1: error
*/
int
reload_services_check(const int debuglvl, struct ServicesData_ *ser_ptr)
{
	int			retval = 0,
				result = 0,
				status = 0;
	struct ServicesData_	*new_ser_ptr;
	/* these are for the comparisson between the portranges */
	d_list_node		*list_node = NULL,
				*temp_node = NULL;
	struct portdata		*list_port = NULL,
				*temp_port = NULL;


	/* safety */
	if(!ser_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: reload_services_check).");
		return(-1);
	}


	/* alloc the temp mem */
	if(!(new_ser_ptr = service_malloc()))
	{
		(void)vrprint.error(-1, "Error", "Claiming memory for new_ser_ptr failed.");
		return(-1);
	}
	if(d_list_setup(debuglvl, &new_ser_ptr->PortrangeList, free) < 0)
		return(-1);


	/* read the service from the backend again */
	result = read_service(debuglvl, ser_ptr->name, new_ser_ptr);
	if(result != 0)
	{
		/* error! memory is freed at the end of this function */
		(void)vrprint.error(-1, "Error", "getting info for service '%s' failed (in: reload_services_check).", ser_ptr->name);
		status = ST_REMOVED;
	}
	else
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug("Debug", "%s: service: %12s.", __FUNCTION__, ser_ptr->name);

		/* we asume that the service did not change, if so we change it below */
		status = ST_KEEP;

		/* active */
		if(ser_ptr->active == new_ser_ptr->active)
		{
			/* compare the protocol helpers */
			if(strcmp(ser_ptr->helper, new_ser_ptr->helper) == 0)
			{
				if(ser_ptr->PortrangeList.len == new_ser_ptr->PortrangeList.len)
				{
					list_node = ser_ptr->PortrangeList.top;
					temp_node = new_ser_ptr->PortrangeList.top;

					if(!list_node && !temp_node)
					{
						/* no change */
					}
					/* if eitherone is NULL and the other not there must be a change */
					else if((!list_node && temp_node) || (list_node && !temp_node))
					{
						/* change */
						if(!list_node)
							(void)vrprint.info("Info", "Service '%s': the service now has (a) portrange(s).", ser_ptr->name);

						if(!temp_node)
							(void)vrprint.info("Info", "Service '%s': the service no longer has portranges.", ser_ptr->name);

						status = ST_CHANGED;
					}
					else
					{
						for(; list_node && temp_node; list_node = list_node->next, temp_node = temp_node->next)
						{
							if(!(list_port = list_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "reload_services_check: list_port == NULL!");
								return(-1);
							}

							if(!(temp_port = temp_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "reload_services_check: temp_port == NULL!");
								return(-1);
							}

							if((list_port->protocol == temp_port->protocol) && (list_port->src_low == temp_port->src_low) && (list_port->src_high == temp_port->src_high) && (list_port->dst_low == temp_port->dst_low) && (list_port->dst_high == temp_port->dst_high))
							{
								/* nothing changed */
							}
							else
							{
								(void)vrprint.info("Info", "Service '%s': one of the portranges has been changed.", ser_ptr->name);
								status = ST_CHANGED;

								break;
							}
						}
					}
				}
				else
				{
					(void)vrprint.info("Info", "Service '%s': the number of portranges has been changed.", ser_ptr->name);
					status = ST_CHANGED;
				}
			}
			else
			{
				(void)vrprint.info("Info", "Service '%s': helper has been set to: '%s'.", ser_ptr->name, new_ser_ptr->helper);
				status = ST_CHANGED;
			}
		}
		else
		{
			if(new_ser_ptr->active == 1)
			{
				status = ST_ACTIVATED;
				(void)vrprint.info("Info", "Service '%s' has been activated.", ser_ptr->name);
			}
			else
			{
				status = ST_DEACTIVATED;
				(void)vrprint.info("Info", "Service '%s' has been deactivated.", ser_ptr->name);
			}
		}
	}


	/* */
	if(status == ST_CHANGED || status == ST_ACTIVATED || status == ST_DEACTIVATED)
	{
		(void)vrprint.info("Info", "Service '%s' has been changed.", ser_ptr->name);

		/* delete the old portrange list */
		d_list_cleanup(debuglvl, &ser_ptr->PortrangeList);

		/* copy the data */
		*ser_ptr = *new_ser_ptr;

		/* transfer the status */
		ser_ptr->status = status;

		/* set retval to changes */
		retval = 1;
	}
	else if(status == ST_REMOVED || status == ST_KEEP)
	{
		/* destroy the portrangelist of the temp service */
		d_list_cleanup(debuglvl, &new_ser_ptr->PortrangeList);

		/* set the status */
		ser_ptr->status = status;
	}

	/* free the temp data */
	free(new_ser_ptr);
	new_ser_ptr = NULL;

	return(retval);
}


// reload_zonedata
int
reload_zonedata(const int debuglvl, d_list *d_list, struct rgx_ *reg)
{
	int			retval = 0,
				result = 0;
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	char			name[MAX_HOST_NET_ZONE];
	int			zonetype;

	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "reload_zonedata: ** start **");

	/* check if we have a backend */
	if(!zf)
	{
		(void)vrprint.error(-1, "Internal Error", "backend not open (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* first reset all statusses */
	for(d_node = d_list->top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		zone_ptr->status = ST_UNTOUCHED;
	}

	/* loop trough backend and check */
	while(zf->list(debuglvl, zone_backend, name, &zonetype, CAT_ZONES) != NULL)
	{
		zone_ptr = search_zonedata(debuglvl, name, d_list);
		if(zone_ptr == NULL)
		{
			/* print that we have a new zone */
			if(zonetype == TYPE_ZONE)
				(void)vrprint.info("Info", "Zone '%s' was added.", name);
			else if(zonetype == TYPE_NETWORK)
				(void)vrprint.info("Info", "Network '%s' was added.", name);
			else if(zonetype == TYPE_HOST)
				(void)vrprint.info("Info", "Host '%s' was added.", name);
			else if(zonetype == TYPE_GROUP)
				(void)vrprint.info("Info", "Group '%s' was added.", name);

			/* new zone */
			result = insert_zonedata(debuglvl, d_list, name, zonetype, reg, &InterfacesList);
			if(result != 0)
			{
				(void)vrprint.error(-1, "Internal Error", "inserting data for '%s' into the list failed (reload_zonedata).", name);
				return(-1);
			}

			retval = 1;
		}
		else
		{
			result = reload_zonedata_check(debuglvl, d_list, zone_ptr, reg);
			if(result < 0)
			{
				(void)vrprint.debug("Debug", "reload_zonedata: < 0");
				return(-1);
			}
			else if(result == 1)
				retval = 1;
		}
	}

	/* untouched means to be removed */
	for(d_node = d_list->top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		if(zone_ptr->status == ST_UNTOUCHED)
		{
			/* print that we have a zone to remove */
			if(zone_ptr->type == TYPE_ZONE)
				(void)vrprint.info("Info", "Zone '%s' was removed.", zone_ptr->name);
			else if(zone_ptr->type == TYPE_NETWORK)
				(void)vrprint.info("Info", "Network '%s' was removed.", zone_ptr->name);
			else if(zone_ptr->type == TYPE_HOST)
				(void)vrprint.info("Info", "Host '%s' was removed.", zone_ptr->name);
			else if(zone_ptr->type == TYPE_GROUP)
				(void)vrprint.info("Info", "Group '%s' was removed.", zone_ptr->name);
			
			zone_ptr->status = ST_REMOVED;
			retval = 1;
		}
	}

	//zonedata_print_list(&ZonedataList);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "reload_zonedata: ** end **, retval=%d", retval);

	return(retval);
}


/*	reload_zonedata_check

	Checks the content of an host, group, network or zone for changes.
	
	Returncodes:
		 1: changes
		 0: no changes
		-1: error
*/
int
reload_zonedata_check(const int debuglvl, d_list *d_list, struct ZoneData_ *zone_ptr, struct rgx_ *reg)
{
	int			result = 0,
				retval = 0;
	struct ZoneData_	*host_ptr_new  =NULL,
				*host_ptr_orig,
				*new_zone_ptr=NULL;
	struct InterfaceData_	*iface_ptr_new,
				*iface_ptr_orig;
	d_list_node		*d_node_orig = NULL,
				*d_node_new  = NULL;
	d_list_node		*protect_d_node_orig = NULL,
				*protect_d_node_new  = NULL;
	struct RuleData_	*org_rule_ptr = NULL,
				*new_rule_ptr = NULL;
	int			status=-1;


	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "reload_zonedata_check: ** start **");


	/* safety */
	if(!d_list || !zone_ptr || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: reload_zonedata_check).");
		return(-1);
	}


	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "reload_zonedata_check: zone: %s, type: %d", zone_ptr->name, zone_ptr->type);


	/* alloc mem for new zone */
	if(!(new_zone_ptr = zone_malloc(debuglvl)))
	{
		(void)vrprint.error(-1, "Error", "allocating memory failed: %s.", strerror(errno));
		return(-1);
	}


	switch(zone_ptr->type)
	{
		case TYPE_ZONE:

			/* set the zone up */
			result = read_zonedata(debuglvl, zone_ptr->name, TYPE_ZONE, new_zone_ptr, d_list, reg, &InterfacesList);
			if(result != 0)
			{
				/* error! memory is freed at the end of this function */
				(void)vrprint.error(-1, "Error", "getting info for zone '%s' failed (in: reload_zonedata_check).", zone_ptr->name);
				status = ST_REMOVED;
			}

			/* now go check for differences */
			else
			{
				/* active */
				if(zone_ptr->active == new_zone_ptr->active)
				{
					status = ST_KEEP;
				}
				else
				{
					if(new_zone_ptr->active == 1)
					{
						status = ST_ACTIVATED;
						(void)vrprint.info("Info", "Zone '%s' has been activated.", zone_ptr->name);
					}
					else
					{
						status = ST_DEACTIVATED;
						(void)vrprint.info("Info", "Zone '%s' has been deactivated.", zone_ptr->name);
					}
				}
			}
			break;

		case TYPE_NETWORK:

			result = read_zonedata(debuglvl, zone_ptr->name, TYPE_NETWORK, new_zone_ptr, d_list, reg, &InterfacesList);
			if(result != 0)
			{
				/* error! memory is freed at the end of this function */
				(void)vrprint.error(-1, "Error", "getting info for network '%s' failed (in: reload_zonedata_check).", zone_ptr->name);
				status = ST_REMOVED;
			}
			else
			{
				/* we start at keep */
				status = ST_KEEP;

				/* active */
				if(zone_ptr->active == new_zone_ptr->active)
				{
					/* network */
					if(strcmp(zone_ptr->ipv4.network, new_zone_ptr->ipv4.network) == 0)
					{
						/* netmask */
						if(strcmp(zone_ptr->ipv4.netmask, new_zone_ptr->ipv4.netmask) == 0)
						{
							/* interfaces */
							if(zone_ptr->interface_count != new_zone_ptr->interface_count)
							{
								(void)vrprint.info("Info", "Network '%s': the number of interfaces has been changed.", zone_ptr->name);
								status = ST_CHANGED;
							}

							/* now loop through the member to see if they have changes */
							d_node_new  = new_zone_ptr->InterfaceList.top;
							d_node_orig = zone_ptr->InterfaceList.top;

							if(!d_node_new && !d_node_orig)
							{
								/* no change */
							}
							/* if eitherone is NULL and the other not there must be a change */
							else if((!d_node_new && d_node_orig) || (d_node_new && !d_node_orig))
							{
								/* change */
								if(!d_node_orig)
									(void)vrprint.info("Info", "Network '%s': network now has (an) interface(s).", zone_ptr->name);
									
								if(!d_node_new)
									(void)vrprint.info("Info", "Network '%s': network now has (an) interface(s).", zone_ptr->name);
								
								status = ST_CHANGED;
							}
							else
							{
								for(; d_node_new && d_node_orig; d_node_new = d_node_new->next, d_node_orig = d_node_orig->next)
								{
									if(!(iface_ptr_new = d_node_new->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
									if(!(iface_ptr_orig = d_node_orig->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}

									if(strcmp(iface_ptr_orig->name, iface_ptr_new->name) != 0)
									{
										(void)vrprint.info("Info", "Network '%s': interfaces not in the same order.", zone_ptr->name);
										status = ST_CHANGED;
									}

									if(iface_ptr_new->status != ST_KEEP)
									{
										(void)vrprint.info("Info", "Network '%s': interface '%s' has been changed.", zone_ptr->name, iface_ptr_new->name);
										status = ST_CHANGED;
									}
								}
							}


							/* protect rules */
							if(zone_ptr->ProtectList.len != new_zone_ptr->ProtectList.len)
							{
								(void)vrprint.info("Info", "Network '%s': the number of protectrules has been changed.", zone_ptr->name);
								status = ST_CHANGED;
							}

							/* now loop through the member to see if they have changes */
							protect_d_node_new  = new_zone_ptr->ProtectList.top;
							protect_d_node_orig = zone_ptr->ProtectList.top;

							if(!protect_d_node_new && !protect_d_node_orig)
							{
								/* no change */
							}
							/* if eitherone is NULL and the other not there must be a change */
							else if((!protect_d_node_new && protect_d_node_orig) || (protect_d_node_new && !protect_d_node_orig))
							{
								/* change */
								if(!protect_d_node_orig)
									(void)vrprint.info("Info", "Network '%s': network now has (an) protectrule(s).", zone_ptr->name);
									
								if(!protect_d_node_new)
									(void)vrprint.info("Info", "Network '%s': network now has (an) protectrule(s).", zone_ptr->name);
								
								status = ST_CHANGED;
							}
							else
							{
								for(; protect_d_node_new && protect_d_node_orig; protect_d_node_new = protect_d_node_new->next, protect_d_node_orig = protect_d_node_orig->next)
								{
									if(!(new_rule_ptr = protect_d_node_new->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
									if(!(org_rule_ptr = protect_d_node_orig->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}

									if(	strcmp(org_rule_ptr->danger, new_rule_ptr->danger) != 0 ||
										strcmp(org_rule_ptr->source, new_rule_ptr->source) != 0)
									{
										(void)vrprint.info("Info", "Network '%s': protectrules not in the same order.", zone_ptr->name);
										status = ST_CHANGED;
									}
								}
							}

						}
						else
						{
							(void)vrprint.info("Info", "Network '%s': netmask changed to: '%s'.", zone_ptr->name, new_zone_ptr->ipv4.netmask);
							status = ST_CHANGED;
						}
					}
					else
					{
						(void)vrprint.info("Info", "Network '%s': network address changed to: '%s'.", zone_ptr->name, new_zone_ptr->ipv4.network);
						status = ST_CHANGED;
					}
				}
				else
				{
					if(new_zone_ptr->active == 1)
					{
						(void)vrprint.info("Info", "Network '%s' has been activated.", zone_ptr->name);
						status = ST_ACTIVATED;
					}
					else
					{
						(void)vrprint.info("Info", "Network '%s' has been deactivated.", zone_ptr->name);
						status = ST_DEACTIVATED;
					}
				}
			}
			break;

		case TYPE_HOST:

			result = read_zonedata(debuglvl, zone_ptr->name, TYPE_HOST, new_zone_ptr, d_list, reg, &InterfacesList);
			if(result != 0)
			{
				/* error! memory is freed at the end of this function */
				(void)vrprint.error(-1, "Error", "getting info for host '%s' failed (in: reload_zonedata_check).", zone_ptr->name);
				status = ST_REMOVED;
			}
			else
			{
				/* active */
				if(zone_ptr->active == new_zone_ptr->active)
				{
					/* ipaddress */
					if(strcmp(zone_ptr->ipv4.ipaddress, new_zone_ptr->ipv4.ipaddress) == 0)
					{
						/* have mac */
						if(zone_ptr->has_mac == new_zone_ptr->has_mac)
						{
							/* mac */
							if(zone_ptr->has_mac == 1)
							{
								if(strcmp(zone_ptr->mac, new_zone_ptr->mac) == 0)
								{
									status = ST_KEEP;
								}
								else
								{
									(void)vrprint.info("Info", "Host '%s' has a new mac-address: '%s'.", zone_ptr->name, new_zone_ptr->mac);
									status = ST_CHANGED;
								}
							}
							else
							{
								status = ST_KEEP;
							}
						}
						else
						{
							if(zone_ptr->has_mac == 0)
							{
								(void)vrprint.info("Info", "Host '%s' now has a mac-address (%s).", zone_ptr->name, new_zone_ptr->mac);
							}
							else
							{
								(void)vrprint.info("Info", "Host '%s' no longer has a mac-address.", zone_ptr->name);
							}
							
							status = ST_CHANGED;
						}
					}
					else
					{
						(void)vrprint.info("Info", "Host '%s' has a new ipaddress: '%s'.", zone_ptr->name, new_zone_ptr->ipv4.ipaddress);
						status = ST_CHANGED;
					}
				}
				else
				{
					if(new_zone_ptr->active == 1)
					{
						(void)vrprint.info("Info", "Host '%s' has been activated.", zone_ptr->name);
						status = ST_ACTIVATED;
					}
					else
					{
						(void)vrprint.info("Info", "Host '%s' has been deactivated.", zone_ptr->name);
						status = ST_DEACTIVATED;
					}
				}
			}
			break;

		case TYPE_GROUP:

			result = read_zonedata(debuglvl, zone_ptr->name, TYPE_GROUP, new_zone_ptr, d_list, reg, &InterfacesList);
			if(result != 0)
			{
				/* error! memory is freed at the end of this function */
				(void)vrprint.error(-1, "Error", "getting info for group '%s' failed (in: reload_zonedata_check).", zone_ptr->name);

				status = ST_REMOVED;
			}
			else
			{
				status = ST_KEEP;

				/* active */
				if(zone_ptr->active == new_zone_ptr->active)
				{
					/* member count */
					if(zone_ptr->group_member_count != new_zone_ptr->group_member_count)
					{
						(void)vrprint.info("Info", "Group '%s': the number of members changed.", zone_ptr->name);
						status = ST_CHANGED;
					}

					/* now loop through the member to see if they have changes */
					d_node_new  = new_zone_ptr->GroupList.top;
					d_node_orig = zone_ptr->GroupList.top;

					if(!d_node_new && !d_node_orig)
					{
						/* no change */
					}
					/* if eitherone is NULL and the other not there must be a change */
					else if((!d_node_new && d_node_orig) || (d_node_new && !d_node_orig))
					{
						/* change */
						status = ST_CHANGED;

						/* change */
						if(!d_node_orig)
							(void)vrprint.info("Info", "Group '%s': group now has (a) member(s).", zone_ptr->name);

						if(!d_node_new)
							(void)vrprint.info("Info", "Group '%s': group no longer has members.", zone_ptr->name);
								
					}
					else
					{
						for(; d_node_new && d_node_orig; d_node_new = d_node_new->next, d_node_orig = d_node_orig->next)
						{
							if(!(host_ptr_new = d_node_new->data))
							{
								(void)vrprint.error(-1, "Internal Error", "reload_zonedata_check: host_ptr_new == NULL (and it shouldn't).");
								return(-1);
							}

							if(!(host_ptr_orig = d_node_orig->data))
							{
								(void)vrprint.error(-1, "Internal Error", "reload_zonedata_check: host_ptr_orig == NULL (and it shouldn't).");
								return(-1);
							}

							if(strcmp(host_ptr_orig->name, host_ptr_new->name) != 0)
							{
								(void)vrprint.info("Info", "Group '%s': members not in the same order.", zone_ptr->name);
								status = ST_CHANGED;
							}

							if(host_ptr_new->status != ST_KEEP)
							{
								(void)vrprint.info("Info", "Group '%s': member '%s' has been changed.", zone_ptr->name, host_ptr_new->name);
								status = ST_CHANGED;
							}
						}
					}
				}
				else
				{
					if(new_zone_ptr->active == 1)
					{
						(void)vrprint.info("Info", "Group '%s' has been activated.", zone_ptr->name);
						status = ST_ACTIVATED;
					}
					else
					{
						(void)vrprint.info("Info", "Group '%s' has been deactivated.", zone_ptr->name);
						status = ST_DEACTIVATED;
					}
				}
			}
			break;
		
		default:
		
			(void)vrprint.error(-1, "Error", "unknown zone type: %d for zone %s (in: %s).", zone_ptr->type, zone_ptr->name, __FUNCTION__);

			retval = -1;
			break;
	}


	/* update the data in memory */
	if(status == ST_CHANGED || status == ST_ACTIVATED || status == ST_DEACTIVATED)
	{
		/* update the data */

		/* first destroy the old lists */
		if(zone_ptr->type == TYPE_GROUP)
			d_list_cleanup(debuglvl, &zone_ptr->GroupList);
		if(zone_ptr->type == TYPE_NETWORK)
		{
			d_list_cleanup(debuglvl, &zone_ptr->InterfaceList);
			d_list_cleanup(debuglvl, &zone_ptr->ProtectList);
		}

		/* copy the zone */
		*zone_ptr = *new_zone_ptr;

		/* transfer the status */
		zone_ptr->status = status;

		/* tell the caller we have changes */
		retval = 1;
	}
	else if(status == ST_KEEP || status == ST_REMOVED)
	{
		/* first destroy the new lists, the struct will be free'd later */
		if(new_zone_ptr->type == TYPE_GROUP)
			d_list_cleanup(debuglvl, &new_zone_ptr->GroupList);
		if(new_zone_ptr->type == TYPE_NETWORK)
		{
			d_list_cleanup(debuglvl, &new_zone_ptr->InterfaceList);
			d_list_cleanup(debuglvl, &new_zone_ptr->ProtectList);
		}
                        
		/* status to keep */
		zone_ptr->status = status;
	}

	/* now free new_zone_ptr */
	free(new_zone_ptr);
	new_zone_ptr = NULL;

	return(retval);
}

/*	reload_interfaces

	Returncodes:
		 1: changes
		 0: no changes
		-1: error
*/
int
reload_interfaces(const int debuglvl, d_list *interfaces_list)
{
	int			retval = 0,
				result = 0;
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	char			name[MAX_INTERFACE] = "";
	int			zonetype = 0;


	/* safety first */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* check if we have a backend */
	if(!af)
	{
		(void)vrprint.error(-1, "Internal Error", "backend not open (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	
	/* first reset all statusses */
	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		iface_ptr->status = ST_UNTOUCHED;
	}


	/* now loop trough the interfaces and check them */
	while(af->list(debuglvl, ifac_backend, name, &zonetype, CAT_INTERFACES) != NULL)
	{
		iface_ptr = search_interface(debuglvl, name, interfaces_list);
		if(iface_ptr == NULL)
		{
			(void)vrprint.info("Info", "Interface '%s' is added.", name);

			/* this is a new interface */
			result = insert_interface(debuglvl, interfaces_list, name);
			if(result != 0)
			{
				(void)vrprint.error(-1, "Internal Error", "inserting interface '%s' into the list failed (in: %s).", name, __FUNCTION__);
				return(-1);
			}

			retval = 1;
		}
		else
		{
			/* existing interface, so check it for changes */
			result = reload_interfaces_check(debuglvl, iface_ptr);
			if(result == 1)
				retval = 1;
			else if(retval < 0)
				return(-1);
		}
	}


	/* tag untouched interfaces as (to be) 'removed' */
	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(iface_ptr->status == ST_UNTOUCHED)
		{
			(void)vrprint.info("Info", "Interface '%s' is removed.", iface_ptr->name);
			iface_ptr->status = ST_REMOVED;
			
			retval = 1;
		}
	}


	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "reload_interfaces: ** end **, result = %d", retval);

	return(retval);
}

/*	reload_interfaces_check

	Checks an individual interface for changes.

	Returncodes:
		 1: changes
		 0: no changes
		-1: error
*/
int
reload_interfaces_check(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	int			retval = 0;
	struct InterfaceData_	*new_iface_ptr = NULL;
	int			status = 0;
	d_list_node		*protect_d_node_orig = NULL,
				*protect_d_node_new  = NULL;
	struct RuleData_	*org_rule_ptr = NULL,
				*new_rule_ptr = NULL;


	/* safety */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* alloc mem for a temporary interface */
	if(!(new_iface_ptr = interface_malloc(debuglvl)))
	{
		(void)vrprint.error(-1, "Error", "claiming memory for new_iface_ptr failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}


	/* set the name */
	if(strlcpy(new_iface_ptr->name, iface_ptr->name, sizeof(new_iface_ptr->name)) >= sizeof(new_iface_ptr->name))
	{
		(void)vrprint.error(-1, "Error", "buffer overflow in interfacename (in: %s).", __FUNCTION__);
		return(-1);
	}


	/* first we asume that the interface did not change, if so we change it below */
	status = ST_KEEP;


	/* get the info from the backend */
	if(read_interface_info(debuglvl, new_iface_ptr) != 0)
	{
		(void)vrprint.error(-1, "Error", "getting interface information for '%s' failed (in: %s).", iface_ptr->name, __FUNCTION__);
		status = ST_REMOVED;
	}
	else
	{
		/*
			NOW CHECK for changes
		*/

		/* active */
		if(iface_ptr->active == new_iface_ptr->active)
		{
			/* compare device */
			if(strcmp(iface_ptr->device, new_iface_ptr->device) == 0)
			{
				if(iface_ptr->device_virtual == new_iface_ptr->device_virtual)
				{
					/* compare ipaddress */
					if(strcmp(iface_ptr->ipv4.ipaddress, new_iface_ptr->ipv4.ipaddress) == 0)
					{
						/* check 'up' */
						if(iface_ptr->up == new_iface_ptr->up)
						{
							/* protect rules */
							if(iface_ptr->ProtectList.len != new_iface_ptr->ProtectList.len)
							{
								(void)vrprint.info("Info", "Interface '%s': the number of protectrules has been changed.", iface_ptr->name);
								status = ST_CHANGED;
							}

							/* now loop through the member to see if they have changes */
							protect_d_node_new  = new_iface_ptr->ProtectList.top;
							protect_d_node_orig = iface_ptr->ProtectList.top;

							if(!protect_d_node_new && !protect_d_node_orig)
							{
								/* no change */
							}
							/* if eitherone is NULL and the other not there must be a change */
							else if((!protect_d_node_new && protect_d_node_orig) || (protect_d_node_new && !protect_d_node_orig))
							{
								/* change */
								if(!protect_d_node_orig)
									(void)vrprint.info("Info", "Interface '%s': interface now has (an) protectrule(s).", iface_ptr->name);
									
								if(!protect_d_node_new)
									(void)vrprint.info("Info", "Interface '%s': interface now has (an) protectrule(s).", iface_ptr->name);
								
								status = ST_CHANGED;
							}
							else
							{
								for(; protect_d_node_new && protect_d_node_orig; protect_d_node_new = protect_d_node_new->next, protect_d_node_orig = protect_d_node_orig->next)
								{
									if(!(new_rule_ptr = protect_d_node_new->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}
									if(!(org_rule_ptr = protect_d_node_orig->data))
									{
										(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}

									if(	strcmp(org_rule_ptr->danger, new_rule_ptr->danger) != 0 ||
										strcmp(org_rule_ptr->source, new_rule_ptr->source) != 0)
									{
										(void)vrprint.info("Info", "Interface '%s': protectrules not in the same order.", iface_ptr->name);
										status = ST_CHANGED;
									}
								}
							}

						}
						else
						{
							(void)vrprint.info("Info", "Interface '%s' is now: %s.", iface_ptr->name, new_iface_ptr->up ? "up" : "down");
							status = ST_CHANGED;
						}
					}
					else
					{
						(void)vrprint.info("Info", "Interface '%s' has a new ipaddress: '%s'.", iface_ptr->name, new_iface_ptr->ipv4.ipaddress);
						status = ST_CHANGED;
					}
				}
				else
				{
					if(new_iface_ptr->device_virtual == 1)
						(void)vrprint.info("Info", "Interface '%s' is now 'virtual'.", iface_ptr->name);
					else
						(void)vrprint.info("Info", "Interface '%s' is no longer 'virtual'.", iface_ptr->name);

					status = ST_CHANGED;
				}
			}
			else
			{
				(void)vrprint.info("Info", "Interface '%s' has a new system device: '%s'.", iface_ptr->name, new_iface_ptr->device);
				status = ST_CHANGED;
			}
		}
		else
		{
			if(new_iface_ptr->active == 1)
			{
				(void)vrprint.info("Info", "Interface '%s' has been activated.", iface_ptr->name);
				status = ST_ACTIVATED;
			}
			else
			{
				(void)vrprint.info("Info", "Interface '%s' has been deactivated.", iface_ptr->name);
				status = ST_DEACTIVATED;
			}
		}
	}

	if(status == ST_CHANGED || status == ST_ACTIVATED || status == ST_DEACTIVATED)
	{
		d_list_cleanup(debuglvl, &iface_ptr->ProtectList);

		/* copy the data */
		*iface_ptr = *new_iface_ptr;

		/* inform the caller that there were changes */
		retval = 1;
	}
	else if(status == ST_KEEP || status == ST_REMOVED)
	{
		d_list_cleanup(debuglvl, &new_iface_ptr->ProtectList);
	}

	/* free the temp data */
	free(new_iface_ptr);
	new_iface_ptr = NULL;

	/* transfer the status */
	iface_ptr->status = status;

	return(retval);
}
