/***************************************************************************
 *   Copyright (C) 2002-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/***************************************************************************
 * Here we try to create the actual rule.                                  *
 ***************************************************************************/
#include "main.h"


/*	analyze_all_rules

	Analyzes all rules :-)

	Returncodes:
		 0: ok
		-1: error
*/
int
analyze_all_rules(const int debuglvl, d_list *rules_list, d_list *zones_list, d_list *services_list, d_list *interfaces_list)
{
	struct RuleData_	*rule_ptr = NULL;
	int			rulescount = 0,
				rulesfailedcount = 0;
	d_list_node		*d_node = NULL,
				*next_d_node = NULL,
				*net_d_node = NULL,
				*if_d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	struct InterfaceData_	*iface_ptr = NULL;

	/* safety */
	if(!rules_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	(void)vrprint.info("Info", "Analyzing the rules... ");

	/* first analyze the protectrules in the interfaces */
	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		for(if_d_node = iface_ptr->ProtectList.top; if_d_node; if_d_node = if_d_node->next)
		{
			if(!(rule_ptr = if_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services_list, zones_list, interfaces_list) == 0)
			{
				(void)vrprint.debug(__FUNCTION__, "analizing protectrule success, active = 1.");
				rule_ptr->active = 1;
			}
			else
			{
				(void)vrprint.debug(__FUNCTION__, "analizing protectrule failed, active = 0.");
				rule_ptr->active = 0;
			}
		}
	}

	/* first analyze the protectrules in the network */
	for(d_node = zones_list->top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		if(zone_ptr->type == TYPE_NETWORK)
		{
			for(net_d_node = zone_ptr->ProtectList.top; net_d_node; net_d_node = net_d_node->next)
			{
				if(!(rule_ptr = net_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services_list, zones_list, interfaces_list) == 0)
				{
					(void)vrprint.debug(__FUNCTION__, "analizing protectrule success, active = 1.");
					rule_ptr->active = 1;
				}
				else
				{
					(void)vrprint.debug(__FUNCTION__, "analizing protectrule failed, active = 0.");
					rule_ptr->active = 0;
				}
			}
		}
	}

	/* check if the list is not empty. If it is, d_node will be NULL. */
	if(rules_list->len > 0)
	{
		/*	Get the top of the list. This should never fail because
			we already checked the listsize.
		*/
		if(!(d_node = rules_list->top))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	/* if we have a node, continue */
	for(;d_node;)
	{
		rulescount++;

		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* analyze! */
		if(rules_analyze_rule(debuglvl, rule_ptr, &rule_ptr->rulecache, services_list, zones_list, interfaces_list) == 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "main: analyze_rule %3d ok.", rulescount);
				
			/* update d_node */
			d_node = d_node->next;
		}
		else
		{
			(void)vrprint.warning("Warning", "Analyzing rule %d failed.", rulescount);
			rulesfailedcount++;
				
			/* update node before removing */
			next_d_node = d_node->next;

			/* remove the failed rule from the list */
			if(d_list_remove_node(debuglvl, rules_list, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "removing node from list failed (in: analyze_all_rules).");
				return(-1);
			}

			/* now update the number in the list */
			query_update_numbers(rulescount - rulesfailedcount + 1, rules_list, 0);

			free_options(debuglvl, rule_ptr->opt);
			free(rule_ptr);
			rule_ptr = NULL;
				
			/* set d_node to the next_d_node */
			d_node = next_d_node;
		}
	}

	(void)vrprint.info("Info", "Analyzing %d rules succesfull, %d failed.", rulescount - rulesfailedcount, rulesfailedcount);
	return(0);
}


/*	create_all_rules

	Creates all rules.

	If 'create_prerules' is set to 1, prerules() will be called.

	Returncodes:
		 0: ok
		-1: error
*/
int
create_all_rules(const int debuglvl, d_list *rules_list, d_list *block_list, int create_prerules)
{
	int			result = 0,
				active = 0,
				rulescount = 0,
				forward_rules = 0;
	struct RuleData_	*rule_ptr = 0;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!rules_list || !block_list || create_prerules < 0 || create_prerules > 1)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: create_all_rules).");
		return(-1);
	}

	(void)vrprint.info("Info", "Creating the rules... (rules to create: %d)", rules_list->len);

	/* create the prerules if were called with it */
	if(create_prerules)
	{
		result = pre_rules(debuglvl, &InterfacesList);
		if(result < 0)
			return(-1);
	}

	/* create the blocklist */
	if(conf.use_blocklist)
	{
		if(create_block_rules(debuglvl, block_list) < 0)
			(void)vrprint.error(-1, "Error", "create blocklist failed.");
	}

	/* create the protect rules */
	if(create_network_protect_rules(debuglvl, &ZonedataList) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}
	/* create the protect rules */
	if(create_interface_protect_rules(debuglvl, &InterfacesList) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}
	if(create_system_protectrules(debuglvl, &conf) < 0)
	{
		(void)vrprint.error(-1, "Error", "create protectrules failed.");
	}


	/* walk trough the ruleslist and create the rules */
	for(d_node = rules_list->top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		/* asume active when we begin */
		active = 1;

		/* count the rules */
		rulescount++;

		/* check normal rule */
		if(rule_ptr->rulecache.from && rule_ptr->rulecache.to && rule_ptr->rulecache.service)
		{
			if(!rule_ptr->rulecache.from->active || !rule_ptr->rulecache.to->active || !rule_ptr->rulecache.service->active)
			{
				/* if one of above is not active, set active to NO */
				active = 0;
			}
		}

		/* check protect rule */
		if(rule_ptr->rulecache.who)
		{
			if(rule_ptr->rulecache.who->active == 0)
			{
				active = 0;
			}
		}

		/* create the rule */
		if(active)
		{
			if(strcasecmp(rule_ptr->action, "protect") == 0)
			{
				/* iptables protect */
				if(rule_ptr->rulecache.danger.solution == PROT_IPTABLES)
				{
					if(rule_ptr->rulecache.who)
					{
						if(create_rule(debuglvl, &rule_ptr->rulecache) < 0)
						{
							(void)vrprint.warning("Warning", "Creating rule %d failed.", rulescount);
						}
					}
					else
					{
						(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: create_all_rules).");
						return(-1);
					}
				}

				/* proc protect */
				else
				{
					if (create_rule(debuglvl, &rule_ptr->rulecache) < 0)
					{
						(void)vrprint.warning("Warning", "Creating rule %d failed.", rulescount);
					}
				}
			}

			/* normal rule */
			else
			{
				if(rule_ptr->rulecache.from && rule_ptr->rulecache.to && rule_ptr->rulecache.service)
				{
					if(create_rule(debuglvl, &rule_ptr->rulecache) == 0)
					{
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "rule created succesfully.");
							
						if(rule_ptr->rulecache.iptcount.forward > 0)
							forward_rules = 1;
					}
					else
					{
						(void)vrprint.warning("Warning", "Creating rule %d failed.", rulescount);
					}
				}
				else
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: create_all_rules).");
					return(-1);
				}
			}
		}
		else
		{
			(void)vrprint.warning("Warning", "Rule %d not created: inactive.", rulescount);
		}
	}

	/* post rules: enable logging */
	if(post_rules(debuglvl, forward_rules) < 0)
		return(-1);

	(void)vrprint.info("Info", "Creating rules succesfull.");

	return(0);
}


static int
create_rule_set_ports(struct RuleCreateData_ *rule, struct portdata *portrange_ptr)
{
	/* from */
	if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
	{
		if(portrange_ptr->src_high == 0)
			snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), "--sport %d", portrange_ptr->src_low);
		else
			snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), "--sport %d:%d", portrange_ptr->src_low, portrange_ptr->src_high);
	}
	else if(portrange_ptr->protocol == 1 || portrange_ptr->protocol == 47)
		snprintf(rule->temp_src_port, sizeof(rule->temp_src_port), " ");

	/* to */
	if(portrange_ptr->protocol == 6 || portrange_ptr->protocol == 17)
	{
		if(portrange_ptr->dst_high == 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", portrange_ptr->dst_low, portrange_ptr->dst_high);
	}
	else if(portrange_ptr->protocol == 1)
	{
		if(portrange_ptr->dst_high == -1)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--icmp-type %d", portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--icmp-type %d/%d", portrange_ptr->dst_low, portrange_ptr->dst_high);
	}
	else if(portrange_ptr->protocol == 47)
		snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), " ");

	return(0);
}


static int
create_rule_set_proto(struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(rule->portrange_ptr)
	{
		/* tcp */
		if(rule->portrange_ptr->protocol == 6)
			strlcpy(rule->proto, "-p tcp --syn", sizeof(rule->proto));
		/* udp */
		else if(rule->portrange_ptr->protocol == 17)
			strlcpy(rule->proto, "-p udp", sizeof(rule->proto));
		/* icmp */
		else if(rule->portrange_ptr->protocol == 1)
			strlcpy(rule->proto, "-p icmp", sizeof(rule->proto));
		/* gre */
		else if(rule->portrange_ptr->protocol == 47)
			strlcpy(rule->proto, "-p gre", sizeof(rule->proto));
		else
		{
			if(rule->portrange_ptr->protocol > 0 && rule->portrange_ptr->protocol <= 255)
				snprintf(rule->proto, sizeof(rule->proto), "-p %d", rule->portrange_ptr->protocol);
			else
				strlcpy(rule->proto, "", sizeof(rule->proto));
		}
	}
	
	/* if we have no protocol (usefull for service 'all') */
	if((create->service->count == 1) && (create->service->tcp_count+create->service->udp_count+create->service->icmp_count+create->service->gre_count == 0))
	{
		strcpy(rule->proto, "");
		strcpy(rule->temp_src_port, "");
		strcpy(rule->temp_dst_port, "");
	}

	return(0);
}


static int
create_rule_set_action_and_ip(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	char	action[9] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* clear rule->limit because we only use it with log rules and if loglimit > 0 */
	bzero(rule->limit, sizeof(rule->limit));

	/* if we want to log a rule, but havent done it yet: */
	if(create->option.rule_log && !rule->logged)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: log the rule.", __FUNCTION__);

		sscanf(create->action, "%s", action);

		/* set the loglimit */
		if(create->option.loglimit > 0)
			snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d", create->option.loglimit, create->option.logburst);

		/* with prefix */
		if(create->option.rule_logprefix == 1)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: log with user prefix.", __FUNCTION__);

			/* see if we use a loglevel */
			if(strcmp(conf.loglevel, "") == 0)
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s %s '", action, create->option.logprefix);
			else
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s %s ' --log-level %s", action, create->option.logprefix, conf.loglevel);
		}
		else
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: log without user prefix.", __FUNCTION__);

			/* see if we use a loglevel */
			if(strcmp(conf.loglevel, "") == 0)
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s '", action);
			else
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s ' --log-level %s", action, conf.loglevel);
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.ipaddress, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, rule->ipv4_to.netmask,   sizeof(rule->to_netmask));
		
		/* tag logged done */
		rule->logged = 1;
	}
	/* now create the normal rule */
	else if(!rule->normal)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: create the normal rule.", __FUNCTION__);

		sscanf(create->action, "%s", action);

		/* action LOG requires some extra attention */
		if(strncasecmp(create->action, "LOG", 3) == 0)
		{
			/* set the loglimit */
			if(create->option.loglimit > 0)
			{
				snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d",
										create->option.loglimit,
										create->option.logburst);
			}

			snprintf(rule->action, sizeof(rule->action), "%s --log-prefix 'vrmr: %s %s '",
									create->action,
									action,
									create->option.logprefix);
		}
		else
		{
			strlcpy(rule->action, create->action, sizeof(rule->action));
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.ipaddress, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, rule->ipv4_to.netmask,   sizeof(rule->to_netmask));
		
		/* tag normal rule done */
		rule->normal = 1;
	}
	/* if we have a broadcasting protocol and want logging, and haven't logged yet */
	else if((create->service->broadcast == 1) && (create->option.rule_log == 1) && (rule->broadlog == 0))
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: create the log rule for broadcast.", __FUNCTION__);
		
		/* set the loglimit */
		if(create->option.loglimit > 0)
			snprintf(rule->limit, sizeof(rule->limit), "-m limit --limit %d/s --limit-burst %d", create->option.loglimit, create->option.logburst);

		if(create->option.rule_logprefix == 1)
		{
			if(strcmp(conf.loglevel, "") == 0)
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s %s '", create->action, create->option.logprefix);
			else
				snprintf(rule->action, sizeof(rule->action), "LOG --log-prefix 'vrmr: %s %s ' --log-level %s", create->action,  create->option.logprefix, conf.loglevel);
		}
		else
		{
			if(strcmp(conf.loglevel, "") == 0)
				strlcpy(rule->action, "LOG", sizeof(rule->action));
			else
				snprintf(rule->action, sizeof(rule->action), "LOG --log-level %s", conf.loglevel);
		}

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.broadcast, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, "255.255.255.255", sizeof(rule->to_netmask));

		/* tag broadcast logging done */
		rule->broadlog = 1;
	}
	/* broadcasting */
	else if((create->service->broadcast == 1) && (rule->broadcasted == 0))
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "%s: create the broadcast rule.", __FUNCTION__);

		strlcpy(rule->action, create->action, sizeof(rule->action));

		/* set ip and netmask */
		strlcpy(rule->to_ip,      rule->ipv4_to.broadcast, sizeof(rule->to_ip));
		strlcpy(rule->to_netmask, "255.255.255.255", sizeof(rule->to_netmask));

		/* tag broadcasting done */
		rule->broadcasted = 1;
	}

	/* all the above stuff is only for 'to'. */
	strlcpy(rule->from_ip,      rule->ipv4_from.ipaddress, sizeof(rule->from_ip));
	strlcpy(rule->from_netmask, rule->ipv4_from.netmask, sizeof(rule->from_netmask));

	return(0);
}

/*	create_rule

	This fuctions creates the actual rule.

	Returncodes:
		 0: ok
		-1: error
*/
int
create_rule(const int debuglvl, struct RuleCache_ *create)
{
	int			retval = 0;
	int			z = 0;
	int			from,
				to,
				from_nor,
				to_nor;
	int			linescreated = 0,
				createloop = 0;

	struct RuleCreateData_ 	*rule = NULL;

	char			from_mac[19] = "";
	int			from_has_mac = 0;

	struct InterfaceData_	*from_if_ptr = NULL,
				*to_if_ptr = NULL;

	d_list_node		*from_d_node = NULL,
				*to_d_node = NULL;

	d_list_node		*from_if_node = NULL,
				*to_if_node = NULL;

	struct ZoneData_ 	*from_zone_ptr = NULL,
				*to_zone_ptr = NULL,
				*from_member_ptr = NULL,
				*to_member_ptr = NULL;

	int 			from_active = 0,
				to_active = 0,
				active = 0;

	d_list_node		*port_d_node = NULL,
				*listenport_d_node = NULL,
				*remoteport_d_node = NULL;

	struct portdata		*portrange_ptr = NULL;


	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start ** (create->action: %s).", __FUNCTION__, create->action);

	/* alloc the temp rule data */
	if(!(rule = malloc(sizeof(struct RuleCreateData_))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* initialize all values */
	rule->logged = 0;
	rule->normal = 0;
	rule->broadcasted = 0;
	rule->broadlog = 0;
	
	strcpy(rule->action, "");
	strcpy(rule->chain, "");
	strcpy(rule->from_ip, "");
	strcpy(rule->to_ip, "");

	/* here we print the description if we are in bashmode */
	if(bash_out == 1 && create->description != NULL)
	{
		fprintf(stdout, "\n# %s\n", create->description);
		
		free(create->description);
		create->description = NULL;
	}

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////// PROTECT RULE /////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	if(strcasecmp(create->action, "protect") == 0)
	{
	}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////// NORMAL RULE //////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	else
	{

		/* safety checks */
		if(!create->from || !create->to || !create->service)
		{
			(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* clear counters */
		create->iptcount.input = 0,
		create->iptcount.output = 0,
		create->iptcount.forward = 0,
		create->iptcount.preroute = 0,
		create->iptcount.postroute = 0;

		/* at start we asume that both from and to are active */
		from_active = 1;
		to_active = 1;
		active = create->active;

		if(debuglvl >= MEDIUM)
			(void)vrprint.debug("Debug", "create_rule: count: %d, tcp_count: %d, udp_count: %d, icmp_count: %d.", create->service->count, create->service->tcp_count, create->service->udp_count, create->service->icmp_count);

		/* if bash, print comment (if any) */
		if(create->option.rule_comment && bash_out)
		{
			fprintf(stdout, "# %s\n", create->option.comment);
		}
		
		/* copy the helper */
		strlcpy(rule->helper, create->service->helper, sizeof(rule->helper));

////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////// LOOP CONTROL /////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

		/* FROM zone, network, host or group */
		if(create->from->type == TYPE_GROUP)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: groupmembercount: from: %d.", __FUNCTION__, create->from->group_member_count);

			/* set the from-loop counter to the group-member-count */
			from = create->from->group_member_count;

			/* set the first member */
			from_d_node = create->from->GroupList.top;

			/* see if it is not NULL */
			if(from_d_node)
			{
				if(!(from_member_ptr = from_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
					return(-1);
				}

				/* active check */
				if(!from_member_ptr->active)
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "%s: groupmember '%s' is inactive.", __FUNCTION__, from_member_ptr->name);

					from_active = 0;
				}
				else
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "%s: groupmember '%s' is active.", __FUNCTION__, from_member_ptr->name);

					from_active = 1;
				}
			}
			/* safety check */
			else if(!from_d_node && from > 0)
			{
				(void)vrprint.error(-1, "Internal Error", "from: %d, but from_d_node == NULL for group %s (in: %s).", from, create->from->name, __FUNCTION__);
				return(-1);
			}

		}
		/* not a group */
		else
		{
			/* just one from */
			from = 1;
		}

		/* TO zone, network, host or group */
		if(create->to->type == TYPE_GROUP)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "%s: groupmembercount: to: %d.", __FUNCTION__, create->to->group_member_count);

			/* set the to-loop counter to the group-member-count */
			to = create->to->group_member_count;

			/* set the first member */
			to_d_node = create->to->GroupList.top;

			/* see if it is not NULL */
			if(to_d_node)
			{
				if(!(to_member_ptr = to_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
					return(-1);
				}

				if(!to_member_ptr->active)
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "%s: groupmember '%s' is inactive.", __FUNCTION__, to_member_ptr->name);

					to_active = 0;
				}
				else
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "%s: groupmember '%s' is active.", __FUNCTION__, to_member_ptr->name);

					to_active = 1;
				}
			}
			/* safety check */
			else if(!to_d_node && to > 0)
			{
				(void)vrprint.error(-1, "Internal Error", "to: %d, but to_d_node == NULL for group %s (in: %s).", to, create->to->name, __FUNCTION__);
				return(-1);
			}
		}
		/* not a group */
		else
		{
			/* to is one if it's not a group */
			to = 1;
		}

		/*	determine how many times we are going to run the next loops
		
			multiply by the number of interfaces - we need to make sure that interfaces are not counted twice */

		/* from zone */
		if(create->from->type == TYPE_HOST || create->from->type == TYPE_GROUP)
		{
			/* check for counting interfaces twice */
			if((create->to->type == TYPE_NETWORK) && (create->to == create->from->network_parent))
			{
				/* do nothing */
			}
			else
			{
				from = from * create->from->network_parent->interface_count;
			}
		}
		else if(create->from->type == TYPE_NETWORK)
		{
			from = from * create->from->interface_count;
		}
		else
		{
			(void)vrprint.error(-1, "Internal Error", "invalid from type %d (in: %s).", create->from->type, __FUNCTION__);
			return(-1);
		}

		/* to zone */
		if(create->to->type == TYPE_HOST || create->to->type == TYPE_GROUP)
		{
			/* check for counting twice: compare from parentnetwork with our to parentnetwork */
			if((create->from->type == TYPE_HOST || create->from->type == TYPE_GROUP) && create->from->network_parent == create->to->network_parent)
			{
				/* do nothing */
			}
			/* check for twice:  compare from network with our parentnetwork */
			else if(create->from->type == TYPE_NETWORK && create->from == create->to->network_parent)
			{
				/* do nothing */
			}
			/* check for counting twice: else multiply */
			else
			{
				to = to * create->to->network_parent->interface_count;
			}
		}
		else if(create->to->type == TYPE_NETWORK)
		{
			/* check for counting twice: compare from network with our network */
			if(create->from->type == TYPE_NETWORK && create->from == create->to)
			{
				/* do nothing */
			}
			/* check for counting twice:  else multiply */
			else
			{
				to = to * create->to->interface_count;
			}
		}
		else
		{
			(void)vrprint.error(-1, "Internal Error", "invalid to type %d (in: %s).", create->from->type, __FUNCTION__);
			return(-1);
		}


		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "create_rule: loop: from: %d, to: %d.", from, to);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////// MAIN LOOP ///////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////
// BEGIN LOOP INTERFACES
////

	/*
	 * ==== FROM LOOP ====
	 */
		/* run this loop 'from' times */
		for(from_nor = 0;  from_nor < from; from_nor++)
		{
			/* network or host, group */
			if(create->from->type == TYPE_NETWORK)
				from_zone_ptr = create->from;
			else
				from_zone_ptr = create->from->network_parent;

			if(!from_zone_ptr)
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			/* get the ifac ptr */
			if(!from_if_node)
				from_if_node = from_zone_ptr->InterfaceList.top;
			else
				from_if_node = from_if_node->next;

			/* safety check */
			if(!from_if_node)
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return -1;
			}

			/* reset from_zone_ptr */
			from_zone_ptr = NULL;

			if(!(from_if_ptr = from_if_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
				return(-1);
			}

			/*	set the interface
			
				if the device is virtual we don't want it in our
				iptables commands
			*/
			if(!from_if_ptr->device_virtual)
				strlcpy(rule->from_int, from_if_ptr->device, sizeof(rule->from_int));
			else
				bzero(rule->from_int, sizeof(rule->from_int));

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "create_rule: rule->from_int: '%s', virtual? %s.", rule->from_int, from_if_ptr->device_virtual ? "Yes" : "No");

			/*	if the interface is not active we set the to zone
				inactive. But only if its already active. */
			if(active)
				active = from_if_ptr->active;

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "create_rule: active = %d (because of from_if_ptr->active).", active);

			/* if were at the end of the ifac list, set to null */
			if(d_list_node_is_bot(debuglvl, from_if_node))
				from_if_node = NULL;

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "create_rule: create->from.group_member_count: %d.", create->from->group_member_count);


		/*
		 * ==== TO LOOP ====
		 */
			/* run this loop 'to' times */
			for(to_nor = 0; to_nor < to; to_nor++)
			{
				/* set to zone pointer to the network of this zone */
				if(create->to->type == TYPE_NETWORK)
					to_zone_ptr = create->to;
				else
					to_zone_ptr = create->to->network_parent;

				/* check */
				if(!to_zone_ptr)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
					return(-1);
				}
				
				/* get the ifac node */
				if(!to_if_node)
					to_if_node = to_zone_ptr->InterfaceList.top;
				else
					to_if_node = to_if_node->next;

				/* safety check */
				if(!to_if_node)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return -1;
				}
				
				/* reset to_zone_ptr */
				to_zone_ptr = NULL;
				
				/* get the interface */
				if(!(to_if_ptr = to_if_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				/*	set the interface

					if the device is virtual we don't want it in our
					iptables commands
				*/
				if(!to_if_ptr->device_virtual)
					strlcpy(rule->to_int, to_if_ptr->device, sizeof(rule->to_int));
				else
					bzero(rule->to_int, sizeof(rule->to_int));

				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "create_rule: rule->to_int: '%s', virtual? %s.", rule->to_int, to_if_ptr->device_virtual ? "Yes" : "No");

				/* if the interface is not active we set the rule inactive. */
				if(active)
				{
					active = to_if_ptr->active;

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: active = %d (because of to_if_ptr->active).", __FUNCTION__, active);
				}
				else
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: active = %d (so not checking the to_int).", __FUNCTION__, active);
				}

				/* if were at the end of the ifac list, set to null so we can rewind at the next run */
				if(d_list_node_is_bot(debuglvl, to_if_node))
					to_if_node = NULL;


			/*
			 * ==== ACTIVE CHECK ====
			 */
				if(from_active && to_active && active)
				{
					/* SET PORTS */
					if(create->service->PortrangeList.len > 0)
					{
						/* top of the list */
						if(!(port_d_node = create->service->PortrangeList.top))
						{
							(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
							return -1;
						}

						/* listenport option */
						if(create->option.listenport == 1)
							listenport_d_node = create->option.ListenportList.top;
						else
							listenport_d_node = NULL;

						/* remoteport option */
						if(create->option.remoteport == 1)
							remoteport_d_node = create->option.RemoteportList.top;
						else
							remoteport_d_node = NULL;
					}
					else
					{
						/* if there are no portranges set port_d_node to NULL */
						port_d_node = NULL;
					}


					for(linescreated = 0; create->service->count > linescreated; linescreated++)
					{
						/* start with creating 1 rule */
						createloop = 1;

						/* set ports */
						if(port_d_node)
						{
							/* get the current portrange */
							if(!(portrange_ptr = port_d_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
								return -1;
							}
							/* set portrange_ptr into rule */
							rule->portrange_ptr = portrange_ptr;

							/* set rule->listenport_ptr */
							if(create->option.listenport && listenport_d_node)
								rule->listenport_ptr = listenport_d_node->data;
							else
								rule->listenport_ptr = NULL;

							/* set rule->remoteport_ptr */
							if(create->option.remoteport && remoteport_d_node)
								rule->remoteport_ptr = remoteport_d_node->data;
							else
								rule->remoteport_ptr = NULL;

							/* now load the ports to the rule struct */
							if(create_rule_set_ports(rule, portrange_ptr) < 0)
							{
								(void)vrprint.error(-1, "Internal Error", "setting up the ports failed (in: %s).", __FUNCTION__);
								retval = -1;
							}
						}
						else
						{
							/* no portrange */
							rule->portrange_ptr = NULL;
						}

						if(debuglvl >= MEDIUM)
						{
							(void)vrprint.debug("Debug", "create_rule: create->service.count: %d, linescreated: %d", create->service->count, linescreated);
							(void)vrprint.debug("Debug", "create_rule: create->service.tcp_count: %d, create->service.udp_count: %d, create->service.icmp_count: %d, create->service.gre_count: %d", create->service->tcp_count, create->service->udp_count, create->service->icmp_count, create->service->gre_count);
						}

						/* set protocol */
						if(create_rule_set_proto(rule, create) < 0)
						{
							(void)vrprint.error(-1, "Error", "create_rule_set_proto failed.");
							retval = -1;
						}

					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					/////////////////////////////// okay, we are going to create something////////////////////////////////////////////////////////
					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

						/*	set the ipaddress and netmask and mac
							
							FROM

							first host
						*/
						from_has_mac = 0;
						if(create->from->type == TYPE_HOST)
						{
							strlcpy(rule->ipv4_from.ipaddress, create->from->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   create->from->ipv4.netmask,   sizeof(rule->ipv4_from.netmask));
							if(create->from->has_mac)
							{
								from_has_mac = 1;
								strlcpy(from_mac, create->from->mac, sizeof(from_mac));
							}
						}
						/* network */
						else if(create->from->type == TYPE_NETWORK && create->from_firewall == 0)
						{
							strlcpy(rule->ipv4_from.ipaddress, create->from->ipv4.network, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   create->from->ipv4.netmask, sizeof(rule->ipv4_from.netmask));
						}
						/* firewall (this host) */
						else if(create->from->type == TYPE_NETWORK && create->from_firewall == 1)
						{
							strlcpy(rule->ipv4_from.ipaddress, from_if_ptr->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   "255.255.255.255",           sizeof(rule->ipv4_from.netmask));
						}
						/* finally group */
						else if(create->from->type == TYPE_GROUP)
						{
							strlcpy(rule->ipv4_from.ipaddress, from_member_ptr->ipv4.ipaddress, sizeof(rule->ipv4_from.ipaddress));
							strlcpy(rule->ipv4_from.netmask,   from_member_ptr->ipv4.netmask,   sizeof(rule->ipv4_from.netmask));
							if(from_member_ptr->has_mac)
							{
								from_has_mac = 1;
								strlcpy(from_mac, from_member_ptr->mac, sizeof(from_mac));
							}
						}
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: (from) ipaddress: %s, netmask: %s (pre createloop)", __FUNCTION__, rule->ipv4_from.ipaddress, rule->ipv4_from.netmask);

						/*	TO
							
							first host
						*/
						if(create->to->type == TYPE_HOST)
						{
							strlcpy(rule->ipv4_to.ipaddress, create->to->ipv4.ipaddress, sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   create->to->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
							/* to has no mac, because it is not supported by iptables */
						}
						/* network */
						else if(create->to->type == TYPE_NETWORK && create->to_firewall == 0)
						{
							strlcpy(rule->ipv4_to.ipaddress, create->to->ipv4.network,   sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   create->to->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
						}
						/* firewall (this host) */
						else if(create->to->type == TYPE_NETWORK && create->to_firewall == 1)
						{
							strlcpy(rule->ipv4_to.ipaddress, to_if_ptr->ipv4.ipaddress,  sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   "255.255.255.255",          sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast, sizeof(rule->ipv4_to.broadcast));
						}
						else if(create->to->type == TYPE_GROUP)
						{
							strlcpy(rule->ipv4_to.ipaddress, to_member_ptr->ipv4.ipaddress, sizeof(rule->ipv4_to.ipaddress));
							strlcpy(rule->ipv4_to.netmask,   to_member_ptr->ipv4.netmask,   sizeof(rule->ipv4_to.netmask));
							strlcpy(rule->ipv4_to.broadcast, create->to->ipv4.broadcast,    sizeof(rule->ipv4_to.broadcast));
							/* to has no mac, because it is not supported by iptables */
						}
						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: ( to ) ipaddress: %s, netmask: %s (pre createloop)", __FUNCTION__, rule->ipv4_to.ipaddress, rule->ipv4_to.netmask);


						/* some special actions for a few actions */
						if(create->ruletype == RT_SNAT)
						{
							/* copy the ipaddress of the to-interface to rule->serverip so snat can use it */
							if(!to_if_ptr)
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer: to_if_ptr (in: %s).", __FUNCTION__);
								return(-1);
							}
								
							snprintf(rule->serverip, sizeof(rule->serverip), "%s", to_if_ptr->ipv4.ipaddress);
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: SNAT: rule->serverip = '%s' (interface: %s).", __FUNCTION__, rule->serverip, to_if_ptr->name);
						}
						else if(create->ruletype == RT_PORTFW)
						{
							/* copy the ipaddress of the from-interface to rule->serverip so portfw can use it */
							if(!from_if_ptr)
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer: from_if_ptr (in: %s).", __FUNCTION__);
								return(-1);
							}
									
							snprintf(rule->serverip, sizeof(rule->serverip), "%s", from_if_ptr->ipv4.ipaddress);
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: PORTFW: rule->serverip = '%s' (interface: %s).", __FUNCTION__, rule->serverip, from_if_ptr->name);
						}


						/* add mac-address if we happen to know it, only 'from' is supported by iptables */
						if(from_has_mac)
							snprintf(rule->from_mac, sizeof(rule->from_mac), "-m mac --mac-source %s", from_mac);
						else
							strcpy(rule->from_mac, "");

						if(portrange_ptr && portrange_ptr->protocol == 17 && create->service->broadcast == 1)
							createloop = createloop * 2;
						
						/* if we log we need to create twice as many rules */
						if(create->option.rule_log)
							createloop = createloop * 2;

						if(debuglvl >= HIGH)
							(void)vrprint.debug("Debug", "%s: createloop = %d", __FUNCTION__, createloop);

						/* reset counters */
						rule->logged = 0, rule->normal = 0, rule->broadlog = 0, rule->broadcasted = 0;

					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
					////////////////////////////////////////// enter createloop //////////////////////////////////////////////////////
					//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
						for(z = 0; z < createloop; z++)
						{
							if(create_rule_set_action_and_ip(debuglvl, rule, create) < 0)
							{
								(void)vrprint.error(-1, "Error", "create_rule_set_action_and_ip failed.");
								retval = -1;
							}

							/* rule count */
							ipt_rulecount++;

							if(debuglvl >= HIGH)
							{
								(void)vrprint.debug("Debug", "%s: (from) ipaddress: %s, netmask: %s (in createloop)", __FUNCTION__, rule->from_ip, rule->from_netmask);
								(void)vrprint.debug("Debug", "%s: ( to ) ipaddress: %s, netmask: %s (in createloop)", __FUNCTION__, rule->to_ip, rule->to_netmask);
							}

							/* here we create the iptables commands */
							if(debuglvl >= HIGH)
								(void)vrprint.debug("Debug", "%s: action: %s, chain: %s, fromip: %s, toip: %s, ruletype: %d", __FUNCTION__, rule->action, rule->chain, rule->from_ip, rule->to_ip, create->ruletype);

							/* normal input rules */
							if(create->ruletype == RT_INPUT)
							{
								if(create_rule_input(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating input rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* normal output rules */
							else if(create->ruletype == RT_OUTPUT)
							{
								if(create_rule_output(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating output rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* normal forward rules */
							else if(create->ruletype == RT_FORWARD)
							{
								if(create_rule_forward(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating forward rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* masq rules */
							else if(create->ruletype == RT_MASQ)
							{
								if(create_rule_masq(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating masq rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* snat rules */
							else if(create->ruletype == RT_SNAT)
							{
								if(create_rule_snat(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating snat rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							/* portforward rules */
							else if(create->ruletype == RT_PORTFW)
							{
								if(create_rule_portfw(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating portfw rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							else if(create->ruletype == RT_REDIRECT)
							{
								if(create_rule_redirect(debuglvl, rule, create) < 0)
								{
									(void)vrprint.error(-1, "Error", "creating redirect rule failed (in: %s).", __FUNCTION__);
									retval = -1;
								}
							}
							else
							{
								(void)vrprint.error(-1, "Internal Error", "unknown ruletype '%d' (in: %s).", create->ruletype, __FUNCTION__);
								return(-1);
							}
						} /* end createloop */

						/* get the next service/portrange */
						if(create->service->PortrangeList.len > 0)
						{
							if(d_list_node_is_bot(debuglvl, port_d_node))
							{
								/* rewind */
								port_d_node = create->service->PortrangeList.top;
							}
							else
							{
								/* next */
								port_d_node = port_d_node->next;
							}

							/* listport option */
							if(create->option.listenport == 1)
							{
								if(listenport_d_node->next != NULL)
									listenport_d_node = listenport_d_node->next;
							}

							/* remoteport  option */
							if(create->option.remoteport == 1)
							{
								if(remoteport_d_node->next != NULL)
									remoteport_d_node = remoteport_d_node->next;
							}
						}
					}
				} /* end if active */
				else
				{
					if(debuglvl >= LOW)
						(void)vrprint.debug("Debug", "%s: rule not created, not active.", __FUNCTION__);
				}

				/* set the next group member TO GROUP */
				if(create->to->type == TYPE_GROUP)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "%s: TO: set the next (or prev) group member.", __FUNCTION__);

					/* get the new to_d_node

					   if we have reached the bot of the list, start all over again
					 */
					if(!d_list_node_is_bot(debuglvl, to_d_node))
						to_d_node = to_d_node->next;
					else
						to_d_node = create->to->GroupList.top;

					if(to_d_node)
					{
						if(!(to_member_ptr = to_d_node->data))
						{
							(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
							return(-1);
						}

						if(to_member_ptr->active == 0)
						{
							if(debuglvl >= MEDIUM)
								(void)vrprint.debug("Debug", "%s: group member '%s' is inactive", __FUNCTION__, to_member_ptr->name);
							
							to_active = 0;
						}
						else
						{
							if(debuglvl >= MEDIUM)
								(void)vrprint.debug("Debug", "%s: group member '%s' is active", __FUNCTION__, to_member_ptr->name);
							
							to_active = 1;
						}
					}
				}
			}

			/* set the next group member of the FROM GROUP */
			if(create->from->type == TYPE_GROUP)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "%s: FROM: set the next (or prev) group member.", __FUNCTION__);
				
				/* get the new from_d_node

				   if we have reached the bot of the list, start all over again
				 */
				if(!d_list_node_is_bot(debuglvl, from_d_node))
					from_d_node = from_d_node->next;
				else
					from_d_node = create->from->GroupList.top;

				if(from_d_node)
				{
					if(!(from_member_ptr = from_d_node->data))
					{
						(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
						return(-1);
					}

					if(from_member_ptr->active == 0)
					{
						if(debuglvl >= MEDIUM)
							(void)vrprint.debug("Debug", "%s: group member '%s' is inactive", __FUNCTION__, from_member_ptr->name);

						from_active = 0;
					}
					else
					{
						if(debuglvl >= MEDIUM)
							(void)vrprint.debug("Debug", "%s: group member '%s' is active", __FUNCTION__, from_member_ptr->name);

						from_active = 1;
					}
				}
			}
		} // end normal rule
//
//// END LOOP INTERFACES
//
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug("Debug", "%s: rules created: (i: %2d, o: %2d, f: %2d, e: %2d, p: %2d)", __FUNCTION__, create->iptcount.input, create->iptcount.output, create->iptcount.forward, create->iptcount.preroute, create->iptcount.postroute);
	} //end if protect

	/* free the temp data */
	free(rule);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** end **, return=%d", __FUNCTION__, retval);
	
	return(retval);
}


/*	update_synlimit_rules

	Creates/updates the the rules in the SYNLIMIT chain.
*/
int
update_synlimit_rules(const int debuglvl)
{
	int			retval = 0,
				result = 0;
	char			cmd[MAX_PIPE_COMMAND] = "";


	/* first flush the chain */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush SYNLIMIT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;

	/* create the return rule */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A SYNLIMIT -m limit --limit %d/s --limit-burst %d -j RETURN", conf.iptables_location, conf.syn_limit, conf.syn_limit_burst);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;

	/* the log rule */
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A SYNLIMIT -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP SYNLIMIT reach.'", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A SYNLIMIT -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP SYNLIMIT reach.' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;

	/* and finally the drop rule */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A SYNLIMIT -j DROP", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;

	return(retval);
}



/* pre_rules

	Cleanup
		chains
	Sets up
		default policies
		connection tracking
		portscan detection
*/
int
pre_rules(const int debuglvl, d_list *interfaces_list)
{
	int			retval = 0,
				result = 0;
	char			cmd[MAX_PIPE_COMMAND] = "";
	d_list_node		*d_node = NULL;
	struct InterfaceData_ 	*iface_ptr = NULL;


	/* safety */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	/*
		first flush the chains
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Flushing chains...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Flushing chains...");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval = -1;


	/*
		allow local loopback
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Allowing local loopback...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Allowing local loopback...");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -i lo -j ACCEPT", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -o lo -j ACCEPT", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;


	/*
		create an empty rule in INPUT, OUTPUT and FORWARD, which count
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Creating interface counters...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Creating interface counters...");

	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		iface_ptr = d_node->data;

		/* if the interface active? */
		if(iface_ptr->active)
		{
			/* does the interface have an ipaddress? */
			if(strcmp(iface_ptr->ipv4.ipaddress, "") != 0)
			{
				/* Check for empty device string and virtual device. */
				if(strcmp(iface_ptr->device, "") != 0 && !iface_ptr->device_virtual)
				{
					/*
						first in the input chain
					*/
					snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -i %s", conf.iptables_location, iface_ptr->device);
					result = pipe_command(debuglvl, cmd, 0);
					if(result < 0)
						retval=-1;
					else
						ipt_input_count++;


					/*
						then in the output chain
					*/
					snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -o %s", conf.iptables_location, iface_ptr->device);
					result = pipe_command(debuglvl, cmd, 0);
					if(result < 0)
						retval=-1;
					else
						ipt_output_count++;

					/*
						then in the forward chain, in
					*/
					snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -i %s", conf.iptables_location, iface_ptr->device);
					result = pipe_command(debuglvl, cmd, 0);
					if(result < 0)
						retval=-1;
					else
						ipt_forward_count++;


					/*
						then in the forward chain, out
					*/
					snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -o %s", conf.iptables_location, iface_ptr->device);
					result = pipe_command(debuglvl, cmd, 0);
					if(result < 0)
						retval=-1;
					else
						ipt_forward_count++;
				}
			}
		}
	}


	/*
		set default policies to DROP
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Setting default policies...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Setting default policies...");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy INPUT DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy OUTPUT DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy FORWARD DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;


	/*
		stealthscan protection
	*/
	if(bash_out==1) fprintf(stdout, "\n# Setting up stealth scan protection...\n");
	if((debuglvl >= LOW) || (bash_out == 1)) (void)vrprint.debug("Debug", "pre_rules: Setting up stealth scan protection...");
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ALL NONE -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe ALL '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ALL NONE -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe ALL ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ALL NONE -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,FIN SYN,FIN -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe SYN-FIN '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,FIN SYN,FIN -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe SYN-FIN ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,FIN SYN,FIN -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,RST SYN,RST -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe SYN-RST '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,RST SYN,RST -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe SYN-RST ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags SYN,RST SYN,RST -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags FIN,RST FIN,RST -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe FIN-RST '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags FIN,RST FIN,RST -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe FIN-RST ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags FIN,RST FIN,RST -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,FIN FIN -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe FIN '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,FIN FIN -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe FIN ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,FIN FIN -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,PSH PSH -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe PSH '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,PSH PSH -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe PSH ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,PSH PSH -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,URG URG -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe URG '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,URG URG -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP probe URG ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --tcp-flags ACK,URG URG -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp ! --syn -m state --state NEW -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP no SYN '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp ! --syn -m state --state NEW -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP no SYN ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp ! --syn -m state --state NEW -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	/*
		Fragmented packets
	*/
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -f -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP FRAG '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -f -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP FRAG ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -f -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	/*
		set up connectiontracking for QUEUE
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Setting up connection-tracking for QUEUE targets...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Setting up connection-tracking for QUEUE targets...");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m mark --mark 3 -m state --state ESTABLISHED,RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;


	/*
		set up connectiontracking for QUEUE with IptState for Snort_inline.
	*/
	if (bash_out==1) fprintf(stdout, "\n# Setting up connection-tracking for QUEUE with IptState for Snort_inline.\n");
	if (debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Setting up connection-tracking for QUEUE with IptState for Snort_inline..");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m mark --mark 1 -m state --state RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m mark --mark 2 -m state --state ESTABLISHED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m mark --mark 1 -m state --state RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m mark --mark 2 -m state --state ESTABLISHED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m mark --mark 1 -m state --state RELATED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m mark --mark 2 -m state --state ESTABLISHED -j QUEUE", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;


	/*
		set up connectiontracking
	*/
	if(bash_out == 1) fprintf(stdout, "\n# Setting up connection-tracking...\n");
	if(debuglvl >= LOW) (void)vrprint.debug("Debug", "pre_rules: Setting up connection-tracking...");
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m state --state ESTABLISHED,RELATED -j ACCEPT", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m state --state ESTABLISHED,RELATED -j ACCEPT", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m state --state ESTABLISHED,RELATED -j ACCEPT", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;


	/*
		invalid input
	*/
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP in INVALID '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP in INVALID ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -m state --state INVALID -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_input_count++;

	/*
		invalid output
	*/
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP out INVALID '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP out INVALID ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -m state --state INVALID -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_output_count++;

	/*
		invalid forward
	*/
	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP fw INVALID '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m state --state INVALID -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP fw INVALID ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -m state --state INVALID -j DROP", conf.iptables_location);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0) retval=-1;
	ipt_forward_count++;


	/*
		Setup Block lists
	*/
	if(bash_out == 1)
		fprintf(stdout, "\n# Setting up blocklist...\n");
	if(debuglvl >= LOW)
		(void)vrprint.debug("Debug", "pre_rules: Setting up blocklist...");

	/* create the chain and insert it into input, output and forward.
	
	   NOTE: we ignore the returncode and want no output (although we get some
	   in the errorlog) because if we start vuurmuur when a ruleset is already in
	   place, the chain will exist and iptables will complain. */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -N BLOCKLIST 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -j BLOCKLIST", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -j BLOCKLIST", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_output_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -j BLOCKLIST", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_forward_count++;

	/* create the BLOCK action
	
	   NOTE: see BLOCKLIST creation. */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -N BLOCK 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	if(strcmp(conf.loglevel, "") == 0)
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A BLOCK -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP BLOCKED '", conf.iptables_location);
	else
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A BLOCK -m limit --limit 1/s --limit-burst 2 -j LOG --log-prefix 'vrmr: DROP BLOCKED ' --log-level %s", conf.iptables_location, conf.loglevel);
	result=pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A BLOCK -j DROP", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;


	/*
		syn-flooding protection
	*/
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -N SYNLIMIT 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	/* create the rules */
	if(update_synlimit_rules(debuglvl) < 0)
		retval = -1;

	/* create the chain and insert it into input, output and forward.
	
	   NOTE: we ignore the returncode and want no output (although we get some
	   in the errorlog) because if we start vuurmuur when a ruleset is already in
	   place, the chain will exist and iptables will complain. */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -p tcp --syn -j SYNLIMIT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -p tcp --syn -j SYNLIMIT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_output_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -p tcp --syn -j SYNLIMIT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_forward_count++;


	/* anti spoof rules */
	if(bash_out == 1)
		fprintf(stdout, "\n# Setting up anti-spoofing rules...\n");

	/* create the chain and insert it into input, output and forward.
	
	   NOTE: we ignore the returncode and want no output (although we get some
	   in the errorlog) because if we start vuurmuur when a ruleset is already in
	   place, the chain will exist and iptables will complain. */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -N ANTISPOOF 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -j ANTISPOOF", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_input_count++;

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -j ANTISPOOF", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	ipt_forward_count++;

	return(retval);
}

/*	post_rules

	Enables logging in the INPUT, OUTPUT and FORWARD chains.
	
	If forward_rules == 1, then /proc/sys/net/ipv4/ip_forward will be enabled. Otherwise it will be disabled.

	Returncode:
		 0: ok
		-1: error
*/
int
post_rules(const int debuglvl, int forward_rules)
{
	int	retval=0,
		result=0;
	char	cmd[MAX_PIPE_COMMAND];
	char	limit[42] = "";


	/* do we want to log the default policy? */
	if(conf.log_policy)
	{
		/* see if we want to limit the logging of the default policy */
		if(conf.log_policy_limit > 0)
		{
			snprintf(limit, sizeof(limit), "-m limit --limit %d/s --limit-burst %d",
									conf.log_policy_limit,
									conf.log_policy_burst);
		}

		/* enable logging for all packets which don't have a rule */
		if(bash_out == 1) fprintf(stdout, "\n# Enabling logging...\n");
		if(debuglvl >= LOW) (void)vrprint.debug("Debug", "post_rules: Enabling logging...");

		if(strcmp(conf.loglevel, "") == 0)
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -j LOG --log-prefix 'vrmr: DROP in policy ' %s",
									conf.iptables_location,
									limit);
		}
		else
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A INPUT -j LOG --log-prefix 'vrmr: DROP in policy ' --log-level %s %s",
									conf.iptables_location,
									conf.loglevel,
									limit);
		}
		result = pipe_command(debuglvl, cmd, 0);
		if(result < 0)
			retval = -1;

		if(strcmp(conf.loglevel, "") == 0)
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -j LOG --log-prefix 'vrmr: DROP out policy ' %s",
									conf.iptables_location,
									limit);
		}
		else
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A OUTPUT -j LOG --log-prefix 'vrmr: DROP out policy ' --log-level %s %s",
									conf.iptables_location,
									conf.loglevel,
									limit);
		}
		result = pipe_command(debuglvl, cmd, 0);
		if(result < 0)
			retval = -1;

		if(strcmp(conf.loglevel, "") == 0)
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -j LOG --log-prefix 'vrmr: DROP fw policy ' %s",
									conf.iptables_location,
									limit);
		}
		else
		{
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -A FORWARD -j LOG --log-prefix 'vrmr: DROP fw policy ' --log-level %s %s",
									conf.iptables_location,
									conf.loglevel,
									limit);
		}
		result = pipe_command(debuglvl, cmd, 0);
		if(result < 0)
			retval = -1;
	}

	/* enable or disable ip-forwarding */
	if(forward_rules)
	{
		(void)vrprint.info("Info", "Created forward-rules: enabling ip-forwarding.");

		result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/ip_forward", 1, NULL);
		if(result != 0)
		{
			/* if it fails, we dont really care, its not fatal */
			(void)vrprint.error(-1, "Error", "enabling ip-forwarding failed.");
		}
	}
	else
	{
		(void)vrprint.info("Info", "Created no forward-rules: disabling ip-forwarding.");
		
		result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/ip_forward", 0, NULL);
		if(result != 0)
		{
			/* if it fails, we dont really care, its not fatal */
			(void)vrprint.error(-1, "Error", "enabling ip-forwarding failed.");
		}
	}

	return(retval);
}

/*	remove_rule

	Function for removing iptables rules.

	Returncodes:
		 0: ok
		-1: error
*/
int
remove_rule(const int debuglvl, int chaintype, int first_ipt_rule, int rules)
{
	int	retval=0,
		i;
	char	cmd[MAX_PIPE_COMMAND];
	char	chain[sizeof(RuleCache.chain)];

	if(debuglvl >= HIGH)
	{
		(void)vrprint.debug("Debug", "remove_rule: ** start **");
		(void)vrprint.debug("Debug", "chain: %d, ipt: %d, rules: %d", chaintype, first_ipt_rule, rules);
	}

	/* determine from which chain to delete */
	if(chaintype == RT_INPUT)
	{
		strlcpy(chain, "-D INPUT", sizeof(chain));
	}
	else if(chaintype == RT_OUTPUT)
	{
		strlcpy(chain, "-D OUTPUT", sizeof(chain));
	}
	else if(chaintype == RT_FORWARD)
	{
		strlcpy(chain, "-D FORWARD", sizeof(chain));
	}
	else if(chaintype == RT_PORTFW || chaintype == RT_REDIRECT)
	{
		strlcpy(chain, "-t nat -D PREROUTING", sizeof(chain));
	}
	else if(chaintype == RT_MASQ || chaintype == RT_SNAT)
	{
		strlcpy(chain, "-t nat -D POSTROUTING", sizeof(chain));
	}
	else
	{
		(void)vrprint.error(-1, "Error", "unknown chaintype %d (remove_rule).", chaintype);
		return(-1);
	}

	for(i = 0; i < rules; i++)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug("Debug", "remove_rule: cmd: %s %s %d", conf.iptables_location, chain, first_ipt_rule);

		snprintf(cmd, sizeof(cmd), "%s %s %d", conf.iptables_location, chain, first_ipt_rule);
		if(pipe_command(debuglvl, cmd, 0) != 0)
		{
			(void)vrprint.error(-1, "Error", "remove_rule: pipe error. This command failed: '%s'.", cmd);
			return(-1);
		}
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "remove_rule: ** end **, return=%d", retval);

	return(retval);
}


/*	create_rule_input

	Creates a rule in the input chain.

*/
int
create_rule_input(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	int	retval=0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "";
	char	reverse_input_device[sizeof(rule->from_int) + 3] = "";


	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);


	/* we use rule->to_int here because it describes the interface of the firewall */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A %s %s %s -s %s/%s %s -d %s/%s %s %s %s -m state --state NEW -j %s",
						conf.iptables_location,
						rule->chain,
						input_device,
						rule->proto,
						rule->from_ip,
						rule->from_netmask,
						rule->temp_src_port,
						rule->to_ip,
						rule->to_netmask,
						rule->temp_dst_port,
						rule->from_mac,
						rule->limit,
						rule->action);
	if(pipe_command(debuglvl, cmd, 0) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating 'forward' rule failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	create->iptcount.input++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_input_device, input_device, sizeof(reverse_input_device));
		reverse_input_device[1] = 'o';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								input_device,
								rule->proto,
								rule->from_ip,
								rule->from_netmask,
								rule->temp_src_port,
								rule->to_ip,
								rule->to_netmask,
								rule->temp_dst_port,
								rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* related */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								reverse_input_device,
								rule->proto,
								rule->to_ip,
								rule->to_netmask,
								temp_dst_port,
								rule->from_ip,
								rule->from_netmask,
								temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
	
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
			
			/* establised */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->temp_src_port,
								rule->to_ip,
								rule->to_netmask,
								rule->temp_dst_port,
								rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								temp_dst_port,
								rule->from_ip,
								rule->from_netmask,
								temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
		
			if(strcmp(rule->helper, "") != 0)
			{
				/* RELATED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			
				/* ESTABLISHED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
		
			/* new, related, established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							input_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* related, established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}

			if(strcmp(rule->helper, "") != 0)
			{
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								input_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for input rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
		}
	}

	return(retval);
}

int
create_rule_output(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	output_device[sizeof(rule->to_int) + 3] = "";
	char	reverse_output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);

	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A %s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW -j %s",
					conf.iptables_location,
					rule->chain,
					output_device,
					rule->proto,
					rule->from_ip,
					rule->from_netmask,
					rule->temp_src_port,
					rule->to_ip,
					rule->to_netmask,
					rule->temp_dst_port,
					rule->limit,		/* log limit */
					rule->action);
	if(pipe_command(debuglvl, cmd, 0) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating 'output' rule failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* update rule counter */
	create->iptcount.output++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_output_device, output_device, sizeof(reverse_output_device));
		reverse_output_device[1] = 'i';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
							conf.iptables_location,
							output_device,
							rule->proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* REVERSE! related */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
							conf.iptables_location,
							reverse_output_device,
							rule->proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
		
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
		
			/* established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							conf.iptables_location,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* REVERSE! */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							conf.iptables_location,
							reverse_output_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}

			if(strcmp(rule->helper, "") != 0)
			{
				/* RELATED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				
				/* ESTABLISHED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));
		
			/* new, related, established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* REVERSE! related, established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							reverse_output_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}

			if(strcmp(rule->helper, "") != 0)
			{
				/* RELATED,ESTABLISHED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A OUTPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A INPUT %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								reverse_output_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for output rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
		}
	}

	return(retval);
}

int
create_rule_forward(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND] = "",
		stripped_proto[16] = "";				/* proto stripped from --syn */
	char	temp_src_port[sizeof(rule->temp_src_port)] = "",
		temp_dst_port[sizeof(rule->temp_dst_port)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "",
		output_device[sizeof(rule->to_int) + 3] = "";
	char	reverse_input_device[sizeof(rule->from_int) + 3] = "",
		reverse_output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);


	/* create the rule */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -A %s %s %s %s -s %s/%s %s -d %s/%s %s %s %s -m state --state NEW -j %s",
					conf.iptables_location,
					rule->chain,
					input_device,
					output_device,
					rule->proto,
					rule->from_ip,
					rule->from_netmask,
					rule->temp_src_port,
					rule->to_ip,
					rule->to_netmask,
					rule->temp_dst_port,
					rule->from_mac,
					rule->limit,
					rule->action);
	if(pipe_command(debuglvl, cmd, 0) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating 'forward' rule failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	create->iptcount.forward++;

	/* if the target is QUEUE we mark the traffic */
	if(strcasecmp(rule->action, "QUEUE") == 0)
	{
		/* swap source ports and dest ports for the rules in the opposite direction */
		strlcpy(temp_src_port, rule->temp_src_port, sizeof(temp_src_port));
		temp_src_port[2] = 'd';
		strlcpy(temp_dst_port, rule->temp_dst_port, sizeof(temp_dst_port));
		temp_dst_port[2] = 's';

		/* swap devices */
		strlcpy(reverse_input_device, input_device, sizeof(reverse_input_device));
		reverse_input_device[1] = 'o';
		strlcpy(reverse_output_device, output_device, sizeof(reverse_output_device));
		reverse_output_device[1] = 'i';

		/* create mangle rules for markiptstate */
		if(create->option.markiptstate)
		{
			/* new, related. from -> dst */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED -j MARK --set-mark 1",
							conf.iptables_location,
							input_device,
							output_device,
							rule->proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}

			/* REVERSE! related. dst -> from  */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED -j MARK --set-mark 1",
							conf.iptables_location,
							reverse_output_device,
							reverse_input_device,
							rule->proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
	
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* established. both ways */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							conf.iptables_location,
							input_device,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* REVERSE! */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s -m state --state ESTABLISHED -j MARK --set-mark 2",
							conf.iptables_location,
							reverse_output_device,
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
		
			if(strcmp(rule->helper, "") != 0)
			{
				/* RELATED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state RELATED -j MARK --set-mark 1",
								conf.iptables_location,
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}

				/* ESTABLISHED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED -j MARK --set-mark 2",
								conf.iptables_location,
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}

		}
		/* create mangle rules for 'normal' QUEUE */
		else
		{
			/* we dont want --syn in the next rules */
			if(strcmp(rule->proto, "-p tcp --syn") == 0)
				strlcpy(stripped_proto, "-p tcp", sizeof(stripped_proto));
			else
				strlcpy(stripped_proto, rule->proto, sizeof(stripped_proto));

			/* new,related,established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s %s -m state --state NEW,RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							input_device,
							output_device,
							stripped_proto,
							rule->from_ip,
							rule->from_netmask,
							rule->temp_src_port,
							rule->to_ip,
							rule->to_netmask,
							rule->temp_dst_port,
							rule->from_mac);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
			/* REVERSE! related,established */
			snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s %s -d %s/%s %s -m state --state RELATED,ESTABLISHED -j MARK --set-mark 3",
							conf.iptables_location,
							reverse_output_device,
							reverse_input_device,
							stripped_proto,
							rule->to_ip,
							rule->to_netmask,
							temp_dst_port,
							rule->from_ip,
							rule->from_netmask,
							temp_src_port);
			if(pipe_command(debuglvl, cmd, 0) < 0)
			{
				(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
				retval = -1;
			}
		
			if(strcmp(rule->helper, "") != 0)
			{
				/* RELATED & ESTABLISHED */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s %s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								input_device,
								output_device,
								stripped_proto,
								rule->from_ip,
								rule->from_netmask,
								rule->to_ip,
								rule->to_netmask,
								rule->from_mac,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
				/* REVERSE! */
				snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle -A FORWARD %s %s %s -s %s/%s -d %s/%s -m helper --helper \"%s\" -m state --state ESTABLISHED,RELATED -j MARK --set-mark 3",
								conf.iptables_location,
								reverse_output_device,
								reverse_input_device,
								stripped_proto,
								rule->to_ip,
								rule->to_netmask,
								rule->from_ip,
								rule->from_netmask,
								rule->helper);
				if(pipe_command(debuglvl, cmd, 0) < 0)
				{
					(void)vrprint.error(-1, "Error", "creating mangle rules for forward rule failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
		}
	}

	return(retval);
}


int
create_rule_masq(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set the chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);

	/* assemble the string */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat -A %s %s -s %s/%s -d %s/%s -j %s", conf.iptables_location, rule->chain, output_device, rule->from_ip, rule->from_netmask, rule->to_ip, rule->to_netmask, rule->action);
	if(pipe_command(debuglvl, cmd, 0) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating 'masq' rule failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* update the chain counter */
	create->iptcount.postroute++;

	return(0);
}


/*
	TODO: maybe we want an option to use only one interface.
*/
int
create_rule_snat(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	char	cmd[MAX_PIPE_COMMAND];
	char	output_device[sizeof(rule->to_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->to_int[0] != '\0')
		snprintf(output_device, sizeof(output_device), "-o %s", rule->to_int);
	
	/* assemble SNAT string - we do this here because LOG can't handle --to-source */
	if(strcmp(rule->action, "SNAT") == 0)
	{
		snprintf(rule->action, sizeof(rule->action), "SNAT --to-source %s", rule->serverip);
	}

	/* assemble the string */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat -A %s %s %s -s %s/%s %s -d %s/%s %s -j %s", conf.iptables_location, rule->chain, output_device, rule->proto, rule->from_ip, rule->from_netmask, rule->temp_src_port, rule->to_ip, rule->to_netmask, rule->temp_dst_port, rule->action);
	if(pipe_command(debuglvl, cmd, 0) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating 'snat' rule failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* update the counter */
	create->iptcount.postroute++;
	return(0);
}


/*	create_rule_portfw

	Here we create the iptablesrules for portfw. They concist of a PREROUTING rule and a FORWARD rule.
	Both are created for PORTFW. We also create REDIRECT rules here, but then without the forward rule.

	For PORTFW we handle both listenport and remoteport options.
*/
int
create_rule_portfw(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	int	retval = 0;
	char	cmd[MAX_PIPE_COMMAND];
	char	input_device[sizeof(rule->from_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* see pp 278 linux firewall 2nd edition for portforwarding && 274-275 for redirect */
	
	/* assembeling rule->action, we start with rule->remoteip */

	/* for remote port use, thats easy we want to use the given remoteport as --to-destination ports */
	if(create->option.remoteport == 1 && rule->remoteport_ptr != NULL)
	{
		if(rule->remoteport_ptr->dst_high == -1)
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d", create->to->ipv4.ipaddress, rule->remoteport_ptr->dst_low);
		else
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d-%d", create->to->ipv4.ipaddress, rule->remoteport_ptr->dst_low, rule->remoteport_ptr->dst_high);
	}
	/* if we use listenport, we want --to-destination to be the original port(s) of the service. */
	else if(create->option.listenport == 1 && rule->portrange_ptr != NULL)
	{
		if(rule->portrange_ptr->dst_high <= 0)
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d", create->to->ipv4.ipaddress, rule->portrange_ptr->dst_low);
		else
			snprintf(rule->remoteip, sizeof(rule->remoteip), "%s:%d-%d", create->to->ipv4.ipaddress, rule->portrange_ptr->dst_low, rule->portrange_ptr->dst_high);
	}
	/* if no listenport or remoteport --to-destination is just the ip. */
	else
	{
		snprintf(rule->remoteip, sizeof(rule->remoteip), "%s", create->to->ipv4.ipaddress);
	}

	/* we set this here because we need remoteip */
	if(strncmp(rule->action, "DNAT", 4) == 0)
	{
		snprintf(rule->action, sizeof(rule->action), "DNAT --to-destination %s", rule->remoteip);
	}

	/* set --dport here, but only if we need to change it. */
	if(create->option.listenport == 1 && rule->listenport_ptr != NULL)
	{
		if(rule->listenport_ptr->dst_high == -1)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->listenport_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->listenport_ptr->dst_low, rule->listenport_ptr->dst_high);
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);
	
	/* here we pipe the rule, but only if its not a log rule, because we only log the forward rule for portfw */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
	{
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat -A %s %s %s -s %s/%s %s -d %s/255.255.255.255 %s %s -m state --state NEW -j %s", conf.iptables_location, rule->chain, input_device, rule->proto, rule->from_ip, rule->from_netmask, rule->temp_src_port, rule->serverip, rule->temp_dst_port, rule->from_mac, rule->action);
		if(pipe_command(debuglvl, cmd, 0) < 0)
		{
			(void)vrprint.error(-1, "Error", "creating 'portfw' rule failed (in: %s).", __FUNCTION__);
			return(-1);
		}
		
		create->iptcount.preroute++;
	}

	/* the forward rule */

	/* if we use remoteport, it will be our destination */
	if (create->option.remoteport == 1 && rule->remoteport_ptr != NULL)
	{
		if(rule->remoteport_ptr->dst_high <= 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->remoteport_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->remoteport_ptr->dst_low, rule->remoteport_ptr->dst_high);
	}

	/*	if we have a listenport option temp_dst_port for the DNAT rule is different from the FORWARD rule, so we fix that here
		we only do this if remoteport == 0, otherwise we use the remoteport
	*/
	if(create->option.listenport == 1 && create->option.remoteport == 0 && rule->portrange_ptr != NULL)
	{
		if(rule->portrange_ptr->dst_high <= 0)
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", rule->portrange_ptr->dst_low);
		else
			snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d:%d", rule->portrange_ptr->dst_low, rule->portrange_ptr->dst_high);
	}

	/* set rule->chain */
	strlcpy(create->chain, "FORWARD", sizeof(create->chain));

	/* set the action */
	if(strncmp(rule->action, "DNAT", 4) == 0)
	{
		snprintf(rule->action, sizeof(rule->action), "ACCEPT");
	}

	if(create_rule_forward(debuglvl, rule, create) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating forward rule for portfw failed (in: %s).", __FUNCTION__);
		retval = -1;
	}

	/* reset rule->chain */
	strlcpy(create->chain, "PREROUTING", sizeof(create->chain));
	
	return(retval);
}


/*	create_rule_redirect

	see pp 278 linux firewall 2nd edition 274-275 for redirect
*/
int
create_rule_redirect(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create)
{
	char	cmd[MAX_PIPE_COMMAND] = "";
	char	tmp_chain[sizeof(create->chain)] = "",
		tmp_port[sizeof(rule->temp_dst_port)] = "",
		tmp_action[sizeof(rule->action)] = "";
	char	input_device[sizeof(rule->from_int) + 3] = "";

	/* safety */
	if(!rule || !create)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set rule->chain */
	strlcpy(rule->chain, create->chain, sizeof(rule->chain));

	/* handle empty device (virtual) */
	if(rule->from_int[0] != '\0')
		snprintf(input_device, sizeof(input_device), "-i %s", rule->from_int);

	/* here we pipe the rule, but only if its not a log rule, because we only log the forward rule for portfw */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
	{
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat -A %s %s %s -s %s/%s %s -d %s/255.255.255.255 %s %s -m state --state NEW -j %s", conf.iptables_location, rule->chain, input_device, rule->proto, rule->from_ip, rule->from_netmask, rule->temp_src_port, rule->to_ip, rule->temp_dst_port, rule->from_mac, rule->action);
		if(pipe_command(debuglvl, cmd, 0) < 0)
		{
			(void)vrprint.error(-1, "Error", "creating 'redirect' rule failed (in: %s).", __FUNCTION__);
			return(-1);
		}

		create->iptcount.preroute++;
	}
	
	/*
		create the corresponding input rule
	*/
	
	/* temp store the realports */
	strlcpy(tmp_port, rule->temp_dst_port, sizeof(tmp_port));
	/* temp store the chain */
	strlcpy(tmp_chain, create->chain, sizeof(tmp_chain));
	/* temp store the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
		strlcpy(tmp_action, rule->action, sizeof(tmp_action));

	/* set the redirectport to rule->temp_dst_port */
	snprintf(rule->temp_dst_port, sizeof(rule->temp_dst_port), "--dport %d", create->option.redirectport);
	/* set the chain */
	snprintf(create->chain, sizeof(create->chain), "INPUT");
	/* set the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
		snprintf(rule->action, sizeof(rule->action), "ACCEPT");
		
	/* now create the input rule */
	if(create_rule_input(debuglvl, rule, create) < 0)
	{
		(void)vrprint.error(-1, "Error", "creating input rule for redirect failed (in: %s).", __FUNCTION__);
		return(-1);
	}
		
	/* restore the realports */
	strlcpy(rule->temp_dst_port, tmp_port, sizeof(rule->temp_dst_port));
	/* restore the chain */
	strlcpy(create->chain, tmp_chain, sizeof(create->chain));
	/* restore the action */
	if(strncasecmp(rule->action, "LOG", 3) != 0)
		strlcpy(rule->action, tmp_action, sizeof(rule->action));

	return(0);
}


int
create_block_rules(const int debuglvl, d_list *blocklist)
{
	char		cmd[MAX_PIPE_COMMAND] = "",
			*ipaddress = NULL;
	d_list_node	*d_node = NULL;


	/* safety */
	if(!blocklist)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading Blocklist...\n");

	if(blocklist->len == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "no items in blocklist.");

		return(0);
	}

	for(d_node = blocklist->top; d_node; d_node = d_node->next)
	{
		if(!(ipaddress = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "ipaddress to add: '%s'.", ipaddress);

		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A BLOCKLIST -s %s/255.255.255.255 -j BLOCK", conf.iptables_location, ipaddress);
		if(pipe_command(debuglvl, cmd, 0) < 0)
		{
			(void)vrprint.error(-1, "Error", "creating 'blocklist' rule failed for ipaddress '%s' (in: %s:%d).", ipaddress, __FUNCTION__, __LINE__);
			return(-1);
		}
		snprintf(cmd, MAX_PIPE_COMMAND, "%s -A BLOCKLIST -d %s/255.255.255.255 -j BLOCK", conf.iptables_location, ipaddress);
		if(pipe_command(debuglvl, cmd, 0) < 0)
		{
			(void)vrprint.error(-1, "Error", "creating 'blocklist' rule failed for ipaddress '%s' (in: %s:%d).", ipaddress, __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	return(0);
}


/*
	we don't care if the network is active
*/
int
create_network_protect_rules(const int debuglvl, d_list *zones_list)
{
	struct RuleCache_	*create = NULL;
	d_list_node		*d_node = NULL,
				*net_d_node = NULL,
				*from_if_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	struct RuleData_	*rule_ptr = NULL;
	int			retval = 0,
				result = 0;
	struct InterfaceData_	*from_if_ptr = NULL;
	char			cmd[MAX_PIPE_COMMAND] = "";
	char			input_device[16 + 3] = "";	/* 16 + '-i ' */


	/* safety */
	if(!zones_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading anti-spoofing rules...\n");

	/* loop through the zones to look for networks */
	for(d_node = zones_list->top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		if(zone_ptr->type == TYPE_NETWORK)
		{
			/* now loop through the ruleslist */
			for(net_d_node = zone_ptr->ProtectList.top; net_d_node; net_d_node = net_d_node->next)
			{
				if(!(rule_ptr = net_d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
				if(!(create = &rule_ptr->rulecache))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}
				if(!create->who)
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				/* here we print the description if we are in bashmode */
				if(bash_out == 1 && create->description != NULL)
				{
					fprintf(stdout, "\n# %s\n", create->description);
		
					free(create->description);
					create->description = NULL;
				}

				/* iptables protect */
				if(create->danger.solution == PROT_IPTABLES)
				{
					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "create_rule: protect iptables.");

					/* if the network has no interfaces, don't bother */
					if(create->who->InterfaceList.len > 0)
					{
						for(from_if_node = create->who->InterfaceList.top; from_if_node; from_if_node = from_if_node->next)
						{
							if(!(from_if_ptr = from_if_node->data))
							{
								(void)vrprint.error(-1, "Internal Error", "from_if_ptr == NULL! (in: create_rule, PROT_IPTABLES).");
								return(-1);
							}

							/*	see if the interface is active */
							if(from_if_ptr->active == 0)
							{
								/* here we print the description if we are in bashmode */
								if(bash_out == 1)
								{
									fprintf(stdout, "# anti-spoof rule for interface '%s' not created. The interface is inactive.\n", from_if_ptr->name);
								}
							}
							else
							{
								/*	get the input_device

									if the device is virtual, we don't add it to the iptables string.
								*/
								if(!from_if_ptr->device_virtual)
									snprintf(input_device, sizeof(input_device), "-i %s", from_if_ptr->device);
								else
									memset(input_device, 0, sizeof(input_device));


								/*	only create the rule if the interface is active
	
									NOTE: this is not the same as an interface that is down!
								*/
								if(from_if_ptr->active)
								{
									/* create the log rule */
									if(strcmp(conf.loglevel, "") == 0)
									{
										snprintf(cmd, MAX_PIPE_COMMAND, "%s -A ANTISPOOF %s -s %s/%s -d %s/255.255.255.255 -m limit --limit 1/s --limit-burst 5 -j LOG --log-prefix 'vrmr: DROP %s %s '",
														conf.iptables_location,
														input_device,
														create->danger.source_ip.ipaddress,
														create->danger.source_ip.netmask,
														from_if_ptr->ipv4.ipaddress,
														create->danger.type,
														create->danger.source);
									}
									else
									{
										snprintf(cmd, MAX_PIPE_COMMAND, "%s -A ANTISPOOF %s -s %s/%s -d %s/255.255.255.255 -m limit --limit 1/s --limit-burst 5 -j LOG --log-prefix 'vrmr: DROP %s %s ' --log-level %s",
														conf.iptables_location,
														input_device,
														create->danger.source_ip.ipaddress,
														create->danger.source_ip.netmask,
														from_if_ptr->ipv4.ipaddress,
														create->danger.type,
														create->danger.source,
														conf.loglevel);
									}
									result = pipe_command(debuglvl, cmd, 0);
									if(result != 0)
									{
										(void)vrprint.error(-1, "Error", "Creating protect rule failed. Rule: '%s' (in: %s).", cmd, __FUNCTION__);
										retval = -1;
									}

									/* and the DROP rule */
									snprintf(cmd, MAX_PIPE_COMMAND, "%s -A ANTISPOOF %s -s %s/%s -d %s/255.255.255.255 -j DROP",
													conf.iptables_location,
													input_device,
													create->danger.source_ip.ipaddress,
													create->danger.source_ip.netmask,
													from_if_ptr->ipv4.ipaddress);
									result = pipe_command(debuglvl, cmd, 0);
									if(result != 0)
									{
										(void)vrprint.error(-1, "Error", "Creating protect rule failed. Rule: '%s' (in: %s).", cmd, __FUNCTION__);
										retval = -1;
									}
								}
							}
						}
					}
				}
				/* Whoops, this is serious. */
				else
				{
					(void)vrprint.error(-1, "Internal Error", "unknown protect danger type %d (in: %s).", create->danger.solution, __FUNCTION__);
					return(-1);
				}
			}
		}
	}

	return(0);
}


int
create_interface_protect_rules(const int debuglvl, d_list *interfaces_list)
{
	struct RuleCache_	*create = NULL;
	d_list_node		*d_node = NULL,
				*if_d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;
	struct InterfaceData_	*iface_ptr = NULL;


	/* safety */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(bash_out == 1)
		fprintf(stdout, "\n# Loading interfaces protection rules...\n");

	/* loop through the zones to look for networks */
	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	
		/* now loop through the ruleslist */
		for(if_d_node = iface_ptr->ProtectList.top; if_d_node; if_d_node = if_d_node->next)
		{
			if(!(rule_ptr = if_d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}
			if(!(create = &rule_ptr->rulecache))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}
			if(!create->who_int)
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			/* here we print the description if we are in bashmode */
			if(bash_out == 1 && create->description != NULL)
			{
				fprintf(stdout, "\n# %s\n", create->description);
		
				free(create->description);
				create->description = NULL;
			}

			/*
				prot rule, proc only for interface
			*/
			if(create->danger.solution == PROT_PROC_INT)
			{
				if(debuglvl >= HIGH)
					(void)vrprint.debug("Debug", "protect proc (int)... ");

				/*	only set if int is active and up.
		
					NOTE: this is different from iptables rules, which are also created if the interface is
					down. However the proc entries are only available when the interface is up!
				*/
				if(create->who_int->active && create->who_int->up)
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "Setting '%d' to '%s' (where * is %s)... ", create->danger.proc_set_on, create->danger.proc_entry, create->who_int->device);

					if(set_proc_entry(debuglvl, create->danger.proc_entry, create->danger.proc_set_on, create->who_int->device) != 0)
					{
						/* if it fails, we dont really care, its not fatal */
						(void)vrprint.error(-1, "Error", "setting proc entry failed (in: %s:%d).", __FUNCTION__, __LINE__);
					}
				}
				else
				{
					if(bash_out == 1)
					{
						fprintf(stdout, "# not created: interface is inactive or down.\n");
					}
				}
			}
			/* Whoops, this is serious. */
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown protect danger type %d (in: %s:%d).", create->danger.solution, __FUNCTION__, __LINE__);
				return(-1);
			}
		}
	}

	return(0);
}


int
create_system_protectrules(const int debuglvl, struct vuurmuur_config *conf)
{
	int	result = 0;

	/* safety */
	if(!conf)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "protect proc systemwide... ");


	/* syncookies */
	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "Setting '%d' to '%s'... ", conf->protect_syncookie, "/proc/sys/net/ipv4/tcp_syncookies");

	result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/tcp_syncookies", conf->protect_syncookie, NULL);
	if(result != 0)
	{
		/* if it fails, we dont really care, its not fatal */
		(void)vrprint.error(-1, "Error", "set_proc_entry failed (in: create_rule, prot_proc_sys).");
	}


	/* echo broadcasts */
	if(debuglvl >= MEDIUM)
		(void)vrprint.debug(__FUNCTION__, "Setting '%d' to '%s'... ", conf->protect_echobroadcast, "/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts");

	result = set_proc_entry(debuglvl, "/proc/sys/net/ipv4/icmp_echo_ignore_broadcasts", conf->protect_echobroadcast, NULL);
	if(result != 0)
	{
		/* if it fails, we dont really care, its not fatal */
		(void)vrprint.error(-1, "Error", "set_proc_entry failed (in: create_rule, prot_proc_sys).");
	}

	return(0);
}


int
clear_all_iptables_rule(const int debuglvl)
{
	int	retval = 0,
		result = 0;
	char	cmd[MAX_PIPE_COMMAND] = "";

	/* flush everything */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t nat --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -t mangle --flush", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval = -1;

	/* flush anti spoof old rules */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush ANTISPOOF 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	/* flush the BLOCKLIST */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush BLOCKLIST 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	/* flush the BLOCKLIST */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush BLOCK 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	/* flush the synlimit chain */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --flush SYNLIMIT 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	/* delete the chains */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X ANTISPOOF 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X SYNLIMIT 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X BLOCKLIST 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);
	snprintf(cmd, MAX_PIPE_COMMAND, "%s -X BLOCK 2>/dev/null", conf.iptables_location);
	(void)pipe_command(debuglvl, cmd, 1);

	/* set default polices to ACCEPT */
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy INPUT ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy OUTPUT ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;
	snprintf(cmd, MAX_PIPE_COMMAND, "%s --policy FORWARD ACCEPT", conf.iptables_location);
	result = pipe_command(debuglvl, cmd, 0);
	if(result < 0)
		retval=-1;

	return(retval);
}
