/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


struct ConfigSection_
{
	PANEL   *panel[1];
	WINDOW  *win;
	FIELD   **fields;
	FORM    *form;
	int     n_fields;
} ConfigSection;


/* clean up for all config windows */
static int
edit_config_destroy(void)
{
	int	i;

	/* Un post form and free the memory */
	unpost_form(ConfigSection.form);
	free_form(ConfigSection.form);
	for(i=0;i<ConfigSection.n_fields;i++)
	{
		free_field(ConfigSection.fields[i]);
	}
	free(ConfigSection.fields);

	del_panel(ConfigSection.panel[0]);
	destroy_win(ConfigSection.win);
	update_panels();
	doupdate();

	return(0);
}


struct
{
	FIELD	*ruleslocfld,
		*blocklistlocfld,
		
		*iptableslocfld,
//		*ifconfiglocfld,
		
		*scriptslocfld;
//		*helpfilelocfld;

} GenConfig;



static int
edit_genconfig_init(int height, int width, int starty, int startx)
{
	int	retval=0,
		i,
		rows,
		cols;

	ConfigSection.n_fields = 4;
	ConfigSection.fields = (FIELD **)calloc(ConfigSection.n_fields + 1, sizeof(FIELD *));

	/* plugindir & rulesfile */
	GenConfig.ruleslocfld =     (ConfigSection.fields[0] = new_field(1, 64, 2, 1, 0, 0));  /* rulesfile */
	GenConfig.blocklistlocfld = (ConfigSection.fields[1] = new_field(1, 64, 4, 1, 0, 0));  /* blocklistfile */

	/* external programs */
	GenConfig.iptableslocfld =  (ConfigSection.fields[2] = new_field(1, 64, 7, 1, 0, 0));  /* iptables */
//	GenConfig.ifconfiglocfld =  (ConfigSection.fields[3] = new_field(1, 64, 9, 1, 0, 0));  /* ifconfig */

	GenConfig.scriptslocfld =   (ConfigSection.fields[3] = new_field(1, 64, 12, 1, 0, 0)); /* scripts loc */

	/* terminate */
	ConfigSection.fields[ConfigSection.n_fields] = NULL;

	ConfigSection.win = create_newwin(height, width, starty, startx, "Edit Config: General", COLOR_PAIR(5));
	ConfigSection.panel[0] = new_panel(ConfigSection.win);

	/* set buffers - first the visible, then the label */
	set_field_buffer(GenConfig.ruleslocfld, 0, conf.rules_location);
	set_field_buffer(GenConfig.blocklistlocfld, 0, conf.blocklist_location);
	set_field_buffer(GenConfig.iptableslocfld, 0, conf.iptables_location);
//	set_field_buffer(GenConfig.ifconfiglocfld, 0, conf.ifconfig_location);
	set_field_buffer(GenConfig.scriptslocfld, 0, conf.scripts_location);

	/* set buffers done */
	for(i = 0; i < ConfigSection.n_fields; i++)
	{
		set_field_back(ConfigSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
		field_opts_off(ConfigSection.fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(ConfigSection.fields[i], FALSE);
	}

	// Create the form and post it
	ConfigSection.form = new_form(ConfigSection.fields);
	// Calculate the area required for the form
	scale_form(ConfigSection.form, &rows, &cols);
	keypad(ConfigSection.win, TRUE);
	// Set main window and sub window
	set_form_win(ConfigSection.form, ConfigSection.win);
	set_form_sub(ConfigSection.form, derwin(ConfigSection.win, rows, cols, 1, 2));

	post_form(ConfigSection.form);

	/* print labels */
	mvwprintw(ConfigSection.win, 2, 2,  "Rulesfile (full path):");
	mvwprintw(ConfigSection.win, 4, 2,  "BlockList file (full path):");
	mvwprintw(ConfigSection.win, 7, 2,  "Iptables location (full path):");
	mvwprintw(ConfigSection.win, 12, 2, "Scripts location (full path):");

	return(retval);
}


static int
edit_genconfig_save(const int debuglvl)
{
	int	retval = 0,
		i = 0;
	char	tmp_scripts_location[sizeof(conf.scripts_location)] = "";

	/* check for changed fields */
	for(i=0; i < ConfigSection.n_fields; i++)
	{
		/* we only act if a field is changed */
		if(field_status(ConfigSection.fields[i]) == TRUE)
		{
			if(ConfigSection.fields[i] == GenConfig.ruleslocfld)
			{
				/* rulesfile location */
				if(!(copy_field2buf(	conf.rules_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.rules_location))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == GenConfig.blocklistlocfld)
			{
				/* blocklist location */
				if(!(copy_field2buf(	conf.blocklist_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.blocklist_location))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == GenConfig.iptableslocfld)
			{
				/* iptables location */
				if(!(copy_field2buf(	conf.iptables_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.iptables_location))))
					return(-1);
			}
//			else if(ConfigSection.fields[i] == GenConfig.ifconfiglocfld)
//			{
				/* ifconfig location */
//				if(!(copy_field2buf(	conf.ifconfig_location,
//							field_buffer(ConfigSection.fields[i], 0),
//							sizeof(conf.ifconfig_location))))
//					return(-1);
//			}
			else if(ConfigSection.fields[i] == GenConfig.scriptslocfld)
			{
				/* temp store the value in case the check fails */
				strlcpy(tmp_scripts_location, conf.scripts_location, sizeof(tmp_scripts_location));

				/* scripts dir location */
				if(!(copy_field2buf(	conf.scripts_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.scripts_location))))
					return(-1);

				if(config_check_vuurmuurdir(debuglvl, conf.scripts_location) < 0)
				{
					/* restore the old value */
					strlcpy(conf.scripts_location, tmp_scripts_location, sizeof(conf.scripts_location));
					set_field_buffer(ConfigSection.fields[i], 0, conf.scripts_location);

					retval = -1;
				}
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown field.");
				retval = -1;
			}
		}
	}

	return(retval);
}


int
edit_genconfig(const int debuglvl)
{
	int	ch,
		retval=0,
		quit=0;
	int	height,
		width,
		startx,
		starty,
		max_height,
		max_width;
	FIELD	*cur = NULL,
		*prev = NULL;
	char	not_defined = 0;

	getmaxyx(stdscr, max_height, max_width);

	height = 18;
	width = 76;

	startx = (max_width - width)/2;
	starty = (max_height - height)/2;

	edit_genconfig_init(height, width, starty, startx);
	cur = current_field(ConfigSection.form);
	update_panels();
	doupdate();

	/* Loop through to get user requests */
	while(quit == 0)
	{
		draw_field_active_mark(cur, prev, ConfigSection.win, ConfigSection.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		ch = wgetch(ConfigSection.win);

		not_defined = 0;

		if(	cur == GenConfig.ruleslocfld ||
			cur == GenConfig.blocklistlocfld ||
			cur == GenConfig.iptableslocfld ||
//			cur == GenConfig.ifconfiglocfld ||
			cur == GenConfig.scriptslocfld)
		{
			if(nav_field_simpletext(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case KEY_LEFT:
				case 'q':

					/* save the field to the conf struct */
					if(edit_genconfig_save(debuglvl) < 0)
					{
						if(confirm("Saving config failed", "Saving the config failed. Sure you want to quit?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0))
						{
							retval = -1;
							quit = 1;
						}
					}
					else
					{
						quit = 1;
						retval = 0;
					}

					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					// Go to next field
					form_driver(ConfigSection.form, REQ_NEXT_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ConfigSection.form, REQ_PREV_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case 127:
				case KEY_BACKSPACE:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_DC:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:CONFIG:GENERAL]:");
					break;

				default:
					/* If this is a normal character, it gets printed into the field */
					form_driver(ConfigSection.form, ch);
					break;
			}
		}

		prev = cur;
		cur = current_field(ConfigSection.form);
	}

	/* write configfile */
	if(retval == 0)
	{
		if(write_configfile(debug, conf.configfile_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "Writing config file failed.");
			retval = -1;
		}
	}

	/* cleanup */
	edit_config_destroy();

	return(retval);
}


struct
{
	FIELD	*servbackfld,
		*zonebackfld,
		*ifacbackfld,
		*pluginsdirfld;

} PlugConfig;


static int
edit_plugconfig_init(int height, int width, int starty, int startx)
{
	int	retval=0,
		i,
		rows,
		cols;

	ConfigSection.n_fields = 5;
	ConfigSection.fields = (FIELD **)calloc(ConfigSection.n_fields + 1, sizeof(FIELD *));

	/* backends */
	PlugConfig.servbackfld = (ConfigSection.fields[0] = new_field(1, 16, 1, 1, 0, 0));  /* servbackend */
	PlugConfig.zonebackfld = (ConfigSection.fields[1] = new_field(1, 16, 1, 21, 0, 0)); /* zonebackend */
	PlugConfig.ifacbackfld = (ConfigSection.fields[2] = new_field(1, 16, 1, 41, 0, 0)); /* ifacbackend */

	/* plugindir & rulesfile */
	PlugConfig.pluginsdirfld = (ConfigSection.fields[3] = new_field(1, 64, 4, 1, 0, 0)); /* plugindir */

	/* terminate */
	ConfigSection.fields[ConfigSection.n_fields] = NULL;

	ConfigSection.win = create_newwin(height, width, starty, startx, "Edit Config: Plugins", COLOR_PAIR(5));
	ConfigSection.panel[0] = new_panel(ConfigSection.win);

	/* set buffers - first the visible, then the label */
	set_field_buffer(PlugConfig.servbackfld, 0, conf.serv_backend_name);
	set_field_buffer(PlugConfig.zonebackfld, 0, conf.zone_backend_name);
	set_field_buffer(PlugConfig.ifacbackfld, 0, conf.ifac_backend_name);
	set_field_buffer(PlugConfig.pluginsdirfld, 0, conf.plugindir);

	/* set buffers done */
	for(i = 0; i < ConfigSection.n_fields; i++)
	{
		set_field_back(ConfigSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
		field_opts_off(ConfigSection.fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(ConfigSection.fields[i], FALSE);
	}

	// Create the form and post it
	ConfigSection.form = new_form(ConfigSection.fields);
	// Calculate the area required for the form
	scale_form(ConfigSection.form, &rows, &cols);
	keypad(ConfigSection.win, TRUE);
	// Set main window and sub window
	set_form_win(ConfigSection.form, ConfigSection.win);
	set_form_sub(ConfigSection.form, derwin(ConfigSection.win, rows, cols, 1, 2));

	post_form(ConfigSection.form);

	/* print labels */
	mvwprintw(ConfigSection.win, 1, 2,  "Services Backend:");
	mvwprintw(ConfigSection.win, 1, 22, "Zones Backend:");
	mvwprintw(ConfigSection.win, 1, 42, "Interfaces Backend:");
	mvwprintw(ConfigSection.win, 4, 2,  "Plugin directory (full path):");

	return(retval);
}


static int
edit_plugconfig_save(const int debuglvl)
{
	int	retval = 0,
		i = 0;
	char	tmp_plugindir[sizeof(conf.plugindir)] = "";

	/* check for changed fields */
	for(i=0; i < ConfigSection.n_fields; i++)
	{
		/* we only act if a field is changed */
		if(field_status(ConfigSection.fields[i]) == TRUE)
		{
			if(ConfigSection.fields[i] == PlugConfig.servbackfld)
			{
				/* services backend */
				if(!(copy_field2buf(	conf.serv_backend_name,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.serv_backend_name))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == PlugConfig.zonebackfld)
			{
				/* zones backend */
				if(!(copy_field2buf(	conf.zone_backend_name,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.zone_backend_name))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == PlugConfig.ifacbackfld)
			{
				/* interfaces backend */
				if(!(copy_field2buf(	conf.ifac_backend_name,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.ifac_backend_name))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == PlugConfig.pluginsdirfld)
			{
				/* temp store the value in case the check fails */
				strlcpy(tmp_plugindir, conf.plugindir, sizeof(tmp_plugindir));

				/* plugindir */
				if(!(copy_field2buf(	conf.plugindir,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.plugindir))))
					return(-1);

				if(config_check_vuurmuurdir(debuglvl, conf.plugindir) < 0)
				{
					/* ok restore old value */
					strlcpy(conf.plugindir, tmp_plugindir, sizeof(conf.plugindir));
					set_field_buffer(ConfigSection.fields[i], 0, conf.plugindir);

					retval = -1;
				}
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown field.");
				retval = -1;
			}
		}
	}

	return(retval);
}


int
edit_plugconfig(const int debuglvl)
{
	int	ch,
		retval=0,
		quit=0;
	int	height,
		width,
		startx,
		starty,
		max_height,
		max_width;
	FIELD	*cur = NULL,
		*prev = NULL;
	char	not_defined = 0;

	getmaxyx(stdscr, max_height, max_width);

	height = 18;
	width = 76;

	startx = (max_width - width)/2;
	starty = (max_height - height)/2;

	edit_plugconfig_init(height, width, starty, startx);
	cur = current_field(ConfigSection.form);
	update_panels();
	doupdate();

	/* Loop through to get user requests */
	while(quit == 0)
	{
		draw_field_active_mark(cur, prev, ConfigSection.win, ConfigSection.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		ch = wgetch(ConfigSection.win);

		not_defined = 0;

		if(	cur == PlugConfig.servbackfld ||
			cur == PlugConfig.zonebackfld ||
			cur == PlugConfig.ifacbackfld ||
			cur == PlugConfig.pluginsdirfld
			)
		{
			if(nav_field_simpletext(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case KEY_LEFT:
				case 'q':

					/* save the field to the conf struct */
					if(edit_plugconfig_save(debuglvl) < 0)
					{
						if(confirm("Saving config failed", "Saving the config failed. Sure you want to quit?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0))
						{
							retval = -1;
							quit = 1;
						}
					}
					else
					{
						quit = 1;
						retval = 0;
					}

					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					// Go to next field
					form_driver(ConfigSection.form, REQ_NEXT_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ConfigSection.form, REQ_PREV_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case 127:
				case KEY_BACKSPACE:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_DC:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:CONFIG:PLUGINS]:");
					break;

				default:
					/* If this is a normal character, it gets printed into the field */
					form_driver(ConfigSection.form, ch);
					break;
			}
		}

		prev = cur;
		cur = current_field(ConfigSection.form);
	}

	/* write configfile */
	if(retval == 0)
	{
		if(write_configfile(debug, conf.configfile_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "Writing config file failed.");
			retval = -1;
		}
	}

	/* cleanup */
	edit_config_destroy();

	return(retval);
}


struct
{
	FIELD	*synlimitfld,
		*synburstfld;
	char	number[8];

} ConConfig;


static int
edit_conconfig_init(int height, int width, int starty, int startx)
{
	int	i,
		rows,
		cols;

	ConfigSection.n_fields = 2;
	ConfigSection.fields = (FIELD **)calloc(ConfigSection.n_fields + 1, sizeof(FIELD *));

	/* fields */
	ConConfig.synlimitfld = (ConfigSection.fields[0] = new_field(1, 8, 4, 1, 0, 1)); /* SYN-limit */
	ConConfig.synburstfld = (ConfigSection.fields[1] = new_field(1, 8, 6, 1, 0, 1)); /* SYN-limit-burst */
	ConfigSection.fields[ConfigSection.n_fields] = NULL;

	/* create win & pan */
	if(!(ConfigSection.win = create_newwin(height, width, starty, startx, "Edit Config: Connections", COLOR_PAIR(5))))
	{
		(void)vrprint.error(-1, "Error", "could not create window.");
		return(-1);
	}
	if(!(ConfigSection.panel[0] = new_panel(ConfigSection.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel.");
		return(-1);
	}

	/* set fields */
	snprintf(ConConfig.number, sizeof(ConConfig.number), "%d", conf.syn_limit);
	set_field_buffer(ConConfig.synlimitfld, 0, ConConfig.number);

	snprintf(ConConfig.number, sizeof(ConConfig.number), "%d", conf.syn_limit_burst);
	set_field_buffer(ConConfig.synburstfld, 0, ConConfig.number);

	/* set the field options */
	for(i = 0; i < ConfigSection.n_fields; i++)
	{
		/* background */
		set_field_back(ConfigSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
		/* no autoskip */
		field_opts_off(ConfigSection.fields[i], O_AUTOSKIP);
		/* set status to false */
		set_field_status(ConfigSection.fields[i], FALSE);
	}

	/* Create the form and post it */
	if(!(ConfigSection.form = new_form(ConfigSection.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form.");
		return(-1);
	}
	/* Calculate the area required for the form */
	scale_form(ConfigSection.form, &rows, &cols);
	keypad(ConfigSection.win, TRUE);
	/* Set main window and sub window */
	set_form_win(ConfigSection.form, ConfigSection.win);
	set_form_sub(ConfigSection.form, derwin(ConfigSection.win, rows, cols, 1, 2));
	post_form(ConfigSection.form);

	/* print labels */
	mvwprintw(ConfigSection.win, 2, 2, "You can limit the number of new tcp connections per second:");

	mvwprintw(ConfigSection.win, 4, 2, "Number of SYN-packets per second:");
	mvwprintw(ConfigSection.win, 6, 2, "Burst-rate:");

	return(0);
}


static int
edit_conconfig_save(void)
{
	int	retval = 0,
		i = 0;
	int	syn = 0;

	/* check for changed fields */
	for(i=0; i < ConfigSection.n_fields; i++)
	{
		/* we only act if a field is changed */
		if(field_status(ConfigSection.fields[i]) == TRUE)
		{
			if(ConfigSection.fields[i] == ConConfig.synlimitfld)
			{
				/* synlimit */
				if(!(copy_field2buf(	ConConfig.number,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(ConConfig.number))))
					return(-1);

				syn = atoi(ConConfig.number);
				if(syn < 0)
				{
					(void)vrprint.error(-1, "Error", "the SYN-limit value must be positive.");
				}
				else if(syn == 0)
				{
					(void)vrprint.error(-1, "Error", "the SYN-limit value must be bigger than 0.");
				}
				else
				{
					conf.syn_limit = (unsigned int)syn;
				}
			}
			else if(ConfigSection.fields[i] == ConConfig.synburstfld)
			{
				/* synlimit */
				if(!(copy_field2buf(	ConConfig.number,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(ConConfig.number))))
					return(-1);

				syn = atoi(ConConfig.number);
				if(syn < 0)
				{
					(void)vrprint.error(-1, "Error", "the SYN-limit-burst value must be positive.");
				}
				else if(syn == 0)
				{
					(void)vrprint.error(-1, "Error", "the SYN-limit-burst value must be bigger than 0.");
				}
				else
				{
					conf.syn_limit_burst = (unsigned int)syn;
				}
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown field.");
				retval = -1;
			}
		}
	}

	return(retval);
}


int
edit_conconfig(const int debuglvl)
{
	int	ch,
		retval = 0,
		quit = 0,
		not_defined = 0;
	int	height,
		width,
		startx,
		starty,
		max_height,
		max_width;
	FIELD	*cur = NULL,
		*prev = NULL;

	/* window dimentions */
	getmaxyx(stdscr, max_height, max_width);

	height = 18;
	width = 76;

	startx = (max_width - width)/2;
	starty = (max_height - height)/2;

	/* setup */
	if(edit_conconfig_init(height, width, starty, startx) < 0)
		return(-1);

	cur = current_field(ConfigSection.form);
	update_panels();
	doupdate();

	/* Loop through to get user requests */
	while(quit == 0)
	{
		/* visual support */
		draw_field_active_mark(cur, prev, ConfigSection.win, ConfigSection.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		/* keyboard input */
		ch = wgetch(ConfigSection.win);

		not_defined = 0;

		if(cur == ConConfig.synlimitfld ||
		   cur == ConConfig.synburstfld)
		{
			if(nav_field_simpletext(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case KEY_LEFT:
				case 'q':
					quit = 1;
					break;

				case KEY_F(5): //help
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					// Go to next field
					form_driver(ConfigSection.form, REQ_NEXT_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ConfigSection.form, REQ_PREV_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case 127:
				case KEY_BACKSPACE:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_DC:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:CONFIG:CONNECTIONS]:");
					break;

				default:
					/* If this is a normal character, it gets printed into the field */
					form_driver(ConfigSection.form, ch);
					break;
			}
		}

		prev = cur;
		cur = current_field(ConfigSection.form);
	}

	/* save the field to the conf struct */
	if(edit_conconfig_save() < 0)
	{
		retval = -1;
	}

	/* write configfile */
	if(retval == 0)
	{
		if(write_configfile(debug, conf.configfile_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "Writing config file failed.");
			retval=-1;
		}
	}

	/* cleanup */
	edit_config_destroy();

	return(retval);
}


struct
{
	FIELD	*newrule_loglimitfld,
		*newrule_logfld,

		*logview_bufsizefld,
		
		*helpfilelocfld,
		
		*advancedmodefld,
		
		*mainmenu_statusfld;

	char	number[7];

} VcConfig;


static int
edit_vcconfig_init(int height, int width, int starty, int startx)
{
	int	i,
		rows,
		cols;

	ConfigSection.n_fields = 6;
	ConfigSection.fields = (FIELD **)calloc(ConfigSection.n_fields + 1, sizeof(FIELD *));

	/* fields */
	VcConfig.newrule_logfld      = (ConfigSection.fields[0] = new_field(1, 1,  3, 23, 0, 0));
	VcConfig.newrule_loglimitfld = (ConfigSection.fields[1] = new_field(1, 3,  4, 22, 0, 0));

	VcConfig.logview_bufsizefld  = (ConfigSection.fields[2] = new_field(1, 6,  6, 22, 0, 0));
	VcConfig.helpfilelocfld      = (ConfigSection.fields[3] = new_field(1, 64, 9, 1, 0, 0));

	VcConfig.advancedmodefld     = (ConfigSection.fields[4] = new_field(1, 1,  11, 27, 0, 0));
	VcConfig.mainmenu_statusfld  = (ConfigSection.fields[5] = new_field(1, 1,  13, 27, 0, 0));

	ConfigSection.fields[ConfigSection.n_fields] = NULL;

	/* create win & pan */
	if(!(ConfigSection.win = create_newwin(height, width, starty, startx, "Vuurmuur-conf Settings", COLOR_PAIR(5))))
	{
		(void)vrprint.error(-1, "Error", "could not create window.");
		return(-1);
	}
	if(!(ConfigSection.panel[0] = new_panel(ConfigSection.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel.");
		return(-1);
	}

	/* set fields */
	set_field_buffer(VcConfig.newrule_logfld, 0, vccnf.newrule_log ? "X" : " ");

	snprintf(VcConfig.number, sizeof(VcConfig.number), "%d", vccnf.newrule_loglimit);
	set_field_buffer(VcConfig.newrule_loglimitfld, 0, VcConfig.number);

	snprintf(VcConfig.number, sizeof(VcConfig.number), "%d", vccnf.logview_bufsize);
	set_field_buffer(VcConfig.logview_bufsizefld, 0, VcConfig.number);

	set_field_buffer(VcConfig.helpfilelocfld, 0, vccnf.helpfile_location);

	set_field_buffer(VcConfig.advancedmodefld, 0, vccnf.advanced_mode ? "X" : " ");

	set_field_buffer(VcConfig.mainmenu_statusfld, 0, vccnf.draw_status ? "X" : " ");

	/* set the field options */
	for(i = 0; i < ConfigSection.n_fields; i++)
	{
		/* background */
		set_field_back(ConfigSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
		/* no autoskip */
		field_opts_off(ConfigSection.fields[i], O_AUTOSKIP);
		/* set status to false */
		set_field_status(ConfigSection.fields[i], FALSE);
	}
	set_field_back(VcConfig.newrule_logfld,     COLOR_PAIR(CP_BLUE_WHITE));
	set_field_back(VcConfig.advancedmodefld,    COLOR_PAIR(CP_BLUE_WHITE));
	set_field_back(VcConfig.mainmenu_statusfld, COLOR_PAIR(CP_BLUE_WHITE));

	/* Create the form and post it */
	if(!(ConfigSection.form = new_form(ConfigSection.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form.");
		return(-1);
	}
	/* Calculate the area required for the form */
	scale_form(ConfigSection.form, &rows, &cols);
	keypad(ConfigSection.win, TRUE);
	/* Set main window and sub window */
	set_form_win(ConfigSection.form, ConfigSection.win);
	set_form_sub(ConfigSection.form, derwin(ConfigSection.win, rows, cols, 1, 2));
	post_form(ConfigSection.form);

	/* print labels */
	mvwprintw(ConfigSection.win, 2, 2, "Defaults for creating new rules:");

	mvwprintw(ConfigSection.win, 4, 2,   "Log the rule");
	mvwprintw(ConfigSection.win, 4, 24,  "[");
	mvwprintw(ConfigSection.win, 4, 26,  "]");
	
	mvwprintw(ConfigSection.win, 5, 2,   "Loglimit per second");

	mvwprintw(ConfigSection.win, 7, 2,   "Buffersize logviewer");
	mvwprintw(ConfigSection.win, 9, 2,   "Helpfile location (full path)");

	mvwprintw(ConfigSection.win, 12, 2,  "Advanced mode:");
	mvwprintw(ConfigSection.win, 12, 28, "[");
	mvwprintw(ConfigSection.win, 12, 30, "]");

	mvwprintw(ConfigSection.win, 14, 2,  "Draw status in Main Menu:");
	mvwprintw(ConfigSection.win, 14, 28, "[");
	mvwprintw(ConfigSection.win, 14, 30, "]");

	return(0);
}


static int
edit_vcconfig_save(void)
{
	int	retval = 0,
		i = 0;
	int	syn = 0;
	int	bufsize = 0;

	/* check for changed fields */
	for(i=0; i < ConfigSection.n_fields; i++)
	{
		/* we only act if a field is changed */
		if(field_status(ConfigSection.fields[i]) == TRUE)
		{
			if(ConfigSection.fields[i] == VcConfig.newrule_loglimitfld)
			{
				/* synlimit */
				if(!(copy_field2buf(	VcConfig.number,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(VcConfig.number))))
					return(-1);

				syn = atoi(VcConfig.number);
				if(syn < 0)
				{
					(void)vrprint.error(-1, "Error", "the loglimit value must be positive.");
				}
				else if(syn == 0)
				{
					(void)vrprint.error(-1, "Error", "the loglimit value must be bigger than 0.");
				}
				else
				{
					vccnf.newrule_loglimit = (unsigned int)syn;
				}
			}
			else if(ConfigSection.fields[i] == VcConfig.newrule_logfld)
			{
				if(field_buffer(ConfigSection.fields[i], 0)[0] == 'X')
					vccnf.newrule_log = 1;
				else
					vccnf.newrule_log = 0;
			}
			else if(ConfigSection.fields[i] == VcConfig.advancedmodefld)
			{
				if(field_buffer(ConfigSection.fields[i], 0)[0] == 'X')
					vccnf.advanced_mode = 1;
				else
					vccnf.advanced_mode = 0;
			}
			else if(ConfigSection.fields[i] == VcConfig.mainmenu_statusfld)
			{
				if(field_buffer(ConfigSection.fields[i], 0)[0] == 'X')
					vccnf.draw_status = 1;
				else
					vccnf.draw_status = 0;
			}
			else if(ConfigSection.fields[i] == VcConfig.logview_bufsizefld)
			{
				/* bufsize */
				if(!(copy_field2buf(	VcConfig.number,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(VcConfig.number))))
					return(-1);

				bufsize = atoi(VcConfig.number);
				if(bufsize < 0)
				{
					(void)vrprint.error(-1, "Error", "the buffersize value must be positive.");
				}
				else if(bufsize == 0)
				{
					(void)vrprint.error(-1, "Error", "the buffersize value must be bigger than 0.");
				}
				else
				{
					vccnf.logview_bufsize = (unsigned int)bufsize;
				}
			}
			else if(ConfigSection.fields[i] == VcConfig.helpfilelocfld)
			{
				/* synlimit */
				if(!(copy_field2buf(	vccnf.helpfile_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(vccnf.helpfile_location))))
					return(-1);
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown field.");
				retval = -1;
			}
		}
	}

	return(retval);
}


int
edit_vcconfig(const int debuglvl)
{
	int	ch,
		retval = 0,
		quit = 0,
		not_defined = 0;
	int	height,
		width,
		startx,
		starty,
		max_height,
		max_width;
	FIELD	*cur = NULL,
		*prev = NULL;

	/* window dimentions */
	getmaxyx(stdscr, max_height, max_width);

	height = 18;
	width = 76;

	startx = (max_width - width)/2;
	starty = (max_height - height)/2;

	/* setup */
	if(edit_vcconfig_init(height, width, starty, startx) < 0)
		return(-1);

	cur = current_field(ConfigSection.form);
	update_panels();
	doupdate();

	/* Loop through to get user requests */
	while(quit == 0)
	{
		/* visual support */
		draw_field_active_mark(cur, prev, ConfigSection.win, ConfigSection.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		/* keyboard input */
		ch = wgetch(ConfigSection.win);

		not_defined = 0;

		if(cur == VcConfig.newrule_loglimitfld ||
		   cur == VcConfig.logview_bufsizefld ||
		   cur == VcConfig.helpfilelocfld)
		{
			if(nav_field_simpletext(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == VcConfig.newrule_logfld ||
			cur == VcConfig.advancedmodefld ||
			cur == VcConfig.mainmenu_statusfld)
		{
			if(nav_field_toggleX(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case KEY_LEFT:
				case 'q':
					quit = 1;
					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					// Go to next field
					form_driver(ConfigSection.form, REQ_NEXT_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ConfigSection.form, REQ_PREV_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case 127:
				case KEY_BACKSPACE:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_DC:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUURCONF:SETTINGS]:");
					break;

				default:
					/* If this is a normal character, it gets printed into the field */
					form_driver(ConfigSection.form, ch);
					break;
			}
		}

		prev = cur;
		cur = current_field(ConfigSection.form);
	}

	/* save the field to the conf struct */
	if(edit_vcconfig_save() < 0)
	{
		retval = -1;
	}

	/* write configfile */
	if(retval == 0)
	{
		if(write_vcconfigfile(debug, vccnf.configfile_location, &vccnf) < 0)
		{
			(void)vrprint.error(-1, "Error", "Writing config file failed.");
			retval=-1;
		}
	}

	/* cleanup */
	edit_config_destroy();

	return(retval);
}


struct
{
	FIELD	*logdirfld,
		*loglevelfld,
		*systemlogfld,

		*logpolicyfld,
		*logpolicylimitfld;

	char	number[8];

} LogConfig;


static int
edit_logconfig_init(int height, int width, int starty, int startx)
{
	int	retval = 0,
		i,
		rows,
		cols;
	char	limit_string[4] = "";

	ConfigSection.n_fields = 6;
	ConfigSection.fields = (FIELD **)calloc(ConfigSection.n_fields + 1, sizeof(FIELD *));

	/* fields */
	LogConfig.logdirfld    = (ConfigSection.fields[0] = new_field(1, 64, 2, 1, 0, 0));	/* vuurmuur_logdir */
	LogConfig.loglevelfld  = (ConfigSection.fields[1] = new_field(1, 8,  4, 1, 0, 0));	/* loglevel */
	LogConfig.systemlogfld = (ConfigSection.fields[2] = new_field(1, 64, 6, 1, 0, 0));	/* systemlog */

	LogConfig.logpolicyfld      = (ConfigSection.fields[3] = new_field(1, 1, 9,  61, 0, 0));	/* log policy */
	LogConfig.logpolicylimitfld = (ConfigSection.fields[4] = new_field(1, 3, 11, 60, 0, 0));	/* log policy limit */

	ConfigSection.fields[ConfigSection.n_fields] = NULL;


	if(!(ConfigSection.win = create_newwin(height, width, starty, startx, "Edit Config: Logging", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "couldn't create window.");
		return(-1);
	}
	if(!(ConfigSection.panel[0] = new_panel(ConfigSection.win)))
	{
		(void)vrprint.error(-1, "Error", "couldn't create panel.");
		return(-1);
	}

	/* fill the fields */
	set_field_buffer(LogConfig.logdirfld, 0, conf.vuurmuur_logdir_location);
	set_field_buffer(LogConfig.loglevelfld, 0, conf.loglevel);
	set_field_buffer(LogConfig.systemlogfld, 0, conf.systemlog_location);
	set_field_buffer(LogConfig.logpolicyfld, 0, conf.log_policy ? "X" : " ");
	if(conf.log_policy_limit > 0)
	{
		snprintf(limit_string, sizeof(limit_string), "%d", conf.log_policy_limit);
		set_field_buffer(LogConfig.logpolicylimitfld, 0, limit_string);
	}

	for(i = 0; i < ConfigSection.n_fields; i++)
	{
		// set field options
		set_field_back(ConfigSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);
		field_opts_off(ConfigSection.fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(ConfigSection.fields[i], FALSE);
	}
	set_field_back(LogConfig.logpolicyfld, COLOR_PAIR(CP_BLUE_WHITE));

	/* Create the form and post it */
	ConfigSection.form = new_form(ConfigSection.fields);
	/* Calculate the area required for the form */
	scale_form(ConfigSection.form, &rows, &cols);
	keypad(ConfigSection.win, TRUE);
	/* Set main window and sub window */
	set_form_win(ConfigSection.form, ConfigSection.win);
	set_form_sub(ConfigSection.form, derwin(ConfigSection.win, rows, cols, 1, 2));

	post_form(ConfigSection.form);

	/* print labels */
	mvwprintw(ConfigSection.win, 2, 2, "Vuurmuur logfiles location:");
	mvwprintw(ConfigSection.win, 4, 2, "Loglevel (for use with syslog, requires vuurmuur restart):");
	mvwprintw(ConfigSection.win, 6, 2, "System log (for vuurmuur-log):");

	mvwprintw(ConfigSection.win, 10, 2, "Log the default policy? (recommended):");
	mvwprintw(ConfigSection.win, 10, 62, "[");
	mvwprintw(ConfigSection.win, 10, 64, "]");

	mvwprintw(ConfigSection.win, 12, 2, "Limit of the number of logs per second (0 for no limit):");
	return(retval);
}


static int
edit_logconfig_save(const int debuglvl)
{
	int	retval=0,
		i = 0;
	char	limit_string[4] = "";
	int	result = 0;

	/* check for changed fields */
	for(i=0; i < ConfigSection.n_fields; i++)
	{
		/* we only act if a field is changed */
		if(field_status(ConfigSection.fields[i]) == TRUE)
		{
			if(ConfigSection.fields[i] == LogConfig.logdirfld)
			{
				/* vuurmuurlog location */
				if(!(copy_field2buf(	conf.vuurmuur_logdir_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.vuurmuur_logdir_location))))
					return(-1);
				
				if(strlen(conf.vuurmuur_logdir_location) > 0)
				{
					/* cut of the trailing slash if we have any */
					if(conf.vuurmuur_logdir_location[strlen(conf.vuurmuur_logdir_location)-1] == '/')
						conf.vuurmuur_logdir_location[strlen(conf.vuurmuur_logdir_location)-1] = '\0';
				}

				if(config_check_logdir(debuglvl, conf.vuurmuur_logdir_location) < 0)
				{
					retval = -1;
				}
				else
				{
					/* print a warning about apply changes won't work for loglevel */
					(void)vrprint.warning("Warning", "changing the logdir requires applying changes to get into effect in the logviewer.");
				}
			}
			else if(ConfigSection.fields[i] == LogConfig.loglevelfld)
			{
				/* loglevel */
				if(!(copy_field2buf(	conf.loglevel,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.loglevel))))
					return(-1);

				/* print a warning about apply changes won't work for loglevel */
				(void)vrprint.warning("Warning", "changing the loglevel requires restarting of Vuurmuur to take effect.");
			}
			else if(ConfigSection.fields[i] == LogConfig.systemlogfld)
			{
				/* systemlog */
				if(!(copy_field2buf(	conf.systemlog_location,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(conf.systemlog_location))))
					return(-1);
			}
			else if(ConfigSection.fields[i] == LogConfig.logpolicyfld)
			{
				/* log policy */
				if(field_buffer(ConfigSection.fields[i], 0)[0] == 'X')
					conf.log_policy = 1;
				else
					conf.log_policy = 0;
			}
			else if(ConfigSection.fields[i] == LogConfig.logpolicylimitfld)
			{
				/* log policy limit */
				if(!(copy_field2buf(	limit_string,
							field_buffer(ConfigSection.fields[i], 0),
							sizeof(limit_string))))
					return(-1);

				result = atoi(limit_string);
				if(result < 0 || result > 999)
				{
					(void)vrprint.error(-1, "Error", "limit must be between 0-999.");

					if(conf.log_policy_limit > 0)
					{
						snprintf(limit_string, sizeof(limit_string), "%d", conf.log_policy_limit);
						set_field_buffer(LogConfig.logpolicylimitfld, 0, limit_string);
					}
				}
				else
				{
					conf.log_policy_limit = result;
				}
			}
			else
			{
				(void)vrprint.error(-1, "Internal Error", "unknown field.");
				return(-1);
			}
		}
	}

	return(retval);
}


int
edit_logconfig(const int debuglvl)
{
	int	ch,
		retval = 0,
		quit = 0,
		not_defined = 0;
	int	height,
		width,
		startx,
		starty,
		max_height,
		max_width;
	FIELD	*cur = NULL,
		*prev = NULL;

	// window dimentions
	getmaxyx(stdscr, max_height, max_width);

	height = 18;
	width = 76;

	startx = (max_width - width)/2;
	starty = (max_height - height)/2;

	// setup
	edit_logconfig_init(height, width, starty, startx);
	cur = current_field(ConfigSection.form);
	update_panels();
	doupdate();

	/* Loop through to get user requests */
	while(quit == 0)
	{
		/* visual support */
		draw_field_active_mark(cur, prev, ConfigSection.win, ConfigSection.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		/* keyboard input */
		ch = wgetch(ConfigSection.win);

		not_defined = 0;

		if(cur == LogConfig.logdirfld ||
		   cur == LogConfig.loglevelfld ||
		   cur == LogConfig.systemlogfld ||
		   cur == LogConfig.logpolicylimitfld)
		{
			if(nav_field_simpletext(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == LogConfig.logpolicyfld)
		{
			if(nav_field_toggleX(ConfigSection.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(10):
				case 27:
				case KEY_LEFT:
				case 'q':

					/* save the field to the conf struct */
					if(edit_logconfig_save(debuglvl) < 0)
					{
						if(confirm("Saving config failed", "Saving the config failed. Sure you want to quit?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0))
						{
							retval = -1;
							quit = 1;
						}
					}
					else
					{
						quit = 1;
						retval = 0;
					}

					break;
			
				case KEY_F(5): //help
					status_print(status_win, "FIXME: help not yet implemented");
					break;
			
				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					// Go to next field
					form_driver(ConfigSection.form, REQ_NEXT_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;
				
				case KEY_UP:
					// Go to previous field
					form_driver(ConfigSection.form, REQ_PREV_FIELD);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case 127:
				case KEY_BACKSPACE:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_DC:
					form_driver(ConfigSection.form, REQ_PREV_CHAR);
					form_driver(ConfigSection.form, REQ_DEL_CHAR);
					form_driver(ConfigSection.form, REQ_END_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:CONFIG:LOGGING]:");
					break;

				default:
					// If this is a normal character, it gets printed into the field
					form_driver(ConfigSection.form, ch);
					break;
			}
		}

		prev = cur;
		cur = current_field(ConfigSection.form);
	}

	/* write configfile */
	if(retval == 0)
	{
		if(write_configfile(debug, conf.configfile_location) < 0)
		{
			(void)vrprint.error(-1, "Error", "Writing config file failed.");
			retval=-1;
		}
	}

	/* cleanup */
	edit_config_destroy();

	return(retval);
}

int
config_menu(const int debuglvl)
{
	int	retval = 0,
		n_choices = 7,
		i = 0,
		ch = 0,
		quit = 0;
	ITEM	**menu_items;
	MENU	*main_menu;
	WINDOW	*mainmenu_win;
	PANEL	*my_panels[1];

	// menu
	char *choice_ptr = NULL;

	char *choices[] = {
			"General            ",
			"Connections        ",
			"System Protection  ",
			"Logging            ",
			"Plugins            ",
			"Back               ",
			NULL
	};

	char *descriptions[] = {
			" ",
			" ",
			" ",
			" ",
			" ",
			" ",
			NULL
	};

	int	x = 50,
		y = 9 + n_choices,
		startx = 0,
		starty = 0,
		maxx = 0,
		maxy = 0;

	getmaxyx(stdscr, maxy, maxx);
	startx = (maxx-x)/2;
	starty = (maxy-y)/2;

	if(!(menu_items = (ITEM **)calloc(n_choices + 1, sizeof(ITEM *))))
		return(-1);

	for(i = 0; i < n_choices; ++i){
		menu_items[i] = new_item(choices[i], descriptions[i]);
	}
	menu_items[n_choices] = (ITEM *)NULL;

	main_menu = new_menu((ITEM **)menu_items);

	mainmenu_win = create_newwin(y, x, starty, startx, "Configuration Menu", COLOR_PAIR(5));
	keypad(mainmenu_win, TRUE);
	wrefresh(mainmenu_win);

	my_panels[0] = new_panel(mainmenu_win);

	// menu settings
	set_menu_win(main_menu, mainmenu_win);
	set_menu_sub(main_menu, derwin(mainmenu_win, y-8, x-12, 6, 6));
	set_menu_format(main_menu, y-4, 1);
	set_menu_back(main_menu, COLOR_PAIR(5));
	set_menu_fore(main_menu, COLOR_PAIR(3));

	post_menu(main_menu);

	// welcome message
	mvwprintw(mainmenu_win, 3, 6, "Select a section.");

	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(mainmenu_win);
		switch(ch)
		{
			case 27:
			case KEY_LEFT:
			case 'q':
			case KEY_F(10):
				quit=1;
				retval=1;
				break;
			
			case KEY_DOWN:
				menu_driver(main_menu, REQ_DOWN_ITEM);
				break;

			case KEY_UP:
				menu_driver(main_menu, REQ_UP_ITEM);
				break;

			case KEY_RIGHT:
			case 32: // space
			case 10: // enter
			{
				ITEM *cur;
				cur = current_item(main_menu);

				choice_ptr = malloc(strlen((char *)item_name(cur))+1);
				strcpy(choice_ptr, (char *)item_name(cur));

				/* we now have a choice, so quit the loop */
				quit = 1;
				break;
			}
		}
	}

	unpost_menu(main_menu);
	free_menu(main_menu);
	for(i = 0; i < n_choices; ++i)
		free_item(menu_items[i]);
	free(menu_items);

	del_panel(my_panels[0]);

	destroy_win(mainmenu_win);

	update_panels();
	doupdate();

	if(choice_ptr != NULL)
	{
		if(strncasecmp(choice_ptr, "general", 7) == 0)
		{
			edit_genconfig(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "connections", 11) == 0)
		{
			edit_conconfig(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "system protection", 17) == 0)
		{
			edit_sysopt(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "logging", 7) == 0)
		{
			edit_logconfig(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "plugins", 7) == 0)
		{
			edit_plugconfig(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "back", 4) == 0)
		{
			retval=1;
		}

		free(choice_ptr);
	}

	return(retval);
}
