/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


struct InterfacesSection_
{
	WINDOW  *win;
	PANEL   *panel[1];
	MENU    *menu;
	ITEM    **items;

	int     list_items;

	struct EditInterface_
	{
		WINDOW  *win;
		PANEL   *panel[1];
		FORM    *form;
		FIELD   **fields;
		int     n_fields;
	} EditInterface;

	char comment[241];

} InterfacesSection;


struct
{
	FIELD	*activefld,
		*activelabelfld,

		*ipaddressfld,
		*ipaddresslabelfld,

		*dynamicfld,
		*dynamiclabelfld,
		*dynamicbracketsfld,

		*devicefld,
		*devicelabelfld,

		*devicevirtualfld,
		*devicevirtuallabelfld,
		*devicevirtualbracketsfld,

		*interfaceupfld,
		*interfaceuplabelfld,

		*commentfld,
		*commentlabelfld;

	FIELD	*srcrtpktsfld,
		*srcrtpktslabelfld,
		*srcrtpktsbracketsfld,
		
		*icmpredirectfld,
		*icmpredirectlabelfld,
		*icmpredirectbracketsfld,

		*sendredirectfld,
		*sendredirectlabelfld,
		*sendredirectbracketsfld,
		
		*rpfilterfld,
		*rpfilterlabelfld,
		*rpfilterbracketsfld,
		
		*logmartiansfld,
		*logmartianslabelfld,
		*logmartiansbracketsfld;
} IfSec;


int
protectrule_loaded(const int debuglvl, d_list *rules_list, char *danger, char *source)
{
	struct RuleData_	*rule_ptr = NULL;
	d_list_node		*d_node = NULL;

	if(!rules_list || !danger)
	{
		return(0);
	}

	if(rules_list->len == 0)
		return(0);

	for(d_node = rules_list->top; d_node; d_node = d_node->next)
	{
		if(!(rule_ptr = d_node->data))
		{
			return(0);
		}

		if(strcmp(danger, rule_ptr->danger) == 0)
		{
			if(!source)
			{
				return(1);
			}
			else
			{
				if(strcmp(source, rule_ptr->source) == 0)
				{
					return(1);
				}
			}
		}
	}

	return(0);
}

static int
edit_interface_init(const int debuglvl, char *name, int height, int width, int startx, int starty, struct InterfaceData_ *iface_ptr)
{
	int	retval=0,	// return value
		i,
		rows,
		cols,
		max_height,
		max_width,
		comment_y=0,	// number of lines of the commentfield
		comment_x=0;	// number of colums of the commentfield
	int	field_num = 0;
		
	getmaxyx(stdscr, max_height, max_width);

	InterfacesSection.EditInterface.n_fields = 31;
	InterfacesSection.EditInterface.fields = (FIELD **)calloc(InterfacesSection.EditInterface.n_fields + 1, sizeof(FIELD *));

	/*
		create the fields
	*/

	/* active */
	IfSec.activelabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 6, 2, 0, 0, 0));
	field_num++;
	set_field_buffer(IfSec.activelabelfld, 0, "Active");
	field_opts_off(IfSec.activelabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.activefld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 3, 1, 0, 0));
	field_num++;
	set_field_buffer(IfSec.activefld, 0, iface_ptr->active ? "Yes" : "No");


	/* ipaddress */
	IfSec.ipaddresslabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 9, 5, 0, 0, 0));
	field_num++;
	set_field_buffer(IfSec.ipaddresslabelfld, 0, "IPAddress");
	field_opts_off(IfSec.ipaddresslabelfld, O_AUTOSKIP | O_ACTIVE);
	
	IfSec.ipaddressfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 16, 6, 1, 0, 0));
	field_num++;
	set_field_type(IfSec.ipaddressfld, TYPE_IPV4);
	set_field_buffer(IfSec.ipaddressfld, 0, iface_ptr->ipv4.ipaddress);

	/* if ipaddress is dynamic, we don't want to edit the ipaddress */
	if(iface_ptr->dynamic)
		field_opts_off(IfSec.ipaddressfld, O_AUTOSKIP | O_ACTIVE);


	/* dynamic ip toggle */
	IfSec.dynamicbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 6, 21, 0, 0));
	field_num++;
	set_field_buffer(IfSec.dynamicbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.dynamicbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.dynamiclabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 8, 5, 20, 0, 0));
	field_num++;
	set_field_buffer(IfSec.dynamiclabelfld, 0, "Dynamic");
	field_opts_off(IfSec.dynamiclabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.dynamicfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 6, 22, 0, 0));
	field_num++;
	set_field_buffer(IfSec.dynamicfld, 0, iface_ptr->dynamic ? "X" : " ");


	/* device */
	IfSec.devicelabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 6, 7, 0, 0, 0));
	field_num++;
	set_field_buffer(IfSec.devicelabelfld, 0, "Device");
	field_opts_off(IfSec.devicelabelfld, O_AUTOSKIP | O_ACTIVE);
	
	IfSec.devicefld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 12, 8, 1, 0, 0));
	field_num++;
	set_field_buffer(IfSec.devicefld, 0, iface_ptr->device);


	/* is the device virtual */
	IfSec.devicevirtualbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 8, 21, 0, 0));
	field_num++;
	set_field_buffer(IfSec.devicevirtualbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.devicevirtualbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.devicevirtuallabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 8, 7, 20, 0, 0));
	field_num++;
	set_field_buffer(IfSec.devicevirtuallabelfld, 0, "Virtual");
	field_opts_off(IfSec.devicevirtuallabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.devicevirtualfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 8, 22, 0, 0));
	field_num++;
	set_field_buffer(IfSec.devicevirtualfld, 0, iface_ptr->device_virtual ? "X" : " ");


	/* source routed packets */
	IfSec.srcrtpktsbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 4, 54, 0, 0));
	field_num++;
	set_field_buffer(IfSec.srcrtpktsbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.srcrtpktsbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.srcrtpktslabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 14, 4, 38, 0, 0));
	field_num++;
	set_field_buffer(IfSec.srcrtpktslabelfld, 0, "Src-rt-pkts");
	field_opts_off(IfSec.srcrtpktslabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.srcrtpktsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 4, 55, 0, 0));
	field_num++;
	set_field_buffer(IfSec.srcrtpktsfld, 0, protectrule_loaded(debuglvl, &iface_ptr->ProtectList, "source-routed-packets", NULL) ? "X" : " ");

	/* icmp redirects */
	IfSec.icmpredirectbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 5, 54, 0, 0));
	field_num++;
	set_field_buffer(IfSec.icmpredirectbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.icmpredirectbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.icmpredirectlabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 14, 5, 38, 0, 0));
	field_num++;
	set_field_buffer(IfSec.icmpredirectlabelfld, 0, "Icmp-redirect");
	field_opts_off(IfSec.icmpredirectlabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.icmpredirectfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 5, 55, 0, 0));
	field_num++;
	set_field_buffer(IfSec.icmpredirectfld, 0, protectrule_loaded(debuglvl, &iface_ptr->ProtectList, "icmp-redirect", NULL) ? "X" : " ");

	/* send redirect */
	IfSec.sendredirectbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 6, 54, 0, 0));
	field_num++;
	set_field_buffer(IfSec.sendredirectbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.sendredirectbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.sendredirectlabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 14, 6, 38, 0, 0));
	field_num++;
	set_field_buffer(IfSec.sendredirectlabelfld, 0, "Send-redirect");
	field_opts_off(IfSec.sendredirectlabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.sendredirectfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 6, 55, 0, 0));
	field_num++;
	set_field_buffer(IfSec.sendredirectfld, 0, protectrule_loaded(debuglvl, &iface_ptr->ProtectList, "send-redirect", NULL) ? "X" : " ");

	/* rp filter */
	IfSec.rpfilterbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 7, 54, 0, 0));
	field_num++;
	set_field_buffer(IfSec.rpfilterbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.rpfilterbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.rpfilterlabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 14, 7, 38, 0, 0));
	field_num++;
	set_field_buffer(IfSec.rpfilterlabelfld, 0, "Rp-filter");
	field_opts_off(IfSec.rpfilterlabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.rpfilterfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 7, 55, 0, 0));
	field_num++;
	set_field_buffer(IfSec.rpfilterfld, 0, protectrule_loaded(debuglvl, &iface_ptr->ProtectList, "rp-filter", NULL) ? "X" : " ");

	/* log martians */
	IfSec.logmartiansbracketsfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 8, 54, 0, 0));
	field_num++;
	set_field_buffer(IfSec.logmartiansbracketsfld, 0, "[ ]");
	field_opts_off(IfSec.logmartiansbracketsfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.logmartianslabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 14, 8, 38, 0, 0));
	field_num++;
	set_field_buffer(IfSec.logmartianslabelfld, 0, "Log martians");
	field_opts_off(IfSec.logmartianslabelfld, O_AUTOSKIP | O_ACTIVE);

	IfSec.logmartiansfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 1, 8, 55, 0, 0));
	field_num++;
	set_field_buffer(IfSec.logmartiansfld, 0, protectrule_loaded(debuglvl, &iface_ptr->ProtectList, "log-martians", NULL) ? "X" : " ");


	/* comment */
	IfSec.commentlabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 7, 10, 0, 0, 0));
	field_num++;
	set_field_buffer(IfSec.commentlabelfld, 0, "Comment");
	field_opts_off(IfSec.commentlabelfld, O_AUTOSKIP | O_ACTIVE);

	comment_y = 5;
	comment_x = 48;

	IfSec.commentfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(comment_y, comment_x, 11, 1, 0, 0));
	field_num++;
	if(af->ask(debug, ifac_backend, iface_ptr->name, "COMMENT", InterfacesSection.comment, sizeof(InterfacesSection.comment), TYPE_INTERFACE, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	set_field_buffer(IfSec.commentfld, 0, InterfacesSection.comment);


	/* up? */
	IfSec.interfaceuplabelfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 16, 2, 20, 0, 0));
	field_num++;
	set_field_buffer(IfSec.interfaceuplabelfld, 0, "Interface up?");
	field_opts_off(IfSec.interfaceuplabelfld, O_ACTIVE);

	IfSec.interfaceupfld = (InterfacesSection.EditInterface.fields[field_num] = new_field(1, 3, 3, 21, 0, 0));
	field_num++;
	set_field_buffer(IfSec.interfaceupfld, 0, iface_ptr->up ? "Yes" : "No");
	field_opts_off(IfSec.interfaceupfld, O_ACTIVE);

	/* terminate the fields */
	InterfacesSection.EditInterface.fields[InterfacesSection.EditInterface.n_fields] = NULL;

	if(InterfacesSection.EditInterface.n_fields != field_num)
		(void)vrprint.error(-1, "Internal Error", "InterfacesSection.EditInterface.n_fields != field_num.");

	/* create the window & panel */
	InterfacesSection.EditInterface.win = create_newwin(height, width, startx, starty, "Edit Interface", vccnf.color_win);
	InterfacesSection.EditInterface.panel[0] = new_panel(InterfacesSection.EditInterface.win);
	keypad(InterfacesSection.EditInterface.win, TRUE);

	for(i = 0; i < InterfacesSection.EditInterface.n_fields; i++)
	{
		// set field options
		set_field_back(InterfacesSection.EditInterface.fields[i], vccnf.color_win_field);

		field_opts_off(InterfacesSection.EditInterface.fields[i], O_AUTOSKIP);
		// set status to false
		set_field_status(InterfacesSection.EditInterface.fields[i], FALSE);
	}

	/* labels are blue-white */
	set_field_back(IfSec.activelabelfld, vccnf.color_win);
	set_field_back(IfSec.ipaddresslabelfld, vccnf.color_win);
	set_field_back(IfSec.devicelabelfld, vccnf.color_win);
	set_field_back(IfSec.commentlabelfld, vccnf.color_win);
	set_field_back(IfSec.interfaceuplabelfld, vccnf.color_win);

	set_field_back(IfSec.dynamicfld, vccnf.color_win);
	set_field_back(IfSec.dynamiclabelfld, vccnf.color_win);
	set_field_back(IfSec.dynamicbracketsfld, vccnf.color_win);

	set_field_back(IfSec.devicevirtualfld, vccnf.color_win);
	set_field_back(IfSec.devicevirtuallabelfld, vccnf.color_win);
	set_field_back(IfSec.devicevirtualbracketsfld, vccnf.color_win);


	/* the toggles */

	set_field_back(IfSec.srcrtpktsfld, vccnf.color_win);
	set_field_back(IfSec.srcrtpktslabelfld, vccnf.color_win);
	set_field_back(IfSec.srcrtpktsbracketsfld, vccnf.color_win);
		
	set_field_back(IfSec.icmpredirectfld, vccnf.color_win);
	set_field_back(IfSec.icmpredirectlabelfld, vccnf.color_win);
	set_field_back(IfSec.icmpredirectbracketsfld, vccnf.color_win);

	set_field_back(IfSec.sendredirectfld, vccnf.color_win);
	set_field_back(IfSec.sendredirectlabelfld, vccnf.color_win);
	set_field_back(IfSec.sendredirectbracketsfld, vccnf.color_win);
		
	set_field_back(IfSec.rpfilterfld, vccnf.color_win);
	set_field_back(IfSec.rpfilterlabelfld, vccnf.color_win);
	set_field_back(IfSec.rpfilterbracketsfld, vccnf.color_win);
		
	set_field_back(IfSec.logmartiansfld, vccnf.color_win);
	set_field_back(IfSec.logmartianslabelfld, vccnf.color_win);
	set_field_back(IfSec.logmartiansbracketsfld, vccnf.color_win);

	// Create the form and post it
	InterfacesSection.EditInterface.form = new_form(InterfacesSection.EditInterface.fields);
	scale_form(InterfacesSection.EditInterface.form, &rows, &cols);
	set_form_win(InterfacesSection.EditInterface.form, InterfacesSection.EditInterface.win);
	set_form_sub(InterfacesSection.EditInterface.form, derwin(InterfacesSection.EditInterface.win, rows, cols, 1, 2));

	post_form(InterfacesSection.EditInterface.form);

	mvwprintw(InterfacesSection.EditInterface.win, 1, 2,  "Name: %s", iface_ptr->name);
	mvwprintw(InterfacesSection.EditInterface.win, 3, 40,  "Protection");
	
	wrefresh(InterfacesSection.EditInterface.win);

	return(retval);
}


static int
edit_interface_destroy(void)
{
	int retval=0, i;

	// Un post form and free the memory
	unpost_form(InterfacesSection.EditInterface.form);
	free_form(InterfacesSection.EditInterface.form);

	for(i=0;i<InterfacesSection.EditInterface.n_fields;i++)
	{
		free_field(InterfacesSection.EditInterface.fields[i]);
	}
	free(InterfacesSection.EditInterface.fields);

	del_panel(InterfacesSection.EditInterface.panel[0]);
	destroy_win(InterfacesSection.EditInterface.win);

	update_panels();
	doupdate();

	strcpy(InterfacesSection.comment, "");
	
	return(retval);
}


/*	interfaces_save_protectrules

	Save the protectrules to the backend.
	
	Returncodes:
		 0: ok
		-1: error
*/
int
interfaces_save_protectrules(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;
	char			rule_str[MAX_RULE_LENGTH] = "";
	
	/* safety */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* safety */
	if(iface_ptr->ProtectList.len < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "negative number of protectrules (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* write to backend */
	if(iface_ptr->ProtectList.len == 0)
	{
		/* clear */
		if(af->tell(debuglvl, ifac_backend, iface_ptr->name, "RULE", "", 1, TYPE_INTERFACE) < 0)
		{
			(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}
	else
	{
		/* write to backend */
		for(d_node = iface_ptr->ProtectList.top; d_node; d_node = d_node->next)
		{
			if(!(rule_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			snprintf(rule_str, sizeof(rule_str), "protect against %s", rule_ptr->danger);

			if(d_node == iface_ptr->ProtectList.top)
			{
				/* save to backend */
				if(af->tell(debuglvl, ifac_backend, iface_ptr->name, "RULE", rule_str, 1, TYPE_INTERFACE) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
			else
			{
				/* save to backend */
				if(af->tell(debuglvl, ifac_backend, iface_ptr->name, "RULE", rule_str, 0, TYPE_INTERFACE) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
		}
	}
	
	return(0);
}


struct RuleData_ *
create_protect_rule(const int debuglvl, char *who, char *danger, /*@null@*/ char *source)
{
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!who || !danger)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(NULL);
	}

	/* get mem (rule_malloc will report error) */
	if(!(rule_ptr = rule_malloc()))
		return(NULL);

	/* who do we protect */
	strlcpy(rule_ptr->who, who, sizeof(rule_ptr->who));

	/* and against what? */
	strlcpy(rule_ptr->danger, danger, sizeof(rule_ptr->danger));

	if(source != NULL)
	{
		/* from which source */
		strlcpy(rule_ptr->source, source, sizeof(rule_ptr->source));
	}

	return(rule_ptr);
}


static int
edit_interface_save_protectrules(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);	
	}

	d_list_cleanup(debuglvl, &iface_ptr->ProtectList);

	if(field_buffer(IfSec.srcrtpktsfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, iface_ptr->name, "source-routed-packets", NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(IfSec.icmpredirectfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, iface_ptr->name, "icmp-redirect", NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(IfSec.sendredirectfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, iface_ptr->name, "send-redirect", NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(IfSec.rpfilterfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, iface_ptr->name, "rp-filter", NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(IfSec.logmartiansfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, iface_ptr->name, "log-martians", NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &iface_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	/* now let try to write this to the backend */
	if(interfaces_save_protectrules(debuglvl, iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


static int
edit_interface_save(const int debuglvl, struct InterfaceData_ *iface_ptr)
{
	int			retval = 0,
				i,
				result = 0,
				status = 0;
	struct InterfaceData_	*tempiface_ptr = NULL;

	/* safety */
	if(!iface_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);	
	}

	/* get a temp interface */
	tempiface_ptr = interface_malloc(debuglvl);
	if(!tempiface_ptr)
		return(-1);

	/* copy the interface to the temp one */
	*tempiface_ptr = *iface_ptr;

	/* check for changed fields */
	for(i=0; i < InterfacesSection.EditInterface.n_fields; i++)
	{
		if(field_status(InterfacesSection.EditInterface.fields[i]) == TRUE)
		{
			/* active */
			if(InterfacesSection.EditInterface.fields[i] == IfSec.activefld)
			{
				status = ST_CHANGED;
				
				if(strncasecmp(field_buffer(InterfacesSection.EditInterface.fields[i], 0), "yes", 3) == 0)
				{
					tempiface_ptr->active = 1;
				}
				else if(strncasecmp(field_buffer(InterfacesSection.EditInterface.fields[i], 0), "no", 2) == 0)
				{
					tempiface_ptr->active = 0;
				}
				else
				{
					tempiface_ptr->active = -1;
				}

				result = af->tell(debug, ifac_backend, tempiface_ptr->name, "ACTIVE", tempiface_ptr->active ? "Yes" : "No", 1, TYPE_INTERFACE);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, active).");
					retval=-1;
				}
			}
			else if(InterfacesSection.EditInterface.fields[i] == IfSec.ipaddressfld)
			{
				// ipaddress
				status = ST_CHANGED;

				if(!(copy_field2buf(	tempiface_ptr->ipv4.ipaddress,
							field_buffer(InterfacesSection.EditInterface.fields[i], 0),
							sizeof(tempiface_ptr->ipv4.ipaddress))))
					return(-1);

				result = af->tell(debug, ifac_backend, tempiface_ptr->name, "IPADDRESS", tempiface_ptr->ipv4.ipaddress, 1, TYPE_INTERFACE);
				if(result < 0)
				{
					vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, ipaddress).");
					retval=-1;
				}
			}
			else if(InterfacesSection.EditInterface.fields[i] == IfSec.dynamicfld)
			{
				// active
				status = ST_CHANGED;

				if(strncasecmp(field_buffer(InterfacesSection.EditInterface.fields[i], 0), "X", 1) == 0)
				{
					tempiface_ptr->dynamic = 1;

					result = af->tell(debug, ifac_backend, tempiface_ptr->name, "IPADDRESS", "dynamic", 1, TYPE_INTERFACE);
					if(result < 0)
					{
						vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, dynamic).");
						retval=-1;
					}
				}
				else if(strncasecmp(field_buffer(InterfacesSection.EditInterface.fields[i], 0), " ", 1) == 0)
				{
					tempiface_ptr->dynamic = 0;

					result = af->tell(debug, ifac_backend, tempiface_ptr->name, "IPADDRESS", tempiface_ptr->ipv4.ipaddress, 1, TYPE_INTERFACE);
					if(result < 0)
					{
						vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, dynamic).");
						retval=-1;
					}
				}
				else
				{
					tempiface_ptr->dynamic = -1;
				}
			}
			else if(InterfacesSection.EditInterface.fields[i] == IfSec.devicefld)
			{
				status = ST_CHANGED;

				if(!(copy_field2buf(	tempiface_ptr->device,
							field_buffer(InterfacesSection.EditInterface.fields[i], 0),
							sizeof(tempiface_ptr->device))))
					return(-1);

				result = af->tell(debug, ifac_backend, tempiface_ptr->name, "INTERFACE", tempiface_ptr->device, 1, TYPE_INTERFACE);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, interface).");
					retval=-1;
				}
			}
			else if(InterfacesSection.EditInterface.fields[i] == IfSec.commentfld)
			{
				result = af->tell(debug, ifac_backend, tempiface_ptr->name, "COMMENT", field_buffer(InterfacesSection.EditInterface.fields[i], 0), 1, TYPE_INTERFACE);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, interface).");
					retval=-1;
				}
			}
			else if(InterfacesSection.EditInterface.fields[i] == IfSec.devicevirtualfld)
			{
				status = ST_CHANGED;

				if(strncasecmp(field_buffer(InterfacesSection.EditInterface.fields[i], 0), "X", 1) == 0)
				{
					tempiface_ptr->device_virtual = 1;
				}
				else
				{
					tempiface_ptr->device_virtual = 0;
				}

				result = af->tell(debug, ifac_backend, tempiface_ptr->name, "VIRTUAL", tempiface_ptr->device_virtual ? "Yes" : "No", 1, TYPE_INTERFACE);
				if(result < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, dynamic).");
					retval = -1;
				}
			}
			/*

			*/
			else if(	InterfacesSection.EditInterface.fields[i] == IfSec.srcrtpktsfld		||
					InterfacesSection.EditInterface.fields[i] == IfSec.icmpredirectfld	|| 
					InterfacesSection.EditInterface.fields[i] == IfSec.sendredirectfld	||
					InterfacesSection.EditInterface.fields[i] == IfSec.rpfilterfld		||
					InterfacesSection.EditInterface.fields[i] == IfSec.logmartiansfld)
			{
				status = ST_CHANGED;

				if(edit_interface_save_protectrules(debuglvl, tempiface_ptr) < 0)
				{
					(void)vrprint.error(-1, "Error", "Saving to backend failed (in: edit_interfaces_save, protectrules).");
					retval = -1;
				}
			}
		}
	}

	if(status == ST_CHANGED)
	{
		*iface_ptr = *tempiface_ptr;
	}
	free(tempiface_ptr);

	return(retval);
}


static void
edit_interface(const int debuglvl, char *name)
{
	int			ch;			/* keystroke catcher */
	int			height,
				width,
				startx,
				starty,
				max_height,
				max_width,
				not_defined;		/* 1 if a key is undefined */

	struct InterfaceData_	*iface_ptr = NULL,
				TempIface;

	FIELD			*cur  = NULL,
				*prev = NULL;

	char			quit = 0,
				advanced_mode = vccnf.advanced_mode;

	/* get screen dimentions */
	getmaxyx(stdscr, max_height, max_width);

	height = 20;
	if(height > max_height - 6)
		height = max_height - 6;

	width  = 62;
	startx = 3;
	starty = 16;

	/* TODO: advanced option */

	/* search the interface in memory */
	iface_ptr = search_interface(debuglvl, name, &InterfacesList);
	if(iface_ptr != NULL)
	{
		TempIface = *iface_ptr;
	}
	else
	{
		(void)vrprint.error(-1, "Error", "Interface not found in memory");
		return;
	}

	edit_interface_init(debuglvl, name, height, width, startx, starty, iface_ptr);
	cur = current_field(InterfacesSection.EditInterface.form);

	// Loop through to get user requests
	while(quit == 0)
	{
		if(advanced_mode)
		{
			field_opts_on(IfSec.devicevirtuallabelfld, O_VISIBLE);
			field_opts_on(IfSec.devicevirtualbracketsfld, O_VISIBLE);
			field_opts_on(IfSec.devicevirtualfld, O_VISIBLE);
		}
		else
		{
			field_opts_off(IfSec.devicevirtualfld, O_VISIBLE);
			field_opts_off(IfSec.devicevirtuallabelfld, O_VISIBLE);
			field_opts_off(IfSec.devicevirtualbracketsfld, O_VISIBLE);
		}

		draw_field_active_mark(cur, prev, InterfacesSection.EditInterface.win, InterfacesSection.EditInterface.form, vccnf.color_win_mark|A_BOLD);

		not_defined = 0;

		// get key
		ch = wgetch(InterfacesSection.EditInterface.win);

/*
			else if(	InterfacesSection.EditInterface.fields[i] == IfSec.srcrtpktsfld		||
					InterfacesSection.EditInterface.fields[i] == IfSec.icmpredirectfld	|| 
					InterfacesSection.EditInterface.fields[i] == IfSec.sendredirectfld	||
					InterfacesSection.EditInterface.fields[i] == IfSec.rpfilterfld		||
					InterfacesSection.EditInterface.fields[i] == IfSec.logmartiansfld)
			{

*/
		/* comment */
		if(cur == IfSec.commentfld)
		{
			if(nav_field_comment(InterfacesSection.EditInterface.form, ch) < 0)
				not_defined = 1;
		}
		/* active */
		else if(cur == IfSec.activefld)
		{
			if(nav_field_yesno(InterfacesSection.EditInterface.form, ch) < 0)
				not_defined = 1;
		}
		/* dynamic */
		else if(cur == IfSec.dynamicfld)
		{
			if(nav_field_toggleX(InterfacesSection.EditInterface.form, ch) < 0)
				not_defined = 1;

			/* set the ipaddress field to active/inactive */
			if(strncmp(field_buffer(cur, 0), "X", 1) == 0)
				field_opts_off(IfSec.ipaddressfld, O_AUTOSKIP | O_ACTIVE);
			else
				field_opts_on(IfSec.ipaddressfld, O_ACTIVE);

		}
		/* device virtual */
		else if(	cur == IfSec.devicevirtualfld	||
				cur == IfSec.srcrtpktsfld	||
				cur == IfSec.icmpredirectfld	||
				cur == IfSec.sendredirectfld	||
				cur == IfSec.rpfilterfld	||
				cur == IfSec.logmartiansfld)
		{
			if(nav_field_toggleX(InterfacesSection.EditInterface.form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == IfSec.ipaddressfld ||
			cur == IfSec.devicefld)
		{
			if(nav_field_simpletext(InterfacesSection.EditInterface.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		// Keys special for this window
		if(not_defined == 1)
		{
			switch(ch)
			{
				case 9:		// tab
				case KEY_DOWN:
					// Go to next field
					form_driver(InterfacesSection.EditInterface.form, REQ_NEXT_FIELD);
					// Go to the end of the present buffer
					// Leaves nicely at the last character
					form_driver(InterfacesSection.EditInterface.form, REQ_END_LINE);
					break;
				case 10: // enter
					// Go to next field
					form_driver(InterfacesSection.EditInterface.form, REQ_NEXT_FIELD);
					form_driver(InterfacesSection.EditInterface.form, REQ_END_LINE);
					break;
				case KEY_UP:
					// Go to previous field
					form_driver(InterfacesSection.EditInterface.form, REQ_PREV_FIELD);
					form_driver(InterfacesSection.EditInterface.form, REQ_END_LINE);
					break;

				case 27: //esc
				case KEY_F(10):
					quit = 1;
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:INTERFACES:EDIT]:");
					break;

				/* enable advanced mode */
				case KEY_F(5):

					if(!advanced_mode)
						advanced_mode = 1;
					else
						advanced_mode = 0;

					break;
			}
		}

		prev = cur;
		cur = current_field(InterfacesSection.EditInterface.form);

		wrefresh(InterfacesSection.EditInterface.win);
		pos_form_cursor(InterfacesSection.EditInterface.form);
	}

	/* update the screen */
	update_panels();
	doupdate();

	/* save changes */
	edit_interface_save(debuglvl, iface_ptr);

	/* destroy */
	edit_interface_destroy();

	return;
}


static int
init_interfaces_section(void)
{
	int			retval = 0,
				i;
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	int			height,
				width,
				startx,
				starty;

	// TODO
	height = 0;
	width = 34;
	startx = 4;
	starty = 1;

	InterfacesSection.list_items = InterfacesList.len;
	InterfacesSection.items = (ITEM **)calloc(InterfacesSection.list_items + 1, sizeof(ITEM *));

	i=0;
	for(d_node = InterfacesList.top; d_node ; d_node = d_node->next)
	{
		iface_ptr = d_node->data;
		//TODO

		// load all interfaces into memory
		InterfacesSection.items[i] = new_item(iface_ptr->name, NULL);
		i++;
	}
	InterfacesSection.items[InterfacesSection.list_items] = (ITEM *)NULL;

	height = InterfacesSection.list_items + 8;
	
	InterfacesSection.win = newwin(height, width, startx, starty);
	wbkgd(InterfacesSection.win, vccnf.color_win);
	keypad(InterfacesSection.win, TRUE);
	wrefresh(InterfacesSection.win);

	InterfacesSection.panel[0] = new_panel(InterfacesSection.win);

	InterfacesSection.menu = new_menu((ITEM **)InterfacesSection.items);
	set_menu_win(InterfacesSection.menu, InterfacesSection.win);
	set_menu_sub(InterfacesSection.menu, derwin(InterfacesSection.win, height-3, width-2, 3, 1));

	set_menu_format(InterfacesSection.menu, height-4, 1);

	box(InterfacesSection.win, 0, 0);
	print_in_middle(InterfacesSection.win, 1, 0, width, "Interfaces", vccnf.color_win);
	mvwaddch(InterfacesSection.win, 2, 0, ACS_LTEE);
	mvwhline(InterfacesSection.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(InterfacesSection.win, 2, width-1, ACS_RTEE);

	set_menu_back(InterfacesSection.menu, vccnf.color_win);
	set_menu_fore(InterfacesSection.menu, vccnf.color_win_rev);

	post_menu(InterfacesSection.menu);

	mvwaddch(InterfacesSection.win, height-5, 0, ACS_LTEE);
	mvwhline(InterfacesSection.win, height-5, 1, ACS_HLINE, width-2);
	mvwaddch(InterfacesSection.win, height-5, width-1, ACS_RTEE);

	mvwprintw(InterfacesSection.win, height-4, 2, "<RET> to edit an interface");
	mvwprintw(InterfacesSection.win, height-3, 2, "<INS> to insert an interface");
	mvwprintw(InterfacesSection.win, height-2, 2, "<DEL> to remove an interface");

	return(retval);
}

static int
destroy_interfaces_section(void)
{
	int	retval = 0,
		i;

	unpost_menu(InterfacesSection.menu);
	free_menu(InterfacesSection.menu);
	for(i = 0; i < InterfacesSection.list_items; ++i)
		free_item(InterfacesSection.items[i]);

	free(InterfacesSection.items);

	del_panel(InterfacesSection.panel[0]);
	destroy_win(InterfacesSection.win);

	return(retval);
}

void
interfaces_section(const int debuglvl, struct rgx_ *reg)
{
	int	result = 0,
		quit = 0,
		reload = 0;
	int	ch;
	int	height,
		width,
		startx,
		starty;
	char	*new_name_ptr = NULL;
	ITEM	*cur = NULL;

	// todo
	height = LINES-8;
	width = 34;
	startx = 4;
	starty = 1;

	result=init_interfaces_section();
	if(result < 0)
		return;

	// refresh screen
	update_panels();
	doupdate();

	while(quit == 0)
	{
		if(reload == 1)
		{
			result=destroy_interfaces_section();
			if(result < 0)
				return;

			result=init_interfaces_section();
			if(result < 0)
				return;

			// refresh screen
			update_panels();
			doupdate();
			reload=0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(InterfacesSection.win);
			switch(ch)
			{
				case KEY_LEFT:
				case 27: //esc
				case 'q':
				case KEY_F(10): //quit
					quit = 1;
					break;

				case KEY_IC: //insert
					
					new_name_ptr = input_box(32, "New Interface", "Please enter the name of the new interface:");
					if(new_name_ptr != NULL)
					{
						if(validate_interfacename(debuglvl, new_name_ptr, reg->interfacename) == 0)
						{
							result = new_interface(debug, new_name_ptr, &InterfacesList);
							if(result == 0)
							{
								reload = 1;
								edit_interface(debuglvl, new_name_ptr);
							}
							else
							{
								(void)vrprint.error(-1, "Error", "creating new interface failed.");
							}
						}
						free(new_name_ptr);
					}
					break;

				case KEY_DC: // delete
					
					cur = current_item(InterfacesSection.menu);
					if(cur)
					{
						if(confirm("Delete", "Are you sure?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{
						
							result = delete_interface(debug, (char *)item_name(cur), &InterfacesList);
							if(result < 0)
							{
								(void)vrprint.error(-1, "Error", "deleting interface %s failed.", (char *)item_name(cur));
							}
							
							reload = 1;
						}
					}
					break;

				case KEY_DOWN:
					menu_driver(InterfacesSection.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(InterfacesSection.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(InterfacesSection.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(InterfacesSection.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(InterfacesSection.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(InterfacesSection.menu, REQ_LAST_ITEM); // page down
					break;

				case 32: // space
				case KEY_RIGHT:
				case 10: // enter
				{
					cur = current_item(InterfacesSection.menu);
					if(cur)
					{
						draw_top_menu_edit_interface(top_win);
						(void)edit_interface(debuglvl, (char *)item_name(cur));
						draw_top_menu_interfaces(top_win);
					}
					break;
				}

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:INTERFACES]:");
					break;
			}
		}
	}

//TODO
	result = destroy_interfaces_section();

	// refresh screen
	update_panels();
	doupdate();
}
