/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "main.h"

int
mm_select_logfile(const int debuglvl)
{
	int	retval = 0,
		n_choices = 5,
		i = 0,
		ch = 0,
		quit = 0;
	ITEM	**menu_items;
	MENU	*main_menu;
	WINDOW	*mainmenu_win;
	PANEL	*my_panels[1];

	/* menu */
	char *choice_ptr = NULL;

	char *choices[] = {
			"Traffic.log     ",
			"Vuurmuur.log    ",
			"Error.log       ",
			"Debug.log       ",
			"Back            ",
			NULL
	};

	char *descriptions[] = {
			" ",
			" ",
			" ",
			" ",
			" ",
			NULL
	};

	int	x = 40,
		y = 9 + n_choices,
		startx = 0,
		starty = 0,
		maxx = 0,
		maxy = 0;

	getmaxyx(stdscr, maxy, maxx);
	startx = (maxx-x)/2;
	starty = (maxy-y)/2;

	if(!(menu_items = (ITEM **)calloc(n_choices + 1, sizeof(ITEM *))))
		return(-1);

	for(i = 0; i < n_choices; ++i){
		menu_items[i] = new_item(choices[i], descriptions[i]);
	}
	menu_items[n_choices] = (ITEM *)NULL;

	main_menu = new_menu((ITEM **)menu_items);

	mainmenu_win = create_newwin(y, x, starty, startx, "Logview Menu", COLOR_PAIR(5));
	keypad(mainmenu_win, TRUE);
	wrefresh(mainmenu_win);

	my_panels[0] = new_panel(mainmenu_win);

	// menu settings
	set_menu_win(main_menu, mainmenu_win);
	set_menu_sub(main_menu, derwin(mainmenu_win, y-8, x-12, 6, 6));
	set_menu_format(main_menu, y-4, 1);
	set_menu_back(main_menu, COLOR_PAIR(5));
	set_menu_fore(main_menu, COLOR_PAIR(3));

	post_menu(main_menu);

	// welcome message
	mvwprintw(mainmenu_win, 3, 6, "Select a log to view.");

	update_panels();
	doupdate();

	while(quit == 0)
	{
		ch = wgetch(mainmenu_win);
		switch(ch)
		{
			case 27:
			case KEY_LEFT:
			case 'q':
			case KEY_F(10):
				quit=1;
				retval=1;
				break;
			
			case KEY_DOWN:
				menu_driver(main_menu, REQ_DOWN_ITEM);
				break;

			case KEY_UP:
				menu_driver(main_menu, REQ_UP_ITEM);
				break;

			case KEY_RIGHT:
			case 32: // space
			case 10: // enter
			{
				ITEM *cur;
				cur = current_item(main_menu);

				choice_ptr = malloc(strlen((char *)item_name(cur))+1);
				strcpy(choice_ptr, (char *)item_name(cur));

				/* we now have a choice, so quit the loop */
				quit = 1;
				break;
			}
		}
	}

	unpost_menu(main_menu);
	free_menu(main_menu);
	for(i = 0; i < n_choices; ++i)
		free_item(menu_items[i]);
	free(menu_items);

	del_panel(my_panels[0]);

	destroy_win(mainmenu_win);

	update_panels();
	doupdate();

	if(choice_ptr != NULL)
	{
		if(strncasecmp(choice_ptr, "traffic.log", 11) == 0)
		{
			logview_section(debuglvl, &ZonedataList, &BlockList, "traffic.log");
		}
		else if(strncasecmp(choice_ptr, "error.log", 9) == 0)
		{
			logview_section(debuglvl, &ZonedataList, &BlockList, "error.log");
		}
		else if(strncasecmp(choice_ptr, "vuurmuur.log", 12) == 0)
		{
			logview_section(debuglvl, &ZonedataList, &BlockList, "vuurmuur.log");
		}
		else if(strncasecmp(choice_ptr, "debug.log", 9) == 0)
		{
			logview_section(debuglvl, &ZonedataList, &BlockList, "debug.log");
		}
		else if(strncasecmp(choice_ptr, "back", 4) == 0)
		{
			retval = 1;
		}

		free(choice_ptr);
	}

	return(retval);
}


void
set_colors(const int debuglvl, vc_cnf *cnf)
{
	if(!cnf)
		return;

	init_pair(CP_WIN,      cnf->win_fore, cnf->win_back);
	init_pair(CP_WIN_MARK, COLOR_RED,     cnf->win_back);
	init_pair(CP_WIN_REV,  cnf->win_back, cnf->win_fore);

	cnf->color_win      = COLOR_PAIR(CP_WIN);
	cnf->color_win_rev  = COLOR_PAIR(CP_WIN_REV);
	cnf->color_win_mark = COLOR_PAIR(CP_WIN_MARK);

	return;
}


static void
mm_shm_connect_vuurmuur(const int debuglvl)
{
	/* first try to detach */
	if(vuurmuur_shmp != NULL)
		(void)shmdt(vuurmuur_shmp);

	/* reset */
	vuurmuur_shmid = -1;
	vuurmuur_semid = -1;
	vuurmuur_shmtable = NULL;
	vuurmuur_shmp = NULL;

	/* reconnect */
	vuurmuur_pid = get_vuurmuur_pid("/var/run/vuurmuur.pid", &vuurmuur_shmid);
	if(vuurmuur_shmid > 0)
	{
		/* attach to shared memory */
		vuurmuur_shmp = shmat(vuurmuur_shmid, 0, 0);
		if(vuurmuur_shmp == (char *)(-1))
		{
			(void)vrprint.error(-1, "Error", "Attaching to shared memory failed: %s.", strerror(errno));
			vuurmuur_shmp = NULL;
		}
		else
		{
			vuurmuur_shmtable = (struct SHM_TABLE *)vuurmuur_shmp;
			vuurmuur_semid = vuurmuur_shmtable->sem_id;

			/* now try to connect to the shared memory */
			if(SILENT_LOCK(vuurmuur_semid))
			{
				vuurmuur_shmtable->configtool.connected = 1;
				snprintf(vuurmuur_shmtable->configtool.name, sizeof(vuurmuur_shmtable->configtool.name), "Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
				SILENT_UNLOCK(vuurmuur_semid);
			}
			else
			{
				vuurmuur_shmp = NULL;
			}
		}
	}
}


static void
mm_shm_connect_vuurmuurlog(const int debuglvl)
{
	/* first try to detach */
	if(vuurmuurlog_shmp != NULL)
		(void)shmdt(vuurmuurlog_shmp);

	/* reset */
	vuurmuurlog_shmid = -1;
	vuurmuurlog_semid = -1;
	vuurmuurlog_shmtable = NULL;
	vuurmuurlog_shmp = NULL;

	vuurmuurlog_pid = get_vuurmuur_pid("/var/run/vuurmuur_log.pid", &vuurmuurlog_shmid);
	if(vuurmuurlog_shmid > 0)
	{
		/* attach to shared memory */
		vuurmuurlog_shmp = shmat(vuurmuurlog_shmid, 0, 0);
		if(vuurmuurlog_shmp == (char *)(-1))
		{
			(void)vrprint.error(-1, "Error", "attaching to shared memory failed: %s.", strerror(errno));
			vuurmuurlog_shmp = NULL;
		}
		else
		{
			vuurmuurlog_shmtable = (struct SHM_TABLE *)vuurmuurlog_shmp;
			vuurmuurlog_semid = vuurmuurlog_shmtable->sem_id;

			/* now try to connect to the shared memory */
			if(SILENT_LOCK(vuurmuurlog_semid))
			{
				vuurmuurlog_shmtable->configtool.connected = 1;
				snprintf(vuurmuurlog_shmtable->configtool.name, sizeof(vuurmuurlog_shmtable->configtool.name), "Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
				SILENT_UNLOCK(vuurmuurlog_semid);
			}
			else
			{
				vuurmuurlog_shmp = NULL;
			}
		}
	}
}

static void
queue_status_msg(const int debuglvl, /*@null@*/ d_list *status_list, char *line, int status)
{
	if(!status_list)
		return;

	(void)read_helpline(debuglvl, status_list, line);
	if(status == -1)
		(void)read_helpline(debuglvl, status_list, "(fail).\n");
	else if(status == 0)
		(void)read_helpline(debuglvl, status_list, "(warn).\n");
	else
		(void)read_helpline(debuglvl, status_list, ".\n");
	
	/* one final newline */
	(void)read_helpline(debuglvl, status_list, "\n");
}

/*
TODO: check search script

*/
static void
mm_check_status_settings(const int debuglvl, /*@null@*/ d_list *status_list)
{
	FILE	*fp = NULL;

	/* asume ok */
	VuurmuurStatus.settings = 1;

	if(strcmp(vccnf.helpfile_location, "") == 0)
	{
		VuurmuurStatus.settings = 0;
		queue_status_msg(debuglvl, status_list, "- The path to the Vuurmuur helpfile was not specified\n", VuurmuurStatus.settings);
	}
	else
	{
		if(!(fp = fopen(vccnf.helpfile_location, "r")))
		{
			VuurmuurStatus.settings = 0;
			queue_status_msg(debuglvl, status_list, "- Opening the helpfile failed. Please check the file\n", VuurmuurStatus.settings);
		}
		else
			fclose(fp);
	}
}


static void
mm_check_status_shm(const int debuglvl, /*@null@*/ d_list *status_list)
{
	/* asume ok */
	VuurmuurStatus.vuurmuur = 1;
	VuurmuurStatus.vuurmuur_log = 1;

	/* shm connection with Vuurmuur */
	if(!vuurmuur_shmp)
	{
		VuurmuurStatus.vuurmuur = -1;
		queue_status_msg(debuglvl, status_list, "- No connection could be established with Vuurmuur. Please make sure that it is running\n", VuurmuurStatus.vuurmuur);
	}
	else
	{
		if(!(SILENT_LOCK(vuurmuur_semid)))
		{
			VuurmuurStatus.vuurmuur = -1;
			queue_status_msg(debuglvl, status_list, "- The connection with Vuurmuur seems to be lost. Please make sure that it is running\n", VuurmuurStatus.vuurmuur);
		}
		else
			SILENT_UNLOCK(vuurmuur_semid);
	}

	/* shm connection with Vuurmuur_log */
	if(!vuurmuurlog_shmp)
	{
		VuurmuurStatus.vuurmuur_log = 0;
		queue_status_msg(debuglvl, status_list, "- No connection could be established with Vuurmuur_log. Please make sure that it is running\n", VuurmuurStatus.vuurmuur_log);
	}
	else
	{
		if(!(SILENT_LOCK(vuurmuurlog_semid)))
		{
			VuurmuurStatus.vuurmuur_log = 0;
			queue_status_msg(debuglvl, status_list, "- The connection with Vuurmuur_log seems to be lost. Please make sure that it is running\n", VuurmuurStatus.vuurmuur_log);
		}
		else
			SILENT_UNLOCK(vuurmuurlog_semid);
	}
}


/*
	TODO:
		check if scripts dir exists
*/
static void
mm_check_status_config(const int debuglvl, /*@null@*/ d_list *status_list)
{
	/* asume ok when we start */
	VuurmuurStatus.config = 1;

	if(strcmp(conf.iptables_location, "") == 0)
	{
		VuurmuurStatus.config = -1;
		queue_status_msg(debuglvl, status_list, "- The path to the 'iptables'-command is not yet specified. Please do so in the 'Vuurmuur Config' section\n", VuurmuurStatus.config);
	}
	else
	{
		if(!check_iptables_command(debuglvl, conf.iptables_location, 1))
		{
			VuurmuurStatus.config = -1;
			queue_status_msg(debuglvl, status_list, "- The path to the 'iptables'-command seems to be wrong. There was an error while testing it. Please check it in your system and correct it in the 'Vuurmuur Config' section\n", VuurmuurStatus.config);
		}
	}

	return;
}


/*

*/
static void
mm_check_status_interfaces(const int debuglvl, /*@null@*/ d_list *status_list, d_list *interfaces_list)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	char			at_least_one_active = 0;

	/* safety */
	if(!interfaces_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		VuurmuurStatus.backend = -1;

		return;
	}

	/* asume ok when we start */
	VuurmuurStatus.interfaces = 1;

	if(interfaces_list->len == 0)
	{
		VuurmuurStatus.interfaces = 0;
		queue_status_msg(debuglvl, &VuurmuurStatus.StatusList, "- No interfaces are defined. Please define one or more interfaces\n", VuurmuurStatus.interfaces);
	}

	for(d_node = interfaces_list->top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			VuurmuurStatus.interfaces = -1;

			return;
		}

		if(iface_ptr->active)
			at_least_one_active = 1;
	}

	if(!at_least_one_active)
	{
		VuurmuurStatus.interfaces = 0;
		queue_status_msg(debuglvl, status_list, "- No interfaces are active. Please make sure that at least one of the interfaces is active\n", VuurmuurStatus.interfaces);
	}

	(void)vrprint.debug(__FUNCTION__, "at_least_one_active: %s.", VuurmuurStatus.interfaces ? "Yes" : "No");
	return;
}


/*
*/
static void
mm_check_status_zones(const int debuglvl, /*@null@*/ d_list *status_list, d_list *zones_list)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*zone_ptr = NULL;
	char			at_least_one_active_network = 0;
	char			at_least_one_network = 0;

	if(!zones_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		VuurmuurStatus.zones = -1;

		return;
	}

	/* asume ok when we start */
	VuurmuurStatus.zones = 1;

	/* we need zones */
	if(zones_list->len == 0)
	{
		VuurmuurStatus.zones = 0;
		queue_status_msg(debuglvl, &VuurmuurStatus.StatusList, "- No zones are defined. Please define one or more zones, and at least one network\n", VuurmuurStatus.zones);
	}

	for(d_node = zones_list->top; d_node; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			VuurmuurStatus.zones = -1;

			return;
		}

		if(zone_ptr->type == TYPE_NETWORK)
			at_least_one_network = 1;

		if(zone_ptr->type == TYPE_NETWORK && zone_ptr->active)
			at_least_one_active_network = 1;
	}

	if(!at_least_one_network)
	{
		VuurmuurStatus.zones = 0;
		queue_status_msg(debuglvl, status_list, "- No networks are defined. Please make sure that you define at least one network\n", VuurmuurStatus.zones);
	}
	else if(!at_least_one_active_network)
	{
		VuurmuurStatus.zones = 0;
		queue_status_msg(debuglvl, status_list, "- No networks are active. Please make sure that at least one of the networks is active\n", VuurmuurStatus.zones);
	}

	return;
}


static void
mm_update_overall_status(const int debuglvl)
{
	/* asume all ok */
	VuurmuurStatus.shm     = 1;
	VuurmuurStatus.backend = 1;
	VuurmuurStatus.overall = 1;

	/* backend */
	if(	VuurmuurStatus.zones == 0		||
		VuurmuurStatus.services == 0		||
		VuurmuurStatus.interfaces == 0		||
		VuurmuurStatus.rules == 0
	)
	{
		VuurmuurStatus.backend = 0;
	}
	if(	VuurmuurStatus.zones == -1		||
		VuurmuurStatus.services == -1		||
		VuurmuurStatus.interfaces == -1		||
		VuurmuurStatus.rules == -1
	)
	{
		VuurmuurStatus.backend = -1;
	}

	/* shm */
	if(	VuurmuurStatus.vuurmuur == 0		||
		VuurmuurStatus.vuurmuur_log == 0
	)
	{
		VuurmuurStatus.shm = 0;
	}
	if(	VuurmuurStatus.vuurmuur == -1		||
		VuurmuurStatus.vuurmuur_log == -1
	)
	{
		VuurmuurStatus.shm = -1;
	}

	/* overall */
	if(	VuurmuurStatus.shm == 0			||
		VuurmuurStatus.backend == 0		||
//		VuurmuurStatus.settings == 0		||
		VuurmuurStatus.config == 0		||
		VuurmuurStatus.system == 0
	)
	{
		VuurmuurStatus.overall = 0;
	}
	if(	VuurmuurStatus.shm == -1		||
		VuurmuurStatus.backend == -1		||
		VuurmuurStatus.config == -1		||
//		VuurmuurStatus.settings == -1		||
		VuurmuurStatus.system == -1
	)
	{
		VuurmuurStatus.overall = -1;
	}

	if(debuglvl >= LOW)
		(void)vrprint.debug(__FUNCTION__, "VuurmuurStatus.all: %d.", VuurmuurStatus.overall);
}


static int
mm_reload_shm(const int debuglvl)
{
	WINDOW	*wait_win;
	PANEL	*panel[1];
	FORM	*form;
	FIELD	**fields;
	FIELD	*vuurmuurfld,
		*vuurmuurlogfld;
	int	n_fields = 0,
		cols = 0,
		rows = 0,
		i = 0;

	int	max_height = 0,
		max_width = 0;
	int	vuurmuur_result = 0,
		vuurmuurlog_result = 0;
	int	waittime = 0;
	
	int	vuurmuur_progress = 0,
		vuurmuurlog_progress = 0;

	char	str[4] = "";

	char	failed = 0;

	getmaxyx(stdscr, max_height, max_width);

	/* create a little wait dialog */
	if(!(wait_win = create_newwin(7, 45, (max_height-7)/2, (max_width-45)/2, "One moment please...", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "creating window failed.");
		return(-1);
	}
	panel[0] = new_panel(wait_win);
	
	n_fields = 2;
	fields = (FIELD **)calloc(n_fields + 1, sizeof(FIELD *));

	/* overall */
	vuurmuurfld 	= (fields[0] = new_field(1, 3, 3, 20, 0, 0));
	set_field_buffer(vuurmuurfld, 0, "  0");
	set_field_back(vuurmuurfld, COLOR_PAIR(CP_BLUE_WHITE));

	vuurmuurlogfld	= (fields[1] = new_field(1, 3, 4, 20, 0, 0));
	set_field_buffer(vuurmuurlogfld, 0, "  0");
	set_field_back(vuurmuurlogfld, COLOR_PAIR(CP_BLUE_WHITE));

	/* terminate */
	fields[n_fields] = NULL;

	/* Create the form and post it */
	form = new_form(fields);
	scale_form(form, &rows, &cols);
	set_form_win(form, wait_win);
	set_form_sub(form, derwin(wait_win, rows, cols, 1, 1));
	post_form(form);

	mvwprintw(wait_win, 2, 4, "Applying changes ...");
	mvwprintw(wait_win, 4, 4, "Vuurmuur:            %%");
	mvwprintw(wait_win, 5, 4, "Vuurmuur_log:        %%");

	update_panels();
	doupdate();

	(void)vrprint.info("Info", "Applying changes.");

	/* notify both vuurmuur and vuurmuurlog */
	if(vuurmuur_semid != -1)
	{
		if(LOCK(vuurmuur_semid))
		{
			vuurmuur_shmtable->backend_changed = 1;
			UNLOCK(vuurmuur_semid);

			vuurmuur_result = VR_RR_NO_RESULT_YET;
		}
	}
	else
	{
		vuurmuur_result   = VR_RR_READY;
		vuurmuur_progress = 100;

		snprintf(str, sizeof(str), " - ");
		set_field_buffer(vuurmuurfld, 0, str);

	}
	if(vuurmuurlog_semid != -1)
	{
		if(LOCK(vuurmuurlog_semid))
		{
			vuurmuurlog_shmtable->backend_changed = 1;
			UNLOCK(vuurmuurlog_semid);

			vuurmuurlog_result = VR_RR_NO_RESULT_YET;
		}
	}
	else
	{
		vuurmuurlog_result = VR_RR_READY;
		vuurmuurlog_progress = 100;

		snprintf(str, sizeof(str), " - ");
		set_field_buffer(vuurmuurlogfld, 0, str);
	}

	/* wait max 60 seconds */
	while(	((vuurmuur_result   == VR_RR_NO_RESULT_YET || vuurmuur_result    == VR_RR_RESULT_ACK) ||
		(vuurmuurlog_result == VR_RR_NO_RESULT_YET || vuurmuurlog_result == VR_RR_RESULT_ACK))
		&& waittime < 60000000)
	{
		if(vuurmuur_progress < 100)
		{
			if(SILENT_LOCK(vuurmuur_semid))
			{
				if(vuurmuur_shmtable->reload_result != VR_RR_READY)
				{
					vuurmuur_result   = vuurmuur_shmtable->reload_result;
				}
				vuurmuur_progress = vuurmuur_shmtable->reload_progress;

				SILENT_UNLOCK(vuurmuur_semid);
			}

			snprintf(str, sizeof(str), "%3d", vuurmuur_progress);
			set_field_buffer(vuurmuurfld, 0, str);
		}

		if(vuurmuur_progress == 100)
		{
			if(vuurmuur_semid == -1)
			{
				wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
				mvwprintw(wait_win, 4, 29, "Not connected");
				wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

				failed = 1;
			}
			else if(SILENT_LOCK(vuurmuur_semid))
			{
				vuurmuur_shmtable->reload_result = VR_RR_RESULT_ACK;
				SILENT_UNLOCK(vuurmuur_semid);

				if(vuurmuur_result == VR_RR_SUCCES)
				{
					wattron(wait_win, COLOR_PAIR(CP_GREEN_WHITE));
					mvwprintw(wait_win, 4, 29, "Success");
					wattroff(wait_win, COLOR_PAIR(CP_GREEN_WHITE));
				}
				else if(vuurmuur_result == VR_RR_NOCHANGES)
				{
					mvwprintw(wait_win, 4, 29, "No changes");
				}
				else
				{
					wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
					mvwprintw(wait_win, 4, 29, "Error");
					wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

					failed = 1;
				}
			}
		}

		if(vuurmuurlog_progress < 100)
		{
			if(SILENT_LOCK(vuurmuurlog_semid))
			{
				if(vuurmuurlog_shmtable->reload_result != VR_RR_READY)
				{
					vuurmuurlog_result = vuurmuurlog_shmtable->reload_result;
				}
				vuurmuurlog_progress = vuurmuurlog_shmtable->reload_progress;

				SILENT_UNLOCK(vuurmuurlog_semid);
			}

			snprintf(str, sizeof(str), "%3d", vuurmuurlog_progress);
			set_field_buffer(vuurmuurlogfld, 0, str);
		}

		if(vuurmuurlog_progress == 100)
		{
			if(vuurmuurlog_semid == -1)
			{
				wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
				mvwprintw(wait_win, 5, 29, "Not connected");
				wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

				failed = 1;
			}
			else if(SILENT_LOCK(vuurmuurlog_semid))
			{
				vuurmuurlog_shmtable->reload_result = VR_RR_RESULT_ACK;
				SILENT_UNLOCK(vuurmuurlog_semid);

				if(vuurmuurlog_result == VR_RR_SUCCES)
				{
					wattron(wait_win, COLOR_PAIR(CP_GREEN_WHITE));
					mvwprintw(wait_win, 5, 29, "Success");
					wattroff(wait_win, COLOR_PAIR(CP_GREEN_WHITE));
				}
				else if(vuurmuur_result == VR_RR_NOCHANGES)
				{
					mvwprintw(wait_win, 5, 29, "No changes");
				}
				else
				{
					wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
					mvwprintw(wait_win, 5, 29, "Error");
					wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

					failed = 1;
				}
			}
		}

		update_panels();
		doupdate();

		/* no result yet, sleep 1 sec, or if the server didn't have a chance to do anything */
		if(	(vuurmuur_result    == VR_RR_NO_RESULT_YET || vuurmuur_result    == VR_RR_RESULT_ACK) ||
			(vuurmuurlog_result == VR_RR_NO_RESULT_YET || vuurmuurlog_result == VR_RR_RESULT_ACK))
		{
			waittime += 1000;
			usleep(1000);
		}
	}
	
	/* timed out */
	if(vuurmuur_progress < 100)
	{
		wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
		mvwprintw(wait_win, 4, 29, "Timed out");
		wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

		failed = 1;
	}

	/* timed out */
	if(vuurmuurlog_progress < 100)
	{
		wattron(wait_win, COLOR_PAIR(CP_RED_WHITE));
		mvwprintw(wait_win, 5, 29, "Timed out");
		wattroff(wait_win, COLOR_PAIR(CP_RED_WHITE));

		failed = 1;
	}

	update_panels();
	doupdate();

	if(failed == 1)
		sleep(4);
	else
		sleep(1);

	/*
		destroy the wait dialog
	*/
	unpost_form(form);
	free_form(form);

	for(i=0; i < n_fields; i++)
	{
		free_field(fields[i]);
	}
	free(fields);
	
	del_panel(panel[0]);
	destroy_win(wait_win);
	update_panels();
	doupdate();

	return(0);
}


struct
{
	FIELD	*overallfld,
		
		*backendfld,
		*configfld,
		*settingsfld,
		
		*shmfld,

		*systemfld;

	FIELD	**fields;
	int	n_fields;
	FORM	*form;

} StatusFlds;


/*
	we don't use set_field_just here because for some reason it
	doesn't work on the first field.
*/
static void
mm_set_status_field(const int debuglvl, int status, FIELD *fld)
{
	if(debuglvl >= HIGH)
		(void)vrprint.debug(__FUNCTION__, "status: %d.", status);

	if(status == 1) /* OK */
	{
		set_field_buffer(fld, 0, " OK ");
		set_field_back(fld, COLOR_PAIR(CP_GREEN_WHITE));
	}
	else if(status == 0) /* Attention */
	{
		set_field_buffer(fld, 0, "Warn");
		set_field_back(fld, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
	}
	else /* Warning */
	{
		set_field_buffer(fld, 0, "Fail");
		set_field_back(fld, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);
	}
}


/*
	the main menu, here you choose between rules, zones, config, logview, etc.
*/
int
main_menu(const int debuglvl, d_list *rules_list, struct rgx_ *reg)
{
	int	retval = 0,
		n_choices = 13,
		i = 0,
		ch = 0,
		quit = 0;
	ITEM	**menu_items;
	MENU	*main_menu;
	WINDOW	*mainmenu_win;
	PANEL	*my_panels[1];

//	short	pair = 0,
//		fore = 0,
//		back = 0;

	char	*choice_ptr = NULL;

	// this are the menu items
	char *choices[] = {	"Rules              ",
				"BlockList          ",

				"Zones              ",
				"Interfaces         ",
				"Services           ",

				"Vuurmuur Config    ",

				"Logview            ",
				"Status             ",
				"Connections        ",

				"Settings           ",

				"Apply Changes      ",

				"Quit               ",
				NULL
			};

	// with their descriptions
	char *descriptions[] = {
				"(F9) ",
				"(b)  ",

				"(F7) ",
				"     ",
				"(F8) ",

				"(F6) ",

				"     ",
				"(s)  ",
				"(c)  ",

				"     ",

				"(F11)",
				
				"(F10)",
				NULL
			};

	int	x=0,
		y=6+n_choices,
		startx=0,
		starty=0,
		maxx=0,
		maxy=0;	// time in sec to wait for vuurmuur in IPC

	int	field_num = 0;
	int	cols = 0;
	int	rows = 0;

	/* update the status */
	mm_update_overall_status(debuglvl);

	/* main menu width */
	if(vccnf.draw_status)
		x = 66;
	else
		x = 45;

	getmaxyx(stdscr, maxy, maxx);
	startx = (maxx - x) / 2;
	starty = (maxy - y) / 2;

	/* alloc the items */
	if(!(menu_items = (ITEM **)calloc(n_choices + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s.", strerror(errno));
		return(-1);
	}
	/* set the items */
	for(i = 0; i < n_choices; ++i)
	{
		menu_items[i] = new_item(choices[i], descriptions[i]);
	}
	menu_items[n_choices] = (ITEM *)NULL;
	/* create menu */
	main_menu = new_menu((ITEM **)menu_items);
	/* create the window and panel */
	mainmenu_win = create_newwin(y, x, starty, startx, "Main Menu", vccnf.color_win);
	keypad(mainmenu_win, TRUE);
	wrefresh(mainmenu_win);
	my_panels[0] = new_panel(mainmenu_win);
	/* menu settings */
	set_menu_win(main_menu, mainmenu_win);
	set_menu_sub(main_menu, derwin(mainmenu_win, y-6, x-12, 5, 6));
	set_menu_format(main_menu, y-4, 1);
	set_menu_back(main_menu, vccnf.color_win);
	set_menu_fore(main_menu, vccnf.color_win_rev);
	post_menu(main_menu);

	/* the form for the status */
	if(vccnf.draw_status)
	{
		StatusFlds.n_fields = 6;
		StatusFlds.fields = (FIELD **)calloc(StatusFlds.n_fields + 1, sizeof(FIELD *));

		/* overall */
		StatusFlds.overallfld = (StatusFlds.fields[field_num] = new_field(1, 4, 5, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.overall, StatusFlds.overallfld);

		/* backend */
		StatusFlds.backendfld = (StatusFlds.fields[field_num] = new_field(1, 4, 8, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.backend, StatusFlds.backendfld);
		/* config */
		StatusFlds.configfld = (StatusFlds.fields[field_num] = new_field(1, 4, 9, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.config, StatusFlds.configfld);
		/* settings */
		StatusFlds.settingsfld = (StatusFlds.fields[field_num] = new_field(1, 4, 10, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.settings, StatusFlds.settingsfld);

		/* shm */
		StatusFlds.shmfld = (StatusFlds.fields[field_num] = new_field(1, 4, 12, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.shm, StatusFlds.shmfld);

		/* system */
		StatusFlds.systemfld = (StatusFlds.fields[field_num] = new_field(1, 4, 14, 0, 0, 0));
		field_num++;
		mm_set_status_field(debuglvl, VuurmuurStatus.system, StatusFlds.systemfld);

		/* terminate */
		StatusFlds.fields[field_num] = NULL;

		/* Create the form and post it */
		StatusFlds.form = new_form(StatusFlds.fields);
		scale_form(StatusFlds.form, &rows, &cols);
		set_form_win(StatusFlds.form, mainmenu_win);
		set_form_sub(StatusFlds.form, derwin(mainmenu_win, rows, cols, 1, 54));

		/* welcome message */
		mvwprintw(mainmenu_win, 2, 16, "Welcome to Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);

		mvwprintw(mainmenu_win, 6,  39, "Overall       [    ]");

		mvwprintw(mainmenu_win, 9,  39, "Backend       [    ]");
		mvwprintw(mainmenu_win, 10, 39, "Config        [    ]");
		mvwprintw(mainmenu_win, 11, 39, "Settings      [    ]");
	
		mvwprintw(mainmenu_win, 13, 39, "Daemons       [    ]");

		mvwprintw(mainmenu_win, 15, 39, "System        [    ]");

		post_form(StatusFlds.form);

		if(VuurmuurStatus.overall != 1)
			mvwprintw(mainmenu_win, 17, 39, "Press F5 for details.");

		/* draw a nice box */
		mvwhline(mainmenu_win,  5,  38, ACS_HLINE, 22);
		mvwaddch(mainmenu_win,  5,  37, ACS_ULCORNER);
		mvwaddch(mainmenu_win,  5,  60, ACS_URCORNER);
		mvwvline(mainmenu_win,  6,  37, ACS_VLINE, 10);
		mvwvline(mainmenu_win,  6,  60, ACS_VLINE, 10);
		mvwhline(mainmenu_win,  16, 38, ACS_HLINE, 22);
		mvwaddch(mainmenu_win,  16, 37, ACS_LLCORNER);
		mvwaddch(mainmenu_win,  16, 60, ACS_LRCORNER);
	}
	else /* dont print status */
	{
		mvwprintw(mainmenu_win, 2, 4, "Welcome to Vuurmuur-conf %d.%d.%d", VUURMUURCONF_VERSION_MAJOR, VUURMUURCONF_VERSION_MINOR, VUURMUURCONF_VERSION_SUB);
	}
	/* restore the cursor */
	pos_menu_cursor(main_menu);

	update_panels();
	doupdate();

	// enter loop, when we quit this loop, we quit the program
	while(quit == 0)
	{
		ch = wgetch(mainmenu_win);
		if(debug >= LOW) status_print(status_win, "ch: %d", ch);
		switch(ch)
		{
			// navigation
			case KEY_DOWN:
				menu_driver(main_menu, REQ_DOWN_ITEM);
				break;

			case KEY_UP:
				menu_driver(main_menu, REQ_UP_ITEM);
				break;

			// handle selecting an item in the menu
			case KEY_RIGHT:
			case 32: // space
			case 10: // enter
			{
				ITEM *cur;
				cur = current_item(main_menu);

				choice_ptr = malloc(strlen((char *)item_name(cur))+1);
				strcpy(choice_ptr, (char *)item_name(cur));

				// we now have a choice, so quit the loop
				quit=1;
				break;
			}

			/* handle the keyboad shortcuts */
			case 27: //ESC
			case 'q':
			case KEY_F(10):

				retval = 1;
				quit = 1;
				break;

			/* rules */
			case KEY_F(5):

				choice_ptr = malloc(sizeof("printstatus")+1);
				strcpy(choice_ptr, "printstatus");
				quit = 1;
				break;

			/* rules */
			case KEY_F(9):

				choice_ptr = malloc(sizeof("rules")+1);
				strcpy(choice_ptr, "rules");
				quit = 1;
				break;

			/* services */
			case KEY_F(8):

				choice_ptr = malloc(sizeof("services")+1);
				strcpy(choice_ptr, "services");
				quit=1;
				break;

			/* zones */
			case KEY_F(7):

				choice_ptr = malloc(sizeof("zones")+1);
				strcpy(choice_ptr, "zones");
				quit=1;
				break;

			case KEY_F(6):

				choice_ptr = malloc(sizeof("Vuurmuur Config")+1);
				strcpy(choice_ptr, "Vuurmuur Config");
				quit=1;
				break;

			case KEY_F(11):

				choice_ptr = malloc(sizeof("apply changes")+1);
				strcpy(choice_ptr, "apply changes");
				quit=1;
				break;

			case 'l':
			case 'L':

				choice_ptr = malloc(sizeof("traffic")+1);
				strcpy(choice_ptr, "traffic");
				quit=1;
				break;

			case 's':
			case 'S':

				choice_ptr = malloc(sizeof("status")+1);
				strcpy(choice_ptr, "status");
				quit=1;
				break;

			case 'c':
			case 'C':

				choice_ptr = malloc(sizeof("connections")+1);
				strcpy(choice_ptr, "connections");
				quit=1;
				break;

			/* BlockList */
			case 'b':
			case 'B':

				choice_ptr = malloc(sizeof("blocklist")+1);
				strcpy(choice_ptr, "blocklist");
				quit=1;
				break;

			case KEY_F(12):

				choice_ptr = malloc(sizeof("showhelp")+1);
				strcpy(choice_ptr, "showhelp");
				quit=1;
				break;
		
			case KEY_F(1):

				choice_ptr = NULL;

				if(vccnf.win_fore > 0)
					vccnf.win_fore--;

				set_colors(debuglvl, &vccnf);

				quit = 1;
				break;

			case KEY_F(2):

				choice_ptr = NULL;

				if(vccnf.win_fore < 7)
					vccnf.win_fore++;

				set_colors(debuglvl, &vccnf);

				quit = 1;
				break;

			case KEY_F(3):

				choice_ptr = NULL;

				if(vccnf.win_back > 0)
					vccnf.win_back--;

				set_colors(debuglvl, &vccnf);

				quit = 1;
				break;

			case KEY_F(4):

				choice_ptr = NULL;

				if(vccnf.win_back < 7)
					vccnf.win_back++;

				set_colors(debuglvl, &vccnf);

				quit = 1;
				break;

		}
	}


	/*
		choice made, now remove the menu, window and form.
	*/
	unpost_menu(main_menu);
	free_menu(main_menu);
	for(i = 0; i < n_choices; ++i)
		free_item(menu_items[i]);
	free(menu_items);
	/* cleanup status form */
	if(vccnf.draw_status)
	{
		/* Un post form and free the memory */
		unpost_form(StatusFlds.form);
		free_form(StatusFlds.form);
		for(i = 0; i < StatusFlds.n_fields; i++)
			free_field(StatusFlds.fields[i]);
		free(StatusFlds.fields);
	}
	/* remove window and panel */
	del_panel(my_panels[0]);
	destroy_win(mainmenu_win);
	update_panels();
	doupdate();


	/* now act */
	if(choice_ptr != NULL)
	{
		if(strncasecmp(choice_ptr, "rules", 5) == 0)
		{
			draw_top_menu_rules(top_win);

			rules_form(debuglvl, rules_list);

			draw_top_menu_main(debuglvl, top_win);
		}
		else if(strncasecmp(choice_ptr, "zones", 5) == 0)
		{
			draw_top_menu_zones(top_win);

			zones_section(debuglvl, reg);

			draw_top_menu_main(debuglvl, top_win);

			/* check for active interfaces */
			mm_check_status_zones(debuglvl, NULL, &ZonedataList);
		}
		else if(strncasecmp(choice_ptr, "interfaces", 10) == 0)
		{
			draw_top_menu_interfaces(top_win);

			interfaces_section(debuglvl, reg);

			draw_top_menu_main(debuglvl, top_win);

			/* check for active networks */
			mm_check_status_interfaces(debuglvl, NULL, &InterfacesList);
		}
		else if(strncasecmp(choice_ptr, "services", 8) == 0)
		{
			draw_top_menu_services(top_win);
			services_section(debuglvl, reg);
			draw_top_menu_main(debuglvl, top_win);
		}
		else if(strncasecmp(choice_ptr, "Vuurmuur Config", 15) == 0)
		{
			draw_top_menu_config(top_win);
			while(config_menu(debuglvl) == 0);
			draw_top_menu_main(debuglvl, top_win);

			mm_check_status_config(debuglvl, NULL);
		}
		else if(strncasecmp(choice_ptr, "traffic", 7) == 0)
		{
			draw_top_menu_logview(top_win);
			logview_section(debuglvl, &ZonedataList, &BlockList, NULL);
			draw_top_menu_main(debuglvl, top_win);
		}
		else if(strncasecmp(choice_ptr, "logview", 7) == 0)
		{
			draw_top_menu_logview(top_win);

			mm_select_logfile(debuglvl);

			draw_top_menu_main(debuglvl, top_win);
			
		}	
		else if(strncasecmp(choice_ptr, "status", 6) == 0)
		{
			status_section(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "connections", 11) == 0)
		{
			draw_top_menu_connections(top_win);
			connections_section(debuglvl);
			draw_top_menu_main(debuglvl, top_win);
		}
		else if(strncasecmp(choice_ptr, "blocklist", 9) == 0)
		{
			draw_top_menu_blocklist(top_win);
			zones_blocklist(debuglvl, &BlockList, &ZonedataList, reg);
			draw_top_menu_main(debuglvl, top_win);
		}
		else if(strncasecmp(choice_ptr, "settings", 8) == 0)
		{
			edit_vcconfig(debuglvl);
			mm_check_status_settings(debuglvl, NULL);
		}
		else if(strncasecmp(choice_ptr, "printstatus", 11) == 0)
		{
			mm_status_checkall(debuglvl, &VuurmuurStatus.StatusList);
			print_status(debuglvl);
		}
		else if(strncasecmp(choice_ptr, "showhelp", 8) == 0)
		{
			print_help(debuglvl, ":[VUURMUUR:GENERAL]:");
		}
		else if(strncasecmp(choice_ptr, "apply changes", 13) == 0)
		{
			/* check shm one last time, and don't write to status list */
			mm_check_status_shm(debuglvl, NULL);
			/* update the status */
			mm_update_overall_status(debuglvl);

			if(VuurmuurStatus.overall == 1)
			{
				/* reload the shm */
				mm_reload_shm(debuglvl);
				/* update the vuurmuurlognames */
				(void)config_set_log_names(debuglvl, &conf);
			}
			else if(VuurmuurStatus.overall == 0)
			{
				if((confirm(	"Apply Changes",
						"The overall status is not OK. Apply anyway?",
						COLOR_PAIR(CP_RED_WHITE),
						COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1))
				{
					/* reload the shm */
					mm_reload_shm(debuglvl);
					/* update the vuurmuurlognames */
					(void)config_set_log_names(debuglvl, &conf);
				}
			}
			else
			{
				(void)vrprint.error(-1, "Error", "will not apply changes if the overall status is 'fail'.");
			}
		}
		else if(strncasecmp(choice_ptr, "quit", 4) == 0)
		{
			retval=1;
		}

		free(choice_ptr);
		choice_ptr = NULL;

		/* status checks */
		mm_check_status_shm(debuglvl, NULL);
		if(VuurmuurStatus.vuurmuur != 1)
		{
			mm_shm_connect_vuurmuur(debuglvl);
			mm_check_status_shm(debuglvl, NULL);
		}
		if(VuurmuurStatus.vuurmuur_log != 1)
		{
			mm_shm_connect_vuurmuurlog(debuglvl);
			mm_check_status_shm(debuglvl, NULL);
		}
	}

	return(retval);
}


/*
	check all the statusses
*/
void
mm_status_checkall(const int debuglvl, /*@null@*/ d_list *status_list)
{
	int	list_len = 0;


	/* if we one, manage the list */
	if(status_list != NULL)
	{
		d_list_cleanup(debuglvl, status_list);
		/* send a status of '1', so no status is printed */
		queue_status_msg(debuglvl, status_list, "One or more problems were detected in your current setup. Below is a list\n", 1);
		/* store the list length so we can check for changes after all the check functions */
		list_len = status_list->len;
	}
	
	
	/* check for (active) interfaces */
	mm_check_status_interfaces(debuglvl, status_list, &InterfacesList);

	/* check for (active) networks */
	mm_check_status_zones(debuglvl, status_list, &ZonedataList);

	/* check config */
	mm_check_status_config(debuglvl, status_list);

	/* check settings */
	mm_check_status_settings(debuglvl, status_list);

	/* shm connections */
	mm_check_status_shm(debuglvl, status_list);

	/* update the status */
	mm_update_overall_status(debuglvl);


	/* check for changes to the list */
	if(status_list != NULL)
	{
		/* if the list_len is still equal, we asume no problems */
		if(status_list->len == list_len)
		{
			/* nothing was added to the list */
			d_list_cleanup(debuglvl, status_list);
			/* send a status of '1', so no status is printed */
			queue_status_msg(debuglvl, status_list, "No problems were detected in your setup\n", 1);
		}
	}
}
