/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


struct StatusSection_
{
	PANEL   *panel[1];
	WINDOW  *win;
	FIELD   **fields;
	FORM    *form;
	int     n_fields;
} StatusSection;


int
get_iface_stats_from_ipt(int debuglvl, const char *iface_name, const char *chain, unsigned long long *recv_bytes, unsigned long long *trans_bytes)
{
	char			line[512],
				interface_in[32],
				interface_out[32],
				command[128],
				proto[16],
				options[16],
				source[36],
				dest[36];
	FILE			*p;
	int			line_count = 0;

	unsigned long long	packets = 0,
				bytes = 0;
	int			trans_done = 0,
				recv_done = 0;

	*trans_bytes = 0;
	*recv_bytes = 0;

	if(strcmp(chain, "INPUT") == 0)
		trans_done = 1;
	else if(strcmp(chain, "OUTPUT") == 0)
		recv_done = 1;

	// set the command
	snprintf(command, sizeof(command), "%s -vnL %s --exact", conf.iptables_location, chain);

	// open the pipe
	if(!(p = popen(command, "r")))
	{
		(void)vrprint.error(-1, "Error", "pipe failed.");
		return(-1);
	}

	while(fgets(line, sizeof(line), p) != NULL && (!recv_done || !trans_done))
	{
		// we start looking after the first two lines
		if(line_count >= 2)
		{
			//            pack  byte  pr op in ou sr ds
			sscanf(line, "%llu %llu %s %s %s %s %s %s", &packets, &bytes, proto, options, interface_in, interface_out, source, dest);

			if(debuglvl >= HIGH)
				(void)vrprint.debug("Debug", "get_iface_stats_from_ipt: %s: packets: %llu, bytes: %llu", iface_name, packets, bytes);

			if(strcmp(source, "0.0.0.0/0") == 0 && strcmp(dest, "0.0.0.0/0") == 0 && strcmp(proto, "all") == 0 && (strcmp(interface_in, "*") == 0 || strcmp(interface_out, "*") == 0))
			{
				if(strcmp(interface_in, "*") == 0 && strcmp(interface_out, iface_name) == 0)
				{
					*trans_bytes = bytes;
					trans_done = 1;

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "get_iface_stats_from_ipt: %s: trans: %llu (%llu) (trans done)", iface_name, *trans_bytes, bytes);
				}
				else if(strcmp(interface_out, "*") == 0 && strcmp(interface_in, iface_name) == 0)
				{
					*recv_bytes = bytes;
					recv_done = 1;

					if(debuglvl >= HIGH)
						(void)vrprint.debug("Debug", "get_iface_stats_from_ipt: %s: recv: %llu (%llu) (recv done)", iface_name, *recv_bytes, bytes);
				}
			}

		}

		line_count++;
	}

	pclose(p);

	return(0);
}
                                                
/*	get_sys_load

	Gets the systemload from /proc/loadavg

	Returncodes:
		 0: ok
		-1: error
*/
int get_sys_load(float *s_load, float *m_load, float *l_load)
{
	FILE		*fp=NULL;
	char		proc_loadavg[] = "/proc/loadavg",
			line[512];

	if(!(fp = fopen(proc_loadavg, "r")))
		return(-1);

	if(fgets(line, sizeof(line), fp) != NULL)
		sscanf(line, "%f %f %f", s_load, m_load, l_load);
	else
		return(-1);

	if(fclose(fp) < 0)
		return(-1);
	
	return(0);
}

int count_host_tcp_conn(int *tcp_count, int *tcp_list_count)
{
	FILE		*fp=NULL;
	char		proc_net_tcp[] = "/proc/net/tcp",
			line[512];
	int		i=0;

	int		sl=0;
	char		localaddr[32],
			state[3],
			remoteaddr[32];

	int		tcp_listen=0;
	
	if(!(fp = fopen(proc_net_tcp, "r")))
		return(-1);

	while(fgets(line, sizeof(line), fp) != NULL)
	{
		if(i > 0)
		{
			sscanf(line, "%d: %s %s %2s", &sl, localaddr, remoteaddr, state);

			if(strcmp(state, "0A") == 0)
				tcp_listen++;
			
		}
		i++;
	}
	i--;
	
	//vrprint.warning("info", "tcp_listen: %d (%d total).", tcp_listen, i);
        
	*tcp_count = i;
	*tcp_list_count = tcp_listen;

	if(fclose(fp) < 0)
		return(-1);

	return(0);
}

int count_host_udp_conn(int *udp_count, int *udp_list_count)
{
	FILE		*fp=NULL;
	char		proc_net_udp[] = "/proc/net/udp",
			line[512];
	int		i=0;

	int		sl=0;
	char		localaddr[32],
			state[3],
			remoteaddr[32];

	int		udp_listen=0;

	if(!(fp = fopen(proc_net_udp, "r")))
		return(-1);

	while(fgets(line, sizeof(line), fp) != NULL)
	{
		if(i > 0)
		{
			sscanf(line, "%d: %s %s %2s", &sl, localaddr, remoteaddr, state);

			if(strcmp(state, "07") == 0)
				udp_listen++;

		}
		i++;
	}
	i--;

	//vrprint.warning("info", "udp_listen: %d (%d total).", udp_listen, i);

	*udp_count = i;
	*udp_list_count = udp_listen;

	if(fclose(fp) < 0)
		return(-1);

	return(0);
}

int count_conntrack_conn(int *conntrack_count, int *tcp_count, int *udp_count, int *other_count)
{
	FILE		*fp=NULL;
	char		proc_net_ipconntrack[] = "/proc/net/ip_conntrack",
			line[512];
	int		i=0,
			tcp=0,
			udp=0,
			other=0;

	if(!(fp = fopen(proc_net_ipconntrack, "r")))
		return(-1);

	while(fgets(line, sizeof(line), fp) != NULL)
	{
		if(strncmp(line, "tcp", 3) == 0)
			tcp++;
		else if(strncmp(line, "udp", 3) == 0)
			udp++;
		else
			other++;

		i++;
	}

	*conntrack_count = i;
	*tcp_count = tcp;
	*udp_count = udp;
	*other_count = other;

	if(fclose(fp) < 0)
		return(-1);

	return(0);
}

int get_conntrack_max(int *conntrack_max)
{
	FILE		*fp=NULL;
	char		proc_conntrack_max[] = "/proc/sys/net/ipv4/ip_conntrack_max",
			line[16];

	if(!(fp = fopen(proc_conntrack_max, "r")))
		return(-1);

	if(fgets(line, sizeof(line), fp) != NULL)
		*conntrack_max = atoi(line);

	if(fclose(fp) < 0)
		return(-1);

	return(0);
}

/*	get_meminfo

	Gets the info about the memory status of the system.

	Returncodes:
		 0: ok
		-1: error
*/
int get_meminfo(int *mem_total, int *mem_free, int *mem_cached, int *mem_buffers)
{
	FILE		*fp=NULL;
	char		proc_meminfo[] = "/proc/meminfo",
			line[64],
			variable[16],
			value[16];

	// open the proc entry
	if(!(fp = fopen(proc_meminfo, "r")))
	{
		(void)vrprint.error(-1, "Error", "opening '%s' failed: %s (in: get_meminfo).", proc_meminfo, strerror(errno));
		return(-1);
	}

	// loop trough the file and get the info
	while(fgets(line, sizeof(line), fp) != NULL)
	{
		sscanf(line, "%s %s", variable, value);

		if(strcmp(variable, "MemTotal:") == 0)
			*mem_total = atoi(value);
		else if(strcmp(variable, "MemFree:") == 0)
			*mem_free = atoi(value);
		else if(strcmp(variable, "Buffers:") == 0)
			*mem_buffers = atoi(value);
		else if(strcmp(variable, "Cached:") == 0)
			*mem_cached = atoi(value);
	}

	if(fclose(fp) < 0)
		return(-1);

	return(0);
}

int get_system_uptime(char *s_day, char *s_hour, char *s_minute, char *s_second)
{
	FILE		*fp=NULL;
	char		proc_uptime[] = "/proc/uptime",
			line[512];
	int		upt_i=0,
			day=0,
			hour=0,
			min=0,
			sec=0;

// param check

	if(!(fp = fopen(proc_uptime, "r")))
		return(-1);

	if(fgets(line, sizeof(line), fp) != NULL)
		sscanf(line, "%d", &upt_i);

	day = upt_i / 86400;
	hour = (upt_i - (day * 86400)) / 3600;
	min =  (upt_i - (day * 86400) - (hour * 3600))/60;
	sec =  (upt_i - (day * 86400) - (hour * 3600) - (min * 60));

	if(fclose(fp) < 0)
		return(-1);

	snprintf(s_day, 5, "%4d", day);
	snprintf(s_hour, 3, "%02d", hour);
	snprintf(s_minute, 3, "%02d", min);
	snprintf(s_second, 3, "%02d", sec);

	return(0);
}

/*	status_section_init

	This function creates the status section window and the fields inside it.
	It also draws - alot of - lines.
	
	Returncodes:
		 0: ok
		-1: error
*/
int status_section_init(int height, int width, int startx, int starty, int ifac_num)
{
	int	i,
		rows,
		cols,
		max_height,
		max_width,
		ifacs=0,
		ifac_fields=0,
		ifac_start=12;

	/* get and check the screen dimentions */
	getmaxyx(stdscr, max_height, max_width);
	if(width > max_width)
		return(-1);

	/* set the number of fields */
	StatusSection.n_fields = 16 + (6 * ifac_num);
	
	/* alloc the needed memory */
	if(!(StatusSection.fields = (FIELD **)calloc(StatusSection.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s.", strerror(errno));
		return(-1);
	}

	/* create the fields */
	StatusSection.fields[0] = new_field(1, 5, 4, 13, 0, 1);
	set_field_buffer(StatusSection.fields[0], 1, "ld_s");
	StatusSection.fields[1] = new_field(1, 5, 4, 19, 0, 1);
	set_field_buffer(StatusSection.fields[1], 1, "ld_m");
	StatusSection.fields[2] = new_field(1, 5, 4, 25, 0, 1);
	set_field_buffer(StatusSection.fields[2], 1, "ld_l");

	StatusSection.fields[3] = new_field(1, 6, 4, 43, 0, 1);
	set_field_buffer(StatusSection.fields[3], 1, "mem_t");
	StatusSection.fields[4] = new_field(1, 6, 4, 51, 0, 1);
	set_field_buffer(StatusSection.fields[4], 1, "mem_f");
	StatusSection.fields[5] = new_field(1, 6, 4, 59, 0, 1);
	set_field_buffer(StatusSection.fields[5], 1, "mem_c");
	StatusSection.fields[6] = new_field(1, 6, 4, 67, 0, 1);
	set_field_buffer(StatusSection.fields[6], 1, "mem_b");

	StatusSection.fields[7] = new_field(1, 4, 1, 61, 0, 1);
	set_field_buffer(StatusSection.fields[7], 1, "up_d");
	StatusSection.fields[8] = new_field(1, 2, 1, 66, 0, 1);
	set_field_buffer(StatusSection.fields[8], 1, "uh");
	StatusSection.fields[9] = new_field(1, 2, 1, 69, 0, 1);
	set_field_buffer(StatusSection.fields[9], 1, "um");
	StatusSection.fields[10] = new_field(1, 2, 1, 72, 0, 1);
	set_field_buffer(StatusSection.fields[10], 1, "us");


	StatusSection.fields[11] = new_field(1, 6, 6, 23, 0, 1);
	set_field_buffer(StatusSection.fields[11], 1, "con_t");
	StatusSection.fields[12] = new_field(1, 6, 7, 23, 0, 1);
	set_field_buffer(StatusSection.fields[12], 1, "con_u");
	StatusSection.fields[13] = new_field(1, 6, 6, 41, 0, 1);
	set_field_buffer(StatusSection.fields[13], 1, "con_o");

	StatusSection.fields[14] = new_field(1, 6, 6, 59, 0, 1);
	set_field_buffer(StatusSection.fields[14], 1, "con_c");
	StatusSection.fields[15] = new_field(1, 6, 7, 59, 0, 1);
	set_field_buffer(StatusSection.fields[15], 1, "con_m");

	/* create iface stats fields */
	for(ifacs = 0, ifac_fields = 16; ifacs < ifac_num; ifacs++)
	{
		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 13, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "recv_s");
		ifac_fields++;

		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 22, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "send_s");
		ifac_fields++;

		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 31, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "rcv_ti");
		ifac_fields++;

		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 40, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "snd_to");
		ifac_fields++;

		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 49, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "rcv_tf");
		ifac_fields++;

		StatusSection.fields[ifac_fields] = new_field(1, 8, ifac_start+ifacs, 58, 0, 1);
		set_field_buffer(StatusSection.fields[ifac_fields], 1, "snd_tf");
		ifac_fields++;
	}

	/* terminate the field array */
	StatusSection.fields[StatusSection.n_fields] = NULL;

	/* create the window and the panel */
	if(!(StatusSection.win = create_newwin(height, width, startx, starty, "Status Section", COLOR_PAIR(CP_BLUE_WHITE))))
		return(-1);
	if(!(StatusSection.panel[0] = new_panel(StatusSection.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel.");
		return(-1);
	}

	/* field options */
	for(i = 0; i < StatusSection.n_fields; i++)
	{
		if(debug >= LOW)
			set_field_back(StatusSection.fields[i], COLOR_PAIR(CP_WHITE_BLUE));
		else
			set_field_back(StatusSection.fields[i], COLOR_PAIR(CP_BLUE_WHITE));
		
		field_opts_off(StatusSection.fields[i], O_AUTOSKIP);
		/* set status to false */
		set_field_status(StatusSection.fields[i], FALSE);
	}

	/* Create the form and post it */
	if(!(StatusSection.form = new_form(StatusSection.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form.");
		return(-1);
	}
	/* Calculate the area required for the form */
	scale_form(StatusSection.form, &rows, &cols);
	keypad(StatusSection.win, TRUE);
	/* Set main window and sub window */
	set_form_win(StatusSection.form, StatusSection.win);
	set_form_sub(StatusSection.form, derwin(StatusSection.win, rows, cols, 1, 2));

	if(post_form(StatusSection.form) != E_OK)
	{
		(void)vrprint.error(-1, "Error", "posting the from failed.");
		return(-1);
	}

	/* print the field labels */
	mvwprintw(StatusSection.win, 1, 2,  "Hostname");
	mvwprintw(StatusSection.win, 2, 2,  "Kernel");

	mvwprintw(StatusSection.win, 1, 64,  "day h  m  s");
	mvwprintw(StatusSection.win, 2, 70,  ":");
	mvwprintw(StatusSection.win, 2, 73,  ":");
	mvwprintw(StatusSection.win, 2, 55,  "Uptime");

	mvwprintw(StatusSection.win, 4, 15,  "1m    5m    15m");

	mvwprintw(StatusSection.win, 5, 2,  "Load");
	mvwprintw(StatusSection.win, 4, 46,  "Total   Free    Cache   Buffer");
	mvwprintw(StatusSection.win, 5, 34,  "Memory");

	mvwprintw(StatusSection.win, 8, 2,  "Connections");

	mvwprintw(StatusSection.win, 7, 16,  "Tcp");
	mvwprintw(StatusSection.win, 8, 16,  "Udp");

	mvwprintw(StatusSection.win, 7, 34,  "Other");

	mvwprintw(StatusSection.win, 7, 52,  "Current");
	mvwprintw(StatusSection.win, 8, 52,  "Maximal");

	mvwprintw(StatusSection.win, 10, 1,  "              Speed/s           Firewall          Forwarded");
	mvwprintw(StatusSection.win, 11, 1,  " Interfaces   Down     Up       In       Out      Recv     Send");

	/*
		DRAW THE LINES
	*/
	
	/* kernel and domainname */
	mvwvline(StatusSection.win,  1, 14, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  1, 53, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  1, 62, ACS_VLINE, 2);
	
	/* T-pieces on top */
	mvwaddch(StatusSection.win,  0, 14, ACS_TTEE);
	mvwaddch(StatusSection.win,  0, 53, ACS_TTEE);
	mvwaddch(StatusSection.win,  0, 62, ACS_TTEE);

	mvwhline(StatusSection.win,  3, 1,  ACS_HLINE, 76);
	mvwaddch(StatusSection.win,  3, 0,  ACS_LTEE);
	mvwaddch(StatusSection.win,  3, 77, ACS_RTEE);
	
	mvwaddch(StatusSection.win,  3, 14, ACS_PLUS);
	mvwaddch(StatusSection.win,  3, 20, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 26, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 32, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 44, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 52, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 53, ACS_BTEE);
	mvwaddch(StatusSection.win,  3, 60, ACS_TTEE);
	mvwaddch(StatusSection.win,  3, 62, ACS_BTEE);
	mvwaddch(StatusSection.win,  3, 68, ACS_TTEE);

	/* load fields */
	mvwvline(StatusSection.win,  4, 14, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 20, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 26, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 32, ACS_VLINE, 2);

	/* memory */
	mvwvline(StatusSection.win,  4, 44, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 52, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 60, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  4, 68, ACS_VLINE, 2);

	mvwhline(StatusSection.win,  6, 1,  ACS_HLINE, 76);
	mvwaddch(StatusSection.win,  6, 0,  ACS_LTEE);
	mvwaddch(StatusSection.win,  6, 77, ACS_RTEE);
	
	mvwaddch(StatusSection.win,  6, 14, ACS_PLUS);

	mvwaddch(StatusSection.win,  6, 20, ACS_BTEE);
	mvwaddch(StatusSection.win,  6, 26, ACS_BTEE);
	mvwaddch(StatusSection.win,  6, 32, ACS_PLUS);
	mvwaddch(StatusSection.win,  6, 44, ACS_BTEE);
	mvwaddch(StatusSection.win,  6, 50, ACS_TTEE);
	mvwaddch(StatusSection.win,  6, 52, ACS_BTEE);
	
	mvwaddch(StatusSection.win,  6, 60, ACS_BTEE);
	mvwaddch(StatusSection.win,  6, 68, ACS_PLUS);
	
	/* uptime fields */
	mvwvline(StatusSection.win,  7, 14, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  7, 32, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  7, 50, ACS_VLINE, 2);
	mvwvline(StatusSection.win,  7, 68, ACS_VLINE, 2);

	/* connection fields */
	mvwhline(StatusSection.win,  9, 1, ACS_HLINE, 76);
	mvwaddch(StatusSection.win,  9, 0,  ACS_LTEE);
	mvwaddch(StatusSection.win,  9, 77, ACS_RTEE);

	mvwvline(StatusSection.win, 10, 14, ACS_VLINE, 2);
	mvwvline(StatusSection.win, 10, 32, ACS_VLINE, 2);
	mvwvline(StatusSection.win, 10, 50, ACS_VLINE, 2);
	mvwvline(StatusSection.win, 10, 68, ACS_VLINE, 2);

	mvwaddch(StatusSection.win,  9, 14, ACS_PLUS);
	mvwaddch(StatusSection.win,  9, 32, ACS_PLUS);
	mvwaddch(StatusSection.win,  9, 50, ACS_PLUS);
	mvwaddch(StatusSection.win,  9, 68, ACS_PLUS);

	/* interface fields */
	mvwhline(StatusSection.win, ifac_start, 14, ACS_HLINE, 55);
	mvwhline(StatusSection.win, ifac_start+ifac_num+1, 1, ACS_HLINE, 76);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 0,  ACS_LTEE);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 77, ACS_RTEE);

	mvwaddch(StatusSection.win, ifac_start, 14, ACS_LTEE);
	mvwaddch(StatusSection.win, ifac_start, 68, ACS_RTEE);

	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 14, ACS_BTEE);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 68, ACS_BTEE);

	mvwvline(StatusSection.win, ifac_start+1, 14, ACS_VLINE, ifac_num);

	mvwaddch(StatusSection.win, ifac_start, 23, ACS_TTEE);
	mvwvline(StatusSection.win, ifac_start+1, 23, ACS_VLINE, ifac_num);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 23, ACS_BTEE);

	mvwaddch(StatusSection.win, ifac_start, 32, ACS_PLUS);
	mvwvline(StatusSection.win, ifac_start+1, 32, ACS_VLINE, ifac_num);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 32, ACS_BTEE);

	mvwaddch(StatusSection.win, ifac_start, 41, ACS_TTEE);
	mvwvline(StatusSection.win, ifac_start+1, 41, ACS_VLINE, ifac_num);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 41, ACS_BTEE);

	mvwaddch(StatusSection.win, ifac_start, 50, ACS_PLUS);
	mvwvline(StatusSection.win, ifac_start+1, 50, ACS_VLINE, ifac_num);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 50, ACS_BTEE);

	mvwaddch(StatusSection.win, ifac_start, 59, ACS_TTEE);
	mvwvline(StatusSection.win, ifac_start+1, 59, ACS_VLINE, ifac_num);
	mvwaddch(StatusSection.win, ifac_start+ifac_num+1, 59, ACS_BTEE);

	mvwvline(StatusSection.win, ifac_start+1, 68, ACS_VLINE, ifac_num);

	return(0);
}

int status_section_destroy(void)
{
	int i;

	// Un post form and free the memory
	unpost_form(StatusSection.form);
	free_form(StatusSection.form);

	for(i = 0; i < StatusSection.n_fields; i++)
	{
		free_field(StatusSection.fields[i]);
	}
	free(StatusSection.fields);

	del_panel(StatusSection.panel[0]);
	destroy_win(StatusSection.win);

	return(0);
}

/*	status_section

	This section shows information about the system.
	
	Returncodes:
		0: ok
		-1: error
*/
int
status_section(int debuglvl)
{
	FIELD		*cur;
	int		retval=0;
	int		quit=0, ch;

	int		max_onscreen=0,
			y=0;

	int		max_height,
			max_width,
			i=0,

			conntrack_conn_max=0,

			conntrack_conn_total=0,
			conntrack_conn_tcp=0,
			conntrack_conn_udp=0,
			conntrack_conn_other=0,

			mem_total=0,
			mem_free=0,
			mem_cached=0,
			mem_bufferd=0,
			cur_interface=0;

	char		hostname[60],
			load_str[6],
			mem_str[7],
			interfacename[13] = "",

			upt_day[5],
			upt_hour[3],
			upt_minute[3],
			upt_second[3],

			conn_max[7],
			conn_total[7],
			conn_tcp[7],
			conn_udp[7],
			conn_other[7],

			recv_host[9],
			send_host[9],

			recv_net[9],
			send_net[9],

			recv_speed[9],
			send_speed[9];

	/* uname struct, for gettig the kernel version */
	struct utsname	uts_name;

	/* the byte counters */
	long long	recv_bytes=0,
			trans_bytes=0,
			delta_bytes=0,
			speed_bytes=0;

	/* load */
	float		s_load=0, // 1 min
			m_load=0, // 5 min
			l_load=0; // 15 min

	/* structure for storing byte counters per interface */
	struct shadow_ifac_
	{
		long long		prev_recv_bytes;
		long long		prev_send_bytes;

		long long		cur_recv_bytes;
		long long		cur_send_bytes;

		unsigned long long	send_host,
					recv_host,

					send_net,
					recv_net;

		/* for the correction of the speed */
		struct timeval		begin_tv;
		struct timeval		end_tv;
	};

	struct shadow_ifac_	*shadow_ptr=NULL;
	struct InterfaceData_	*iface_ptr=NULL;

	d_list_node		*d_node = NULL;
	d_list_node		*shadow_node = NULL;

	// list which will hold the structs analog to the interfaces list
	d_list			shadow_list;

	// hashtables for fast searching
	Hash			ZonedataHash,
				ServicesHash;

	int			result=0;

	// we correct the speed with the time it takes to get all stats
	double			elapse = 0;
	float			correction = 0;

	int	update_interval = 1000000; /* weird, in pratice this seems to be two sec */
	int	slept_so_far    = 1000000; /* time slept since last update */

	// first create our shadow list
//	shadow_list.size = 0;
	if(d_list_setup(debuglvl, &shadow_list, free) < 0)
		return(-1);

	for(i=0; i < InterfacesList.len; i++)
	{
		if(!(shadow_ptr = malloc(sizeof(struct shadow_ifac_))))
			return(-1);

		shadow_ptr->prev_recv_bytes = -1;
		shadow_ptr->prev_send_bytes = -1;
		shadow_ptr->cur_recv_bytes = -1;
		shadow_ptr->cur_send_bytes = -1;

		gettimeofday(&shadow_ptr->begin_tv, 0);

		/*
			append to the list

			TODO: this should be prepend?
		*/
		if(d_list_append(debuglvl, &shadow_list, shadow_ptr) < 0)
			return(-1);
	}

	/* create the service and zone hash for conn_get_stats */

	/*
		insert the interfaces as TYPE_HOST's into the zonelist as 'firewall', so this appears in to log
	*/
	if(ins_iface_into_zonelist(0, &InterfacesList, &ZonedataList) < 0)
	{
		(void)vrprint.error(-1, "Error", "iface_into_zonelist failed (in: main).");
		return(-1);
	}

	(void)vrprint.info("Info", "Creating hash-table for the zones...");
	if(init_zonedata_hashtable(debuglvl, ZonedataList.len * 2, &ZonedataList, hash_ipaddress, compare_ipaddress, &ZonedataHash) < 0)
	{
		(void)vrprint.error(result, "Error", "init_zonedata_hashtable failed.");
		return(-1);
	}

	/*
		create the service hashtable. We use the size listsize * 5 because a lot of services have more than one portrange
	*/
	(void)vrprint.info("Info", "Creating hash-table for the services...");
	if(init_services_hashtable(debuglvl, ServicesList.len * 5, &ServicesList, hash_port, compare_ports, &ServicesHash) < 0)
	{
		(void)vrprint.error(result, "Error", "init_services_hashtable failed.");
		return(-1);
	}

	/*
		set up the statuswin
	*/
	getmaxyx(stdscr, max_height, max_width);
	max_onscreen = max_height-8;

	/*
		init
	*/
	if(status_section_init(max_height-6, 78, 3, 1, InterfacesList.len) < 0)
		return(-1);

	/*
		make sure wgetch doesn't block
	*/
	nodelay(StatusSection.win, TRUE);
	keypad(StatusSection.win, TRUE);

	/*
		get the hostname of the system, or set to error on failure
	*/
	if(gethostname(hostname, sizeof(hostname)) < 0)
		strlcpy(hostname, "error", sizeof(hostname));

	mvwprintw(StatusSection.win, 1, 15, "%s", hostname);

	/*
		uname - get some system information
	*/
	if(uname(&uts_name) < 0)
		(void)vrprint.error(-1, "Error", "uname failed.");

	mvwprintw(StatusSection.win, 2, 15, "%s %s", uts_name.sysname, uts_name.release);

	/*
		get the maximum connections
	*/
	if(get_conntrack_max(&conntrack_conn_max) < 0)
		snprintf(conn_max, sizeof(conn_max), "error");
	else
		snprintf(conn_max, sizeof(conn_max), "%6d", conntrack_conn_max);

	update_panels();
	doupdate();

	/* the main loop */
	while(quit == 0 && retval == 0)
	{
		if(debuglvl >= LOW)
			(void)vrprint.debug(__FUNCTION__, "slept_so_far: %d, update_interval: %d.", slept_so_far, update_interval);

		/* check if we have slept long enough */
		if(slept_so_far >= update_interval)
		{
			
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "slept_so_far: %d -> now print.", slept_so_far);

			slept_so_far = 0;

			/*  update the information */
			if(get_sys_load(&s_load, &m_load, &l_load) < 0)
			{
				(void)vrprint.error(-1, "Error", "getting the system load failed.");
				return(-1);
			}

			if(get_meminfo(&mem_total, &mem_free, &mem_cached, &mem_bufferd) < 0)
			{
				(void)vrprint.error(-1, "Error", "getting memory info failed.");
				return(-1);
			}

			if(get_system_uptime(upt_day, upt_hour, upt_minute, upt_second) < 0)
			{
				(void)vrprint.error(-1, "Error", "getting system uptime failed.");
				return(-1);
			}

			if(count_conntrack_conn(&conntrack_conn_total, &conntrack_conn_tcp, &conntrack_conn_udp, &conntrack_conn_other) < 0)
			{
				//(void)vrprint.error(-1, "Error", "counting conntrack connections failed.");
				//return(-1);
				snprintf(conn_total, sizeof(conn_total), "error");
				snprintf(conn_tcp,   sizeof(conn_tcp),   "error");
				snprintf(conn_udp,   sizeof(conn_udp),   "error");
				snprintf(conn_other, sizeof(conn_other), "error");
			}
			else
			{
				snprintf(conn_total, sizeof(conn_total), "%6d", conntrack_conn_total);
				snprintf(conn_tcp,   sizeof(conn_tcp),   "%6d", conntrack_conn_tcp);
				snprintf(conn_udp,   sizeof(conn_udp),   "%6d", conntrack_conn_udp);
				snprintf(conn_other, sizeof(conn_other), "%6d", conntrack_conn_other);
			}

			/* loop trough the fields and update the information */
			for(i = 0; i < StatusSection.n_fields; i++)
			{
				cur = StatusSection.fields[i];

				if(strncmp(field_buffer(cur, 1), "ld_s", 4) == 0)
				{
					if(s_load > 2 && s_load < 5)
						set_field_fore(cur, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
					else if(s_load >= 5)
						set_field_fore(cur, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);
					else
						set_field_fore(cur, COLOR_PAIR(CP_BLUE_WHITE));

					snprintf(load_str, sizeof(load_str), "%2.2f", s_load);
					set_field_buffer(cur, 0, load_str);
				}
				else if(strncmp(field_buffer(cur, 1), "ld_m", 4) == 0)
				{
					if(m_load > 2 && m_load < 5)
						set_field_fore(cur, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
					else if(m_load >= 5)
						set_field_fore(cur, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);
					else
						set_field_fore(cur, COLOR_PAIR(CP_BLUE_WHITE));

					snprintf(load_str, sizeof(load_str), "%2.2f", m_load);
					set_field_buffer(cur, 0, load_str);
			}
				else if(strncmp(field_buffer(cur, 1), "ld_l", 4) == 0)
				{
					if(l_load > 2 && l_load < 5)
						set_field_fore(cur, COLOR_PAIR(CP_YELLOW_WHITE)|A_BOLD);
					else if(l_load >= 5)
						set_field_fore(cur, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);
					else
						set_field_fore(cur, COLOR_PAIR(CP_BLUE_WHITE));

					snprintf(load_str, sizeof(load_str), "%2.2f", l_load);
					set_field_buffer(cur, 0, load_str);
				}
				else if(strncmp(field_buffer(cur, 1), "mem_t", 5) == 0)
				{
					snprintf(mem_str, sizeof(mem_str), "%6d", mem_total/1024);
					set_field_buffer(cur, 0, mem_str);
				}
				else if(strncmp(field_buffer(cur, 1), "mem_f", 5) == 0)
				{
					snprintf(mem_str, sizeof(mem_str), "%6d", mem_free/1024);
					set_field_buffer(cur, 0, mem_str);
				}
				else if(strncmp(field_buffer(cur, 1), "mem_c", 5) == 0)
				{
					snprintf(mem_str, sizeof(mem_str), "%6d", mem_cached/1024);
					set_field_buffer(cur, 0, mem_str);
				}
				else if(strncmp(field_buffer(cur, 1), "mem_b", 5) == 0)
				{
					snprintf(mem_str, sizeof(mem_str), "%6d", mem_bufferd/1024);
					set_field_buffer(cur, 0, mem_str);
				}
				else if(strncmp(field_buffer(cur, 1), "up_d", 4) == 0)
				{
					set_field_buffer(cur, 0, upt_day);
				}
				else if(strncmp(field_buffer(cur, 1), "uh", 2) == 0)
				{
					set_field_buffer(cur, 0, upt_hour);
				}
				else if(strncmp(field_buffer(cur, 1), "um", 2) == 0)
				{
					set_field_buffer(cur, 0, upt_minute);
				}
				else if(strncmp(field_buffer(cur, 1), "us", 2) == 0)
				{
					set_field_buffer(cur, 0, upt_second);
				}
				else if(strncmp(field_buffer(cur, 1), "con_m", 5) == 0)
				{
					set_field_buffer(cur, 0, conn_max);
				}
				else if(strncmp(field_buffer(cur, 1), "con_c", 5) == 0)
				{
					set_field_buffer(cur, 0, conn_total);
				}
				else if(strncmp(field_buffer(cur, 1), "con_t", 5) == 0)
				{
					set_field_buffer(cur, 0, conn_tcp);
				}
				else if(strncmp(field_buffer(cur, 1), "con_u", 5) == 0)
				{
					set_field_buffer(cur, 0, conn_udp);
				}
				else if(strncmp(field_buffer(cur, 1), "con_o", 5) == 0)
				{
					set_field_buffer(cur, 0, conn_other);
				}
			}

			/* print interfaces, starting at line 13 */
			for(	cur_interface = 0, y = 13, d_node = InterfacesList.top, shadow_node = shadow_list.top;
				d_node && y < max_height-8;
				d_node = d_node->next, shadow_node = shadow_node->next, y++)
			{
				unsigned long long	tmp_ull;

				iface_ptr = d_node->data;
				shadow_ptr = shadow_node->data;

				/* get the counters for determining speed */
				get_iface_stats(iface_ptr->device, &recv_bytes, NULL, &trans_bytes, NULL);

				/* get the real counters from iptables */
				get_iface_stats_from_ipt(debug, iface_ptr->device, "INPUT", &shadow_ptr->recv_host, &tmp_ull);
				get_iface_stats_from_ipt(debug, iface_ptr->device, "OUTPUT", &tmp_ull, &shadow_ptr->send_host);
				get_iface_stats_from_ipt(debug, iface_ptr->device, "FORWARD", &shadow_ptr->recv_net, &shadow_ptr->send_net);

				/* RECV host/firewall */
				if((shadow_ptr->recv_host/(1024*1024)) >= 1000)
					snprintf(recv_host, sizeof(recv_host), "%2.3f GB", (float)shadow_ptr->recv_host/(1024*1024*1024));
				else if((shadow_ptr->recv_host/(1024*1024)) < 1)
					snprintf(recv_host, sizeof(recv_host), "%5d kb", (int)shadow_ptr->recv_host/(1024));
				else
					snprintf(recv_host, sizeof(recv_host), "%5.1f MB", (float)shadow_ptr->recv_host/(1024*1024));

				/* SEND host/firewall */
				if((shadow_ptr->send_host/(1024*1024)) >= 1000)
					snprintf(send_host, sizeof(send_host), "%2.3f GB", (float)shadow_ptr->send_host/(1024*1024*1024));
				else if((shadow_ptr->send_host/(1024*1024)) < 1)
					snprintf(send_host, sizeof(send_host), "%5d kb", (int)shadow_ptr->send_host/(1024));
				else
					snprintf(send_host, sizeof(send_host), "%5.1f MB", (float)shadow_ptr->send_host/(1024*1024));

				/* RECV net/forward */
				if((shadow_ptr->recv_net/(1024*1024)) >= 1000)
					snprintf(recv_net, sizeof(recv_net), "%2.3f GB", (float)shadow_ptr->recv_net/(1024*1024*1024));
				else if((shadow_ptr->recv_net/(1024*1024)) < 1)
					snprintf(recv_net, sizeof(recv_net), "%5d kb", (int)shadow_ptr->recv_net/(1024));
				else
					snprintf(recv_net, sizeof(recv_net), "%5.1f MB", (float)shadow_ptr->recv_net/(1024*1024));

				/* SEND net/forward */
				if((shadow_ptr->send_net/(1024*1024)) >= 1000)
					snprintf(send_net, sizeof(send_net), "%2.3f GB", (float)shadow_ptr->send_net/(1024*1024*1024));
				else if((shadow_ptr->send_net/(1024*1024)) < 1)
					snprintf(send_net, sizeof(send_net), "%5d kb", (int)shadow_ptr->send_net/(1024));
				else
					snprintf(send_net, sizeof(send_net), "%5.1f MB", (float)shadow_ptr->send_net/(1024*1024));

				/* store the number of bytes */
				shadow_ptr->cur_recv_bytes = recv_bytes;
				shadow_ptr->cur_send_bytes = trans_bytes;

				/* get the time we needed for our run */
				gettimeofday(&shadow_ptr->end_tv, 0);
				elapse  = (double)shadow_ptr->end_tv.tv_sec + (double)shadow_ptr->end_tv.tv_usec * 1e-6;
				elapse -= (double)shadow_ptr->begin_tv.tv_sec + (double)shadow_ptr->begin_tv.tv_usec * 1e-6;
				gettimeofday(&shadow_ptr->begin_tv, 0);

				/* this the correction value */
				(float)correction = (float)elapse;

				/* this is the value to be corrected */
				(long long)delta_bytes = (long long)(shadow_ptr->cur_recv_bytes - shadow_ptr->prev_recv_bytes);
				/* now we correct it */
				speed_bytes = ((long long)delta_bytes/correction);

				if(debug >= HIGH)
					(void)vrprint.debug("Debug", "status_section: bytes: %d, corrections: %f", (int)speed_bytes, correction);

				/* calculating the current connection speed */
				if(iface_ptr->up)
				{
					if(shadow_ptr->prev_recv_bytes == -1)
						snprintf(recv_speed, sizeof(recv_speed), "calc");
					else if((speed_bytes/1024) < 1)
						snprintf(recv_speed, sizeof(recv_speed), "%5d b", (int)speed_bytes);
					else if((speed_bytes/1024) >= 1024)
						snprintf(recv_speed, sizeof(recv_speed), "%5.1f mb", (float)speed_bytes/(1024*1024));
					else
						snprintf(recv_speed, sizeof(recv_speed), "%5.1f kb", (float)speed_bytes/1024);
				}
				else
				{
					snprintf(recv_speed, sizeof(recv_speed), "%5s", "-");
				}


				/* this is the value to be corrected */
				(long long)delta_bytes = (long long)(shadow_ptr->cur_send_bytes - shadow_ptr->prev_send_bytes);
				/* now we correct it */
				delta_bytes = (long long)((long long)delta_bytes/(float)correction);

				if(iface_ptr->up)
				{
					if(shadow_ptr->prev_send_bytes == -1)
						snprintf(send_speed, sizeof(send_speed), "calc");
					else if((delta_bytes/1024) < 1)
						snprintf(send_speed, sizeof(send_speed), "%5d b", (int)delta_bytes);
					else if((delta_bytes/1024) >= 1024)
						snprintf(send_speed, sizeof(send_speed), "%5.1f mb", (float)delta_bytes/(1024*1024));
					else
						snprintf(send_speed, sizeof(send_speed), "%5.1f kb", (float)delta_bytes/1024);
				}
				else
				{
					snprintf(send_speed, sizeof(send_speed), "%5s", "-");
				}
		
				/* set the fields to the form */
				for(i = cur_interface; i < StatusSection.n_fields; i++)
				{
					cur = StatusSection.fields[i];

					if(strncmp(field_buffer(cur, 1), "recv_s", 6) == 0)
						set_field_buffer(cur, 0, recv_speed);
					else if(strncmp(field_buffer(cur, 1), "send_s", 6) == 0)
						set_field_buffer(cur, 0, send_speed);

					else if(strncmp(field_buffer(cur, 1), "rcv_ti", 6) == 0)
						set_field_buffer(cur, 0, recv_host);
					else if(strncmp(field_buffer(cur, 1), "snd_to", 6) == 0)
						set_field_buffer(cur, 0, send_host);

					else if(strncmp(field_buffer(cur, 1), "rcv_tf", 6) == 0)
						set_field_buffer(cur, 0, recv_net);
					else if(strncmp(field_buffer(cur, 1), "snd_tf", 6) == 0)
					{
						set_field_buffer(cur, 0, send_net);
						break;
					}

				}
				cur_interface = i + 1;

				/*
					draw the interface name
				*/
				snprintf(interfacename, sizeof(interfacename), "%s", iface_ptr->name);
			
				if(iface_ptr->up)
					wattron(StatusSection.win, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);
			
				mvwprintw(StatusSection.win, y, 2,  "%s", interfacename);

				if(iface_ptr->up)
					wattroff(StatusSection.win, COLOR_PAIR(CP_BLUE_WHITE)|A_BOLD);

				/*
					store the number of bytes
				*/
				shadow_ptr->prev_recv_bytes = recv_bytes;
				shadow_ptr->prev_send_bytes = trans_bytes;
			}

			/*
				finally draw the screen
			*/
			wrefresh(StatusSection.win);
		}

		/* process the keyboard input */
		ch = wgetch(StatusSection.win);
		switch(ch)
		{
			/* quit */
			case 27:
			case 'q':
			case KEY_F(10):
				quit = 1;
				break;

			case KEY_F(12):
				print_help(debuglvl, ":[VUURMUUR:STATUS]:");
				break;
		}

		if(quit == 0)
		{
			usleep(10000);
			slept_so_far = slept_so_far + 10000;

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "just slept: slept_so_far '%d'.", slept_so_far);
		}
	}

	/* remove the interfaces inserted as TYPE_FIREWALL's into the zonelist */
	if(rem_iface_from_zonelist(0, &ZonedataList) < 0)
	{
		(void)vrprint.error(-1, "Error", "rem_iface_from_zonelist failed (in: main).");
		retval=-1;
	}

	/* destroy hashtables and the shadowlist */
	hash_cleanup(debuglvl, &ZonedataHash);
	hash_cleanup(debuglvl, &ServicesHash);
	d_list_cleanup(debuglvl, &shadow_list);

	/* EXIT: cleanup */
	nodelay(StatusSection.win, FALSE);

	/* destroy the window and form */
	status_section_destroy();
	
	update_panels();
	doupdate();

	return(retval);
}
