/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


struct ZonesSection_
{
	/*	first the menus
	
		they each have their own data struct because they can
		be shown at the same time, only hosts and groups share
		the data struct
	*/

	/* zones */
	WINDOW  *win;
	PANEL   *panel[1];
	MENU    *menu;
	ITEM    **zoneitems;
	int     zone_n;

	/* networks */
	WINDOW  *n_win;
	PANEL   *n_panel[1];
	MENU    *n_menu;
	ITEM    **networkitems;
	int     network_n;

	/* hosts and groups */
	WINDOW  *h_win;
	PANEL   *h_panel[1];
	MENU    *h_menu;
	ITEM    **hostitems;
	int     host_n;
	d_list	desc_list;
  
	/* edit a zone/network/host/group */ 
	struct EditZone_
	{
		PANEL   *panel[1];
		WINDOW  *win;
		FIELD   **fields;
		FORM    *form;
		int     n_fields;
	} EditZone;

	/* edit zone interfaces */
	struct EditZoneInt_
	{
		PANEL   *panel[1];
		WINDOW  *win;
		MENU    *menu;
		ITEM    **items;
		int     n_items;
	} EditZoneInt;

	/* edit zone groups */
	struct EditZoneGrp_
	{
		PANEL   *panel[1];
		WINDOW  *win;
		MENU    *menu;
		ITEM    **items;
		int     n_items;
	} EditZoneGrp;

	char comment[241]; /* 5*48 + 1 */

} ZonesSection;


static int
edit_zone_host_init(const int debuglvl, char *name, int height, int width, int startx, int starty, struct ZoneData_ *zone_ptr)
{
	int	i,
		rows,
		cols,
		max_height,
		max_width,
		comment_y = 0,	/* for the dimentions of */
		comment_x = 0;	/* the comment field */

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* get screen dimensions */
	getmaxyx(stdscr, max_height, max_width);

	/* alloc fields */
	ZonesSection.EditZone.n_fields = 4;
	if(!(ZonesSection.EditZone.fields = (FIELD **)calloc(ZonesSection.EditZone.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "could not setup fields (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* create the fields */
	ZonesSection.EditZone.fields[0] = new_field(1, 3, 3, 1, 0, 1);					/* active */
	set_field_buffer(ZonesSection.EditZone.fields[0], 0, zone_ptr->active ? "Yes" : "No");
	set_field_buffer(ZonesSection.EditZone.fields[0], 1, "act");

	ZonesSection.EditZone.fields[1] = new_field(1, 16, 3, 23, 0, 1);				/* ipaddress */
	set_field_type(ZonesSection.EditZone.fields[1], TYPE_IPV4);
	set_field_buffer(ZonesSection.EditZone.fields[1], 0, zone_ptr->ipv4.ipaddress);
	set_field_buffer(ZonesSection.EditZone.fields[1], 1, "ipaddress");
	field_opts_on(ZonesSection.EditZone.fields[1], O_BLANK);

	ZonesSection.EditZone.fields[2] = new_field(1, 19, 6, 23, 0, 1);				/* macaddress */
	set_field_buffer(ZonesSection.EditZone.fields[2], 0, zone_ptr->mac);
	set_field_buffer(ZonesSection.EditZone.fields[2], 1, "mac");
	field_opts_on(ZonesSection.EditZone.fields[2], O_BLANK);

	/* comment field size */
	comment_y = 5;
	comment_x = 48;

	/* create the comment field */
	ZonesSection.EditZone.fields[3] = new_field(comment_y, comment_x, 11, 1, 0, 1); // comment
	
	/* read the comment from backend */
	if(zf->ask(debuglvl, zone_backend, zone_ptr->name, "COMMENT", ZonesSection.comment, sizeof(ZonesSection.comment), TYPE_HOST, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	set_field_buffer(ZonesSection.EditZone.fields[3], 0, ZonesSection.comment);
	set_field_buffer(ZonesSection.EditZone.fields[3], 1, "comment");

	ZonesSection.EditZone.fields[4] = NULL;

	/* create the window & panel */
	if(!(ZonesSection.EditZone.win = create_newwin(height, width, startx, starty, "Edit Zone: Host", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "could not setup window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZone.panel[0] = new_panel(ZonesSection.EditZone.win)))
	{
		(void)vrprint.error(-1, "Error", "could not setup panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	keypad(ZonesSection.EditZone.win, TRUE);

	/* set field options */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		set_field_back(ZonesSection.EditZone.fields[i], COLOR_PAIR(CP_WHITE_BLUE));
		field_opts_off(ZonesSection.EditZone.fields[i], O_AUTOSKIP);
		set_field_status(ZonesSection.EditZone.fields[i], FALSE);
	}

	/* Create the form and post it */
	if(!(ZonesSection.EditZone.form = new_form(ZonesSection.EditZone.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not setup form (in: %s).", __FUNCTION__);
		return(-1);
	}
	scale_form(ZonesSection.EditZone.form, &rows, &cols);
	set_form_win(ZonesSection.EditZone.form, ZonesSection.EditZone.win);
	set_form_sub(ZonesSection.EditZone.form, derwin(ZonesSection.EditZone.win, rows, cols, 1, 2));
	post_form(ZonesSection.EditZone.form);

	/* print labels */
	mvwprintw(ZonesSection.EditZone.win, 1, 2, "Name: %s", zone_ptr->name);
	mvwprintw(ZonesSection.EditZone.win, 3, 2, "Active:");
	mvwprintw(ZonesSection.EditZone.win, 3, 24, "Ipaddress:");
	mvwprintw(ZonesSection.EditZone.win, 6, 24, "MAC:");
	mvwprintw(ZonesSection.EditZone.win, 11, 2, "Comment:");

	/* draw */
	wrefresh(ZonesSection.EditZone.win);
	return(0);
}

static int
edit_zone_host_destroy(void)
{
	int	retval = 0,
		i = 0;

	/* unpost form and free the memory */
	unpost_form(ZonesSection.EditZone.form);
	free_form(ZonesSection.EditZone.form);

	for(i=0;i<ZonesSection.EditZone.n_fields;i++)
	{
		free_field(ZonesSection.EditZone.fields[i]);
	}
	free(ZonesSection.EditZone.fields);

	del_panel(ZonesSection.EditZone.panel[0]);
	destroy_win(ZonesSection.EditZone.win);

	update_panels();
	doupdate();

	/* clear comment */
	strcpy(ZonesSection.comment, "");
	return(retval);
}

static int
edit_zone_host_save(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	int	i = 0,
		active = 0;
	char	ipaddress[16] = "",
		mac[19] = "";

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* check for changed fields */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		/* field changed! */
		if(field_status(ZonesSection.EditZone.fields[i]) == TRUE)
		{
			/* active field */
			if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "act", 3) == 0)
			{
				/* for the log and incase something goes wrong */
				active = zone_ptr->active;
				
				if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[i], 0), "yes", 3) == 0)
					zone_ptr->active = 1;
				else if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[i], 0), "no", 2) == 0)
					zone_ptr->active = 0;
				else
				{
					zone_ptr->active = active;
					return(-1);
				}

				/* save to the backend */
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "ACTIVE", zone_ptr->active ? "Yes" : "No", 1, TYPE_HOST) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				/* for the log */
				(void)vrprint.info("Info", "host '%s' has been changed: active is now set to '%s' (was: '%s').", zone_ptr->name, zone_ptr->active ? "Yes" : "No", active ? "Yes" : "No");
				zone_ptr->status = ST_CHANGED;
			}

			/* ipaddress field */
			else if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "ipaddress", 9) == 0)
			{
				/* for the log and incase something goes wrong */
				if(strlcpy(ipaddress, zone_ptr->ipv4.ipaddress, sizeof(ipaddress)) >= sizeof(ipaddress))
				{
					(void)vrprint.error(-1, "Internal Error", "copying ipaddress failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				if(!(copy_field2buf(	zone_ptr->ipv4.ipaddress,
							field_buffer(ZonesSection.EditZone.fields[i], 0),
							sizeof(zone_ptr->ipv4.ipaddress))))
					return(-1);

				/* we dont check for invalid ip (check_ip == 0), because this is done by the fieldvalidation */
				if(check_ipv4address(debuglvl, zone_ptr->network_parent->ipv4.network, zone_ptr->network_parent->ipv4.netmask, zone_ptr->ipv4.ipaddress, 0))
				{
					if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "IPADDRESS", zone_ptr->ipv4.ipaddress, 1, TYPE_HOST) < 0)
					{
						(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
						return(-1);
					}

					(void)vrprint.info("Info", "host '%s' has been changed: ipaddress is now set to '%s' (was: '%s').", zone_ptr->name, zone_ptr->ipv4.ipaddress, ipaddress);
					zone_ptr->status = ST_CHANGED;
				}
				else
				{
					(void)vrprint.error(-1, "Error", "ipaddress '%s' doesn't belong in the network %s/%s.", zone_ptr->ipv4.ipaddress, zone_ptr->network_parent->ipv4.network, zone_ptr->network_parent->ipv4.netmask);

					/* copy the old ipaddress back */
					if(strlcpy(zone_ptr->ipv4.ipaddress, ipaddress, sizeof(zone_ptr->ipv4.ipaddress)) >= sizeof(zone_ptr->ipv4.ipaddress))
					{
						(void)vrprint.error(-1, "Internal Error", "copying ipaddress failed (in: %s).", __FUNCTION__);
						return(-1);
					}
				}
			}

			/* MAC field */
			else if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "mac", 3) == 0)
			{
				/* for the log and incase something goes wrong */
				if(strlcpy(mac, zone_ptr->mac, sizeof(mac)) >= sizeof(mac))
				{
					(void)vrprint.error(-1, "Internal Error", "copying macaddress failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				if(!(copy_field2buf(	zone_ptr->mac,
							field_buffer(ZonesSection.EditZone.fields[i], 0),
							sizeof(zone_ptr->mac))))
					return(-1);

/* TODO: regex validate MAC */

				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "MAC", zone_ptr->mac, 1, TYPE_HOST) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				(void)vrprint.info("Info", "host '%s' has been changed: macaddress is now set to '%s' (was: '%s').", zone_ptr->name, zone_ptr->mac, mac);
				zone_ptr->status = ST_CHANGED;
			}

			/* comment field */
			else if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "comment", 7) == 0)
			{
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "COMMENT", field_buffer(ZonesSection.EditZone.fields[i], 0), 1, TYPE_HOST) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				(void)vrprint.info("Info", "host '%s' has been changed: the comment was changed.", zone_ptr->name);
			}
		}
	}

	return(0);
}

/*	edit_zone_host

	Returncodes:
		 0: ok
		-1: error
*/
int
edit_zone_host(const int debuglvl, char *name)
{
	int			ch,
				not_defined = 0,
				quit = 0,
				retval = 0;
	struct ZoneData_	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				max_height,
				max_width;
	FIELD			*cur = NULL,
				*prev = NULL;

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* screen dimentions */
	getmaxyx(stdscr, max_height, max_width);
	height = 20;
	if(height > max_height - 6)
		height = max_height-6;

	width = 54;
	startx = 3;
	starty = 24;

	/* search the host in memory */
	if(!(zone_ptr = search_zonedata(debuglvl, name, &ZonedataList)))
	{
		(void)vrprint.error(-1, "Internal Error", "host not found (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* init */
	if(edit_zone_host_init(debuglvl, name, height, width, startx, starty, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up the host window failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(!(cur = current_field(ZonesSection.EditZone.form)))
	{
		(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
		return(-1);
	}

	status_print(status_win, "Ready.");
	
	/* loop through to get user requests */
	while(quit == 0)
	{
		/* draw nice markers */
		draw_field_active_mark(cur, prev, ZonesSection.EditZone.win, ZonesSection.EditZone.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		not_defined = 0;

		/* get user input */
		ch = wgetch(ZonesSection.EditZone.win);

		if(strncmp(field_buffer(cur, 1), "comment", 7) == 0)
		{
			if(nav_field_comment(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else if(strncmp(field_buffer(cur, 1), "act", 3) == 0)
		{
			if(nav_field_yesno(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else if(strncmp(field_buffer(cur, 1), "ipaddress", 9) == 0 ||
			strncmp(field_buffer(cur, 1), "mac", 3) == 0)
		{
			if(nav_field_simpletext(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else
		{
			not_defined = 1;
		}

		/* the rest is handled here */
		if(not_defined)
		{
			switch(ch)
			{
				case KEY_F(5): //help
				
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
				
					form_driver(ZonesSection.EditZone.form, REQ_NEXT_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;

				case KEY_UP:
					
					form_driver(ZonesSection.EditZone.form, REQ_PREV_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;

				case 27:
				case KEY_F(10):
				case KEY_LEFT:
				case 'q':

					quit = 1;
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:HOST:EDIT]:");
					break;
			}
		}

		/* before we get the new 'cur', store cur in prev */
		prev = cur;
		if(!(cur = current_field(ZonesSection.EditZone.form)))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* now give some help message in the status win */
		if(strncasecmp(field_buffer(cur, 1), "ipaddress", 9) == 0)
			status_print(status_win, "Please enter an ipaddress (mandatory).");
		else if(strncasecmp(field_buffer(cur, 1), "mac", 3) == 0)
			status_print(status_win, "Please enter the macaddress (optional).");
		else if(strncasecmp(field_buffer(cur, 1), "act", 3) == 0)
			status_print(status_win, "If set to 'No' no rules will be created for it.");
		else if(strncasecmp(field_buffer(cur, 1), "vrt", 10) == 0)
			status_print(status_win, "If set to 'Yes' it will be used as a virtual interface of the firewall.");
		else if(strncasecmp(field_buffer(cur, 1), "comment", 7) == 0)
			status_print(status_win, "Type away! ;-)");

		/* draw and set cursor */
		wrefresh(ZonesSection.EditZone.win);
		pos_form_cursor(ZonesSection.EditZone.form);
	}

	/* save changes */
	if(edit_zone_host_save(debuglvl, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving host failed.");
		retval = -1;
	}

	/* cleanup */
	if(edit_zone_host_destroy() < 0)
		return(-1);

	return(retval);
}


static int
zones_section_menu_hosts_init(char *zonename, char *networkname)
{
	int			retval=0,
				i = 0;
	struct ZoneData_	*zone_ptr = NULL;
	int			height = 0,
				width = 0,
				startx = 0,
				starty = 0,
				maxx = 0,
				maxy = 0,
				result = 0;
	d_list_node		*d_node = NULL;
	
	/* get the screensize */
	getmaxyx(stdscr, maxy, maxx);

	/* count how many hosts there are */
	for(ZonesSection.host_n = 0, d_node = ZonedataList.top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* only count hosts inside the network and zone */
		if(zone_ptr->type == TYPE_HOST)
		{
			if(strcmp(zone_ptr->zone_name, zonename) == 0 && strcmp(zone_ptr->network_name, networkname) == 0)
			{
				ZonesSection.host_n++;
			}
		}
	}

	i = ZonesSection.host_n - 1;

	/* allow the menu items */
	if(!(ZonesSection.hostitems = (ITEM **)calloc(ZonesSection.host_n + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* create the menu items */
	for(d_node = ZonedataList.bot; d_node ; d_node = d_node->prev)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* only add hosts inside the network and zone */
		if(zone_ptr->type == TYPE_HOST)
		{
			if(strcmp(zone_ptr->zone_name, zonename) == 0 && strcmp(zone_ptr->network_name, networkname) == 0)
			{
				if(!(ZonesSection.hostitems[i] = new_item(zone_ptr->host_name, zone_ptr->ipv4.ipaddress)))
				{
					(void)vrprint.error(-1, "Internal Error", "adding new item to host menu failed (in: %s).", __FUNCTION__);
					return(-1);
				}

				i--;
			}
		}
	}
	/* terminate the items */
	ZonesSection.hostitems[ZonesSection.host_n] = (ITEM *)NULL;

	/* now create the menu */
	if(!(ZonesSection.h_menu = new_menu((ITEM **)ZonesSection.hostitems)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the host menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* now set the size of the window */
	height = ZonesSection.host_n + 9;
	width  = MAX_HOST + 18 + 2;
	startx = 3;
	starty = 8;

	if(maxy < starty + height + 3)
	{
		starty = 3;
		height = maxy - 2 * starty;
	}

	if(!(ZonesSection.h_win = newwin(height, width, starty, startx)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the host win failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.h_win, COLOR_PAIR(5));
	keypad(ZonesSection.h_win, TRUE);
	box(ZonesSection.h_win, 0, 0);
	print_in_middle(ZonesSection.h_win, 1, 0, width, "Hosts", COLOR_PAIR(5));
	wrefresh(ZonesSection.h_win);

	if(!(ZonesSection.h_panel[0] = new_panel(ZonesSection.h_win)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the host panel failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	set_menu_win(ZonesSection.h_menu, ZonesSection.h_win);
	set_menu_sub(ZonesSection.h_menu, derwin(ZonesSection.h_win, height-7, width-2, 3, 1));

	set_menu_format(ZonesSection.h_menu, height-8, 1);

	mvwaddch(ZonesSection.h_win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, 2, width-1, ACS_RTEE);

	set_menu_back(ZonesSection.h_menu, COLOR_PAIR(5));
	set_menu_fore(ZonesSection.h_menu, COLOR_PAIR(3));

	result = post_menu(ZonesSection.h_menu);
	if(result != E_OK && result != E_NOT_CONNECTED)
	{
		(void)vrprint.error(-1, "Internal Error", "creating the host menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	mvwaddch(ZonesSection.h_win, height-5, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, height-5, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, height-5, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.h_win, height-4, 1, "Usage: <RET> to edit the host");
	mvwprintw(ZonesSection.h_win, height-3, 1, "       <INS> to insert a new host");
	mvwprintw(ZonesSection.h_win, height-2, 1, "       <DEL> to removes the host");

	update_panels();
	doupdate();

	return(retval);
}


static int
zones_section_menu_hosts_destroy(void)
{
	int	retval = 0,
		i = 0;

	unpost_menu(ZonesSection.h_menu);
	free_menu(ZonesSection.h_menu);
	for(i = 0; i < ZonesSection.host_n; ++i)
		free_item(ZonesSection.hostitems[i]);

	free(ZonesSection.hostitems);

	del_panel(ZonesSection.h_panel[0]);

	destroy_win(ZonesSection.h_win);

	return(retval);
}


static int
zones_section_menu_hosts(const int debuglvl, char *zonename, char *networkname, struct rgx_ *reg)
{
	int			ch = 0,
				quit = 0,
				reload = 0,
				result = 0,
				retval = 0,
				size = 0;
	struct ZoneData_	*zone_ptr = NULL;
	char			*new_zone_ptr = NULL,
				*temp_ptr = NULL;
	ITEM			*cur = NULL;

	/* safety */
	if(!zonename || !networkname || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* setup */
	if(zones_section_menu_hosts_init(zonename, networkname) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up hosts menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* enter the loop */
	while(quit == 0)
	{
		/* reload the menu */
		if(reload == 1)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, reloading menu.");
		
			/* first destroy */
			if(zones_section_menu_hosts_destroy() < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, old menu destroyed.");

			/* and setup again */
			if(zones_section_menu_hosts_init(zonename, networkname) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, new menu created.");

			/* we are done with reloading */
			reload = 0;
		}

		/* loop for catching user input */
		while(quit == 0 && reload == 0)
		{
			/* get the user input */
			ch = wgetch(ZonesSection.h_win);
			switch(ch)
			{
				case KEY_LEFT:
				case KEY_F(10):
				case 27:
				case 'q':
					quit = 1;
					break;

				case KEY_F(5): //help
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_IC: /* insert key */
					
					new_zone_ptr = input_box(MAX_HOST, "New Host", "Enter the name of the new host");
					if(new_zone_ptr != NULL)
					{
						/* get the size */
						size = strlen(new_zone_ptr) + 1 + strlen(networkname) + 1 + strlen(zonename) + 1;
						if(size <= 0)
						{
							(void)vrprint.error(-1, "Internal Error", "string size <= 0!");
							return(-1);
						}

						/* alloc the memory */
						if(!(temp_ptr = malloc(size)))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
							return(-1);
						}

						/* create the string */
						strlcpy(temp_ptr, new_zone_ptr, size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, networkname, size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, zonename, size);

						if(validate_zonename(debug, temp_ptr, 1, NULL, NULL, NULL, reg->zonename) == 0)
						{
							if(new_zone(debug, temp_ptr, TYPE_HOST, &ZonedataList) >= 0)
							{
								if(edit_zone_host(debuglvl, temp_ptr) < 0)
								{
									(void)vrprint.error(-1, "Error", "editting the host failed.");
									return(-1);
								}
							}

							/* we have a new host, so reload the menu */
							reload = 1;
						}

						free(new_zone_ptr);
						free(temp_ptr);
					}
					break;

				/*
					delete
				*/
				case KEY_DC:
				{
					if(current_item(ZonesSection.h_menu))
					{
						if(confirm("Delete", "This host?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{
							/*
								get the current item
							*/
							if(!(cur = current_item(ZonesSection.h_menu)))
							{
								(void)vrprint.error(-1, "Internal Error", "current_item failed: cur == NULL.");
								return(-1);
							}

							/*
								size
							*/
							size = strlen((char *)item_name(cur))+1+strlen(networkname)+1+strlen(zonename)+1;
							if(size <= 0)
							{
								(void)vrprint.error(-1, "Internal Error", "string size <= 0!");
								return(-1);
							}

							// alloc the memory
							if(!(temp_ptr = malloc(size)))
							{
								(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
								return(-1);
							}

							// create the string
							strlcpy(temp_ptr, (char *)item_name(cur), size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, networkname, size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, zonename, size);

							zone_ptr = search_zonedata(debuglvl, temp_ptr, &ZonedataList);
							if(zone_ptr == NULL)
							{
								(void)vrprint.error(-1, "Error", "couldn't find %s in memory.", temp_ptr);
								retval = -1;
							}

							if(delete_zone(debug, temp_ptr, zone_ptr->type, &ZonedataList) < 0)
							{
								(void)vrprint.error(result, "Error", "deleting zone failed.");
								retval = -1;
							}

							free(temp_ptr);
							reload = 1;
						}
					}
					break;
				}

				case KEY_DOWN:
					menu_driver(ZonesSection.h_menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.h_menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.h_menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.h_menu, REQ_LAST_ITEM); // page down
					break;

				case 32: // space
				case KEY_RIGHT:
				case 10: // enter
				{
					if(current_item(ZonesSection.h_menu))
					{
						// get the current item
						if(!(cur = current_item(ZonesSection.h_menu)))
						{
							(void)vrprint.error(-1, "Internal Error", "current_item failed: cur == NULL.");
							return(-1);
						}

						// size
						size = strlen((char *)item_name(cur))+1+strlen(networkname)+1+strlen(zonename)+1;
						if(size <= 0)
						{
							(void)vrprint.error(-1, "Internal Error", "string size <= 0!");
							return(-1);
						}

						// alloc the memory
						if(!(temp_ptr = malloc(size)))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
							return(-1);
						}

						// create the string
						strlcpy(temp_ptr, (char *)item_name(cur), size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, networkname, size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, zonename, size);

						zone_ptr = search_zonedata(debuglvl, temp_ptr, &ZonedataList);
						if(zone_ptr != NULL)
						{
							if(zone_ptr->type == TYPE_HOST)
							{
								(void)edit_zone_host(debuglvl, zone_ptr->name);
							}
							else
							{
								(void)vrprint.error(-1, "Internal Error", "expected a host but got %d.", zone_ptr->type);
								return(-1);
							}
						}
						else
						{
							(void)vrprint.error(-1, "Internal Error", "%s not found in memory.", temp_ptr);
							return(-1);
						}

						free(temp_ptr);

						reload = 1;
					}
					break;
				}

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:HOSTS]:");
					break;
			}
		}
	}

	if(zones_section_menu_hosts_destroy() < 0)
		retval = -1;

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_zone_group_members_init(const int debuglvl, struct ZoneData_ *group_ptr)
{
	int			retval=0;
	d_list_node		*d_node = NULL;
	struct ZoneData_	*member_ptr = NULL;
	int			i=0;
	int			height = 30,
				width  = 54, /* max width of host_name (32) + box (2) + 4 + 16 */
				startx = 5,
				starty = 5,
				max_height = 0,
				max_width = 0;

	/* safety */
	if(!group_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s)", __FUNCTION__);
		return(-1);
	}
	if(group_ptr->type != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Internal Error", "expected a GROUP (%d), but got a %d (in: %s)", TYPE_GROUP, group_ptr->type, __FUNCTION__);
		return(-1);
	}

	getmaxyx(stdscr, max_height, max_width);
	
	ZonesSection.EditZoneGrp.n_items = group_ptr->GroupList.len;

	if(!(ZonesSection.EditZoneGrp.items = (ITEM **)calloc(ZonesSection.EditZoneGrp.n_items + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "could not setup items (in: %s)", __FUNCTION__);
		return(-1);
	}

	height = ZonesSection.EditZoneGrp.n_items + 7; /* 7 because: 3 above the list, 4 below */
	if(height >= max_height - starty)
	{
		height = max_height - 6;
		starty = 3;
	}

	/* load the items */
	for(i = 0, d_node = group_ptr->GroupList.top; d_node ; d_node = d_node->next, i++)
	{
		if(!(member_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s)", __FUNCTION__);
			return(-1);
		}

		/* load all interfaces into memory */
		if(!(ZonesSection.EditZoneGrp.items[i] = new_item(member_ptr->host_name, member_ptr->ipv4.ipaddress)))
		{
			(void)vrprint.error(-1, "Error", "could not load item (in: %s)", __FUNCTION__);
			return(-1);
		}
	}
	ZonesSection.EditZoneGrp.items[ZonesSection.EditZoneGrp.n_items] = (ITEM *)NULL;

	/* create win and panel */
	if(!(ZonesSection.EditZoneGrp.win = newwin(height, width, starty, startx)))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s)", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZoneGrp.panel[0] = new_panel(ZonesSection.EditZoneGrp.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s)", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.EditZoneGrp.win, COLOR_PAIR(5));
	keypad(ZonesSection.EditZoneGrp.win, TRUE);

	if(!(ZonesSection.EditZoneGrp.menu = new_menu((ITEM **)ZonesSection.EditZoneGrp.items)))
	{
		(void)vrprint.error(-1, "Error", "could not create menu (in: %s)", __FUNCTION__);
		return(-1);
	}
	set_menu_win(ZonesSection.EditZoneGrp.menu, ZonesSection.EditZoneGrp.win);
	set_menu_sub(ZonesSection.EditZoneGrp.menu, derwin(ZonesSection.EditZoneGrp.win, height-6, width-2, 3, 1));
	set_menu_format(ZonesSection.EditZoneGrp.menu, height-7, 1);

	/* markup */
	box(ZonesSection.EditZoneGrp.win, 0, 0);
	print_in_middle(ZonesSection.EditZoneGrp.win, 1, 0, width, "Members", COLOR_PAIR(5));
	mvwaddch(ZonesSection.EditZoneGrp.win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.EditZoneGrp.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.EditZoneGrp.win, 2, width-1, ACS_RTEE);

	set_menu_back(ZonesSection.EditZoneGrp.menu, COLOR_PAIR(CP_BLUE_WHITE));
	set_menu_fore(ZonesSection.EditZoneGrp.menu, COLOR_PAIR(CP_WHITE_BLUE));

	post_menu(ZonesSection.EditZoneGrp.menu);

	mvwaddch(ZonesSection.EditZoneGrp.win, height-4, 0, ACS_LTEE);
	mvwhline(ZonesSection.EditZoneGrp.win, height-4, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.EditZoneGrp.win, height-4, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.EditZoneGrp.win, height-3, 2, "Usage:  <INS> to insert a member");
	mvwprintw(ZonesSection.EditZoneGrp.win, height-2, 2, "        <DEL> to remove a member");

	update_panels();
	doupdate();

	return(retval);
}

static int
edit_zone_group_members_destroy(void)
{
	int	retval = 0,
		i;

	// Un post form and free the memory
	unpost_menu(ZonesSection.EditZoneGrp.menu);
	free_menu(ZonesSection.EditZoneGrp.menu);

	for(i=0;i<ZonesSection.EditZoneGrp.n_items;i++)
	{
		free_item(ZonesSection.EditZoneGrp.items[i]);
	}
	free(ZonesSection.EditZoneGrp.items);

	del_panel(ZonesSection.EditZoneGrp.panel[0]);
	destroy_win(ZonesSection.EditZoneGrp.win);

	update_panels();
	doupdate();

	return(retval);
}

/*	edit_zone_group_members_delmem

	Deletes the member 'member_name' from the GroupList of
	the group 'group_ptr'.

	Returncodes:
		 0: ok
		-1: error
*/
static int
edit_zone_group_members_delmem(const int debuglvl, char *member_name, struct ZoneData_ *group_ptr)
{
	d_list_node		*d_node = NULL;
	struct ZoneData_	*member_ptr = NULL;

	/* safety */
	if(!group_ptr || !member_name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* this should not happen, but it cant hurt checking right? */
	if(group_ptr->type != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Internal Error", "Expected a GROUP (%d), but got a %d! (in: %s)", TYPE_GROUP, group_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* search the member */
	for(d_node = group_ptr->GroupList.top; d_node; d_node = d_node->next)
	{
		if(!(member_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* here is is */
		if(strcmp(member_name, member_ptr->host_name) == 0)
		{
			/* okay, lets remove the hugeassmotherf*cker */
			if(d_list_remove_node(debuglvl, &group_ptr->GroupList, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "unable to remove member from the list (in: %s).", __FUNCTION__);
				return(-1);
			}
			else
				break;
		}
	}

	/* save the new group list */
	if(zones_group_save_members(debuglvl, group_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saveing the new grouplist to the backend failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* for logging */
	(void)vrprint.info("Info", "group '%s' has been changed: the member '%s' has been removed.", group_ptr->name, member_name);

	return(0);
}

/*	edit_zone_group_members_newmem

	Displays a menu with hosts to chose from to add to the
	group 'group_ptr'.

	Returncodes:
		 0: ok
		-1: error
*/
static int
edit_zone_group_members_newmem(const int debuglvl, struct ZoneData_ *group_ptr)
{
	d_list_node		*d_node;
	char 			**choices,
				*choice_ptr=NULL,
				search_name[MAX_HOST_NET_ZONE];
	int			n_choices=0,
				i=0;
	struct ZoneData_	*new_member_ptr,
				*zonelist_ptr;

	/* safety */
	if(!group_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	/* this should not happen, but it cant hurt checking right? */
	if(group_ptr->type != TYPE_GROUP)
	{
		(void)vrprint.error(-1, "Internal Error", "Expected a GROUP (%d), but got a %d! (in: %s)", TYPE_GROUP, group_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* first count the number of hosts in this network */
	for(n_choices = 0, d_node = ZonedataList.top; d_node; d_node = d_node->next)
	{
		if(!(zonelist_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		if(zonelist_ptr->type == TYPE_HOST)
		{
			/* check if the host belongs to our zone */
			if(strcmp(group_ptr->zone_name, zonelist_ptr->zone_name) == 0)
			{
				/* check if the host belongs to our network */
				if(strcmp(group_ptr->network_name, zonelist_ptr->network_name) == 0)
				{
					n_choices++;
				}
			}
		}
	}

	/* alloc the mem */
	if(!(choices = calloc(n_choices + 1, MAX_HOST)))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* now init the new mem */
	for(i = n_choices - 1, d_node = ZonedataList.bot; d_node ; d_node = d_node->prev)
	{
		if(!(zonelist_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		if(zonelist_ptr->type == TYPE_HOST)
		{
			if(strcmp(group_ptr->zone_name, zonelist_ptr->zone_name) == 0)
			{
				if(strcmp(group_ptr->network_name, zonelist_ptr->network_name) == 0)
				{
					choices[i] = zonelist_ptr->host_name;
					i--;
				}
			}
		}
	}

	/* let the user select one. If he/she doesn't select one, fine, bail out. */
	if(!(choice_ptr = selectbox("New member", "Select a host", n_choices, choices, 1)))
	{
		/* no choice was made, so quit quietly. */
		free(choices);
		return(0);
	}

	/* clean up */
	free(choices);

	/* assemble the full zonename... */
	if(snprintf(search_name, sizeof(search_name), "%s.%s.%s", choice_ptr, group_ptr->network_name, group_ptr->zone_name) >= sizeof(search_name))
	{
		(void)vrprint.error(-1, "Error", "buffer overflow attempt (in: %s).", __FUNCTION__);
		free(choice_ptr);
		return(-1);
	}
	
	/* and free the hostname */
	free(choice_ptr);

//TODO: validate the name?

	/* now search the host in memory */
	new_member_ptr = search_zonedata(debuglvl, search_name, &ZonedataList);
	if(!new_member_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "member '%s' is invalid, it was not found in memory.", search_name);
		return(-1);
	}
	else if(new_member_ptr->type != TYPE_HOST)
	{
		(void)vrprint.error(-1, "Internal Error", "member '%s' is not a host!", search_name);
		return(-1);
	}
	/* okay, no objections? */
	else
	{
		/* now append the new at the tail of the list */
		if(d_list_append(debug, &group_ptr->GroupList, new_member_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "unable to append member to groupslist (in: %s).", __FUNCTION__);
			return(-1);
		}
	}

	/* save the new group list */
	if(zones_group_save_members(debuglvl, group_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saveing the new grouplist to the backend failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	(void)vrprint.info("Info", "group '%s' has been changed: the member '%s' has been added.", group_ptr->name, search_name);

	return(0);
}

/*	edit_group_members

	Displays the grouplist and allows the user to add
	and remove members.

	Returncodes:
		 0: ok
		-1: error
*/
int
edit_zone_group_members(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	int	quit = 0,
		reload = 0,
		ch,
		retval = 0;
	ITEM	*cur = NULL;

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s)", __FUNCTION__);
		return(-1);
	}

	if(edit_zone_group_members_init(debuglvl, zone_ptr) < 0)
		return(-1);

	while(quit == 0)
	{
		if(reload == 1)
		{
			/* first destroy */
			if(edit_zone_group_members_destroy() < 0)
				return(-1);

			/* then init again */
			if(edit_zone_group_members_init(debuglvl, zone_ptr) < 0)
				return(-1);

			/* refresh screen */
			update_panels();
			doupdate();

			reload = 0;
		}

		while(quit == 0 && reload == 0)
		{
			/* get user input */
			ch = wgetch(ZonesSection.EditZoneGrp.win);

			switch(ch)
			{
				case KEY_LEFT:
				case 27:	//esc
				case 'q':
				case KEY_F(10): //quit
				
					quit = 1;
					break;

				case KEY_F(5):	//help
				
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_IC:
				
					if(edit_zone_group_members_newmem(debuglvl, zone_ptr) < 0)
					{
						/* if this failes, print error, quit the members screen and set retval */
						(void)vrprint.error(-1, "Internal Error", "adding groupmember failed.");
						quit = 1;
						retval = -1;
					}

					reload=1;
					break;

				case KEY_DC:
				{
					cur = current_item(ZonesSection.EditZoneGrp.menu);
					if(cur)
					{

						if(edit_zone_group_members_delmem(debuglvl, (char *)item_name(cur), zone_ptr) < 0)
						{
							/* if this failes, print error, quit the members screen and set retval */
							(void)vrprint.error(-1, "Internal Error", "removing groupmember failed.");
							quit = 1;
							retval = -1;
						}

						reload=1;
					}
					break;
				}

				case KEY_DOWN:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.EditZoneGrp.menu, REQ_LAST_ITEM); // page down
					break;
			}
		}
	}

	/* cleanup */
	if(edit_zone_group_members_destroy() < 0)
		return(-1);

	return(retval);
}

static int
edit_zone_group_init(int debuglvl, char *name, struct ZoneData_ *zone_ptr)
{
	int			retval = 0,
				rows,
				cols,
				i,
				max_height = 0,
				max_width = 0,
				height = 0,
				width = 0,
				startx = 0,
				starty = 0,
				comment_y = 0,
				comment_x = 0;
	struct ZoneData_	*member_ptr = NULL;
	d_list_node		*d_node = NULL;

	/* safety */
	if(!name || !zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	getmaxyx(stdscr, max_height, max_width);
	height = 20 + zone_ptr->GroupList.len;
	if(height > max_height - 6)
		height = max_height - 6;
	width = 54;
	startx = 3;
	starty = 24;

	ZonesSection.EditZone.n_fields = 2;
	
	if(!(ZonesSection.EditZone.fields = (FIELD **)calloc(ZonesSection.EditZone.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "could not setup fields (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* preload the active field */
	ZonesSection.EditZone.fields[0] = new_field(1, 3, 3, 1, 0, 1);
	set_field_buffer(ZonesSection.EditZone.fields[0], 0, zone_ptr->active ? "Yes" : "No");
	set_field_buffer(ZonesSection.EditZone.fields[0], 1, "act");

	/* comment field size */
	comment_y = 5;
	comment_x = 48;

	/* create the comment field */
	ZonesSection.EditZone.fields[1] = new_field(comment_y, comment_x, 6, 1, 0, 1);

	/* load the comment from the backend */
	if(zf->ask(debuglvl, zone_backend, zone_ptr->name, "COMMENT", ZonesSection.comment, sizeof(ZonesSection.comment), TYPE_GROUP, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	set_field_buffer(ZonesSection.EditZone.fields[1], 0, ZonesSection.comment);
	set_field_buffer(ZonesSection.EditZone.fields[1], 1, "comment");

	ZonesSection.EditZone.fields[2] = NULL;

	if(!(ZonesSection.EditZone.win = create_newwin(height, width, startx, starty, "Edit Zone: Group", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZone.panel[0] = new_panel(ZonesSection.EditZone.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* set field options */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		set_field_back(ZonesSection.EditZone.fields[i], COLOR_PAIR(CP_WHITE_BLUE));
		field_opts_off(ZonesSection.EditZone.fields[i], O_AUTOSKIP);
		set_field_status(ZonesSection.EditZone.fields[i], FALSE);
	}

	/* Create the form and post it */
	if(!(ZonesSection.EditZone.form = new_form(ZonesSection.EditZone.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	/* Calculate the area required for the form */
	scale_form(ZonesSection.EditZone.form, &rows, &cols);
	keypad(ZonesSection.EditZone.win, TRUE);

	/* Set main window and sub window */
	set_form_win(ZonesSection.EditZone.form, ZonesSection.EditZone.win);
	set_form_sub(ZonesSection.EditZone.form, derwin(ZonesSection.EditZone.win, rows, cols, 1, 2));
	post_form(ZonesSection.EditZone.form);

	/* draw labels */
	mvwprintw(ZonesSection.EditZone.win, 1, 2, "Name: %s", zone_ptr->name);
	mvwprintw(ZonesSection.EditZone.win, 3, 2, "Active:");
	mvwprintw(ZonesSection.EditZone.win, 6, 2, "Comment:");
	mvwprintw(ZonesSection.EditZone.win, 13, 2, "Press <F6> to manage the members of this group.");

	if(height > 13 + 5)
	{
		mvwprintw(ZonesSection.EditZone.win, 15, 2, "List of members:");
		if(zone_ptr->GroupList.len == 0)
			mvwprintw(ZonesSection.EditZone.win, 16, 2, "  No members assigned yet.");
		else
		{
			for(d_node = zone_ptr->GroupList.top, i = 1; d_node; d_node = d_node->next, i++)
			{
				if(!(member_ptr = d_node->data))
					return(-1);
					
				mvwprintw(ZonesSection.EditZone.win, 15 + i, 2, "  %s", member_ptr->host_name);
				mvwprintw(ZonesSection.EditZone.win, 15 + i, width - 20, "  %s", member_ptr->ipv4.ipaddress);
				
				if(15 + i == height - 3) /* -1 is for the border */
				{
					if((zone_ptr->GroupList.len - i) > 1)
					{
						mvwprintw(ZonesSection.EditZone.win, 15+i+1, 2, "There are %d more members. Press F6 to manage.", zone_ptr->GroupList.len - i);
						break;
					}
				}
			}
		}
	}


	
	wrefresh(ZonesSection.EditZone.win);

	update_panels();
	doupdate();

	return(retval);
}

static int
edit_zone_group_save(const int debuglvl, struct ZoneData_ *group_ptr)
{
	int	retval = 0,
		i = 0;

	/* safety */
	if(!group_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check for changed fields */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		if(field_status(ZonesSection.EditZone.fields[i]) == TRUE)
		{
			/* active field */
			if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "act", 3) == 0)
			{
				group_ptr->status = ST_CHANGED;

				if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[i], 0), "yes", 3) == 0)
				{
					group_ptr->active = 1;
				}
				else if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[i], 0), "no", 2) == 0)
				{
					group_ptr->active = 0;
				}
				else
				{
					group_ptr->active = -1;
				}

				if(zf->tell(debuglvl, zone_backend, group_ptr->name, "ACTIVE", group_ptr->active ? "Yes" : "No", 1, TYPE_GROUP) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to the backend failed (in: %s).", __FUNCTION__);
					retval = -1;
				}

				(void)vrprint.info("Info", "group '%s' has been changed: active is now set to '%s'.", group_ptr->name, group_ptr->active ? "Yes" : "No");
			}
			else if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "comment", 7) == 0)
			{
				if(zf->tell(debug, zone_backend, group_ptr->name, "COMMENT", field_buffer(ZonesSection.EditZone.fields[i], 0), 1, TYPE_GROUP) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to the backend failed (in: %s).", __FUNCTION__);
					retval = -1;
				}

				(void)vrprint.info("Info", "group '%s' has been changed: the comment was changed.");
			}
			else
			{
				(void)vrprint.error(-1, "Error", "unknown field '%s' (in: %s).", field_buffer(ZonesSection.EditZone.fields[i], 1), __FUNCTION__);
				retval = -1;
			}
		}
	}

	return(retval);
}


static int
edit_zone_group_destroy(void)
{
	int	i = 0;

	/* Un post form and free the memory */
	unpost_form(ZonesSection.EditZone.form);
	free_form(ZonesSection.EditZone.form);

	for(i=0;i<ZonesSection.EditZone.n_fields;i++)
	{
		free_field(ZonesSection.EditZone.fields[i]);
	}
	free(ZonesSection.EditZone.fields);

	del_panel(ZonesSection.EditZone.panel[0]);
	destroy_win(ZonesSection.EditZone.win);

	/* clear comment string */
	strcpy(ZonesSection.comment, "");

	return(0);
}


/*	edit_zone_group

	Edit a group :-)
	
	Returncodes:
		0: ok
		-1: error
*/
static int
edit_zone_group(const int debuglvl, char *name)
{
	int			ch,
				not_defined = 0,
				retval = 0;
	struct ZoneData_	*zone_ptr = NULL;
	int			quit = 0;
	FIELD			*cur = NULL,
				*prev = NULL;

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* search the group in mem */
	if(!(zone_ptr = search_zonedata(debuglvl, name, &ZonedataList)))
	{
		(void)vrprint.error(-1, "Error", "group not found.");
		return(-1);
	}

	/* init */
	if(edit_zone_group_init(debuglvl, name, zone_ptr) < 0)
		return(-1);

	/* print (or not) initial warning about the group being empty. */
	if(zone_ptr->GroupList.len == 0)
	{
		wattron(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		mvwprintw(ZonesSection.EditZone.win, 15, 14, " Warning: group is empty! ");
		wattroff(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
	}

	mvwprintw(ZonesSection.EditZone.win, 4, 37, "Members: %4d", zone_ptr->GroupList.len);

	/* print, set cursor etc */
	pos_form_cursor(ZonesSection.EditZone.form);
	cur = current_field(ZonesSection.EditZone.form);
	
	wrefresh(ZonesSection.EditZone.win);
	update_panels();
	doupdate();

	/* loop through to get user requests */
	while(quit == 0)
	{
		draw_field_active_mark(cur, prev, ZonesSection.EditZone.win, ZonesSection.EditZone.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		/* get user input */
		ch = wgetch(ZonesSection.EditZone.win);

		not_defined = 0;

		/* handle input */
		if(strncmp(field_buffer(cur, 1), "comment", 7) == 0)
		{
			if(nav_field_comment(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else if(strncmp(field_buffer(cur, 1), "act", 3) == 0)
		{
			if(nav_field_yesno(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else
			not_defined = 1;

		/* the rest is handled here */
		if(not_defined == 1)
		{
			switch(ch)
			{
				case KEY_F(6):
					
					/* edit the members */
					if(edit_zone_group_members(debuglvl, zone_ptr) < 0)
					{
						retval = -1;
						quit = 1;
					}
					break;

				case KEY_LEFT:
				case 'q':
				case 27:
				case KEY_F(10): //quit

					quit = 1;
					break;
					
				case KEY_F(5): //help

					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_DOWN:
				case 9:		// tab
				case 10:	// enter

					form_driver(ZonesSection.EditZone.form, REQ_NEXT_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;
					
				case KEY_UP:

					form_driver(ZonesSection.EditZone.form, REQ_PREV_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:GROUP:EDIT]:");
					break;
			}
		}

		/* before we get the new 'cur', store cur in prev */
		prev = cur;
		cur = current_field(ZonesSection.EditZone.form);

		/* draw empty group warning */
		if(zone_ptr->GroupList.len == 0)
		{
			wattron(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
			mvwprintw(ZonesSection.EditZone.win, 15, 14, " Warning: group is empty! ");
			wattroff(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		}
		/* and clear */
		else
			mvwprintw(ZonesSection.EditZone.win, 15, 14, "                          ");
			
		mvwprintw(ZonesSection.EditZone.win, 4, 37, "Members: %4d", zone_ptr->GroupList.len);

		/* refresh and restore cursor. */
		wrefresh(ZonesSection.EditZone.win);
		pos_form_cursor(ZonesSection.EditZone.form);
	}

	/* save to backend */
	if(retval == 0)
	{
		if(edit_zone_group_save(debuglvl, zone_ptr) < 0)
			retval = -1;
	}

	/* cleanup */
	if(edit_zone_group_destroy() < 0)
		retval = -1;
	
	update_panels();
	doupdate();

	return(retval);
}


static int
zones_section_menu_groups_init(const int debuglvl, char *zonename, char *networkname)
{
	int			retval=0,
				i;
	struct ZoneData_	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				maxx,
				maxy;
	d_list_node		*d_node = NULL;
	char			temp[32],
				*desc_ptr = NULL;
	size_t			size = 0;

	getmaxyx(stdscr, maxx, maxy);

	// count how many zones there are
	ZonesSection.host_n=0;
	for(d_node = ZonedataList.top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "zone_ptr == NULL! (in: zones_section_menu_groups_init)");
			return(-1);
		}

		if(zone_ptr->type == TYPE_GROUP)
		{
			if((strcmp(zone_ptr->zone_name, zonename) == 0) && (strcmp(zone_ptr->network_name, networkname) == 0))
			{
				ZonesSection.host_n++;
			}
		}
	}

	if(d_list_setup(debuglvl, &ZonesSection.desc_list, free) < 0)
		return(-1);

	i = ZonesSection.host_n-1;
	ZonesSection.hostitems = (ITEM **)calloc(ZonesSection.host_n + 1, sizeof(ITEM *));

	for(d_node = ZonedataList.bot; d_node ; d_node = d_node->prev)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "zone_ptr == NULL! (in: zones_section_menu_groups_init)");
			return(-1);
		}

		if(zone_ptr->type == TYPE_GROUP)
		{
			if((strcmp(zone_ptr->zone_name, zonename) == 0) && (strcmp(zone_ptr->network_name, networkname) == 0))
			{
				snprintf(temp, sizeof(temp), "%6d members", zone_ptr->GroupList.len);
				size = strlen(temp) + 1;

				if(!(desc_ptr = malloc(size)))
				{
					(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
					return(-1);
				}

				strlcpy(desc_ptr, temp, size);

				if(d_list_append(debuglvl, &ZonesSection.desc_list, desc_ptr) < 0)
					return(-1);

				ZonesSection.hostitems[i] = new_item(zone_ptr->host_name, desc_ptr);
				i--;
			}
		}
	}

	ZonesSection.hostitems[ZonesSection.host_n] = (ITEM *)NULL;
	ZonesSection.h_menu = new_menu((ITEM **)ZonesSection.hostitems);

	height = ZonesSection.host_n+9;
	width = MAX_HOST+2;
	startx = 8;
	starty = 3;

	if(maxy < starty + height + 3)
	{
		starty = 3;
		height = maxy - 2 * starty;
	}
	
	ZonesSection.h_win = newwin(height, width, startx, starty);
	wbkgd(ZonesSection.h_win, COLOR_PAIR(5));
	keypad(ZonesSection.h_win, TRUE);
	box(ZonesSection.h_win, 0, 0);
	print_in_middle(ZonesSection.h_win, 1, 0, width, "Groups", COLOR_PAIR(5));
	wrefresh(ZonesSection.h_win);

	ZonesSection.h_panel[0] = new_panel(ZonesSection.h_win);
	update_panels();

	set_menu_win(ZonesSection.h_menu, ZonesSection.h_win);
	set_menu_sub(ZonesSection.h_menu, derwin(ZonesSection.h_win, height-7, width-2, 3, 1));

	set_menu_format(ZonesSection.h_menu, height-8, 1);

	mvwaddch(ZonesSection.h_win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, 2, width-1, ACS_RTEE);

	set_menu_back(ZonesSection.h_menu, COLOR_PAIR(5));
	set_menu_fore(ZonesSection.h_menu, COLOR_PAIR(3));

	post_menu(ZonesSection.h_menu);
	doupdate();

	mvwaddch(ZonesSection.h_win, height-5, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, height-5, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, height-5, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.h_win, height-4, 1, "Usage: <RET> to edit the group");
	mvwprintw(ZonesSection.h_win, height-3, 1, "       <INS> to insert a new grp");
	mvwprintw(ZonesSection.h_win, height-2, 1, "       <DEL> to removes the grp");

	return(retval);
}

static int
zones_section_menu_groups_destroy(const int debuglvl)
{
	int	retval = 0,
		i;

	unpost_menu(ZonesSection.h_menu);
	free_menu(ZonesSection.h_menu);
	for(i = 0; i < ZonesSection.host_n; ++i)
		free_item(ZonesSection.hostitems[i]);

	d_list_cleanup(debuglvl, &ZonesSection.desc_list);

	free(ZonesSection.hostitems);

	del_panel(ZonesSection.h_panel[0]);

	destroy_win(ZonesSection.h_win);

	return(retval);
}

int
zones_section_menu_groups(const int debuglvl, char *zonename, char *networkname, struct rgx_ *reg)
{
	int			ch = 0,
				quit = 0,
				reload = 0,
				retval = 0,
				size = 0;
	struct ZoneData_	*zone_ptr = NULL;
	char			*new_zone_ptr = NULL,
				*temp_ptr = NULL;
	ITEM			*cur = NULL;

	/* safety */
	if(!zonename || !networkname || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* setup */
	if(zones_section_menu_groups_init(debuglvl, zonename, networkname) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up groups menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	while(quit == 0)
	{
		if(reload == 1)
		{
			/* first destroy */
			if(zones_section_menu_groups_destroy(debuglvl) < 0)
			{
				(void)vrprint.error(-1, "Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			/* and setup again */
			if(zones_section_menu_groups_init(debuglvl, zonename, networkname) < 0)
			{
				(void)vrprint.error(-1, "Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			/* we are done with reloading */
			reload = 0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ZonesSection.h_win);
			switch(ch)
			{
				case KEY_LEFT:
				case KEY_F(10):
				case 'q':
				case 27:
				
					quit = 1;
					break;

				case KEY_IC: //insert
				
					new_zone_ptr = input_box(MAX_HOST, "New Group", "Enter the name of the new group");
					if(new_zone_ptr != NULL)
					{
						size = strlen(new_zone_ptr)+1+strlen(networkname)+1+strlen(zonename)+1;

						temp_ptr = malloc(size);
						if(temp_ptr != NULL)
						{
							strlcpy(temp_ptr, new_zone_ptr, size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, networkname, size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, zonename, size);

							if(validate_zonename(debug, temp_ptr, 1, NULL, NULL, NULL, reg->zonename) == 0)
							{
								if(new_zone(debug, temp_ptr, TYPE_GROUP, &ZonedataList) >= 0)
								{
									(void)edit_zone_group(debuglvl, temp_ptr);
								}
								else
								{
									(void)vrprint.error(-1, "Error", "failed to create new group.");
								}
								
								reload = 1;
							}
							else
							{
								(void)vrprint.error(-1, "Error", "groupname '%s' is invalid.", temp_ptr);
							}
						}
						else
						{
							(void)vrprint.error(-1, "Internal Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
						}

						free(new_zone_ptr);
						free(temp_ptr);
					}
					break;

				case KEY_DC: // delete

					cur = current_item(ZonesSection.h_menu);
					if(cur)
					{
						if(confirm("Delete", "This group?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{
							size = strlen((char *)item_name(cur))+1+strlen(networkname)+1+strlen(zonename)+1;
					
							if(!(temp_ptr = malloc(size)))
							{
								(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
								return(-1);
							}
							
							strlcpy(temp_ptr, (char *)item_name(cur), size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, networkname, size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, zonename, size);

							/* search the zone */
							if(!(zone_ptr = search_zonedata(debuglvl, temp_ptr, &ZonedataList)))
							{
								(void)vrprint.error(-1, "Internal Error", "couldn't find the group '%s' in memory.", temp_ptr);
								retval = -1;
							}

							if(retval == 0)
							{
								/* delete, the memory is freed by delete_zone(). */
								if(delete_zone(debug, temp_ptr, zone_ptr->type, &ZonedataList) < 0)
								{
									(void)vrprint.error(-1, "Error", "deleting group failed.");
								}
								else
								{
									reload = 1;
								}
							}
						
							free(temp_ptr);
						}
					}

					break;

				case KEY_DOWN:
					menu_driver(ZonesSection.h_menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.h_menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.h_menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.h_menu, REQ_LAST_ITEM); // page down
					break;

				case 32: // space
				case KEY_RIGHT:
				case 10: // enter
				{
					cur = current_item(ZonesSection.h_menu);
					if(cur)
					{
						size = strlen((char *)item_name(cur))+1+strlen(networkname)+1+strlen(zonename)+1;

						if(!(temp_ptr = malloc(size)))
						{
							(void)vrprint.error(-1, "Error", "malloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
							return(-1);
						}

						strlcpy(temp_ptr, (char *)item_name(cur), size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, networkname, size);
						strlcat(temp_ptr, ".", size);
						strlcat(temp_ptr, zonename, size);

						zone_ptr = search_zonedata(debuglvl, temp_ptr, &ZonedataList);
						if(zone_ptr != NULL)
						{
							draw_top_menu_editzones_group(top_win);
							(void)edit_zone_group(debuglvl, zone_ptr->name);
							draw_top_menu_main(debuglvl, top_win);
							
							reload = 1;
						}
						else
						{
							(void)vrprint.error(-1, "Internal Error", "group '%s' not found in memory.", temp_ptr);
						}
					
						free(temp_ptr);
					}
					break;
				}

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:GROUPS]:");
					break;
			}
		}
	}
	
	if(zones_section_menu_groups_destroy(debuglvl) < 0)
		retval = -1;

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_zone_network_interfaces_deliface(const int debuglvl, const char *iface_name, struct ZoneData_ *zone_ptr)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;

	/* safety */
	if(!iface_name || !zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* safety: we expect a network */
	if(zone_ptr->type != TYPE_NETWORK)
	{
		(void)vrprint.error(-1, "Internal Error", "expected a NETWORK (%d), but got a %d! (in: %s)", TYPE_NETWORK, zone_ptr->type, __FUNCTION__);
		return(-1);
	}

	/* search the interface, we start searching at the top of the list */
	for(d_node = zone_ptr->InterfaceList.top; d_node; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* check if this is the one */
		if(strcmp(iface_name, iface_ptr->name) == 0)
		{
			if(d_list_remove_node(debuglvl, &zone_ptr->InterfaceList, d_node) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "Unable to remove interface from the list (in: edit_zone_network_interfaces_deliface).");
				return(-1);
			}
			else
			{
				zone_ptr->interface_count--;
				break;
			}
		}
	}
	
	/* save the new interface list */
	if(zones_network_save_interfaces(debuglvl, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving the new interfaceslist to the backend failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	return(0);
}

static int
edit_zone_network_interfaces_newiface(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	d_list_node		*d_node = NULL;
	char			**choices,
				*choice_ptr = NULL;
	int			n_choices = 0,
				i = 0;
	struct InterfaceData_	*new_iface_ptr = NULL,
				*iface_ptr = NULL;

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check if there are interfaces defined to choose from */
	if((n_choices = InterfacesList.len) == 0)
	{
		(void)vrprint.warning("Warning", "no interfaces found. Please define an interface first.");
		return(0);
	}

	/* get some mem */
	if(!(choices = calloc(n_choices + 1, MAX_INTERFACE)))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s).", strerror(errno), __FUNCTION__);
		return(-1);
	}

	/* load the interfaces */
	for(i = n_choices-1, d_node = InterfacesList.top; d_node ; d_node = d_node->next)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			
			free(choices);
			return(-1);
		}
		
		
		choices[i] = iface_ptr->name;
		i--;
	}

	/* ask the user to select an interface */
	if(!(choice_ptr = selectbox("New interface", "Select a interface", n_choices, choices, 1)))
	{
		/* no choice */
		free(choices);
		return(0);
	}

	/* cleanup */
	free(choices);

	/* search the interface in the list */
	if(!(new_iface_ptr = search_interface(debuglvl, choice_ptr, &InterfacesList)))
	{
		(void)vrprint.error(-1, "Internal Error", "interface is invalid, it was not found in memory.");
		
		free(choice_ptr);
		return(-1);
	}
	
	/* cleanup */
	free(choice_ptr);
		
	/* now append */
	if(d_list_append(debuglvl, &zone_ptr->InterfaceList, new_iface_ptr) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "unable to append interface to interfaceslist (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	zone_ptr->interface_count++;

	/* save the new interface list */
	if(zones_network_save_interfaces(debuglvl, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving the new interfaceslist to the backend failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	return(0);
}

static int
edit_zone_network_interfaces_init(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	d_list_node		*d_node = NULL;
	struct InterfaceData_	*iface_ptr = NULL;
	int			i=0;
	int			height = 30,
				width  = 34, // max width of interface (32) + box (2)
				startx = 5,
				starty = 5,
				max_height = 0,
				max_width = 0;

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(zone_ptr->type != TYPE_NETWORK)
	{
		vrprint.error(-1, "Internal Error", "Expected a TYPE_NETWORK (%d), but got a %d (in: %s).", TYPE_NETWORK, zone_ptr->type, __FUNCTION__);
		return(-1);
	}
	if((ZonesSection.EditZoneInt.n_items = zone_ptr->interface_count) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "interface count is negative (in: %s)", __FUNCTION__);
		return(-1);
	}

	if(!(ZonesSection.EditZoneInt.items = (ITEM **)calloc(ZonesSection.EditZoneInt.n_items + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s.", strerror(errno));
		return(-1);
	}

	getmaxyx(stdscr, max_height, max_width);
	
	height = ZonesSection.EditZoneInt.n_items + 7; // 7 because: 3 above the list, 4 below
	if(height >= max_height - starty)
	{
		height = max_height - 6;
		starty = 3;
	}

	for(i = 0, d_node = zone_ptr->InterfaceList.top; d_node; d_node = d_node->next, i++)
	{
		if(!(iface_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		/* load all interfaces into memory */
		if(!(ZonesSection.EditZoneInt.items[i] = new_item(iface_ptr->name, NULL)))
		{
			(void)vrprint.error(-1, "Error", "adding new item failed.");
			return(-1);
		}
	}
	ZonesSection.EditZoneInt.items[ZonesSection.EditZoneInt.n_items] = (ITEM *)NULL;

	/* create the window */
	if(!(ZonesSection.EditZoneInt.win = newwin(height, width, starty, startx)))
	{
		(void)vrprint.error(-1, "Error", "creating new window failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZoneInt.panel[0] = new_panel(ZonesSection.EditZoneInt.win)))
	{
		(void)vrprint.error(-1, "Error", "creating new panel failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.EditZoneInt.win, COLOR_PAIR(CP_BLUE_WHITE));
	keypad(ZonesSection.EditZoneInt.win, TRUE);
	wrefresh(ZonesSection.EditZoneInt.win);

	if(!(ZonesSection.EditZoneInt.menu = new_menu((ITEM **)ZonesSection.EditZoneInt.items)))
	{
		(void)vrprint.error(-1, "Error", "creating new menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	set_menu_win(ZonesSection.EditZoneInt.menu, ZonesSection.EditZoneInt.win);
	set_menu_sub(ZonesSection.EditZoneInt.menu, derwin(ZonesSection.EditZoneInt.win, height-6, width-2, 3, 1));
	set_menu_format(ZonesSection.EditZoneInt.menu, height-7, 1);

	box(ZonesSection.EditZoneInt.win, 0, 0);
	print_in_middle(ZonesSection.EditZoneInt.win, 1, 0, width, "Interfaces", COLOR_PAIR(CP_BLUE_WHITE));
	mvwaddch(ZonesSection.EditZoneInt.win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.EditZoneInt.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.EditZoneInt.win, 2, width - 1, ACS_RTEE);
	set_menu_back(ZonesSection.EditZoneInt.menu, COLOR_PAIR(CP_BLUE_WHITE));
	set_menu_fore(ZonesSection.EditZoneInt.menu, COLOR_PAIR(CP_WHITE_BLUE));
	post_menu(ZonesSection.EditZoneInt.menu);

	mvwaddch(ZonesSection.EditZoneInt.win, height-4, 0, ACS_LTEE);
	mvwhline(ZonesSection.EditZoneInt.win, height-4, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.EditZoneInt.win, height-4, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.EditZoneInt.win, height-3, 2, "<INS> to insert an interface");
	mvwprintw(ZonesSection.EditZoneInt.win, height-2, 2, "<DEL> to remove an interface");

	update_panels();
	doupdate();

	return(0);
}

static int
edit_zone_network_interfaces_destroy(void)
{
	int	retval = 0,
		i;

	/* un post form and free the memory */
	unpost_menu(ZonesSection.EditZoneInt.menu);
	free_menu(ZonesSection.EditZoneInt.menu);

	for(i=0;i<ZonesSection.EditZoneInt.n_items;i++)
	{
		free_item(ZonesSection.EditZoneInt.items[i]);
	}
	free(ZonesSection.EditZoneInt.items);

	del_panel(ZonesSection.EditZoneInt.panel[0]);
	destroy_win(ZonesSection.EditZoneInt.win);
	return(retval);
}

int
edit_zone_network_interfaces(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	int	quit = 0,
		reload = 0,
		result = 0,
		ch,
		retval = 0;
	ITEM	*cur = NULL;

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	if(edit_zone_network_interfaces_init(debuglvl, zone_ptr) < 0)
		return(-1);

	while(quit == 0)
	{
		if(reload == 1)
		{
			if(edit_zone_network_interfaces_destroy() < 0)
				return(-1);

			if(edit_zone_network_interfaces_init(debuglvl, zone_ptr) < 0)
				return(-1);

			/* refresh screen */
			update_panels();
			doupdate();
			
			reload = 0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ZonesSection.EditZoneInt.win);
			
			switch(ch)
			{
				case KEY_LEFT:
				case 'q':
				case KEY_F(10): //quit
				
					quit = 1;
					break;

				case KEY_IC:

					if(edit_zone_network_interfaces_newiface(debuglvl, zone_ptr) < 0)
					{
						retval = -1;
						quit = 1;
					}
					
					reload = 1;
					break;
					
				case KEY_DC:
				{
					cur = current_item(ZonesSection.EditZoneInt.menu);
					if(cur)
					{
						if(edit_zone_network_interfaces_deliface(debuglvl, (char *)item_name(cur), zone_ptr) < 0)
						{
							retval = -1;
							quit = 1;
						}

						reload = 1;
					}
					break;
				}
				
				case KEY_DOWN:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.EditZoneInt.menu, REQ_LAST_ITEM); // page down
					break;
			}
		}
	}

	result = edit_zone_network_interfaces_destroy();
	if(result < 0)
		return(-1);

	return(retval);
}

int
edit_zone_network_init_togglefield(const int debuglvl, char *action, char *who, char *danger, char *source, FIELD *field)
{
	struct RuleData_	*query_ptr = NULL,
				*old_query_ptr = NULL;
	int retval=0;

	if(!(query_ptr = rule_malloc()))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(-1);
	}
	old_query_ptr = query_ptr;

	strcpy(query_ptr->action, action);
	strcpy(query_ptr->who, who);
	strcpy(query_ptr->danger, danger);
	strcpy(query_ptr->source, source);

	if(!(search_rule(debuglvl, &RulesList, old_query_ptr)))
	{
		set_field_buffer(field, 0, " ");
		free(query_ptr);
		query_ptr = NULL;
		old_query_ptr = NULL;
	}
	else
	{
		set_field_buffer(field, 0, "X");
		free(old_query_ptr);
		query_ptr = NULL;
		old_query_ptr = NULL;
	}

	return(retval);
}


struct
{
	FIELD	*activefld,
		*activelabelfld,
		
		*networkfld,
		*networklabelfld,
		
		*netmaskfld,
		*netmasklabelfld,
		
		*commentfld,
		*commentlabelfld,


		/* anti spoofing */
		*loopbackfld,
		*loopbacklabelfld,
		*loopbackbracketsfld,

		*classafld,
		*classalabelfld,
		*classabracketsfld,

		*classbfld,
		*classblabelfld,
		*classbbracketsfld,

		*classcfld,
		*classclabelfld,
		*classcbracketsfld,

		*classdfld,
		*classdlabelfld,
		*classdbracketsfld,

		*classefld,
		*classelabelfld,
		*classebracketsfld,

		*testnetfld,
		*testnetlabelfld,
		*testnetbracketsfld;

} NetworkSec;


/*	interfaces_save_protectrules

	Save the protectrules to the backend.
	
	Returncodes:
		 0: ok
		-1: error
*/
int
zones_network_save_protectrules(const int debuglvl, struct ZoneData_ *network_ptr)
{
	d_list_node		*d_node = NULL;
	struct RuleData_	*rule_ptr = NULL;
	char			rule_str[MAX_RULE_LENGTH] = "";
	
	/* safety */
	if(!network_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* safety */
	if(network_ptr->ProtectList.len < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "negative number of protectrules (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* write to backend */
	if(network_ptr->ProtectList.len == 0)
	{
		/* clear */
		if(zf->tell(debuglvl, zone_backend, network_ptr->name, "RULE", "", 1, TYPE_NETWORK) < 0)
		{
			(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}
	else
	{
		/* write to backend */
		for(d_node = network_ptr->ProtectList.top; d_node; d_node = d_node->next)
		{
			if(!(rule_ptr = d_node->data))
			{
				(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
				return(-1);
			}

			snprintf(rule_str, sizeof(rule_str), "protect against %s from %s", rule_ptr->danger, rule_ptr->source);

			/* the first one needs to be in overwrite mode */
			if(d_node == network_ptr->ProtectList.top)
			{
				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, network_ptr->name, "RULE", rule_str, 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
			else
			{
				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, network_ptr->name, "RULE", rule_str, 0, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					return(-1);
				}
			}
		}
	}
	
	return(0);
}


static int
edit_zone_network_save_protectrules(const int debuglvl, struct ZoneData_ *network_ptr)
{
	struct RuleData_	*rule_ptr = NULL;

	/* safety */
	if(!network_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);	
	}

	/* cleanup the existing list */
	d_list_cleanup(debuglvl, &network_ptr->ProtectList);

	if(field_buffer(NetworkSec.loopbackfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "loopback")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.classafld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "class-a")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.classbfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "class-b")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.classcfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "class-c")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.classdfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "class-d")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.classefld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "class-e")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}

	if(field_buffer(NetworkSec.testnetfld, 0)[0] == 'X')
	{
		if(!(rule_ptr = create_protect_rule(debuglvl, network_ptr->name, "spoofing", "test-net")))
		{
			(void)vrprint.error(-1, "Internal Error", "creating protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
		if(d_list_append(debuglvl, &network_ptr->ProtectList, rule_ptr) < 0)
		{
			(void)vrprint.error(-1, "Internal Error", "appending protect rule failed (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}
	}


	/* now let try to write this to the backend */
	if(zones_network_save_protectrules(debuglvl, network_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	return(0);
}


static int
edit_zone_network_init(const int debuglvl, char *name, int height, int width, int startx, int starty, struct ZoneData_ *zone_ptr)
{
	int	i = 0,
		rows,
		cols,
		max_height,
		max_width,
		comment_y=0,
		comment_x=0;
	int	field_num = 0;

	/* safety */
	if(!name || !zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* get screen size */
	getmaxyx(stdscr, max_height, max_width);

	ZonesSection.EditZone.n_fields = 29;
	if(!(ZonesSection.EditZone.fields = (FIELD **)calloc(ZonesSection.EditZone.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "could not setup fields (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* active toggle */
	NetworkSec.activelabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 6, 2, 0, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.activelabelfld, 0, "Active");
	field_opts_off(NetworkSec.activelabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.activefld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 3, 1, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.activefld, 0, zone_ptr->active ? "Yes" : "No");

	/* network */
	NetworkSec.networklabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 5, 0, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.networklabelfld, 0, "Network");
	field_opts_off(NetworkSec.networklabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.networkfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 16, 6, 1, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.networkfld, 0, zone_ptr->ipv4.network);

	/* network */
	NetworkSec.netmasklabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 7, 0, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.netmasklabelfld, 0, "Network");
	field_opts_off(NetworkSec.netmasklabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.netmaskfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 16, 8, 1, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.netmaskfld, 0, zone_ptr->ipv4.netmask);


	/* anti-spoof loopback */
	NetworkSec.loopbacklabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 4, 20, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.loopbacklabelfld, 0, "Loopback");
	field_opts_off(NetworkSec.loopbacklabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.loopbackbracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 4, 32, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.loopbackbracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.loopbackbracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.loopbackfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 4, 33, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.loopbackfld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "loopback") ? "X" : " ");

	/* anti-spoof class-a */
	NetworkSec.classalabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 5, 20, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classalabelfld, 0, "Class A");
	field_opts_off(NetworkSec.classalabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classabracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 5, 32, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classabracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.classabracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classafld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 5, 33, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classafld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "class-a") ? "X" : " ");

	/* anti-spoof class-b */
	NetworkSec.classblabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 6, 20, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classblabelfld, 0, "Class B");
	field_opts_off(NetworkSec.classblabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classbbracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 6, 32, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classbbracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.classbbracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classbfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 6, 33, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classbfld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "class-b") ? "X" : " ");

	/* anti-spoof class-c */
	NetworkSec.classclabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 7, 20, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classclabelfld, 0, "Class C");
	field_opts_off(NetworkSec.classclabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classcbracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 7, 32, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classcbracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.classcbracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classcfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 7, 33, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classcfld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "class-c") ? "X" : " ");

	/* anti-spoof class-d */
	NetworkSec.classdlabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 4, 40, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classdlabelfld, 0, "Class D");
	field_opts_off(NetworkSec.classdlabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classdbracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 4, 52, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classdbracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.classdbracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classdfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 4, 53, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classdfld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "class-d") ? "X" : " ");

	/* anti-spoof class-e */
	NetworkSec.classelabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 5, 40, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classelabelfld, 0, "Class E");
	field_opts_off(NetworkSec.classelabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classebracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 5, 52, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classebracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.classebracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.classefld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 5, 53, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.classefld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "class-e") ? "X" : " ");
	
	/* anti-spoof testnet */
	NetworkSec.testnetlabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 6, 40, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.testnetlabelfld, 0, "Test-net");
	field_opts_off(NetworkSec.testnetlabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.testnetbracketsfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 3, 6, 52, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.testnetbracketsfld, 0, "[ ]");
	field_opts_off(NetworkSec.testnetbracketsfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.testnetfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 1, 6, 53, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.testnetfld, 0, protectrule_loaded(debuglvl, &zone_ptr->ProtectList, "spoofing", "test-net") ? "X" : " ");


	/* comment field */
	comment_y = 5;
	comment_x = 48;

	/* comment */
	NetworkSec.commentlabelfld = (ZonesSection.EditZone.fields[field_num] = new_field(1, 8, 10, 0, 0, 0));
	field_num++;
	set_field_buffer(NetworkSec.commentlabelfld, 0, "Comment");
	field_opts_off(NetworkSec.commentlabelfld, O_AUTOSKIP | O_ACTIVE);

	NetworkSec.commentfld = (ZonesSection.EditZone.fields[field_num] = new_field(comment_y, comment_x, 11, 1, 0, 0));
	field_num++;

	/* terminate */
	ZonesSection.EditZone.fields[field_num] = NULL;

	if(ZonesSection.EditZone.n_fields != field_num)
		(void)vrprint.error(-1, "Internal Error", "ZonesSection.EditZone.n_fields != field_num.");

	/* read the comment from backend */
	if(zf->ask(debuglvl, zone_backend, zone_ptr->name, "COMMENT", ZonesSection.comment, sizeof(ZonesSection.comment), TYPE_NETWORK, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	set_field_buffer(NetworkSec.commentfld, 0, ZonesSection.comment);

	/* create window and panel */
	if(!(ZonesSection.EditZone.win = create_newwin(height, width, startx, starty, "Edit Zone: Network", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZone.panel[0] = new_panel(ZonesSection.EditZone.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	keypad(ZonesSection.EditZone.win, TRUE);

	/* set field options */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		set_field_back(ZonesSection.EditZone.fields[i], COLOR_PAIR(CP_BLUE_WHITE));
		field_opts_off(ZonesSection.EditZone.fields[i], O_AUTOSKIP);
		set_field_status(ZonesSection.EditZone.fields[i], FALSE);
	}

	set_field_back(NetworkSec.activefld, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_back(NetworkSec.networkfld, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_back(NetworkSec.netmaskfld, COLOR_PAIR(CP_WHITE_BLUE));
	set_field_back(NetworkSec.commentfld, COLOR_PAIR(CP_WHITE_BLUE));

	/* Create the form and post it */
	if(!(ZonesSection.EditZone.form = new_form(ZonesSection.EditZone.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form (in: %s).", __FUNCTION__);
		return(-1);
	}
	scale_form(ZonesSection.EditZone.form, &rows, &cols);
	set_form_win(ZonesSection.EditZone.form, ZonesSection.EditZone.win);
	set_form_sub(ZonesSection.EditZone.form, derwin(ZonesSection.EditZone.win, rows, cols, 1, 2));
	post_form(ZonesSection.EditZone.form);

	/* the labels */
	mvwprintw(ZonesSection.EditZone.win, 1, 2, "Name: %s", zone_ptr->name);
	mvwprintw(ZonesSection.EditZone.win, 3, 30, "Anti-spoofing");

	mvwprintw(ZonesSection.EditZone.win, 12, 54, "Press <F6> to assign");
	mvwprintw(ZonesSection.EditZone.win, 13, 54, "(an) interface(s) to");
	mvwprintw(ZonesSection.EditZone.win, 14, 54, "this network.");

	mvwprintw(ZonesSection.EditZone.win, 5, 62, "Hosts  : %4d", count_zones(debug, &ZonedataList, TYPE_HOST, zone_ptr->network_name, zone_ptr->zone_name));
	mvwprintw(ZonesSection.EditZone.win, 6, 62, "Groups : %4d", count_zones(debug, &ZonedataList, TYPE_GROUP, zone_ptr->network_name, zone_ptr->zone_name));

	wrefresh(ZonesSection.EditZone.win);
	return(0);
}

/*	edit_zone_network_save

	
	Returncodes:
		0: ok
		-1: error
*/
static int
edit_zone_network_save(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	int			i;

	/* check for changed fields */
	for(i=0; i < ZonesSection.EditZone.n_fields; i++)
	{
		if(field_status(ZonesSection.EditZone.fields[i]) == TRUE)
		{
			if(ZonesSection.EditZone.fields[i] == NetworkSec.activefld)
			{
				/* active */
				zone_ptr->status = ST_CHANGED;

				if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[i], 0), "yes", 3) == 0)
				{
					zone_ptr->active = 1;
				}
				else
				{
					zone_ptr->active = 0;
				}

				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "ACTIVE", zone_ptr->active ? "Yes" : "No", 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed.");
					return(-1);
				}
			}
			else if(ZonesSection.EditZone.fields[i] == NetworkSec.networkfld)
			{
				/* network */
				zone_ptr->status = ST_CHANGED;

				if(!(copy_field2buf(	zone_ptr->ipv4.network,
							field_buffer(ZonesSection.EditZone.fields[i], 0),
							sizeof(zone_ptr->ipv4.network))))
					return(-1);

				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "NETWORK", zone_ptr->ipv4.network, 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed.");
					return(-1);
				}
			}
			else if(ZonesSection.EditZone.fields[i] == NetworkSec.netmaskfld)
			{
				/* netmask */
				zone_ptr->status = ST_CHANGED;

				if(!(copy_field2buf(	zone_ptr->ipv4.netmask,
							field_buffer(ZonesSection.EditZone.fields[i], 0),
							sizeof(zone_ptr->ipv4.netmask))))
					return(-1);

				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "NETMASK", zone_ptr->ipv4.netmask, 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed.");
					return(-1);
				}
			}
			/* save the comment to the backend */
			else if(ZonesSection.EditZone.fields[i] == NetworkSec.commentfld)
			{
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "COMMENT", field_buffer(ZonesSection.EditZone.fields[i], 0), 1, TYPE_NETWORK) < 0)
				{
					(void)vrprint.error(-1, "Error", "writing to backend failed.");
					return(-1);
				}
			}
			else if(ZonesSection.EditZone.fields[i] == NetworkSec.loopbackfld	||
				ZonesSection.EditZone.fields[i] == NetworkSec.classafld		||
				ZonesSection.EditZone.fields[i] == NetworkSec.classbfld		||
				ZonesSection.EditZone.fields[i] == NetworkSec.classcfld		||
				ZonesSection.EditZone.fields[i] == NetworkSec.classdfld		||
				ZonesSection.EditZone.fields[i] == NetworkSec.classefld		||
				ZonesSection.EditZone.fields[i] == NetworkSec.testnetfld)
			{
				if(edit_zone_network_save_protectrules(debuglvl, zone_ptr) < 0)
				{
					return(-1);
				}
			}
		}
	}

	return(0);
}

static int
edit_zone_network_destroy(void)
{
	int	retval=0,
		i;

	/* unpost form and free the memory */
	unpost_form(ZonesSection.EditZone.form);
	free_form(ZonesSection.EditZone.form);

	for(i=0;i<ZonesSection.EditZone.n_fields;i++)
	{
		free_field(ZonesSection.EditZone.fields[i]);
	}
	free(ZonesSection.EditZone.fields);

	del_panel(ZonesSection.EditZone.panel[0]);
	destroy_win(ZonesSection.EditZone.win);

	/* reset the comment field */
	strcpy(ZonesSection.comment, "");
	return(retval);
}


/*	edit_zone_network

	The TmpZone crap here is the fault of edit_zone_network_save.
	See the comment above that function for more horror...
	
	Returncodes:
		0: ok
		-1: error
*/
static int
edit_zone_network(const int debuglvl, char *name)
{
	int			ch,
				not_defined = 0,
				quit = 0,
				retval = 0;
	struct ZoneData_	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				max_height,
				max_width;
	FIELD			*cur = NULL,
				*prev = NULL;

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* get current screen size */
	getmaxyx(stdscr, max_height, max_width);

	height = 21;
	if(height > max_height - 6)
		height = max_height-6;

	width = 78;
	startx = 1;
	starty = 3;

	if(!(zone_ptr = search_zonedata(debuglvl, name, &ZonedataList)))
	{
		(void)vrprint.error(-1, "Error", "network '%s' not found in memory.", name);
		return(-1);
	}

	if(edit_zone_network_init(debuglvl, name, height, width, starty, startx, zone_ptr) < 0)
		return(-1);

	/* print warning if no interfaces have been assigned to this network */
	if(zone_ptr->InterfaceList.len == 0)
	{
		wattron(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		if(height == 18 || height == 19) /* default xterm height && linux console */
		{
			mvwprintw(ZonesSection.EditZone.win, 15, 53, " Warning:       ");
			mvwprintw(ZonesSection.EditZone.win, 16, 53, " no interfaces! ");
		}
		else
		{
			mvwprintw(ZonesSection.EditZone.win, 18, 26,  " Warning: no interfaces! ");
		}
		wattroff(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
	}

	wrefresh(ZonesSection.EditZone.win);
	cur = current_field(ZonesSection.EditZone.form);
	pos_form_cursor(ZonesSection.EditZone.form);

	update_panels();
	doupdate();

	/* loop through to get user requests */
	while(quit == 0)
	{
		/* get current field */
		cur = current_field(ZonesSection.EditZone.form);

		/* draw the arrow around the active field */
		draw_field_active_mark(cur, prev, ZonesSection.EditZone.win, ZonesSection.EditZone.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		not_defined = 0;

		/* get the input */
		ch = wgetch(ZonesSection.EditZone.win);

		if(cur == NetworkSec.commentfld)
		{
			if(nav_field_comment(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == NetworkSec.activefld)
		{
			if(nav_field_yesno(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else if(cur == NetworkSec.networkfld ||
			cur == NetworkSec.netmaskfld)
		{
			if(nav_field_simpletext(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		/* this one needs to be last */
		else if(cur == NetworkSec.loopbackfld	||
			cur == NetworkSec.classafld	||
			cur == NetworkSec.classbfld	||
			cur == NetworkSec.classcfld	||
			cur == NetworkSec.classdfld	||
			cur == NetworkSec.classefld	||
			cur == NetworkSec.testnetfld)
		{
			if(nav_field_toggleX(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else
			not_defined = 1;

		/* the rest is handled here */
		if(not_defined == 1)
		{
			switch(ch)
			{
				case KEY_F(6):

					if(edit_zone_network_interfaces(debuglvl, zone_ptr) < 0)
						retval = -1;
					
					break;

				case KEY_F(5): //help

					status_print(status_win, "FIXME: help not yet implemented");
					break;
				
				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
				
					form_driver(ZonesSection.EditZone.form, REQ_NEXT_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;
				
				case KEY_UP:
				
					form_driver(ZonesSection.EditZone.form, REQ_PREV_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;
				
				case 27:
				case KEY_F(10):
				
					quit = 1;
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:NETWORK:EDIT]:");
					break;
			}
		}

		prev = cur;
		cur = current_field(ZonesSection.EditZone.form);

		/* print warning if no interfaces have been assigned to this network */
		if(zone_ptr->InterfaceList.len == 0)
		{
			wattron(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
			if(height == 18 || height == 19) /* default xterm height && linux console */
			{
				mvwprintw(ZonesSection.EditZone.win, 15, 53, " Warning:       ");
				mvwprintw(ZonesSection.EditZone.win, 16, 53, " no interfaces! ");
			}
			else
			{
				mvwprintw(ZonesSection.EditZone.win, 18, 26,  " Warning: no interfaces! ");
			}
			wattroff(ZonesSection.EditZone.win, COLOR_PAIR(CP_YELLOW_RED) | A_BOLD);
		}
		/* and clear it again */
		else
		{
			if(height == 18 || height == 19) /* default xterm height && linux console */
			{
				mvwprintw(ZonesSection.EditZone.win, 15, 53, "                ");
				mvwprintw(ZonesSection.EditZone.win, 16, 53, "                ");
			}
			else
			{
				mvwprintw(ZonesSection.EditZone.win, 18, 26,  "                         ");
			}
		}

		/* draw and set cursor */
		wrefresh(ZonesSection.EditZone.win);
		pos_form_cursor(ZonesSection.EditZone.form);
	}

	/* save */
	if(edit_zone_network_save(debuglvl, zone_ptr) < 0)
	{
		(void)vrprint.error(-1, "Error", "saving changes in the network failed (in: %s).", __FUNCTION__);
		retval = -1;
	}

	/* cleanup */
	if(edit_zone_network_destroy() < 0)
		retval = -1;
		
	update_panels();
	doupdate();

	return(retval);
}


/*

*/
static int
zones_section_menu_networks_init(const int debuglvl, char *zonename)
{
	int			retval=0, i;
	struct ZoneData_	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				maxx,
				maxy;
	d_list_node		*d_node = NULL;

	/* safety */
	if(!zonename)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* get screen dimentions */
	getmaxyx(stdscr, maxx, maxy);

	/* count how many networks there are */
	ZonesSection.network_n = 0;
	
	/* count the networks */
	for(d_node = ZonedataList.top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}
		
		if(zone_ptr->type == TYPE_NETWORK && strcmp(zone_ptr->zone_name, zonename)== 0)
			ZonesSection.network_n++;
	}

	i = ZonesSection.network_n - 1;
	
	if(!(ZonesSection.networkitems = (ITEM **)calloc(ZonesSection.network_n + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "could not setup items (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* now load the items */
	for(d_node = ZonedataList.bot; d_node ; d_node = d_node->prev)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}
		
		if(zone_ptr->type == TYPE_NETWORK  && strcmp(zone_ptr->zone_name, zonename)== 0)
		{
			if(!(ZonesSection.networkitems[i] = new_item(zone_ptr->network_name, NULL)))
			{
				(void)vrprint.error(-1, "Error", "could not load item (in: %s).", __FUNCTION__);
				return(-1);
			}
			
			i--;
		}
	}
	ZonesSection.networkitems[ZonesSection.network_n] = (ITEM *)NULL;
	
	if(!(ZonesSection.n_menu = new_menu((ITEM **)ZonesSection.networkitems)))
	{
		(void)vrprint.error(-1, "Error", "could not setup menu (in: %s).", __FUNCTION__);
		return(-1);
	}

	height = ZonesSection.network_n + 10;
	width = MAX_NETWORK + 2;
	startx = 6;
	starty = 2;

	if(maxy < starty + height + 3)
	{
		starty = 3;
		height = maxy - 2 * starty;
	}

	if(!(ZonesSection.n_win = newwin(height, width, startx, starty)))
	{
		(void)vrprint.error(-1, "Error", "could not setup window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.n_panel[0] = new_panel(ZonesSection.n_win)))
	{
		(void)vrprint.error(-1, "Error", "could not setup panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.n_win, COLOR_PAIR(CP_BLUE_WHITE));
	keypad(ZonesSection.n_win, TRUE);

	set_menu_win(ZonesSection.n_menu, ZonesSection.n_win);
	set_menu_sub(ZonesSection.n_menu, derwin(ZonesSection.n_win, height-8, width-2, 3, 1));
	set_menu_format(ZonesSection.n_menu, height-9, 1);

	box(ZonesSection.n_win, 0, 0);
	print_in_middle(ZonesSection.n_win, 1, 0, width, "Networks", COLOR_PAIR(CP_BLUE_WHITE));
	mvwaddch(ZonesSection.n_win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.n_win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.n_win, 2, width-1, ACS_RTEE);

	set_menu_back(ZonesSection.n_menu, COLOR_PAIR(CP_BLUE_WHITE));
	set_menu_fore(ZonesSection.n_menu, COLOR_PAIR(CP_WHITE_BLUE));
	post_menu(ZonesSection.n_menu);
	
	update_panels();
	doupdate();

	mvwaddch(ZonesSection.n_win, height-6, 0, ACS_LTEE);
	mvwhline(ZonesSection.n_win, height-6, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.n_win, height-6, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.n_win, height-5, 2, "<RET> to enter the network");
	mvwprintw(ZonesSection.n_win, height-4, 2, "<INS> to insert a new network");
	mvwprintw(ZonesSection.n_win, height-3, 2, "<DEL> to removes the network");
	mvwprintw(ZonesSection.n_win, height-2, 2, "< e > to edit the network");

	return(retval);
}

static int
zones_section_menu_networks_destroy(void)
{
	int	retval = 0,
		i;

	unpost_menu(ZonesSection.n_menu);
	free_menu(ZonesSection.n_menu);

	for(i = 0; i < ZonesSection.network_n; ++i)
		free_item(ZonesSection.networkitems[i]);

	free(ZonesSection.networkitems);

	del_panel(ZonesSection.n_panel[0]);
	destroy_win(ZonesSection.n_win);

	return(retval);
}

int
zones_section_menu_networks(const int debuglvl, char *zonename, struct rgx_ *reg)
{
	int		ch,
			quit = 0,
			reload = 0,
			result = 0,
			retval = 0;
	unsigned int	size = 0;
	char		*new_zone_ptr = NULL,
			*zonename_ptr = NULL,
			*choices[] = { "Hosts", "Groups", "Network" },
			*temp_ptr = NULL,
			*choice_ptr = NULL;
	ITEM		*cur = NULL;

	/* safety */
	if(!zonename || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(zones_section_menu_networks_init(debuglvl, zonename) < 0)
		return(-1);

	while(quit == 0)
	{
		if(reload == 1)
		{
			if(zones_section_menu_networks_destroy() < 0)
				return(-1);
			
			if(zones_section_menu_networks_init(debuglvl, zonename) < 0)
				return(-1);
			
			reload = 0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ZonesSection.n_win);

			switch(ch)
			{
				case KEY_LEFT:
				case KEY_F(10): // exit/back
				case 27:
				case 'q':

					quit = 1;
					break;

				case KEY_IC: //insert

					new_zone_ptr = input_box(MAX_NETWORK, "New Network", "Enter the name of the new network");
					if(new_zone_ptr != NULL)
					{
						size = strlen(new_zone_ptr)+1+strlen(zonename)+1;

						temp_ptr = malloc(size);
						if(temp_ptr != NULL)
						{
							strlcpy(temp_ptr, new_zone_ptr, size);
							strlcat(temp_ptr, ".", size);
							strlcat(temp_ptr, zonename, size);

							if(validate_zonename(debug, temp_ptr, 1, NULL, NULL, NULL, reg->zonename) == 0)
							{
								if(new_zone(debug, temp_ptr, TYPE_NETWORK, &ZonedataList) >= 0)
								{
									(void)edit_zone_network(debuglvl, temp_ptr);
								}
							}
							else
							{
								(void)vrprint.error(-1, "Error", "invalid name: '%s'.", new_zone_ptr);
							}
							
							free(temp_ptr);
						}
						
						free(new_zone_ptr);
						
						reload = 1;
					}
					break;
				
				case KEY_DC: // delete
				{
					cur = current_item(ZonesSection.n_menu);
					if(cur)
					{
						if(count_zones(debuglvl, &ZonedataList, TYPE_HOST, (char *)item_name(cur), zonename) <= 0 && count_zones(debuglvl, &ZonedataList, TYPE_GROUP, (char *)item_name(cur), zonename) <= 0)
						{
							if(confirm("Delete", "This zone?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
							{
								size = strlen((char *)item_name(cur))+1+strlen(zonename)+1;

								if((zonename_ptr = malloc(size)))
								{
									strlcpy(zonename_ptr, (char *)item_name(cur), size);
									strlcat(zonename_ptr, ".", size);
									strlcat(zonename_ptr, zonename, size);

									result = delete_zone(debug, zonename_ptr, TYPE_NETWORK, &ZonedataList);
									if(result < 0)
										(void)vrprint.error(result, "Error", "deleting zone failed.");
									else
										reload = 1;

									free(zonename_ptr);
								}
							}
						}
						else
						{
							(void)vrprint.error(-1, "Error", "unable to delete: network not empty.");
						}
					}
					break;
				}
				
				case KEY_DOWN:
					menu_driver(ZonesSection.n_menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.n_menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.n_menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.n_menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.n_menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.n_menu, REQ_LAST_ITEM); // page down
					break;
				
				case KEY_RIGHT:
				case 10: // enter
				{
					cur = current_item(ZonesSection.n_menu);
					if(cur)
					{
						choice_ptr = selectbox("Select", "Hosts, Groups or this Network", 3, choices, 1);
						if(choice_ptr != NULL)
						{
							if(strcmp(choice_ptr, "Hosts") == 0)
							{
								(void)zones_section_menu_hosts(debuglvl, zonename, (char *)item_name(cur), reg);
							}
							else if(strcmp(choice_ptr, "Groups") == 0)
							{
								(void)zones_section_menu_groups(debuglvl, zonename, (char *)item_name(cur), reg);
							}
							else
							{
								size = strlen((char *)item_name(cur))+1+strlen(zonename)+1;
								if(size > 0)
								{
									if(!(zonename_ptr = malloc(size)))
									{
										(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
										retval = -1;
									}

									strlcpy(zonename_ptr, (char *)item_name(cur), size);
									strlcat(zonename_ptr, ".", size);
									strlcat(zonename_ptr, zonename, size);
								
									/*	edit the network. We don't care about the result.
										If there is an error, its up to the user to decide
										what to do. */
									(void)edit_zone_network(debuglvl, zonename_ptr);
								
									free(zonename_ptr);
								}
							}
							free(choice_ptr);
						}
					}
				        break;
				}

				case 'g': // group quick key

					cur = current_item(ZonesSection.n_menu);
					if(cur)
					{
						(void)zones_section_menu_groups(debuglvl, zonename, (char *)item_name(cur), reg);
					}
					break;

				case 'h': // host quick key

					cur = current_item(ZonesSection.n_menu);
					if(cur)
					{
						(void)zones_section_menu_hosts(debuglvl, zonename, (char *)item_name(cur), reg);
					}
					break;
				
				case 'e':
				case 32: // spacebar

					cur = current_item(ZonesSection.n_menu);
					if(cur)
					{
						size = strlen((char *)item_name(cur))+1+strlen(zonename)+1;
						if(size > 0)
						{
							if(!(zonename_ptr = malloc(size)))
							{
								(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
								retval = -1;
							}

							strlcpy(zonename_ptr, (char *)item_name(cur), size);
							strlcat(zonename_ptr, ".", size);
							strlcat(zonename_ptr, zonename, size);
							
							/*	edit the network. We don't care about the result.
								If there is an error, its up to the user to decide
								what to do. */
							(void)edit_zone_network(debuglvl, zonename_ptr);
						
							free(zonename_ptr);
						}
					}

					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:NETWORKS]:");
					break;

			}
		}
	}

	if(zones_section_menu_networks_destroy() < 0)
		retval = -1;

	update_panels();
	doupdate();

	return(retval);
}


static int
edit_zone_zone_init(const int debuglvl, char *name, int height, int width, int startx, int starty, struct ZoneData_ *zone_ptr)
{
	int	retval = 0;
	int	rows,
		cols,
		i;
	int	max_height,
		max_width;
	int	comment_y = 0,
		comment_x = 0;

	/* safety */
	if(!name || !zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* get screen dimentions */
	getmaxyx(stdscr, max_height, max_width);

	/* alloc fields */
	ZonesSection.EditZone.n_fields = 2;
	
	if(!(ZonesSection.EditZone.fields = (FIELD **)calloc(ZonesSection.EditZone.n_fields + 1, sizeof(FIELD *))))
	{
		(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
		return(-1);
	}

	/* preload the field */
	ZonesSection.EditZone.fields[0] = new_field(1, 3, 3, 1, 0, 1);
	set_field_buffer(ZonesSection.EditZone.fields[0], 0, zone_ptr->active ? "Yes" : "No");
	set_field_buffer(ZonesSection.EditZone.fields[0], 1, "act");

	/* comment field size */
	comment_y = 5;
	comment_x = 48;

	/* create and label the comment field */
	ZonesSection.EditZone.fields[1] = new_field(comment_y, comment_x, 6, 1, 0, 1);
	set_field_buffer(ZonesSection.EditZone.fields[1], 1, "comment");
	
	/* load the comment from the backend */
	if(zf->ask(debug, zone_backend, zone_ptr->name, "COMMENT", ZonesSection.comment, sizeof(ZonesSection.comment), TYPE_ZONE, 0) < 0)
		(void)vrprint.error(-1, "Error", "error while loading the comment.");

	set_field_buffer(ZonesSection.EditZone.fields[1], 0, ZonesSection.comment);

	ZonesSection.EditZone.fields[2] = NULL;

	/* create the window and panel */
	if(!(ZonesSection.EditZone.win = create_newwin(height, width, startx, starty, "Edit Zone: Zone", COLOR_PAIR(CP_BLUE_WHITE))))
	{
		(void)vrprint.error(-1, "Error", "could not create window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.EditZone.panel[0] = new_panel(ZonesSection.EditZone.win)))
	{
		(void)vrprint.error(-1, "Error", "could not create panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	keypad(ZonesSection.EditZone.win, TRUE);

	/* set the options */
	for(i=0; i < ZonesSection.EditZone.n_fields; i++)
	{
		set_field_back(ZonesSection.EditZone.fields[i], COLOR_PAIR(CP_WHITE_BLUE));
		field_opts_off(ZonesSection.EditZone.fields[i], O_AUTOSKIP);
		set_field_status(ZonesSection.EditZone.fields[i], FALSE);
	}

	/* Create the form and post it */
	if(!(ZonesSection.EditZone.form = new_form(ZonesSection.EditZone.fields)))
	{
		(void)vrprint.error(-1, "Error", "could not create form (in: %s).", __FUNCTION__);
		return(-1);
	}
	scale_form(ZonesSection.EditZone.form, &rows, &cols);
	set_form_win(ZonesSection.EditZone.form, ZonesSection.EditZone.win);
	set_form_sub(ZonesSection.EditZone.form, derwin(ZonesSection.EditZone.win, rows, cols, 1, 2));
	post_form(ZonesSection.EditZone.form);

	/* draw labels and other information */
	mvwprintw(ZonesSection.EditZone.win, 1, 2, "Name: %s", zone_ptr->name);
	mvwprintw(ZonesSection.EditZone.win, 3, 2, "Active:");
	mvwprintw(ZonesSection.EditZone.win, 6, 2, "Comment:");

	mvwprintw(ZonesSection.EditZone.win, 3, 35, "Networks : %4d", count_zones(debuglvl, &ZonedataList, TYPE_NETWORK, NULL, zone_ptr->name));
	mvwprintw(ZonesSection.EditZone.win, 4, 35, "Hosts    : %4d", count_zones(debuglvl, &ZonedataList, TYPE_HOST, NULL, zone_ptr->name));
	mvwprintw(ZonesSection.EditZone.win, 5, 35, "Groups   : %4d", count_zones(debuglvl, &ZonedataList, TYPE_GROUP, NULL, zone_ptr->name));

	/* draw and set cursor */
	wrefresh(ZonesSection.EditZone.win);
	pos_form_cursor(ZonesSection.EditZone.form);
	
	update_panels();
	doupdate();

	return(retval);
}

static int
edit_zone_zone_save(const int debuglvl, struct ZoneData_ *zone_ptr)
{
	int	retval = 0,
		i = 0;

	/* safety */
	if(!zone_ptr)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* check for changed fields */
	for(i = 0; i < ZonesSection.EditZone.n_fields; i++)
	{
		/* check is field is changed */
		if(field_status(ZonesSection.EditZone.fields[i]) == TRUE)
		{
			/* active */
			if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "act", 3) == 0)
			{
				zone_ptr->status = ST_CHANGED;

				if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[0], 0), "yes", 3) == 0)
				{
					zone_ptr->active = 1;
				}
				else if(strncasecmp(field_buffer(ZonesSection.EditZone.fields[0], 0), "no", 2) == 0)
				{
					zone_ptr->active = 0;
				}
				else
				{
					zone_ptr->active = -1;
				}

				/* save to backend */
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "ACTIVE", zone_ptr->active ? "Yes" : "No", 1, TYPE_ZONE) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
			/* comment */
			else if(strncmp(field_buffer(ZonesSection.EditZone.fields[i], 1), "comment", 7) == 0)
			{
				/* save the comment field to the backend */
				if(zf->tell(debuglvl, zone_backend, zone_ptr->name, "COMMENT", field_buffer(ZonesSection.EditZone.fields[i], 0), 1, TYPE_ZONE) < 0)
				{
					(void)vrprint.error(-1, "Error", "saving to backend failed (in: %s).", __FUNCTION__);
					retval = -1;
				}
			}
			else
			{
				(void)vrprint.error(-1, "Error", "unknown field '%s' (in: %s).", field_buffer(ZonesSection.EditZone.fields[i], 1), __FUNCTION__);
				retval = -1;
			}
		}
	}

	return(retval);
}

static void
edit_zone_zone_destroy(void)
{
	int	i = 0;

	// Un post form and free the memory
	unpost_form(ZonesSection.EditZone.form);
	free_form(ZonesSection.EditZone.form);

	for(i=0;i<ZonesSection.EditZone.n_fields;i++)
	{
		free_field(ZonesSection.EditZone.fields[i]);
	}
	free(ZonesSection.EditZone.fields);

	del_panel(ZonesSection.EditZone.panel[0]);
	destroy_win(ZonesSection.EditZone.win);

	// clear comment string
	strcpy(ZonesSection.comment, "");
}

/*	edit_zone_zone

	Edits a zone!
	
	Returncodes:
		0: ok
		-1: error
*/
static int
edit_zone_zone(const int debuglvl, char *name)
{
	int			ch,		/* for the keys */
				not_defined = 0,/* 1 is a key is defined */
				quit = 0,
				retval = 0;
	struct ZoneData_ 	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				max_height,
				max_width;
	FIELD			*cur = NULL,
				*prev = NULL;

	/* safety */
	if(!name)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}
	
	/* get the screen size */
	getmaxyx(stdscr, max_height, max_width);
	height = max_height-6;
	width = 54;
	startx = 3;
	starty = 24;

	/* look for the zone in the list */
	if(!(zone_ptr = search_zonedata(debuglvl, name, &ZonedataList)))
	{
		(void)vrprint.error(-1, "Error", "zone not found in memory.");
		return(0);
	}

	/* setup the window and fields */
	if(edit_zone_zone_init(debuglvl, name, height, width, startx, starty, zone_ptr) < 0)
		return(-1);

	cur = current_field(ZonesSection.EditZone.form);

	update_panels();
	doupdate();

	/* Loop through to get user requests/commands */
	while(quit == 0)
	{
		draw_field_active_mark(cur, prev, ZonesSection.EditZone.win, ZonesSection.EditZone.form, COLOR_PAIR(CP_RED_WHITE)|A_BOLD);

		not_defined = 0;

		/* get user input */
		ch = wgetch(ZonesSection.EditZone.win);

		/* user fields */

		/* comment */
		if(strncmp(field_buffer(cur, 1), "comment", 7) == 0)
		{
			if(nav_field_comment(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		/* active */
		else if(strncmp(field_buffer(cur, 1), "act", 3) == 0)
		{
			if(nav_field_yesno(ZonesSection.EditZone.form, ch) < 0)
				not_defined = 1;
		}
		else
			not_defined = 1;
		
		/* keys special for this window */
		if(not_defined == 1)
		{
			switch(ch)
			{
				case KEY_DOWN:
				case 10:	// enter
				case 9:		// tab
					form_driver(ZonesSection.EditZone.form, REQ_NEXT_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;

				case KEY_UP:
					// Go to previous field
					form_driver(ZonesSection.EditZone.form, REQ_PREV_FIELD);
					form_driver(ZonesSection.EditZone.form, REQ_BEG_LINE);
					break;

				case 27:
				case KEY_F(10):
					quit = 1;
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:ZONE:EDIT]:");
					break;
			}
		}

		/* set current field to prev */
		prev = cur;
		cur = current_field(ZonesSection.EditZone.form);

		/* draw and set cursor */
		wrefresh(ZonesSection.EditZone.win);
		pos_form_cursor(ZonesSection.EditZone.form);
	}

	/* save changes (if any) */
	if(edit_zone_zone_save(debuglvl, zone_ptr) < 0)
		retval = -1;

	/* cleanup */
	edit_zone_zone_destroy();
	update_panels();
	doupdate();

	return(retval);
}


/*
	TODO:
		list as func arg
*/
static int
zones_section_init(void)
{
	int			retval = 0,
				i = 0;
	struct ZoneData_	*zone_ptr = NULL;
	int			height,
				width,
				startx,
				starty,
				maxx,
				maxy;
	int			zones = 0;
	d_list_node		*d_node = NULL;

	getmaxyx(stdscr, maxy, maxx);

//	ZonesSection.n_choices = ZonedataList.len;
	
//	if(!(ZonesSection.items = (ITEM **)calloc(ZonesSection.n_choices + 1, sizeof(ITEM *))))
//	{
//		(void)vrprint.error(-1, "Error", "could not setup items (in: %s).", __FUNCTION__);
//		return(-1);
//	}

	i = ZonedataList.len - 1;

	/* count how many zones there are */
	for(d_node = ZonedataList.top; d_node ; d_node = d_node->next)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}
		
		if(zone_ptr->type == TYPE_ZONE)
			zones++;
	}
	ZonesSection.zone_n = zones;
	i = zones-1;

	if(!(ZonesSection.zoneitems = (ITEM **)calloc(ZonesSection.zone_n + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "unable to allocate memory (in: %s).", __FUNCTION__);
		return(-1);
	}

	for(d_node = ZonedataList.bot; d_node ; d_node = d_node->prev)
	{
		if(!(zone_ptr = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s).", __FUNCTION__);
			return(-1);
		}

		if(zone_ptr->type == TYPE_ZONE)
		{
			if(!(ZonesSection.zoneitems[i] = new_item(zone_ptr->name, NULL)))
			{
				(void)vrprint.error(-1, "Error", "could not add item (in: %s).", __FUNCTION__);
				return(-1);
			}
			
			i--;
		}
	}
	ZonesSection.zoneitems[ZonesSection.zone_n] = (ITEM *)NULL;
	
	if(!(ZonesSection.menu = new_menu((ITEM **)ZonesSection.zoneitems)))
	{
		(void)vrprint.error(-1, "Error", "could not setup menu (in: %s).", __FUNCTION__);
		return(-1);
	}

	height = ZonesSection.zone_n + 10;
	width = MAX_ZONE + 2;
	startx = 4;
	starty = 1;

	if(maxy < starty + height + 3)
	{
		starty = 3;
		height = maxy - 2 * starty;
	}

	if(!(ZonesSection.win = newwin(height, width, startx, starty)))
	{
		(void)vrprint.error(-1, "Error", "could not setup window (in: %s).", __FUNCTION__);
		return(-1);
	}
	if(!(ZonesSection.panel[0] = new_panel(ZonesSection.win)))
	{
		(void)vrprint.error(-1, "Error", "could not setup panel (in: %s).", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.win, COLOR_PAIR(CP_BLUE_WHITE));
	keypad(ZonesSection.win, TRUE);

	set_menu_win(ZonesSection.menu, ZonesSection.win);
	set_menu_sub(ZonesSection.menu, derwin(ZonesSection.win, height-8, width-2, 3, 1));
	set_menu_format(ZonesSection.menu, height-9, 1);

	box(ZonesSection.win, 0, 0);
	print_in_middle(ZonesSection.win, 1, 0, width, "Zones", COLOR_PAIR(CP_BLUE_WHITE));
	mvwaddch(ZonesSection.win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.win, 2, width-1, ACS_RTEE);
	
	set_menu_back(ZonesSection.menu, COLOR_PAIR(CP_BLUE_WHITE));
	set_menu_fore(ZonesSection.menu, COLOR_PAIR(CP_WHITE_BLUE));
	post_menu(ZonesSection.menu);

	mvwaddch(ZonesSection.win, height-6, 0, ACS_LTEE);
	mvwhline(ZonesSection.win, height-6, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.win, height-6, width-1, ACS_RTEE);

	/* print labels */
	mvwprintw(ZonesSection.win, height-5, 2, "<RET> to enter the zone");
	mvwprintw(ZonesSection.win, height-4, 2, "<INS> to insert a new zone");
	mvwprintw(ZonesSection.win, height-3, 2, "<DEL> to removes the zone");
	mvwprintw(ZonesSection.win, height-2, 2, "< e > to edit the zone");

	update_panels();
	doupdate();

	return(retval);
}

int
zones_section_destroy(void)
{
	int	retval = 0,
		i;

	unpost_menu(ZonesSection.menu);
	free_menu(ZonesSection.menu);
	for(i = 0; i < ZonesSection.zone_n; ++i)
		free_item(ZonesSection.zoneitems[i]);

	free(ZonesSection.zoneitems);
	
	del_panel(ZonesSection.panel[0]);
	destroy_win(ZonesSection.win);

	return(retval);
}

int
zones_section(const int debuglvl, struct rgx_ *reg)
{
	int	ch,
		quit = 0,
		reload = 0,
		result = 0,
		retval = 0;
	char	*new_zone_ptr = NULL;
	ITEM	*cur = NULL;

	/* safety */
	if(!reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	if(zones_section_init() < 0)
		return(-1);

	while(quit == 0)
	{
		if(reload == 1)
		{
			if(zones_section_destroy() < 0)
				return(-1);
				
			if(zones_section_init() < 0)
				return(-1);
				
			reload = 0;
		}

		while(quit == 0 && reload == 0)
		{
			ch = wgetch(ZonesSection.win);

			switch(ch)
			{
				case KEY_LEFT:
				case KEY_F(10):
				case 'q':
				case 27:
					quit=1;
					break;

				case KEY_F(5): //help
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_IC: //insert

					new_zone_ptr = input_box(MAX_ZONE, "New Zone", "Enter the name of the new zone");
					if(new_zone_ptr != NULL)
					{
						if(validate_zonename(debug, new_zone_ptr, 1, NULL, NULL, NULL, reg->zonename) == 0)
						{
							if(new_zone(debug, new_zone_ptr, TYPE_ZONE, &ZonedataList) < 0)
							{
								(void)vrprint.error(result, "Error", "adding zone failed (in: %s).", __FUNCTION__);
							}
							else
							{
								if(edit_zone_zone(debuglvl, new_zone_ptr) < 0)
								{
									retval = -1;
									quit = 1;
								}
							}
						}
					
						free(new_zone_ptr);
					}
					
					reload = 1;
					
					break;

				case KEY_DC: // delete

					cur = current_item(ZonesSection.menu);
					if(cur)
					{
						if(confirm("Delete", "This zone?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{
							result = delete_zone(debug, (char *)item_name(cur), TYPE_ZONE, &ZonedataList);
							if(result < 0)
							{
								(void)vrprint.error(result, "Error", "deleting zone failed (in: %s).", __FUNCTION__);
							}
							else
							{
								reload = 1;
							}
						}
					}
					break;

				case KEY_DOWN:
					menu_driver(ZonesSection.menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.menu, REQ_LAST_ITEM); // page down
					break;

				case 'e':
				case 32:

					cur = current_item(ZonesSection.menu);
					if(cur)
					{
						if(edit_zone_zone(debuglvl, (char *)item_name(cur)) < 0)
						{
							retval = -1;
							quit = 1;
						}
					}

					break;

				case KEY_RIGHT:
				case 10:	// enter

					cur = current_item(ZonesSection.menu);
					if(cur)
					{
						if(zones_section_menu_networks(debuglvl, (char *)item_name(cur), reg) < 0)
						{
							retval = -1;
							quit = 1;
						}
					}
					break;

				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:ZONES:ZONES]:");
					break;
			}
		}
	}

	if(zones_section_destroy() < 0)
		retval = -1;

	update_panels();
	doupdate();
	
	return(retval);
}


static int
zones_blocklist_init(const int debuglvl, d_list *block_list)
{
	int			retval=0,
				i = 0,
				result = 0;
	int			height = 0,
				width = 0,
				startx = 0,
				starty = 0,
				maxx = 0,
				maxy = 0;
	d_list_node		*d_node = NULL;
	char			*string = NULL;
	
	/* get the screensize */
	getmaxyx(stdscr, maxy, maxx);

	/* number of items */
	ZonesSection.host_n = block_list->len;

	/* allow the menu items */
	if(!(ZonesSection.hostitems = (ITEM **)calloc(ZonesSection.host_n + 1, sizeof(ITEM *))))
	{
		(void)vrprint.error(-1, "Error", "calloc failed: %s (in: %s:%d).", strerror(errno), __FUNCTION__, __LINE__);
		return(-1);
	}

	/* create the menu items */
	for(d_node = block_list->top, i = 0; d_node ; d_node = d_node->next, i++)
	{
		if(!(string = d_node->data))
		{
			(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
			return(-1);
		}

		if(!(ZonesSection.hostitems[i] = new_item(string, NULL)))
		{
			(void)vrprint.error(-1, "Internal Error", "adding new item to blocklist menu failed (in: %s).", __FUNCTION__);
			return(-1);
		}
	}

	/* terminate the items */
	ZonesSection.hostitems[ZonesSection.host_n] = (ITEM *)NULL;

	/* now create the menu */
	if(!(ZonesSection.h_menu = new_menu((ITEM **)ZonesSection.hostitems)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the blocklist menu failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* now set the size of the window */
	height = ZonesSection.host_n + 9;
	width  = MAX_HOST_NET_ZONE + 2;
	startx = 1;
	starty = 4;

	if(maxy < starty + height + 3)
	{
		starty = 3;
		height = maxy - 2 * starty;
	}

	if(maxx < startx + width + 3)
	{
		startx = 1;
		width = maxx - 2 * startx;
	}

	if(!(ZonesSection.h_win = newwin(height, width, starty, startx)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the blocklist win failed (in: %s).", __FUNCTION__);
		return(-1);
	}
	wbkgd(ZonesSection.h_win, COLOR_PAIR(5));
	keypad(ZonesSection.h_win, TRUE);
	box(ZonesSection.h_win, 0, 0);
	print_in_middle(ZonesSection.h_win, 1, 0, width, "Block List", COLOR_PAIR(5)); 
	wrefresh(ZonesSection.h_win);

	if(!(ZonesSection.h_panel[0] = new_panel(ZonesSection.h_win)))
	{
		(void)vrprint.error(-1, "Internal Error", "creating the blocklist panel failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	set_menu_win(ZonesSection.h_menu, ZonesSection.h_win);
	set_menu_sub(ZonesSection.h_menu, derwin(ZonesSection.h_win, height-7, width-2, 3, 1));

	set_menu_format(ZonesSection.h_menu, height-8, 1);

	mvwaddch(ZonesSection.h_win, 2, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, 2, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, 2, width-1, ACS_RTEE);

	set_menu_back(ZonesSection.h_menu, COLOR_PAIR(5));
	set_menu_fore(ZonesSection.h_menu, COLOR_PAIR(3));

	result = post_menu(ZonesSection.h_menu);
	if(result != E_OK && result != E_NOT_CONNECTED)
	{
		(void)vrprint.error(-1, "Internal Error", "posting the blocklist menu failed (in: %s).", __FUNCTION__);
		return(-1);
	}

	mvwaddch(ZonesSection.h_win, height-4, 0, ACS_LTEE);
	mvwhline(ZonesSection.h_win, height-4, 1, ACS_HLINE, width-2);
	mvwaddch(ZonesSection.h_win, height-4, width-1, ACS_RTEE);

	mvwprintw(ZonesSection.h_win, height-3, 1, "Usage: <INS> to insert a new item");
	mvwprintw(ZonesSection.h_win, height-2, 1, "       <DEL> to removes the item");

	update_panels();
	doupdate();

	return(retval);
}

static int
zones_blocklist_destroy(void)
{
	int	retval = 0,
		i = 0;

	unpost_menu(ZonesSection.h_menu);
	free_menu(ZonesSection.h_menu);
	for(i = 0; i < ZonesSection.host_n; ++i)
		free_item(ZonesSection.hostitems[i]);

	free(ZonesSection.hostitems);

	del_panel(ZonesSection.h_panel[0]);

	destroy_win(ZonesSection.h_win);

	return(retval);
}


int
zones_blocklist_add_one(const int debuglvl, d_list *block_list, d_list *zones_list)
{
	char			*new_ipaddress = NULL,
				*choices[] = { "IPAddress", "Host", "Group" },
				*choice_ptr = NULL,
				choice_type = 0,
				**zone_choices;
	int			i = 0;
	char			changes = 0;
	struct ZoneData_	*zone_ptr = NULL;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!block_list || !zones_list)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}


	choice_ptr = selectbox("Select", "Block a IP/Host/Group", 3, choices, 1);
	if(choice_ptr != NULL)
	{
		if(strcasecmp(choice_ptr, "ipaddress") == 0)
		{
			new_ipaddress = input_box(16, "New IPAddress", "Enter the ipaddress to be blocked");
			if(new_ipaddress != NULL)
			{
				/* validate ip */
				if(check_ipv4address(debuglvl, NULL, NULL, new_ipaddress, 1) != 1)
				{
					(void)vrprint.warning("Warning", "'%s' is not a valid ipaddress.", new_ipaddress);

					free(new_ipaddress);
					new_ipaddress = NULL;
				}
				else
				{
					/* add to list */
					if(d_list_append(debuglvl, block_list, new_ipaddress) < 0)
					{
						(void)vrprint.error(-1, "Internal Error", "adding ipaddress to list failed (in: %s:%d).", __FUNCTION__, __LINE__);
						return(-1);
					}

					changes = 1;
					new_ipaddress = NULL;
				}
			}
		}
		else if(strcasecmp(choice_ptr, "host") == 0 || strcasecmp(choice_ptr, "group") == 0)
		{
			d_list_setup(debuglvl, &ZonesSection.desc_list, free);

			/* get the type */
			if(strcasecmp(choice_ptr, "host") == 0)
				choice_type = TYPE_HOST;
			else
				choice_type = TYPE_GROUP;

			for(d_node = zones_list->top, i = 0; d_node ; d_node = d_node->next)
			{
				if(!(zone_ptr = d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				if(zone_ptr->type == choice_type)
				{
					i++;
				}
			}

			if(!(zone_choices = calloc(i + 1, MAX_HOST_NET_ZONE)))
			{
				(void)vrprint.error(-1, "Error", "malloc failed: %s.", strerror(errno));
				return(-1);
			}

			for(d_node = zones_list->top, i = 0; d_node ; d_node = d_node->next)
			{
				if(!(zone_ptr = d_node->data))
				{
					(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				if(zone_ptr->type == choice_type)
				{
					zone_choices[i] = zone_ptr->name;
					i++;
				}
			}
			zone_choices[i] = NULL;

			/* get the zone */
			if((choice_ptr = selectbox("Select", "Select a host or group to block", i, zone_choices, 2)))
			{
				/* add to list */
				if(d_list_append(debuglvl, block_list, choice_ptr) < 0)
				{
					(void)vrprint.error(-1, "Internal Error", "adding host/group to list failed (in: %s:%d).", __FUNCTION__, __LINE__);
					return(-1);
				}

				changes = 1;
			}

			free(zone_choices);
			zone_choices = NULL;
		}
	}

	if(changes)
		return(1);

	return(0);
}


int
zones_blocklist(const int debuglvl, d_list *block_list, d_list *zones_list, struct rgx_ *reg)
{
	int			ch = 0,
				quit = 0,
				reload = 0,
				retval = 0;
	char			changes = 0;
	char			*itemname = NULL,
				*listitemname = NULL;
	ITEM			*cur = NULL;
	d_list_node		*d_node = NULL;


	/* safety */
	if(!block_list || !zones_list || !reg)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s).", __FUNCTION__);
		return(-1);
	}

	/* setup */
	if(zones_blocklist_init(debuglvl, block_list) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up blocklist menu failed (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	/* enter the loop */
	while(quit == 0)
	{
		/* reload the menu */
		if(reload == 1)
		{
			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, reloading menu.");
		
			/* first destroy */
			if(zones_blocklist_destroy() < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, old menu destroyed.");

			/* and setup again */
			if(zones_blocklist_init(debuglvl, block_list) < 0)
			{
				(void)vrprint.error(-1, "Internal Error", "reinitializing menu failed (in: %s).", __FUNCTION__);
				return(-1);
			}

			if(debuglvl >= HIGH)
				(void)vrprint.debug(__FUNCTION__, "reload == 1, new menu created.");

			/* we are done with reloading */
			reload = 0;
		}

		/* loop for catching user input */
		while(quit == 0 && reload == 0)
		{
			/* get the user input */
			ch = wgetch(ZonesSection.h_win);
			switch(ch)
			{
				case KEY_LEFT:
				case KEY_F(10):
				case 27:
				case 'q':
					quit = 1;
					break;

				case KEY_F(5): //help
					status_print(status_win, "FIXME: help not yet implemented");
					break;

				case KEY_IC: /* insert key */
					
					if(zones_blocklist_add_one(debuglvl, block_list, zones_list) == 1)
					{
						changes = 1;
						reload = 1;
					}

					break;

				/*
					delete
				*/
				case KEY_DC:

					if(block_list->len > 0)
					{
						if(confirm("Remove", "This IP/Host/Group?", COLOR_PAIR(CP_RED_WHITE), COLOR_PAIR(CP_WHITE_RED)|A_BOLD, 0) == 1)
						{
							/* get the current item */
							if(!(cur = current_item(ZonesSection.h_menu)))
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
								return(-1);
							}

							if(!(itemname = (char *)item_name(cur)))
							{
								(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
								return(-1);
							}

							if(debuglvl >= HIGH)
								(void)vrprint.debug(__FUNCTION__, "itemname to remove: '%s'.", itemname);

							/* look the item up in the list */
							for(d_node = block_list->top; d_node; d_node = d_node->next)
							{
								if(!(listitemname = d_node->data))
								{
									(void)vrprint.error(-1, "Internal Error", "NULL pointer (in: %s:%d).", __FUNCTION__, __LINE__);
									return(-1);
								}

								if(strcmp(listitemname, itemname) == 0)
								{
									/* this one needs to be removed */
									if(d_list_remove_node(debuglvl, block_list, d_node) < 0)
									{
										(void)vrprint.error(-1, "Internal Error", "removing item from list failed (in: %s:%d).", __FUNCTION__, __LINE__);
										return(-1);
									}

									itemname = NULL;
									listitemname = NULL;

									changes = 1;
									reload = 1;
									break;
								}
							}
						}
					}
					break;

				case KEY_DOWN:
					menu_driver(ZonesSection.h_menu, REQ_DOWN_ITEM);
					break;
				case KEY_UP:
					menu_driver(ZonesSection.h_menu, REQ_UP_ITEM);
					break;
				case KEY_NPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_DPAGE); // page up
					break;
				case KEY_PPAGE:
					menu_driver(ZonesSection.h_menu, REQ_SCR_UPAGE); // page down
					break;
				case KEY_HOME:
					menu_driver(ZonesSection.h_menu, REQ_FIRST_ITEM); // page up
					break;
				case KEY_END:
					menu_driver(ZonesSection.h_menu, REQ_LAST_ITEM); // page down
					break;
				case KEY_F(12):
					print_help(debuglvl, ":[VUURMUUR:BLOCKLIST]:");
					break;
			}
		}
	}

	if(changes && retval == 0)
	{
		if(debuglvl >= HIGH)
			(void)vrprint.debug(__FUNCTION__, "changes and retval == 0 so save the list to disk.");

		if(blocklist_write_file(debuglvl, block_list) < 0)
			retval = -1;
	}

	if(zones_blocklist_destroy() < 0)
		retval = -1;

	update_panels();
	doupdate();

	return(retval);
}
